@@ -3,8 +3,9 @@
 #include <string.h>
 
 #define CMARK_NO_SHORT_NAMES
-#include "cmark.h"
+#include <cmark-gfm.h>
 #include "node.h"
+#include <cmark-gfm-core-extensions.h>
 
 #include "harness.h"
 #include "cplusplus.h"
@@ -24,7 +25,7 @@ static void test_md_to_html(test_batch_runner *runner, const char *markdown,
                             const char *expected_html, const char *msg);
 
 static void test_content(test_batch_runner *runner, cmark_node_type type,
-                         int allowed_content);
+                         unsigned int *allowed_content);
 
 static void test_char(test_batch_runner *runner, int valid, const char *utf8,
                       const char *msg);
@@ -35,8 +36,8 @@ static void test_incomplete_char(test_batch_runner *runner, const char *utf8,
 static void test_continuation_byte(test_batch_runner *runner, const char *utf8);
 
 static void version(test_batch_runner *runner) {
-  INT_EQ(runner, cmark_version(), CMARK_VERSION, "cmark_version");
-  STR_EQ(runner, cmark_version_string(), CMARK_VERSION_STRING,
+  INT_EQ(runner, cmark_version(), CMARK_GFM_VERSION, "cmark_version");
+  STR_EQ(runner, cmark_version_string(), CMARK_GFM_VERSION_STRING,
          "cmark_version_string");
 }
 
@@ -177,8 +178,7 @@ static void accessors(test_batch_runner *runner) {
   OK(runner, cmark_node_set_literal(string, literal + sizeof("prefix")),
      "set_literal suffix");
 
-  char *rendered_html = cmark_render_html(doc,
-		          CMARK_OPT_DEFAULT | CMARK_OPT_UNSAFE);
+  char *rendered_html = cmark_render_html(doc, CMARK_OPT_DEFAULT | CMARK_OPT_UNSAFE, NULL);
   static const char expected_html[] =
       "<h3>Header</h3>\n"
       "<ol start=\"3\">\n"
@@ -243,21 +243,6 @@ static void accessors(test_batch_runner *runner) {
   cmark_node_free(doc);
 }
 
-static void free_parent(test_batch_runner *runner) {
-  static const char markdown[] = "text\n";
-
-  cmark_node *doc =
-      cmark_parse_document(markdown, sizeof(markdown) - 1, CMARK_OPT_DEFAULT);
-
-  cmark_node *para = cmark_node_first_child(doc);
-  cmark_node *text = cmark_node_first_child(para);
-  cmark_node_unlink(text);
-  cmark_node_free(doc);
-  STR_EQ(runner, cmark_node_get_literal(text), "text",
-         "inline content after freeing parent block");
-  cmark_node_free(text);
-}
-
 static void node_check(test_batch_runner *runner) {
   // Construct an incomplete tree.
   cmark_node *doc = cmark_node_new(CMARK_NODE_DOCUMENT);
@@ -315,7 +300,7 @@ static void iterator_delete(test_batch_runner *runner) {
     }
   }
 
-  char *html = cmark_render_html(doc, CMARK_OPT_DEFAULT);
+  char *html = cmark_render_html(doc, CMARK_OPT_DEFAULT, NULL);
   static const char expected[] = "<p>a  c</p>\n"
                                  "<p>a  c</p>\n";
   STR_EQ(runner, html, expected, "iterate and delete nodes");
@@ -355,7 +340,7 @@ static void create_tree(test_batch_runner *runner) {
   OK(runner, cmark_node_append_child(emph, str2), "append3");
   INT_EQ(runner, cmark_node_check(doc, NULL), 0, "append3 consistent");
 
-  html = cmark_render_html(doc, CMARK_OPT_DEFAULT);
+  html = cmark_render_html(doc, CMARK_OPT_DEFAULT, NULL);
   STR_EQ(runner, html, "<p>Hello, <em>world</em>!</p>\n", "render_html");
   free(html);
 
@@ -391,11 +376,14 @@ static void create_tree(test_batch_runner *runner) {
 
   cmark_node_unlink(emph);
 
-  html = cmark_render_html(doc, CMARK_OPT_DEFAULT);
+  html = cmark_render_html(doc, CMARK_OPT_DEFAULT, NULL);
   STR_EQ(runner, html, "<p>brzz!</p>\n", "render_html after shuffling");
   free(html);
 
   cmark_node_free(doc);
+
+  // TODO: Test that the contents of an unlinked inline are valid
+  // after the parent block was destroyed. This doesn't work so far.
   cmark_node_free(emph);
 }
 
@@ -420,7 +408,7 @@ static void custom_nodes(test_batch_runner *runner) {
   STR_EQ(runner, cmark_node_get_on_exit(cb), "", "get_on_exit (empty)");
   cmark_node_append_child(doc, cb);
 
-  html = cmark_render_html(doc, CMARK_OPT_DEFAULT);
+  html = cmark_render_html(doc, CMARK_OPT_DEFAULT, NULL);
   STR_EQ(runner, html, "<p><ON ENTER|Hello|ON EXIT></p>\n<on enter|\n",
          "render_html");
   free(html);
@@ -447,22 +435,18 @@ void hierarchy(test_batch_runner *runner) {
 
   cmark_node_free(bquote1);
 
-  int max_node_type = CMARK_NODE_LAST_BLOCK > CMARK_NODE_LAST_INLINE
-                          ? CMARK_NODE_LAST_BLOCK
-                          : CMARK_NODE_LAST_INLINE;
-  OK(runner, max_node_type < 32, "all node types < 32");
-
-  int list_item_flag = 1 << CMARK_NODE_ITEM;
-  int top_level_blocks =
-      (1 << CMARK_NODE_BLOCK_QUOTE) | (1 << CMARK_NODE_LIST) |
-      (1 << CMARK_NODE_CODE_BLOCK) | (1 << CMARK_NODE_HTML_BLOCK) |
-      (1 << CMARK_NODE_PARAGRAPH) | (1 << CMARK_NODE_HEADING) |
-      (1 << CMARK_NODE_THEMATIC_BREAK);
-  int all_inlines = (1 << CMARK_NODE_TEXT) | (1 << CMARK_NODE_SOFTBREAK) |
-                    (1 << CMARK_NODE_LINEBREAK) | (1 << CMARK_NODE_CODE) |
-                    (1 << CMARK_NODE_HTML_INLINE) | (1 << CMARK_NODE_EMPH) |
-                    (1 << CMARK_NODE_STRONG) | (1 << CMARK_NODE_LINK) |
-                    (1 << CMARK_NODE_IMAGE);
+  unsigned int list_item_flag[] = {CMARK_NODE_ITEM, 0};
+  unsigned int top_level_blocks[] = {
+    CMARK_NODE_BLOCK_QUOTE, CMARK_NODE_LIST,
+    CMARK_NODE_CODE_BLOCK, CMARK_NODE_HTML_BLOCK,
+    CMARK_NODE_PARAGRAPH, CMARK_NODE_HEADING,
+    CMARK_NODE_THEMATIC_BREAK, 0};
+  unsigned int all_inlines[] = {
+    CMARK_NODE_TEXT, CMARK_NODE_SOFTBREAK,
+    CMARK_NODE_LINEBREAK, CMARK_NODE_CODE,
+    CMARK_NODE_HTML_INLINE, CMARK_NODE_EMPH,
+    CMARK_NODE_STRONG, CMARK_NODE_LINK,
+    CMARK_NODE_IMAGE, 0};
 
   test_content(runner, CMARK_NODE_DOCUMENT, top_level_blocks);
   test_content(runner, CMARK_NODE_BLOCK_QUOTE, top_level_blocks);
@@ -485,7 +469,7 @@ void hierarchy(test_batch_runner *runner) {
 }
 
 static void test_content(test_batch_runner *runner, cmark_node_type type,
-                         int allowed_content) {
+                         unsigned int *allowed_content) {
   cmark_node *node = cmark_node_new(type);
 
   for (int i = 0; i < num_node_types; ++i) {
@@ -493,7 +477,10 @@ static void test_content(test_batch_runner *runner, cmark_node_type type,
     cmark_node *child = cmark_node_new(child_type);
 
     int got = cmark_node_append_child(node, child);
-    int expected = (allowed_content >> child_type) & 1;
+    int expected = 0;
+    if (allowed_content)
+        for (unsigned int *p = allowed_content; *p; ++p)
+            expected |= *p == (unsigned int)child_type;
 
     INT_EQ(runner, got, expected, "add %d as child of %d", child_type, type);
 
@@ -518,17 +505,17 @@ static void render_html(test_batch_runner *runner) {
       cmark_parse_document(markdown, sizeof(markdown) - 1, CMARK_OPT_DEFAULT);
 
   cmark_node *paragraph = cmark_node_first_child(doc);
-  html = cmark_render_html(paragraph, CMARK_OPT_DEFAULT);
+  html = cmark_render_html(paragraph, CMARK_OPT_DEFAULT, NULL);
   STR_EQ(runner, html, "<p>foo <em>bar</em></p>\n", "render single paragraph");
   free(html);
 
   cmark_node *string = cmark_node_first_child(paragraph);
-  html = cmark_render_html(string, CMARK_OPT_DEFAULT);
+  html = cmark_render_html(string, CMARK_OPT_DEFAULT, NULL);
   STR_EQ(runner, html, "foo ", "render single inline");
   free(html);
 
   cmark_node *emph = cmark_node_next(string);
-  html = cmark_render_html(emph, CMARK_OPT_DEFAULT);
+  html = cmark_render_html(emph, CMARK_OPT_DEFAULT, NULL);
   STR_EQ(runner, html, "<em>bar</em>", "render inline with children");
   free(html);
 
@@ -709,6 +696,50 @@ static void render_commonmark(test_batch_runner *runner) {
   cmark_node_free(doc);
 }
 
+static void render_plaintext(test_batch_runner *runner) {
+  char *plaintext;
+
+  static const char markdown[] = "> \\- foo *bar* \\*bar\\*\n"
+                                 "\n"
+                                 "- Lorem ipsum dolor sit amet,\n"
+                                 "  consectetur adipiscing elit,\n"
+                                 "- sed do eiusmod tempor incididunt\n"
+                                 "  ut labore et dolore magna aliqua.\n";
+  cmark_node *doc =
+      cmark_parse_document(markdown, sizeof(markdown) - 1, CMARK_OPT_DEFAULT);
+
+  plaintext = cmark_render_plaintext(doc, CMARK_OPT_DEFAULT, 26);
+  STR_EQ(runner, plaintext, "- foo bar *bar*\n"
+                             "\n"
+                             "  - Lorem ipsum dolor sit\n"
+                             "    amet, consectetur\n"
+                             "    adipiscing elit,\n"
+                             "  - sed do eiusmod tempor\n"
+                             "    incididunt ut labore\n"
+                             "    et dolore magna\n"
+                             "    aliqua.\n",
+         "render document with wrapping");
+  free(plaintext);
+  plaintext = cmark_render_plaintext(doc, CMARK_OPT_DEFAULT, 0);
+  STR_EQ(runner, plaintext, "- foo bar *bar*\n"
+                             "\n"
+                             "  - Lorem ipsum dolor sit amet,\n"
+                             "    consectetur adipiscing elit,\n"
+                             "  - sed do eiusmod tempor incididunt\n"
+                             "    ut labore et dolore magna aliqua.\n",
+         "render document without wrapping");
+  free(plaintext);
+
+  cmark_node *text = cmark_node_new(CMARK_NODE_TEXT);
+  cmark_node_set_literal(text, "Hi");
+  plaintext = cmark_render_plaintext(text, CMARK_OPT_DEFAULT, 0);
+  STR_EQ(runner, plaintext, "Hi\n", "render single inline node");
+  free(plaintext);
+
+  cmark_node_free(text);
+  cmark_node_free(doc);
+}
+
 static void utf8(test_batch_runner *runner) {
   // Ranges
   test_char(runner, 1, "\x01", "valid utf8 01");
@@ -757,6 +788,13 @@ static void utf8(test_batch_runner *runner) {
   STR_EQ(runner, html, "<pre><code>\xef\xbf\xbd\n</code></pre>\n",
          "utf8 with \\0\\n");
   free(html);
+
+  // Test byte-order marker
+  static const char string_with_bom[] = "\xef\xbb\xbf# Hello\n";
+  html = cmark_markdown_to_html(
+      string_with_bom, sizeof(string_with_bom) - 1, CMARK_OPT_DEFAULT);
+  STR_EQ(runner, html, "<h1>Hello</h1>\n", "utf8 with BOM");
+  free(html);
 }
 
 static void test_char(test_batch_runner *runner, int valid, const char *utf8,
@@ -849,11 +887,11 @@ static void numeric_entities(test_batch_runner *runner) {
                   "Valid numeric entity 0x10FFFF");
   test_md_to_html(runner, "&#x110000;", "<p>" UTF8_REPL "</p>\n",
                   "Invalid numeric entity 0x110000");
-  test_md_to_html(runner, "&#x80000000;", "<p>&amp;#x80000000;</p>\n",
+  test_md_to_html(runner, "&#x80000000;", "<p>" UTF8_REPL "</p>\n",
                   "Invalid numeric entity 0x80000000");
-  test_md_to_html(runner, "&#xFFFFFFFF;", "<p>&amp;#xFFFFFFFF;</p>\n",
+  test_md_to_html(runner, "&#xFFFFFFFF;", "<p>" UTF8_REPL "</p>\n",
                   "Invalid numeric entity 0xFFFFFFFF");
-  test_md_to_html(runner, "&#99999999;", "<p>&amp;#99999999;</p>\n",
+  test_md_to_html(runner, "&#99999999;", "<p>" UTF8_REPL "</p>\n",
                   "Invalid numeric entity 99999999");
 
   test_md_to_html(runner, "&#;", "<p>&amp;#;</p>\n",
@@ -900,6 +938,61 @@ static void test_feed_across_line_ending(test_batch_runner *runner) {
   cmark_node_free(document);
 }
 
+#if !defined(_WIN32) || defined(__CYGWIN__)
+#  include <sys/time.h>
+static struct timeval _before, _after;
+static int _timing;
+#  define START_TIMING() \
+       gettimeofday(&_before, NULL)
+
+#  define END_TIMING() \
+        do { \
+          gettimeofday(&_after, NULL); \
+          _timing = (_after.tv_sec - _before.tv_sec) * 1000 + (_after.tv_usec - _before.tv_usec) / 1000; \
+        } while (0)
+
+#  define TIMING _timing
+#else
+#  define START_TIMING()
+#  define END_TIMING()
+#  define TIMING 0
+#endif
+
+static void test_pathological_regressions(test_batch_runner *runner) {
+  {
+    // I don't care what the output is, so long as it doesn't take too long.
+    char path[] = "[a](b";
+    char *input = (char *)calloc(1, (sizeof(path) - 1) * 50000);
+    for (int i = 0; i < 50000; ++i)
+      memcpy(input + i * (sizeof(path) - 1), path, sizeof(path) - 1);
+
+    START_TIMING();
+    char *html = cmark_markdown_to_html(input, (sizeof(path) - 1) * 50000,
+                                        CMARK_OPT_VALIDATE_UTF8);
+    END_TIMING();
+    free(html);
+    free(input);
+
+    OK(runner, TIMING < 1000, "takes less than 1000ms to run");
+  }
+
+  {
+    char path[] = "[a](<b";
+    char *input = (char *)calloc(1, (sizeof(path) - 1) * 50000);
+    for (int i = 0; i < 50000; ++i)
+      memcpy(input + i * (sizeof(path) - 1), path, sizeof(path) - 1);
+
+    START_TIMING();
+    char *html = cmark_markdown_to_html(input, (sizeof(path) - 1) * 50000,
+                                        CMARK_OPT_VALIDATE_UTF8);
+    END_TIMING();
+    free(html);
+    free(input);
+
+    OK(runner, TIMING < 1000, "takes less than 1000ms to run");
+  }
+}
+
 static void source_pos(test_batch_runner *runner) {
   static const char markdown[] =
     "# Hi *there*.\n"
@@ -911,13 +1004,19 @@ static void source_pos(test_batch_runner *runner) {
     ">    Sure.\n"
     ">\n"
     "> 2. Yes, okay.\n"
-    ">    ![ok](hi \"yes\")\n";
+    ">    ![ok](hi \"yes\")\n"
+    "<!-- HTML Comment -->\n"
+    "\n"
+    "what happens if we spread a link [across multiple\n"
+    "lines][anchor]\n"
+    "\n"
+    "[anchor]: http://example.com\n";
 
   cmark_node *doc = cmark_parse_document(markdown, sizeof(markdown) - 1, CMARK_OPT_DEFAULT);
   char *xml = cmark_render_xml(doc, CMARK_OPT_DEFAULT | CMARK_OPT_SOURCEPOS);
   STR_EQ(runner, xml, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n"
                       "<!DOCTYPE document SYSTEM \"CommonMark.dtd\">\n"
-                      "<document sourcepos=\"1:1-10:20\" xmlns=\"http://commonmark.org/xml/1.0\">\n"
+                      "<document sourcepos=\"1:1-16:28\" xmlns=\"http://commonmark.org/xml/1.0\">\n"
                       "  <heading sourcepos=\"1:1-1:13\" level=\"1\">\n"
                       "    <text sourcepos=\"1:3-1:5\" xml:space=\"preserve\">Hi </text>\n"
                       "    <emph sourcepos=\"1:6-1:12\">\n"
@@ -926,8 +1025,8 @@ static void source_pos(test_batch_runner *runner) {
                       "    <text sourcepos=\"1:13-1:13\" xml:space=\"preserve\">.</text>\n"
                       "  </heading>\n"
                       "  <paragraph sourcepos=\"3:1-4:42\">\n"
-                      "    <text sourcepos=\"3:1-3:14\" xml:space=\"preserve\">Hello “ </text>\n"
-                      "    <link sourcepos=\"3:15-3:37\" destination=\"http://www.google.com\">\n"
+                      "    <text sourcepos=\"3:1-3:14\" xml:space=\"preserve\">Hello \xe2\x80\x9c </text>\n"
+                      "    <link sourcepos=\"3:15-3:37\" destination=\"http://www.google.com\" title=\"\">\n"
                       "      <text sourcepos=\"3:16-3:36\" xml:space=\"preserve\">http://www.google.com</text>\n"
                       "    </link>\n"
                       "    <softbreak />\n"
@@ -959,6 +1058,16 @@ static void source_pos(test_batch_runner *runner) {
                       "      </item>\n"
                       "    </list>\n"
                       "  </block_quote>\n"
+                      "  <html_block sourcepos=\"11:1-11:21\" xml:space=\"preserve\">&lt;!-- HTML Comment --&gt;\n"
+                      "</html_block>\n"
+                      "  <paragraph sourcepos=\"13:1-14:14\">\n"
+                      "    <text sourcepos=\"13:1-13:33\" xml:space=\"preserve\">what happens if we spread a link </text>\n"
+                      "    <link sourcepos=\"13:34-14:14\" destination=\"http://example.com\" title=\"\">\n"
+                      "      <text sourcepos=\"13:35-13:49\" xml:space=\"preserve\">across multiple</text>\n"
+                      "      <softbreak />\n"
+                      "      <text sourcepos=\"14:1-14:5\" xml:space=\"preserve\">lines</text>\n"
+                      "    </link>\n"
+                      "  </paragraph>\n"
                       "</document>\n",
          "sourcepos are as expected");
   free(xml);
@@ -969,13 +1078,15 @@ static void source_pos_inlines(test_batch_runner *runner) {
   {
     static const char markdown[] =
       "*first*\n"
-      "second\n";
+      "second\n"
+      "\n"
+      "   <http://example.com>";
 
     cmark_node *doc = cmark_parse_document(markdown, sizeof(markdown) - 1, CMARK_OPT_DEFAULT);
     char *xml = cmark_render_xml(doc, CMARK_OPT_DEFAULT | CMARK_OPT_SOURCEPOS);
     STR_EQ(runner, xml, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n"
                         "<!DOCTYPE document SYSTEM \"CommonMark.dtd\">\n"
-                        "<document sourcepos=\"1:1-2:6\" xmlns=\"http://commonmark.org/xml/1.0\">\n"
+                        "<document sourcepos=\"1:1-4:23\" xmlns=\"http://commonmark.org/xml/1.0\">\n"
                         "  <paragraph sourcepos=\"1:1-2:6\">\n"
                         "    <emph sourcepos=\"1:1-1:7\">\n"
                         "      <text sourcepos=\"1:2-1:6\" xml:space=\"preserve\">first</text>\n"
@@ -983,6 +1094,11 @@ static void source_pos_inlines(test_batch_runner *runner) {
                         "    <softbreak />\n"
                         "    <text sourcepos=\"2:1-2:6\" xml:space=\"preserve\">second</text>\n"
                         "  </paragraph>\n"
+                        "  <paragraph sourcepos=\"4:4-4:23\">\n"
+                        "    <link sourcepos=\"4:4-4:23\" destination=\"http://example.com\" title=\"\">\n"
+                        "      <text sourcepos=\"4:5-4:22\" xml:space=\"preserve\">http://example.com</text>\n"
+                        "    </link>\n"
+                        "  </paragraph>\n"
                         "</document>\n",
                         "sourcepos are as expected");
     free(xml);
@@ -1010,6 +1126,26 @@ static void source_pos_inlines(test_batch_runner *runner) {
     free(xml);
     cmark_node_free(doc);
   }
+  {
+    static const char markdown[] =
+      "` It is one backtick\n"
+      "`` They are two backticks\n";
+
+    cmark_node *doc = cmark_parse_document(markdown, sizeof(markdown) - 1, CMARK_OPT_DEFAULT);
+    char *xml = cmark_render_xml(doc, CMARK_OPT_DEFAULT | CMARK_OPT_SOURCEPOS);
+    STR_EQ(runner, xml, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n"
+                        "<!DOCTYPE document SYSTEM \"CommonMark.dtd\">\n"
+                        "<document sourcepos=\"1:1-2:25\" xmlns=\"http://commonmark.org/xml/1.0\">\n"
+                        "  <paragraph sourcepos=\"1:1-2:25\">\n"
+                        "    <text sourcepos=\"1:1-1:20\" xml:space=\"preserve\">` It is one backtick</text>\n"
+                        "    <softbreak />\n"
+                        "    <text sourcepos=\"2:1-2:25\" xml:space=\"preserve\">`` They are two backticks</text>\n"
+                        "  </paragraph>\n"
+                        "</document>\n",
+                        "sourcepos are as expected");
+    free(xml);
+    cmark_node_free(doc);
+  }
 }
 
 static void ref_source_pos(test_batch_runner *runner) {
@@ -1036,6 +1172,405 @@ static void ref_source_pos(test_batch_runner *runner) {
   cmark_node_free(doc);
 }
 
+static void inline_only_opt(test_batch_runner *runner) {
+  static const char markdown[] =
+    "# My heading\n"
+    "> My block quote\n\n"
+    "- List item\n\n"
+    "[link](https://github.com)\n";
+  
+  cmark_node *doc = cmark_parse_document(markdown, sizeof(markdown) - 1, CMARK_OPT_INLINE_ONLY);
+  char *html = cmark_render_html(doc, CMARK_OPT_DEFAULT, 0);
+  STR_EQ(runner, html, "<p># My heading\n"
+         "&gt; My block quote\n"
+         "\n"
+         "- List item\n"
+         "\n"
+         "<a href=\"https://github.com\">link</a>\n"
+         "</p>\n", "html is as expected");
+  free(html);
+  cmark_node_free(doc);
+}
+
+static void check_markdown_plaintext(test_batch_runner *runner, const char *markdown) {
+  cmark_node *doc = cmark_parse_document(markdown, strlen(markdown), CMARK_OPT_PRESERVE_WHITESPACE);
+  cmark_node *pg = cmark_node_first_child(doc);
+  INT_EQ(runner, cmark_node_get_type(pg), CMARK_NODE_PARAGRAPH, "markdown '%s' did not produce a paragraph node", markdown);
+  cmark_node *textNode = cmark_node_first_child(pg);
+  INT_EQ(runner, cmark_node_get_type(textNode), CMARK_NODE_TEXT, "markdown '%s' did not produce a text node inside the paragraph node", markdown);
+  const char *text = cmark_node_get_literal(textNode);
+  OK(runner, (text != NULL), "Text literal for '%s' was null", markdown);
+  if (text) {
+    STR_EQ(runner, text, markdown, "markdown '%s' resulted in '%s'", markdown, text);
+  } else {
+    SKIP(runner, 1);
+  }
+  cmark_node_free(doc);
+}
+
+static void preserve_whitespace_opt(test_batch_runner *runner) {
+  check_markdown_plaintext(runner, " ");
+  check_markdown_plaintext(runner, "    ");
+  check_markdown_plaintext(runner, "hello");
+  check_markdown_plaintext(runner, "hello ");
+  check_markdown_plaintext(runner, " hello");
+  check_markdown_plaintext(runner, "    hello");
+  check_markdown_plaintext(runner, "hello    ");
+  check_markdown_plaintext(runner, "hel\nlo");
+  check_markdown_plaintext(runner, "hel\n\nlo");
+  check_markdown_plaintext(runner, "hel\nworld\nlo");
+  check_markdown_plaintext(runner, " hel \n world \n lo ");
+  check_markdown_plaintext(runner, "  hello \n  \n world  ");
+  check_markdown_plaintext(runner, "\n");
+  check_markdown_plaintext(runner, "\n\n\n");
+  check_markdown_plaintext(runner, "\nHello");
+  check_markdown_plaintext(runner, "Hello\n");
+  check_markdown_plaintext(runner, "\nHello\n");
+}
+
+static void check_markdown_attributes_node(test_batch_runner *runner, const char *markdown, cmark_node_type expectedType, const char *expectedAttributes) {
+  cmark_node *doc = cmark_parse_document(markdown, strlen(markdown), CMARK_OPT_DEFAULT);
+  cmark_node *pg = cmark_node_first_child(doc);
+  INT_EQ(runner, cmark_node_get_type(pg), CMARK_NODE_PARAGRAPH, "markdown '%s' did not produce a paragraph node", markdown);
+  cmark_node *attributeNode = cmark_node_first_child(pg);
+  cmark_node_type nodeType = cmark_node_get_type(attributeNode);
+  INT_EQ(runner, nodeType, expectedType, "markdown '%s' did not produce the correct node type: got %d, expecting %d", markdown, nodeType, expectedType);
+  const char *attributeContent = cmark_node_get_attributes(attributeNode);
+  if (attributeContent == NULL) {
+    OK(runner, expectedAttributes == NULL, "markdown '%s' produced an unexpected NULL attribute", markdown);
+  } else if (expectedAttributes == NULL) {
+    OK(runner, attributeContent == NULL, "markdown '%s' produced an unexpected NULL attribute", markdown);
+  } else {
+    STR_EQ(runner, attributeContent, expectedAttributes, "markdown '%s' did not produce the correct attributes: got %s, expecting: %s", markdown, attributeContent, expectedAttributes);
+  }
+
+  cmark_node_free(doc);
+}
+
+static void verify_custom_attributes_node(test_batch_runner *runner) {
+  // Should produce a TEXT node since there's no `()` to signify attributes
+  check_markdown_attributes_node(runner, "^[]", CMARK_NODE_TEXT, NULL);
+  check_markdown_attributes_node(runner, "^[](", CMARK_NODE_TEXT, NULL);
+  check_markdown_attributes_node(runner, "^[])", CMARK_NODE_TEXT, NULL);
+  check_markdown_attributes_node(runner, "^[])(", CMARK_NODE_TEXT, NULL);
+  // Should produce an ATTRIBUTE node with no attributes
+  check_markdown_attributes_node(runner, "^[]()", CMARK_NODE_ATTRIBUTE, "");
+  // Should produce an ATTRIBUTE node with attributes
+  check_markdown_attributes_node(runner, "^[](rainbow: 'extreme')", CMARK_NODE_ATTRIBUTE, "rainbow: 'extreme'");
+}
+
+static cmark_node* parse_custom_attributues_footnote(test_batch_runner *runner, const char *markdown, cmark_node **retdoc) {
+  cmark_node *doc = cmark_parse_document(markdown, strlen(markdown), CMARK_OPT_DEFAULT);
+  cmark_node *pg = cmark_node_first_child(doc);
+  INT_EQ(runner, cmark_node_get_type(pg), CMARK_NODE_PARAGRAPH, "markdown '%s' did not produce a paragraph node", markdown);
+  *retdoc = doc;
+  return cmark_node_first_child(pg);
+}
+
+static void verify_custom_attributes_footnote_basic(test_batch_runner *runner) {
+  static const char markdown[] =
+    "^[caffe][1]\n"
+    "\n"
+    "^[1]: rainbow: 'extreme', colors: { r: 255, g: 0, b: 0 }, corgicopter: true";
+  cmark_node *doc;
+  cmark_node *attributeNode = parse_custom_attributues_footnote(runner, markdown, &doc);
+  INT_EQ(runner, cmark_node_get_type(attributeNode), CMARK_NODE_ATTRIBUTE, "markdown '%s' did not produce an attribute node", markdown);
+  STR_EQ(runner, cmark_node_get_attributes(attributeNode),
+         "rainbow: 'extreme', colors: { r: 255, g: 0, b: 0 }, corgicopter: true",
+         "markdown '%s' did not produce the right attribute in footnote", markdown);
+
+  cmark_node_free(doc);
+}
+
+static void verify_custom_attributes_footnote_multiple_footnotes(test_batch_runner *runner) {
+  static const char markdown[] =
+    "^[food][1] and ^[drinks][2]\n"
+    "\n"
+    "^[1]: rainbow: 'fun'\n"
+    "^[2]: magic: 42";
+  cmark_node *doc;
+  cmark_node *attributeNode1 = parse_custom_attributues_footnote(runner, markdown, &doc);
+  INT_EQ(runner, cmark_node_get_type(attributeNode1), CMARK_NODE_ATTRIBUTE, "markdown '%s' did not produce an attribute node", markdown);
+  STR_EQ(runner, cmark_node_get_attributes(attributeNode1), "rainbow: 'fun'", "markdown '%s' did not produce the right attribute in footnote", markdown);
+  cmark_node *textNode = cmark_node_next(attributeNode1); // "and"
+  cmark_node *attributeNode2 = cmark_node_next(textNode);
+  INT_EQ(runner, cmark_node_get_type(attributeNode2), CMARK_NODE_ATTRIBUTE, "markdown '%s' did not produce an attribute node", markdown);
+  STR_EQ(runner, cmark_node_get_attributes(attributeNode2), "magic: 42", "markdown '%s' did not produce the right attribute in footnote", markdown);
+
+  cmark_node_free(doc);
+}
+
+static void verify_custom_attributes_footnote_reuse(test_batch_runner *runner) {
+  static const char markdown[] =
+    "^[pizza][1], ^[sandwich][2], ^[ice cream][2], and ^[salad][1]\n"
+    "\n"
+    "^[1]: has_tomato: true\n"
+    "^[2]: price: 12";
+  cmark_node *doc;
+  cmark_node *pizzaNode = parse_custom_attributues_footnote(runner, markdown, &doc);
+  INT_EQ(runner, cmark_node_get_type(pizzaNode), CMARK_NODE_ATTRIBUTE, "markdown '%s' did not produce an attribute node", markdown);
+  STR_EQ(runner, cmark_node_get_attributes(pizzaNode), "has_tomato: true", "markdown '%s' did not produce the right attribute in footnote", markdown);
+  cmark_node *sandwichNode = cmark_node_next(cmark_node_next(pizzaNode));
+  INT_EQ(runner, cmark_node_get_type(sandwichNode), CMARK_NODE_ATTRIBUTE, "markdown '%s' did not produce an attribute node", markdown);
+  STR_EQ(runner, cmark_node_get_attributes(sandwichNode), "price: 12", "markdown '%s' did not produce the right attribute in footnote", markdown);
+  cmark_node *icecreamNode = cmark_node_next(cmark_node_next(sandwichNode));
+  INT_EQ(runner, cmark_node_get_type(icecreamNode), CMARK_NODE_ATTRIBUTE, "markdown '%s' did not produce an attribute node", markdown);
+  STR_EQ(runner, cmark_node_get_attributes(icecreamNode), "price: 12", "markdown '%s' did not produce the right attribute in footnote", markdown);
+  cmark_node *saladNode = cmark_node_next(cmark_node_next(icecreamNode));
+  INT_EQ(runner, cmark_node_get_type(saladNode), CMARK_NODE_ATTRIBUTE, "markdown '%s' did not produce an attribute node", markdown);
+  STR_EQ(runner, cmark_node_get_attributes(saladNode), "has_tomato: true", "markdown '%s' did not produce the right attribute in footnote", markdown);
+
+  cmark_node_free(doc);
+}
+
+static void verify_custom_attributes_footnote_mixed_content(test_batch_runner *runner) {
+  static const char markdown[] =
+    "^[attribute1][1], [a link][2], ^[attribute2][3], ^[attribute3][1]\n"
+    "\n"
+    "^[1]: rainbow: 'fun'\n"
+    "[2]: https://www.example.com\n"
+    "Lorem ipsum\n"
+    "\n"
+    "^[3]: universe: 42\n";
+  cmark_node *doc;
+  cmark_node *attributeNode1 = parse_custom_attributues_footnote(runner, markdown, &doc);
+  INT_EQ(runner, cmark_node_get_type(attributeNode1), CMARK_NODE_ATTRIBUTE, "markdown '%s' did not produce an attribute node", markdown);
+  STR_EQ(runner, cmark_node_get_attributes(attributeNode1), "rainbow: 'fun'", "markdown '%s' did not produce the right attribute in footnote", markdown);
+  cmark_node *linkNode = cmark_node_next(cmark_node_next(attributeNode1));
+  INT_EQ(runner, cmark_node_get_type(linkNode), CMARK_NODE_LINK, "markdown '%s' did not produce an link node", markdown);
+  STR_EQ(runner, cmark_node_get_url(linkNode), "https://www.example.com", "markdown '%s' did not produce the right link in footnote", markdown);
+  cmark_node *attributeNode2 = cmark_node_next(cmark_node_next(linkNode));
+  INT_EQ(runner, cmark_node_get_type(attributeNode2), CMARK_NODE_ATTRIBUTE, "markdown '%s' did not produce an attribute node", markdown);
+  STR_EQ(runner, cmark_node_get_attributes(attributeNode2), "universe: 42", "markdown '%s' did not produce the right attribute in footnote", markdown);
+  cmark_node *attributeNode3 = cmark_node_next(cmark_node_next(attributeNode2));
+  INT_EQ(runner, cmark_node_get_type(attributeNode3), CMARK_NODE_ATTRIBUTE, "markdown '%s' did not produce an attribute node", markdown);
+  STR_EQ(runner, cmark_node_get_attributes(attributeNode3), "rainbow: 'fun'", "markdown '%s' did not produce the right attribute in footnote", markdown);
+
+  cmark_node_free(doc);
+}
+
+static void verify_custom_attributes_node_with_footnote(test_batch_runner *runner) {
+  verify_custom_attributes_footnote_basic(runner);
+  verify_custom_attributes_footnote_multiple_footnotes(runner);
+  verify_custom_attributes_footnote_reuse(runner);
+  verify_custom_attributes_footnote_mixed_content(runner);
+}
+
+typedef void (*reentrant_call_func) (void);
+
+static cmark_node *reentrant_parse_inline_ext(cmark_syntax_extension *self, cmark_parser *parser,
+                                              cmark_node *parent, unsigned char character,
+                                              cmark_inline_parser *inline_parser) {
+  void *priv = cmark_syntax_extension_get_private(self);
+  if (priv) {
+    reentrant_call_func func = (reentrant_call_func)priv;
+    func();
+    cmark_syntax_extension_set_private(self, NULL, NULL);
+  }
+
+  return NULL;
+}
+
+static void run_inner_parser() {
+  cmark_parser *parser = cmark_parser_new(CMARK_OPT_DEFAULT);
+  cmark_parser_attach_syntax_extension(parser, cmark_find_syntax_extension("strikethrough"));
+
+  static const char markdown[] = "this is the ~~outer~~ inner document";
+  cmark_parser_feed(parser, markdown, sizeof(markdown) - 1);
+
+  cmark_node *doc = cmark_parser_finish(parser);
+  cmark_node_free(doc);
+  cmark_parser_free(parser);
+}
+
+static void parser_interrupt(test_batch_runner *runner) {
+  cmark_gfm_core_extensions_ensure_registered();
+
+  cmark_syntax_extension *my_ext = cmark_syntax_extension_new("interrupt");
+  cmark_syntax_extension_set_private(my_ext, run_inner_parser, NULL);
+  cmark_syntax_extension_set_match_inline_func(my_ext, reentrant_parse_inline_ext);
+
+  cmark_parser *parser = cmark_parser_new(CMARK_OPT_DEFAULT);
+  cmark_parser_attach_syntax_extension(parser, cmark_find_syntax_extension("strikethrough"));
+  cmark_parser_attach_syntax_extension(parser, my_ext);
+
+  static const char markdown[] = "this is the ~~inner~~ outer document";
+  cmark_parser_feed(parser, markdown, sizeof(markdown) - 1);
+
+  cmark_node *doc = cmark_parser_finish(parser);
+  char *xml = cmark_render_xml(doc, CMARK_OPT_DEFAULT);
+  STR_EQ(runner, xml, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n"
+         "<!DOCTYPE document SYSTEM \"CommonMark.dtd\">\n"
+         "<document xmlns=\"http://commonmark.org/xml/1.0\">\n"
+         "  <paragraph>\n"
+         "    <text xml:space=\"preserve\">this is the </text>\n"
+         "    <strikethrough>\n"
+         "      <text xml:space=\"preserve\">inner</text>\n"
+         "    </strikethrough>\n"
+         "    <text xml:space=\"preserve\"> outer document</text>\n"
+         "  </paragraph>\n"
+         "</document>\n", "interrupting the parser should still allow extensions");
+
+  free(xml);
+  cmark_node_free(doc);
+  cmark_parser_free(parser);
+  cmark_syntax_extension_free(cmark_get_default_mem_allocator(), my_ext);
+}
+
+static void compare_table_spans_html(test_batch_runner *runner, const char *markdown, bool use_ditto,
+                                     const char *expected_html, const char *msg) {
+  int options = CMARK_OPT_TABLE_SPANS;
+  if (use_ditto)
+    options |= CMARK_OPT_TABLE_ROWSPAN_DITTO;
+  cmark_parser *parser = cmark_parser_new(options);
+  cmark_parser_attach_syntax_extension(parser, cmark_find_syntax_extension("table"));
+
+  cmark_parser_feed(parser, markdown, strlen(markdown));
+
+  cmark_node *doc = cmark_parser_finish(parser);
+  char *html = cmark_render_html(doc, options, NULL);
+  STR_EQ(runner, html, expected_html, msg);
+
+  free(html);
+  cmark_node_free(doc);
+  cmark_parser_free(parser);
+}
+
+static void table_spans(test_batch_runner *runner) {
+  {
+    static const char markdown[] =
+      "| one | two |\n"
+      "| --- | --- |\n"
+      "| hello    ||\n";
+    static const char html[] =
+      "<table>\n"
+      "<thead>\n"
+      "<tr>\n"
+      "<th>one</th>\n"
+      "<th>two</th>\n"
+      "</tr>\n"
+      "</thead>\n"
+      "<tbody>\n"
+      "<tr>\n"
+      "<td colspan=\"2\">hello</td>\n"
+      "</tr>\n"
+      "</tbody>\n"
+      "</table>\n";
+    compare_table_spans_html(runner, markdown, false, html,
+                             "table colspans should work when enabled");
+  }
+  {
+    static const char markdown[] =
+      "| one | two   |\n"
+      "| --- | ----- |\n"
+      "| big | small |\n"
+      "| ^   | small |\n";
+    static const char html[] =
+      "<table>\n"
+      "<thead>\n"
+      "<tr>\n"
+      "<th>one</th>\n"
+      "<th>two</th>\n"
+      "</tr>\n"
+      "</thead>\n"
+      "<tbody>\n"
+      "<tr>\n"
+      "<td rowspan=\"2\">big</td>\n"
+      "<td>small</td>\n"
+      "</tr>\n"
+      "<tr>\n"
+      "<td>small</td>\n"
+      "</tr>\n"
+      "</tbody>\n"
+      "</table>\n";
+    compare_table_spans_html(runner, markdown, false, html,
+                             "table rowspans should work when enabled");
+  }
+  {
+    static const char markdown[] =
+      "| one | two   |\n"
+      "| --- | ----- |\n"
+      "| big | small |\n"
+      "| \"   | small |\n";
+    static const char html[] =
+      "<table>\n"
+      "<thead>\n"
+      "<tr>\n"
+      "<th>one</th>\n"
+      "<th>two</th>\n"
+      "</tr>\n"
+      "</thead>\n"
+      "<tbody>\n"
+      "<tr>\n"
+      "<td rowspan=\"2\">big</td>\n"
+      "<td>small</td>\n"
+      "</tr>\n"
+      "<tr>\n"
+      "<td>small</td>\n"
+      "</tr>\n"
+      "</tbody>\n"
+      "</table>\n";
+    compare_table_spans_html(runner, markdown, true, html,
+                             "rowspan ditto marks should work when enabled");
+  }
+  {
+    static const char markdown[] =
+      "| one | two | three |\n"
+      "| --- | --- | ----- |\n"
+      "| big      || small |\n"
+      "| ^        || small |\n";
+    static const char html[] =
+      "<table>\n"
+      "<thead>\n"
+      "<tr>\n"
+      "<th>one</th>\n"
+      "<th>two</th>\n"
+      "<th>three</th>\n"
+      "</tr>\n"
+      "</thead>\n"
+      "<tbody>\n"
+      "<tr>\n"
+      "<td colspan=\"2\" rowspan=\"2\">big</td>\n"
+      "<td>small</td>\n"
+      "</tr>\n"
+      "<tr>\n"
+      "<td>small</td>\n"
+      "</tr>\n"
+      "</tbody>\n"
+      "</table>\n";
+    compare_table_spans_html(runner, markdown, false, html,
+                             "colspan and rowspan should combine sensibly");
+  }
+  {
+    static const char markdown[] =
+      "| one | two | three |\n"
+      "| --- | --- | ----- |\n"
+      "| big      || small |\n"
+      "| \"        || small |\n";
+    static const char html[] =
+      "<table>\n"
+      "<thead>\n"
+      "<tr>\n"
+      "<th>one</th>\n"
+      "<th>two</th>\n"
+      "<th>three</th>\n"
+      "</tr>\n"
+      "</thead>\n"
+      "<tbody>\n"
+      "<tr>\n"
+      "<td colspan=\"2\" rowspan=\"2\">big</td>\n"
+      "<td>small</td>\n"
+      "</tr>\n"
+      "<tr>\n"
+      "<td>small</td>\n"
+      "</tr>\n"
+      "</tbody>\n"
+      "</table>\n";
+    compare_table_spans_html(runner, markdown, true, html,
+                             "colspan and rowspan should combine when ditto marks are enabled");
+  }
+}
+
 int main() {
   int retval;
   test_batch_runner *runner = test_batch_runner_new();
@@ -1043,7 +1578,6 @@ int main() {
   version(runner);
   constructor(runner);
   accessors(runner);
-  free_parent(runner);
   node_check(runner);
   iterator(runner);
   iterator_delete(runner);
@@ -1056,15 +1590,23 @@ int main() {
   render_man(runner);
   render_latex(runner);
   render_commonmark(runner);
+  render_plaintext(runner);
   utf8(runner);
   line_endings(runner);
   numeric_entities(runner);
   test_cplusplus(runner);
   test_safe(runner);
   test_feed_across_line_ending(runner);
+  test_pathological_regressions(runner);
   source_pos(runner);
   source_pos_inlines(runner);
   ref_source_pos(runner);
+  inline_only_opt(runner);
+  preserve_whitespace_opt(runner);
+  verify_custom_attributes_node(runner);
+  verify_custom_attributes_node_with_footnote(runner);
+  parser_interrupt(runner);
+  table_spans(runner);
 
   test_print_summary(runner);
   retval = test_ok(runner) ? 0 : 1;
