@@ -23,6 +23,7 @@
 #include "llbuild/BuildSystem/Tool.h"
 #include "llbuild/Core/BuildEngine.h"
 #include "llbuild/Core/DependencyInfoParser.h"
+#include "llbuild/Core/MakefileDepsParser.h"
 
 #include "BuildKey-C-API-Private.h"
 #include "BuildValue-C-API-Private.h"
@@ -31,11 +32,13 @@
 #include "llvm/ADT/SmallString.h"
 #include "llvm/Support/SourceMgr.h"
 #include "llvm/Support/raw_ostream.h"
+#include "llvm/Support/Path.h"
 
 #include <atomic>
 #include <cassert>
 #include <future>
 #include <memory>
+#include <utility>
 
 using namespace llbuild;
 using namespace llbuild::basic;
@@ -468,6 +471,39 @@ public:
     uint64_t count() { return rules.size(); }
   };
 
+  static llb_rule_run_reason_t
+  convertRunReason(core::Rule::RunReason reason) {
+    switch (reason) {
+      case core::Rule::RunReason::NeverBuilt:
+        return llb_rule_run_reason_never_built;
+      case core::Rule::RunReason::SignatureChanged:
+        return llb_rule_run_reason_signature_changed;
+      case core::Rule::RunReason::InvalidValue:
+        return llb_rule_run_reason_invalid_value;
+      case core::Rule::RunReason::InputRebuilt:
+        return llb_rule_run_reason_input_rebuilt;
+      case core::Rule::RunReason::Forced:
+        return llb_rule_run_reason_forced;
+    }
+    assert(0 && "unknown reason");
+    return llb_rule_run_reason_invalid_value;
+  }
+
+  virtual void determinedRuleNeedsToRun(core::Rule* ruleNeedingToRun, core::Rule::RunReason reason, core::Rule* inputRule) override {
+    if (cAPIDelegate.determined_rule_needs_to_run) {
+      auto key = BuildKey::fromData(ruleNeedingToRun->key);
+      auto needsToRun = (llb_build_key_t *)new CAPIBuildKey(key);
+      llb_build_key_t * input;
+      if (inputRule) {
+        auto inputKey = BuildKey::fromData(inputRule->key);
+        input = (llb_build_key_t *)new CAPIBuildKey(inputKey);
+      } else {
+        input = nullptr;
+      }
+      cAPIDelegate.determined_rule_needs_to_run(cAPIDelegate.context, needsToRun, convertRunReason(reason), input);
+      llb_build_key_destroy(needsToRun);
+    }
+  }
 
   virtual void cycleDetected(const std::vector<core::Rule*>& items) override {
     CAPIRulesVector rules(items);
@@ -665,12 +701,21 @@ class CAPIExternalCommand : public ExternalCommand {
   // that the delegates are const and we just carry the context pointer around.
   llb_buildsystem_external_command_delegate_t cAPIDelegate;
   
-  /// The path to the dependency output file, if used.
-  std::string depsPath;
+  /// The paths to the dependency output files, if used.
+  SmallVector<std::string, 1> depsPaths{};
 
-  bool processDiscoveredDependencies(BuildSystem& system,
-                                     core::TaskInterface ti,
-                                     QueueJobContext* context) {
+  /// The working directory used to resolve relative paths in dependency files.
+  std::string workingDirectory;
+  
+  /// Format of the dependencies in `depsPaths`.
+  /// We default to `dependencyinfo` for legacy behavior since not all tasks specify the format.
+  llb_buildsystem_dependency_data_format_t depsFormat =
+          llb_buildsystem_dependency_data_format_dependencyinfo;
+
+  bool processDependencyInfoDiscoveredDependencies(BuildSystem& system,
+                                                   core::TaskInterface ti,
+                                                   QueueJobContext* context,
+                                                   std::string depsPath) {
     // Read the dependencies file.
     auto input = system.getFileSystem().getFileContents(depsPath);
     if (!input) {
@@ -717,17 +762,140 @@ class CAPIExternalCommand : public ExternalCommand {
     core::DependencyInfoParser(input->getBuffer(), actions).parse();
     return actions.numErrors == 0;
   }
+  
+  bool processMakefileDiscoveredDependencies(BuildSystem& system,
+                                             core::TaskInterface ti,
+                                             QueueJobContext* context,
+                                             std::string depsPath) {
+    // Read the dependencies file.
+    auto input = system.getFileSystem().getFileContents(depsPath);
+    if (!input) {
+      system.getDelegate().commandHadError(this, "unable to open dependencies file (" + depsPath + ")");
+      return false;
+    }
+    
+    // Parse the output.
+    //
+    // We just ignore the rule, and add any dependency that we encounter in the
+    // file.
+    struct DepsActions : public core::MakefileDepsParser::ParseActions {
+      BuildSystem& system;
+      core::TaskInterface ti;
+      CAPIExternalCommand* command;
+      StringRef depsPath;
+      unsigned numErrors{0};
+      
+      DepsActions(BuildSystem& system,
+                  core::TaskInterface ti,
+                  CAPIExternalCommand* command, StringRef depsPath)
+      : system(system), ti(ti), command(command), depsPath(depsPath) {}
+      
+      virtual void error(StringRef message, uint64_t position) override {
+        system.getDelegate().commandHadError(command, "error reading dependency file '" + depsPath.str() + "': " + std::string(message));
+        ++numErrors;
+      }
+      
+      virtual void actOnRuleDependency(StringRef dependency,
+                                       StringRef unescapedWord) override {
+        if (llvm::sys::path::is_absolute(unescapedWord)) {
+          ti.discoveredDependency(BuildKey::makeNode(unescapedWord).toData());
+          system.getDelegate().commandFoundDiscoveredDependency(command, unescapedWord, DiscoveredDependencyKind::Input);
+          return;
+        }
 
-  virtual bool configureAttribute(const ConfigureContext& ctx, StringRef name,
-                                  StringRef value) override {
+        // Generate absolute path
+        //
+        // NOTE: This is making the assumption that relative paths coming in a
+        // dependency file are in relation to the explictly set working
+        // directory, or the current working directory when it has not been set.
+        SmallString<PATH_MAX> absPath = StringRef(command->workingDirectory);
+        llvm::sys::path::append(absPath, unescapedWord);
+        llvm::sys::fs::make_absolute(absPath);
+
+        ti.discoveredDependency(BuildKey::makeNode(absPath).toData());
+        system.getDelegate().commandFoundDiscoveredDependency(command, absPath, DiscoveredDependencyKind::Input);
+      }
+
+      virtual void actOnRuleStart(StringRef name,
+                                  const StringRef unescapedWord) override {}
+
+      virtual void actOnRuleEnd() override {}
+    };
+    
+    DepsActions actions(system, ti, this, depsPath);
+    core::MakefileDepsParser(input->getBuffer(), actions).parse();
+    return actions.numErrors == 0;
+  }
+  
+  bool configureAttributeNoContext(StringRef name, StringRef value) {
     if (name == "deps") {
-      depsPath = value;
+      depsPaths.clear();
+      depsPaths.emplace_back(value);
+    } else if (name == "deps-style") {
+      if (value == "unused") {
+        depsFormat = llb_buildsystem_dependency_data_format_unused;
+      } else if (value == "makefile") {
+        depsFormat = llb_buildsystem_dependency_data_format_makefile;
+      } else if (value == "dependency-info") {
+        depsFormat = llb_buildsystem_dependency_data_format_dependencyinfo;
+      } else {
+        return false;
+      }
+    } else if (name == "working-directory") {
+      // Ensure the working directory is absolute. This will make sure any
+      // relative directories are interpreted as relative to the CWD at the time
+      // the rule is defined.
+      SmallString<PATH_MAX> wd = value;
+      llvm::sys::fs::make_absolute(wd);
+      workingDirectory = StringRef(wd);
     } else {
-      return ExternalCommand::configureAttribute(ctx, name, value);
+      return false;
     }
     return true;
   }
+  
+  bool configureAttributeNoContext(StringRef name,
+                                  ArrayRef<StringRef> values) {
+    if (name == "deps") {
+      depsPaths.clear();
+      depsPaths.insert(depsPaths.begin(), values.begin(), values.end());
+    } else {
+      return false;
+    }
+    return true;
+  }
+  bool configureAttributeNoContext(StringRef name,
+      ArrayRef<std::pair<StringRef, StringRef>> values) {
+    return false;
+  }
 
+  virtual bool configureAttribute(const ConfigureContext& ctx, StringRef name,
+                                  StringRef value) override {
+    if (configureAttributeNoContext(name, value)) {
+      return true;
+    } else {
+      return ExternalCommand::configureAttribute(ctx, name, value);
+    }
+  }
+  
+  virtual bool configureAttribute(const ConfigureContext& ctx, StringRef name,
+                                  ArrayRef<StringRef> values) override {
+    if (configureAttributeNoContext(name, values)) {
+      return true;
+    } else {
+      return ExternalCommand::configureAttribute(ctx, name, values);
+    }
+  }
+  virtual bool configureAttribute(
+      const ConfigureContext& ctx, StringRef name,
+      ArrayRef<std::pair<StringRef, StringRef>> values) override {
+        if (configureAttributeNoContext(name, values)) {
+          return true;
+        } else {
+          return ExternalCommand::configureAttribute(ctx, name, values);
+        }
+  }
+  
   virtual void startExternalCommand(BuildSystem& system,
                                     core::TaskInterface ti) override {
     cAPIDelegate.start(cAPIDelegate.context,
@@ -760,20 +928,36 @@ class CAPIExternalCommand : public ExternalCommand {
                                       llvm::Optional<ProcessCompletionFn> completionFn) override {
     auto doneFn = [this, &system, ti, job_context, completionFn](ProcessStatus result) mutable {
       if (result != ProcessStatus::Succeeded) {
-        // If the command failed, there is no need to gather dependencies.
+        // If the command did not succeed, there is no need to gather dependencies.
         if (completionFn.hasValue())
-          completionFn.getValue()(ProcessStatus::Failed);
+          completionFn.getValue()(result);
         return;
       }
 
       // Otherwise, collect the discovered dependencies, if used.
-      if (!depsPath.empty()) {
-        if (!processDiscoveredDependencies(system, ti, job_context)) {
-          // If we were unable to process the dependencies output, report a
-          // failure.
-          if (completionFn.hasValue())
-            completionFn.getValue()(ProcessStatus::Failed);
-          return;
+      bool dependencyParsingResult = false;
+      if (!depsPaths.empty()) {
+        for (const auto& depsPath: depsPaths) {
+          switch (depsFormat) {
+            case llb_buildsystem_dependency_data_format_unused:
+              ti.delegate()->error("No dependency format specified for discovered dependency files.");
+              dependencyParsingResult = false;
+              break;
+            case llb_buildsystem_dependency_data_format_makefile:
+              dependencyParsingResult = processMakefileDiscoveredDependencies(system, ti, job_context, depsPath);
+              break;
+            case llb_buildsystem_dependency_data_format_dependencyinfo:
+              dependencyParsingResult = processDependencyInfoDiscoveredDependencies(system, ti, job_context, depsPath);
+              break;
+          }
+          
+          if (!dependencyParsingResult) {
+            // If we were unable to process the dependencies output, report a
+            // failure.
+            if (completionFn.hasValue())
+              completionFn.getValue()(ProcessStatus::Failed);
+            return;
+          }
         }
       }
 
@@ -805,13 +989,28 @@ class CAPIExternalCommand : public ExternalCommand {
     }
 
     assert(cAPIDelegate.execute_command != nullptr);
-    bool success = cAPIDelegate.execute_command(
+    llb_buildsystem_command_result_t result = cAPIDelegate.execute_command(
       cAPIDelegate.context,
       (llb_buildsystem_command_t*)this,
       (llb_buildsystem_interface_t*)&system,
       *reinterpret_cast<llb_task_interface_t*>(&ti),
       (llb_buildsystem_queue_job_context_t*)job_context);
-    doneFn(success ? ProcessStatus::Succeeded : ProcessStatus::Failed);
+    ProcessStatus status;
+    switch (result) {
+      case llb_buildsystem_command_result_succeeded:
+        status = ProcessStatus::Succeeded;
+        break;
+      case llb_buildsystem_command_result_failed:
+        status = ProcessStatus::Failed;
+        break;
+      case llb_buildsystem_command_result_cancelled:
+        status = ProcessStatus::Cancelled;
+        break;
+      case llb_buildsystem_command_result_skipped:
+        status = ProcessStatus::Skipped;
+        break;
+    }
+    doneFn(status);
   }
 
   BuildValue computeCommandResult(BuildSystem& system, core::TaskInterface ti) override {
@@ -837,7 +1036,29 @@ class CAPIExternalCommand : public ExternalCommand {
 public:
   CAPIExternalCommand(StringRef name,
                       llb_buildsystem_external_command_delegate_t delegate)
-      : ExternalCommand(name), cAPIDelegate(delegate) {}
+      : ExternalCommand(name), cAPIDelegate(delegate) {
+        if (cAPIDelegate.configure) {
+          auto single = [](llb_buildsystem_command_t* ctx, llb_data_t name, llb_data_t data) {
+            ((CAPIExternalCommand *)ctx)->configureAttributeNoContext(StringRef((const char*)name.data, name.length), StringRef((const char*)data.data, data.length));
+          };
+          auto collection = [](llb_buildsystem_command_t* ctx, llb_data_t name, llb_data_t* datas, size_t size) {
+            std::vector<StringRef> values;
+            for (size_t i = 0; i < size; i++) {
+              values.push_back(StringRef((const char*)datas[i].data, datas[i].length));
+            }
+            ((CAPIExternalCommand *)ctx)->configureAttributeNoContext(StringRef((const char*)name.data, name.length), ArrayRef<StringRef>(values));
+          };
+          auto map = [](llb_buildsystem_command_t* ctx, llb_data_t name, llb_data_t* keys, llb_data_t* values, size_t size) {
+            std::vector<std::pair<StringRef, StringRef>> attributeValues;
+            for (size_t i = 0; i < size; i++) {
+              attributeValues.push_back(std::pair<StringRef, StringRef>(StringRef((const char*)keys[i].data, keys[i].length), StringRef((const char*)values[i].data, values[i].length)));
+            }
+            ((CAPIExternalCommand *)ctx)->configureAttributeNoContext(StringRef((const char*)name.data, name.length), ArrayRef<std::pair<StringRef, StringRef>>(attributeValues));
+          };
+          
+          cAPIDelegate.configure(cAPIDelegate.context, (llb_buildsystem_command_t*)this, single, collection, map);
+        }
+      }
 
 
   virtual void getShortDescription(SmallVectorImpl<char> &result) const override {
@@ -994,6 +1215,11 @@ void llb_buildsystem_command_interface_task_needs_input(llb_task_interface_t ti,
   coreti->request(((CAPIBuildKey *)key)->getInternalBuildKey().toData(), inputID);
 }
 
+void llb_buildsystem_command_interface_task_needs_single_use_input(llb_task_interface_t ti, llb_build_key_t* key, uintptr_t inputID) {
+  auto coreti = reinterpret_cast<core::TaskInterface*>(&ti);
+  coreti->requestSingleUse(((CAPIBuildKey *)key)->getInternalBuildKey().toData(), inputID);
+}
+
 llb_build_value_file_info_t llb_buildsystem_command_interface_get_file_info(llb_buildsystem_interface_t* bi_p, const char* path) {
   auto bi = (BuildSystem*)bi_p;
   return llbuild::capi::convertFileInfo(bi->getFileSystem().getFileInfo(path));
@@ -1061,9 +1287,12 @@ bool llb_buildsystem_command_interface_spawn(llb_task_interface_t ti, llb_builds
     }
 
   };
-  
-  coreti->spawn((QueueJobContext*)job_context, ArrayRef<StringRef>(arguments), ArrayRef<std::pair<StringRef, StringRef>>(environment), {true, false, StringRef((const char*)working_dir->data, working_dir->length)}, {commandCompletionFn}, new ForwardingProcessDelegate(delegate));
-  
+
+  auto forwardingDelegate = new ForwardingProcessDelegate(delegate);
+  coreti->spawn((QueueJobContext*)job_context, ArrayRef<StringRef>(arguments), ArrayRef<std::pair<StringRef, StringRef>>(environment), {true, false, StringRef((const char*)working_dir->data, working_dir->length)}, {commandCompletionFn}, forwardingDelegate);
+
+  delete forwardingDelegate;
+
   return result.get().status == ProcessStatus::Succeeded;  
 }
 
