@@ -125,6 +125,17 @@ typedef enum LLBUILD_ENUM_ATTRIBUTES {
   llb_buildsystem_discovered_dependency_kind_output = 2,
 } llb_buildsystem_discovered_dependency_kind_t LLBUILD_SWIFT_NAME(DiscoveredDependencyKind);
 
+typedef enum LLBUILD_ENUM_ATTRIBUTES {
+  /// No discovered dependencies are in use.
+  llb_buildsystem_dependency_data_format_unused = 0,
+  /// "Makefile" style dependencies in the form typically generated by C
+  /// compilers, wherein the dependencies of the first target are treated as
+  /// dependencies of the command.
+  llb_buildsystem_dependency_data_format_makefile = 1,
+  /// Darwin's DependencyInfo format.
+  llb_buildsystem_dependency_data_format_dependencyinfo = 2,
+} llb_buildsystem_dependency_data_format_t LLBUILD_SWIFT_NAME(DependencyDataFormat);
+
 /// Extended result of a command execution
 typedef struct llb_buildsystem_command_extended_result_t_ {
   llb_buildsystem_command_result_t result;  /// Result of command execution
@@ -192,6 +203,24 @@ typedef enum LLBUILD_ENUM_ATTRIBUTES {
     llb_quality_of_service_unspecified = 4
 } llb_quality_of_service_t LLBUILD_SWIFT_NAME(QualityOfService);
 
+/// Rule run reasons.
+typedef enum LLBUILD_ENUM_ATTRIBUTES {
+    /// The rule has never been built.
+    llb_rule_run_reason_never_built LLBUILD_SWIFT_NAME(neverBuilt) = 0,
+
+    /// The rule's signature has changed.
+    llb_rule_run_reason_signature_changed LLBUILD_SWIFT_NAME(signatureChanged) = 1,
+
+    /// The rule has an invalid value.
+    llb_rule_run_reason_invalid_value LLBUILD_SWIFT_NAME(invalidValue) = 2,
+
+    /// An input of the rule has been rebuilt.
+    llb_rule_run_reason_input_rebuilt LLBUILD_SWIFT_NAME(inputRebuilt) = 3,
+
+    /// The rule has never been built.
+    llb_rule_run_reason_forced LLBUILD_SWIFT_NAME(forced) = 4,
+} llb_rule_run_reason_t LLBUILD_SWIFT_NAME(RuleRunReason);
+
 /// Invocation parameters for a build system.
 typedef struct llb_buildsystem_invocation_t_ llb_buildsystem_invocation_t;
 struct llb_buildsystem_invocation_t_ {
@@ -469,6 +498,15 @@ typedef struct llb_buildsystem_delegate_t_ {
                                    llb_buildsystem_process_t* process,
                                    const llb_buildsystem_command_extended_result_t* result);
 
+  /// Called when it's been determined that a rule needs to run.
+  ///
+  ///  Xparam rule_needing_to_run The rule that needs to run.
+  ///
+  ///  Xparam reason Describes why the rule needs to run. For example, because it has never run or because an input was rebuilt.
+  ///
+  ///  Xparam input_rule If `reason` is `InputRebuilt`, the rule for the rebuilt input, else  `nullptr`.
+  void (*determined_rule_needs_to_run)(void* context, llb_build_key_t* rule_needing_to_run, llb_rule_run_reason_t reason, llb_build_key_t* input_rule);
+
   /// Called when a cycle is detected by the build engine and it cannot make
   /// forward progress.
   ///
@@ -633,6 +671,25 @@ typedef struct llb_buildsystem_spawn_delegate_t_ {
 typedef struct llb_buildsystem_external_command_delegate_t_ {
   /// User context pointer.
   void* context;
+  
+  /// Called once on initialization to configure the external command.
+  /// Supports the same properties as in the manifest and allows feature parity between
+  /// upfront planned and dynamic tasks.
+  /// There are different callbacks for different values types.
+  ///
+  /// `single` for values that are represented by a single string represented value
+  /// `collection` for an array of string represented values
+  /// `map` for a map like array of string represented key value pairs
+  ///
+  /// Each callback can be called multiple times. Subsequent calls with the same name
+  /// might override previously set values, including values set in the manifest.
+  void (*configure)(void* context, llb_buildsystem_command_t* command,
+                    void (*single)(llb_buildsystem_command_t* command, llb_data_t name,
+                                   llb_data_t value),
+                    void (*collection)(llb_buildsystem_command_t* command, llb_data_t name,
+                                       llb_data_t* values, size_t value_size),
+                    void (*map)(llb_buildsystem_command_t* command, llb_data_t name,
+                                llb_data_t* keys, llb_data_t* values, size_t keys_values_size));
 
   /// Called to get a signature which represents the internal state of the
   /// command which is not tracked by any other attribute visible to the build
@@ -675,7 +732,7 @@ typedef struct llb_buildsystem_external_command_delegate_t_ {
   /// then try calling the legacy `execute_command` variant if it is defined.
   ///
   /// The C API takes ownership of the value returned by `execute_command_ex`.
-  bool (*execute_command)(void* context,
+  llb_buildsystem_command_result_t (*execute_command)(void* context,
                           llb_buildsystem_command_t* command,
                           llb_buildsystem_interface_t* bi,
                           llb_task_interface_t ti,
@@ -751,6 +808,15 @@ llb_buildsystem_command_get_verbose_description(
 LLBUILD_EXPORT void
 llb_buildsystem_command_interface_task_needs_input(llb_task_interface_t ti, llb_build_key_t* key, uintptr_t inputID);
 
+/// Request a task as a dependency just for the current build iteration.
+/// Once the requesting task finishes, the dependency will be removed so
+/// that incremental builds won't consider it for invalidating the task.
+///
+/// NOTE: This method behaves like `llb_buildsystem_command_interface_task_needs_input`
+/// for the current build.
+LLBUILD_EXPORT void
+llb_buildsystem_command_interface_task_needs_single_use_input(llb_task_interface_t ti, llb_build_key_t* key, uintptr_t inputID);
+
 /// Marks a key as a discovered dependency for the task.
 LLBUILD_EXPORT void
 llb_buildsystem_command_interface_task_discovered_dependency(llb_task_interface_t ti, llb_build_key_t* key);
