@@ -46,7 +46,7 @@ public protocol ProcessDelegate {
     /// Called to report a command processes' (merged) standard output and error.
     ///
     /// - output: The process output.
-    func processHadOutput(output: String)
+    func processHadOutput(output: [UInt8])
     
     /// Called when a command's job has finished executing an external process.
     ///
@@ -64,7 +64,7 @@ private class ProcessDelegateWrapper {
             context: Unmanaged.passUnretained(self).toOpaque(),
             process_started: { BuildSystem.toProcessDelegateWrapper($0!).delegate.processStarted(pid: $1) },
             process_had_error: { BuildSystem.toProcessDelegateWrapper($0!).delegate.processHadError(error: stringFromData($1!.pointee)) },
-            process_had_output: { BuildSystem.toProcessDelegateWrapper($0!).delegate.processHadOutput(output: stringFromData($1!.pointee)) },
+            process_had_output: { BuildSystem.toProcessDelegateWrapper($0!).delegate.processHadOutput(output: bytesFromData($1!.pointee)) },
             process_finished: { BuildSystem.toProcessDelegateWrapper($0!).delegate.processFinished(result: CommandExtendedResult($1!)) }
         )
         self.wrappedDelegate = wrappedDelegate
@@ -89,6 +89,15 @@ public struct BuildSystemCommandInterface {
     public func commandNeedsInput(key: BuildKey, inputID: UInt) {
         llb_buildsystem_command_interface_task_needs_input(_taskInterface, key.internalBuildKey, inputID)
     }
+    
+    /// Request an input as a dependency just for the current build iteration.
+    /// Once the requesting command finishes, the dependency will be removed so that
+    /// incremental builds won't consider it for invalidation.
+    ///
+    /// NOTE: This method behaves like `request` for the current build.
+    public func commandsNeedsSingleUseInput(key: BuildKey, inputID: UInt) {
+        llb_buildsystem_command_interface_task_needs_single_use_input(_taskInterface, key.internalBuildKey, inputID)
+    }
 
     /// Marks a build key as a runtime found dependency for the command.
     public func commandDiscoveredDependency(key: BuildKey) {
@@ -136,9 +145,6 @@ public struct JobContext {
 }
 
 public protocol Tool: AnyObject {
-    @available(*, deprecated, message: "Use the overload that returns an Optional")
-    func createCommand(_ name: String) -> ExternalCommand
-
     /// Called to create a specific command instance of this tool.
     func createCommand(_ name: String) -> ExternalCommand?
 
@@ -147,11 +153,6 @@ public protocol Tool: AnyObject {
 }
 
 public extension Tool {
-    @available(*, deprecated, message: "Use the overload that returns an Optional")
-    func createCommand(_ name: String) -> ExternalCommand? {
-        return createCommand(name) as ExternalCommand
-    }
-
     // Default implementation to allow clients to avoid declaring this method if not required.
     func createCustomCommand(_ buildKey: BuildKey.CustomTask) -> ExternalCommand? {
         return nil
@@ -198,6 +199,7 @@ private final class ToolWrapper {
         self.commandWrappers.append(wrapper)
         var _delegate = llb_buildsystem_external_command_delegate_t()
         _delegate.context = Unmanaged.passUnretained(wrapper).toOpaque()
+        _delegate.configure = { BuildSystem.toCommandWrapper($0!).configure($1!, $2!, $3!, $4!) }
         _delegate.get_signature = { return BuildSystem.toCommandWrapper($0!).getSignature($1!, $2!) }
         _delegate.start = { return BuildSystem.toCommandWrapper($0!).start($1!, $2!, $3) }
         _delegate.provide_value = { return BuildSystem.toCommandWrapper($0!).provideValue($1!, $2!, $3, $4!, $5) }
@@ -227,6 +229,15 @@ public protocol ExternalCommand: AnyObject {
     ///
     /// This is checked to determine if the command needs to rebuild versus the last time it was run.
     func getSignature(_ command: Command) -> [UInt8]
+    
+    /// Paths to files that contain discovered dependencies after command executed successfully for subsequent builds.
+    var dependencyPaths: [String] { get }
+    
+    /// Format of the dependency files listed in `dependencyPaths`.
+    var depedencyDataFormat: DependencyDataFormat { get }
+
+    /// Optional. The command's working directory. Used to resolve relative paths in dependency files.
+    var workingDirectory: String? { get }
 
     /// Called when the command is starting. Commands can request dynamic dependencies using the
     /// commandInterface object.
@@ -246,7 +257,7 @@ public protocol ExternalCommand: AnyObject {
 
     /// Called to execute the given command.
     ///
-    /// This method is deprecated in favor of the execute(Command, BuildSystemCommandInterface) method.
+    /// This method is deprecated in favor of the execute(Command, BuildSystemCommandInterface, JobContext) method.
     ///
     /// - command: A handle to the executing command.
     /// - returns: True on success.
@@ -254,6 +265,8 @@ public protocol ExternalCommand: AnyObject {
 
     /// Called to execute the given command.
     ///
+    /// This method is deprecated in favor of the execute(Command, BuildSystemCommandInterface, JobContext) method.
+    ///
     /// - command: A handle to the executing command.
     /// - commandInterface: A handle to the build system's command interface.
     /// - returns: True on success.
@@ -261,20 +274,39 @@ public protocol ExternalCommand: AnyObject {
     
     /// Called to execute the given command.
     ///
+    /// This method is deprecated in favor of the execute(Command, BuildSystemCommandInterface, JobContext) -> CommandResult method.
+    ///
     /// - command: A handle to the executing command.
     /// - commandInterface: A handle to the build system's command interface.
     /// - jobContext: A handle to opaque context of the executing job for spawning external processes.
     /// - returns: True on success.
     func execute(_ command: Command, _ commandInterface: BuildSystemCommandInterface, _ jobContext: JobContext) -> Bool
+    
+    /// Called to execute the given command.
+    ///
+    /// - command: A handle to the executing command.
+    /// - commandInterface: A handle to the build system's command interface.
+    /// - jobContext: A handle to opaque context of the executing job for spawning external processes.
+    /// - returns: command execution result.
+    func execute(_ command: Command, _ commandInterface: BuildSystemCommandInterface, _ jobContext: JobContext) -> CommandResult
 }
 
 public protocol ProducesCustomBuildValue: AnyObject {
     /// Called to execute the given command that produces a custom build value.
     ///
+    /// This method is deprecated in favor of the execute(Command, BuildSystemCommandInterface, JobContext) method.
+    ///
     /// - command: A handle to the executing command.
     /// - commandInterface: A handle to the build system's command interface.
     /// - returns: Produced build value.
     func execute(_ command: Command, _ commandInterface: BuildSystemCommandInterface) -> BuildValue
+    
+    /// Called to execute the given command that produces a custom build value.
+    ///
+    /// - command: A handle to the executing command.
+    /// - commandInterface: A handle to the build system's command interface.
+    /// - returns: Produced build value.
+    func execute(_ command: Command, _ commandInterface: BuildSystemCommandInterface, _ jobContext: JobContext) -> BuildValue
 
     /// Called to check if the current result for this command remains valid.
     ///
@@ -294,6 +326,10 @@ public extension ExternalCommand {
     func execute(_ command: Command, _ commandInterface: BuildSystemCommandInterface, _ jobContext: JobContext) -> Bool {
         return execute(command, commandInterface)
     }
+    
+    func execute(_ command: Command, _ commandInterface: BuildSystemCommandInterface, _ jobContext: JobContext) -> CommandResult {
+        return execute(command, commandInterface, jobContext) ? .succeeded : .failed
+    }
 
     // If this implementation is invoked, it means that the client implementing ExternalCommand did not
     // implement either of the execute methods, which is a programmer error.
@@ -304,10 +340,27 @@ public extension ExternalCommand {
 
 // Default implementations for these hooks since they're optional to the client.
 public extension ExternalCommand {
+    var depedencyDataFormat: DependencyDataFormat { .unused }
+    var dependencyPaths: [String] { [] }
+    var workingDirectory: String? { nil }
     func start(_ command: Command, _ commandInterface: BuildSystemCommandInterface) {}
     func provideValue(_ command: Command, _ commandInterface: BuildSystemCommandInterface, _ buildValue: BuildValue, _ inputID: UInt) {}
 }
 
+// Extension to provide a default implementation of execute(_ Command, _ commandInterface, _ jobContext)
+// to allow clients to migrate in a staggered manner.
+public extension ProducesCustomBuildValue {
+    // If this implementation is invoked, it means that the client implementing ProducesCustomBuildValue
+    // did not implement either of the execute methods, which is a programmer error.
+    func execute(_ command: Command, _ commandInterface: BuildSystemCommandInterface) -> BuildValue {
+        fatalError("This should never be called.")
+    }
+    
+    func execute(_ command: Command, _ commandInterface: BuildSystemCommandInterface, _ jobContext: JobContext) -> BuildValue {
+        execute(command, commandInterface)
+    }
+}
+
 // FIXME: The terminology is really confusing here, we have ExternalCommand which is divorced from the actual internal command implementation of the same name.
 private final class CommandWrapper {
     let command: ExternalCommand
@@ -317,6 +370,52 @@ private final class CommandWrapper {
         self.command = command
         self._command = Command(handle: nil)
     }
+    
+    func configure(_ context: OpaquePointer, _ single: @convention(c) (OpaquePointer?, llb_data_t, llb_data_t) -> Void, _ collection: @convention(c) (OpaquePointer?, llb_data_t, UnsafeMutablePointer<llb_data_t>?, size_t) -> Void, _ map: @convention(c) (OpaquePointer?, llb_data_t, UnsafeMutablePointer<llb_data_t>?, UnsafeMutablePointer<llb_data_t>?, size_t) -> Void) {
+        if !command.dependencyPaths.isEmpty {
+            var formatKey = copiedDataFromBytes(Array("deps-style".utf8))
+            var value: llb_data_t
+            switch command.depedencyDataFormat {
+            case .makefile:
+                value = copiedDataFromBytes(Array("makefile".utf8))
+            case .dependencyinfo:
+                value = copiedDataFromBytes(Array("dependency-info".utf8))
+            default:
+                value = copiedDataFromBytes(Array("unusued".utf8))
+            }
+            defer {
+                llb_data_destroy(&formatKey)
+                llb_data_destroy(&value)
+            }
+            single(context, formatKey, value)
+            
+            var paths: [llb_data_t] = []
+            for path in command.dependencyPaths {
+                paths.append(copiedDataFromBytes(Array(path.utf8)))
+            }
+            var depsKey = copiedDataFromBytes(Array("deps".utf8))
+            defer {
+                llb_data_destroy(&depsKey)
+                for index in paths.indices {
+                    llb_data_destroy(&paths[index])
+                }
+            }
+            if paths.count == 1 {
+                single(context, depsKey, paths[0])
+            } else {            
+                collection(context, depsKey, &paths, paths.count)
+            }
+        }
+        if let workingDirectory = command.workingDirectory {
+            var workingDirectoryKey = copiedDataFromBytes(Array("working-directory".utf8))
+            var workingDirectoryValue = copiedDataFromBytes(Array(workingDirectory.utf8))
+            defer {
+                llb_data_destroy(&workingDirectoryKey)
+                llb_data_destroy(&workingDirectoryValue)
+            }
+            single(context, workingDirectoryKey, workingDirectoryValue)
+        }
+    }
 
     func getSignature(_: OpaquePointer, _ data: UnsafeMutablePointer<llb_data_t>) {
         data.pointee = copiedDataFromBytes(command.getSignature(_command))
@@ -337,14 +436,14 @@ private final class CommandWrapper {
         return command.provideValue(_command, commandInterface, buildValue, inputID)
     }
 
-    func executeCommand(_: OpaquePointer, _ buildsystemInterface: OpaquePointer, _ taskInterface: llb_task_interface_t, _ jobContext: OpaquePointer) -> Bool {
+    func executeCommand(_: OpaquePointer, _ buildsystemInterface: OpaquePointer, _ taskInterface: llb_task_interface_t, _ jobContext: OpaquePointer) -> CommandResult {
         let commandInterface = BuildSystemCommandInterface(buildsystemInterface, taskInterface)
         return command.execute(_command, commandInterface, JobContext(jobContext))
     }
 
     func executeCommand(_: OpaquePointer, _ buildsystemInterface: OpaquePointer, _ taskInterface: llb_task_interface_t, _ jobContext: OpaquePointer) -> BuildValue {
         let commandInterface = BuildSystemCommandInterface(buildsystemInterface, taskInterface)
-        return (command as! ProducesCustomBuildValue).execute(_command, commandInterface)
+        return (command as! ProducesCustomBuildValue).execute(_command, commandInterface, JobContext(jobContext))
     }
 
     func isResultValid(_: OpaquePointer, _ value: OpaquePointer) -> Bool {
@@ -743,6 +842,15 @@ public protocol BuildSystemDelegate {
     /// - parameter exitStatus: The raw exit status of the process.
     func commandProcessFinished(_ command: Command, process: ProcessHandle, result: CommandExtendedResult)
 
+    /// Called when it's been determined that a rule needs to run.
+    ///
+    ///  - parameter ruleNeedingToRun: The rule that needs to run.
+    ///
+    ///  - parameter reason: Describes why the rule needs to run. For example, because it has never run or because an input was rebuilt.
+    ///
+    ///  - parameter inputRule: If `reason` is `InputRebuilt`, the rule for the rebuilt input, else  `nil`.
+    func determinedRuleNeedsToRun(_ rule: BuildKey, reason: RuleRunReason, inputRule: BuildKey?)
+
     /// Called when a cycle is detected by the build engine and it cannot make
     /// forward progress.
     func cycleDetected(rules: [BuildKey])
@@ -775,6 +883,10 @@ extension BuildSystemDelegate {
         // default implementation for ABI compatibility with older clients
         return nil
     }
+
+    public func determinedRuleNeedsToRun(_ rule: BuildKey, reason: RuleRunReason, inputRule: BuildKey?) {
+        // default implementation for compatibility with older clients
+    }
 }
 
 /// Utility class for constructing a C-style environment.
@@ -924,6 +1036,9 @@ public final class BuildSystem {
             _delegate.command_process_had_error = { BuildSystem.toSystem($0!).commandProcessHadError(Command(handle: $1), ProcessHandle($2!), $3!) }
             _delegate.command_process_had_output = { BuildSystem.toSystem($0!).commandProcessHadOutput(Command(handle: $1), ProcessHandle($2!), $3!) }
             _delegate.command_process_finished = { BuildSystem.toSystem($0!).commandProcessFinished(Command(handle: $1), ProcessHandle($2!), CommandExtendedResult($3!)) }
+            _delegate.determined_rule_needs_to_run = {
+                BuildSystem.toSystem($0!).determinedRuleNeedsToRun(BuildKey.construct(key: $1!), reason: $2, inputRule: $3.map { BuildKey.construct(key: $0) })
+            }
             _delegate.cycle_detected = {
                 var rules = [BuildKey]()
                 UnsafeBufferPointer(start: $1, count: Int($2)).forEach {
@@ -1161,6 +1276,10 @@ public final class BuildSystem {
         delegate.commandProcessFinished(command, process: process, result: result)
     }
 
+    private func determinedRuleNeedsToRun(_ rule: BuildKey, reason: RuleRunReason, inputRule: BuildKey?) {
+        delegate.determinedRuleNeedsToRun(rule, reason: reason, inputRule: inputRule)
+    }
+
     private func cycleDetected(_ rules: [BuildKey]) {
         delegate.cycleDetected(rules: rules)
     }
