@@ -9,6 +9,7 @@
 #include "Config.h"
 #include "ConfigFragment.h"
 #include "ConfigTesting.h"
+#include "Diagnostics.h"
 #include "Feature.h"
 #include "TestFS.h"
 #include "clang/Basic/DiagnosticSema.h"
@@ -121,14 +122,15 @@ TEST_F(ConfigCompileTests, Condition) {
 }
 
 TEST_F(ConfigCompileTests, CompileCommands) {
+  Frag.CompileFlags.Compiler.emplace("tpc.exe");
   Frag.CompileFlags.Add.emplace_back("-foo");
   Frag.CompileFlags.Remove.emplace_back("--include-directory=");
   std::vector<std::string> Argv = {"clang", "-I", "bar/", "--", "a.cc"};
   EXPECT_TRUE(compileAndApply());
-  EXPECT_THAT(Conf.CompileFlags.Edits, SizeIs(2));
+  EXPECT_THAT(Conf.CompileFlags.Edits, SizeIs(3));
   for (auto &Edit : Conf.CompileFlags.Edits)
     Edit(Argv);
-  EXPECT_THAT(Argv, ElementsAre("clang", "-foo", "--", "a.cc"));
+  EXPECT_THAT(Argv, ElementsAre("tpc.exe", "-foo", "--", "a.cc"));
 }
 
 TEST_F(ConfigCompileTests, CompilationDatabase) {
@@ -149,7 +151,7 @@ TEST_F(ConfigCompileTests, CompilationDatabase) {
             Config::CDBSearchSpec::Ancestors)
       << "default value";
   EXPECT_THAT(Diags.Diagnostics,
-              ElementsAre(DiagMessage(
+              ElementsAre(diagMessage(
                   "CompilationDatabase must be an absolute path, because this "
                   "fragment is not associated with any directory.")));
 
@@ -183,7 +185,7 @@ TEST_F(ConfigCompileTests, Index) {
       << "by default";
   EXPECT_THAT(
       Diags.Diagnostics,
-      ElementsAre(DiagMessage(
+      ElementsAre(diagMessage(
           "Invalid Background value 'Foo'. Valid values are Build, Skip.")));
 }
 
@@ -244,6 +246,43 @@ TEST_F(ConfigCompileTests, PathSpecMatch) {
   }
 }
 
+TEST_F(ConfigCompileTests, DiagnosticsIncludeCleaner) {
+  // Defaults to None.
+  EXPECT_TRUE(compileAndApply());
+  EXPECT_EQ(Conf.Diagnostics.UnusedIncludes,
+            Config::UnusedIncludesPolicy::None);
+
+  Frag = {};
+  Frag.Diagnostics.UnusedIncludes.emplace("None");
+  EXPECT_TRUE(compileAndApply());
+  EXPECT_EQ(Conf.Diagnostics.UnusedIncludes,
+            Config::UnusedIncludesPolicy::None);
+
+  Frag = {};
+  Frag.Diagnostics.UnusedIncludes.emplace("Strict");
+  EXPECT_TRUE(compileAndApply());
+  EXPECT_EQ(Conf.Diagnostics.UnusedIncludes,
+            Config::UnusedIncludesPolicy::Strict);
+
+  Frag = {};
+  EXPECT_TRUE(Conf.Diagnostics.Includes.IgnoreHeader.empty())
+      << Conf.Diagnostics.Includes.IgnoreHeader.size();
+  Frag.Diagnostics.Includes.IgnoreHeader.push_back(
+      Located<std::string>("foo.h"));
+  Frag.Diagnostics.Includes.IgnoreHeader.push_back(
+      Located<std::string>(".*inc"));
+  EXPECT_TRUE(compileAndApply());
+  auto HeaderFilter = [this](llvm::StringRef Path) {
+    for (auto &Filter : Conf.Diagnostics.Includes.IgnoreHeader) {
+      if (Filter(Path))
+        return true;
+    }
+    return false;
+  };
+  EXPECT_TRUE(HeaderFilter("foo.h"));
+  EXPECT_FALSE(HeaderFilter("bar.h"));
+}
+
 TEST_F(ConfigCompileTests, DiagnosticSuppression) {
   Frag.Diagnostics.Suppress.emplace_back("bugprone-use-after-move");
   Frag.Diagnostics.Suppress.emplace_back("unreachable-code");
@@ -257,19 +296,20 @@ TEST_F(ConfigCompileTests, DiagnosticSuppression) {
                                    "unreachable-code", "unused-variable",
                                    "typecheck_bool_condition",
                                    "unexpected_friend", "warn_alloca"));
-  EXPECT_TRUE(isBuiltinDiagnosticSuppressed(diag::warn_unreachable,
-                                            Conf.Diagnostics.Suppress));
+  EXPECT_TRUE(isBuiltinDiagnosticSuppressed(
+      diag::warn_unreachable, Conf.Diagnostics.Suppress, LangOptions()));
   // Subcategory not respected/suppressed.
-  EXPECT_FALSE(isBuiltinDiagnosticSuppressed(diag::warn_unreachable_break,
-                                             Conf.Diagnostics.Suppress));
-  EXPECT_TRUE(isBuiltinDiagnosticSuppressed(diag::warn_unused_variable,
-                                            Conf.Diagnostics.Suppress));
+  EXPECT_FALSE(isBuiltinDiagnosticSuppressed(
+      diag::warn_unreachable_break, Conf.Diagnostics.Suppress, LangOptions()));
+  EXPECT_TRUE(isBuiltinDiagnosticSuppressed(
+      diag::warn_unused_variable, Conf.Diagnostics.Suppress, LangOptions()));
   EXPECT_TRUE(isBuiltinDiagnosticSuppressed(diag::err_typecheck_bool_condition,
-                                            Conf.Diagnostics.Suppress));
-  EXPECT_TRUE(isBuiltinDiagnosticSuppressed(diag::err_unexpected_friend,
-                                            Conf.Diagnostics.Suppress));
-  EXPECT_TRUE(isBuiltinDiagnosticSuppressed(diag::warn_alloca,
-                                            Conf.Diagnostics.Suppress));
+                                            Conf.Diagnostics.Suppress,
+                                            LangOptions()));
+  EXPECT_TRUE(isBuiltinDiagnosticSuppressed(
+      diag::err_unexpected_friend, Conf.Diagnostics.Suppress, LangOptions()));
+  EXPECT_TRUE(isBuiltinDiagnosticSuppressed(
+      diag::warn_alloca, Conf.Diagnostics.Suppress, LangOptions()));
 
   Frag.Diagnostics.Suppress.emplace_back("*");
   EXPECT_TRUE(compileAndApply());
@@ -304,9 +344,9 @@ TEST_F(ConfigCompileTests, Tidy) {
   EXPECT_THAT(
       Diags.Diagnostics,
       ElementsAre(
-          DiagMessage(
+          diagMessage(
               "clang-tidy check 'bugprone-use-after-move' was not found"),
-          DiagMessage("clang-tidy check 'llvm-include-order' was not found")));
+          diagMessage("clang-tidy check 'llvm-include-order' was not found")));
 #endif
 }
 
@@ -321,11 +361,11 @@ TEST_F(ConfigCompileTests, TidyBadChecks) {
   EXPECT_THAT(
       Diags.Diagnostics,
       ElementsAre(
-          AllOf(DiagMessage("clang-tidy check 'unknown-check' was not found"),
-                DiagKind(llvm::SourceMgr::DK_Warning)),
+          AllOf(diagMessage("clang-tidy check 'unknown-check' was not found"),
+                diagKind(llvm::SourceMgr::DK_Warning)),
           AllOf(
-              DiagMessage("clang-tidy check 'llvm-includeorder' was not found"),
-              DiagKind(llvm::SourceMgr::DK_Warning))));
+              diagMessage("clang-tidy check 'llvm-includeorder' was not found"),
+              diagKind(llvm::SourceMgr::DK_Warning))));
 }
 
 TEST_F(ConfigCompileTests, ExternalServerNeedsTrusted) {
@@ -335,7 +375,7 @@ TEST_F(ConfigCompileTests, ExternalServerNeedsTrusted) {
   compileAndApply();
   EXPECT_THAT(
       Diags.Diagnostics,
-      ElementsAre(DiagMessage(
+      ElementsAre(diagMessage(
           "Remote index may not be specified by untrusted configuration. "
           "Copy this into user config to use it.")));
   EXPECT_EQ(Conf.Index.External.Kind, Config::ExternalIndexSpec::None);
@@ -352,8 +392,8 @@ TEST_F(ConfigCompileTests, ExternalBlockWarnOnMultipleSource) {
   EXPECT_THAT(
       Diags.Diagnostics,
       Contains(
-          AllOf(DiagMessage("Exactly one of File, Server or None must be set."),
-                DiagKind(llvm::SourceMgr::DK_Error))));
+          AllOf(diagMessage("Exactly one of File, Server or None must be set."),
+                diagKind(llvm::SourceMgr::DK_Error))));
 #else
   ASSERT_TRUE(Conf.Index.External.hasValue());
   EXPECT_EQ(Conf.Index.External->Kind, Config::ExternalIndexSpec::File);
@@ -377,8 +417,8 @@ TEST_F(ConfigCompileTests, ExternalBlockErrOnNoSource) {
   EXPECT_THAT(
       Diags.Diagnostics,
       Contains(
-          AllOf(DiagMessage("Exactly one of File, Server or None must be set."),
-                DiagKind(llvm::SourceMgr::DK_Error))));
+          AllOf(diagMessage("Exactly one of File, Server or None must be set."),
+                diagKind(llvm::SourceMgr::DK_Error))));
 }
 
 TEST_F(ConfigCompileTests, ExternalBlockDisablesBackgroundIndex) {
@@ -416,9 +456,9 @@ TEST_F(ConfigCompileTests, ExternalBlockMountPoint) {
   ASSERT_THAT(
       Diags.Diagnostics,
       ElementsAre(
-          AllOf(DiagMessage("MountPoint must be an absolute path, because this "
+          AllOf(diagMessage("MountPoint must be an absolute path, because this "
                             "fragment is not associated with any directory."),
-                DiagKind(llvm::SourceMgr::DK_Error))));
+                diagKind(llvm::SourceMgr::DK_Error))));
   EXPECT_EQ(Conf.Index.External.Kind, Config::ExternalIndexSpec::None);
 
   auto FooPath = testPath("foo/", llvm::sys::path::Style::posix);
