@@ -12,9 +12,13 @@
 #include "flang/Optimizer/Support/InitFIR.h"
 #include "flang/Optimizer/Support/KindMapping.h"
 
+using namespace mlir;
+
 struct FIRBuilderTest : public testing::Test {
 public:
   void SetUp() override {
+    fir::support::loadDialects(context);
+
     llvm::ArrayRef<fir::KindTy> defs;
     fir::KindMapping kindMap(&context, defs);
     mlir::OpBuilder builder(&context);
@@ -23,13 +27,12 @@ public:
     // Set up a Module with a dummy function operation inside.
     // Set the insertion point in the function entry block.
     mlir::ModuleOp mod = builder.create<mlir::ModuleOp>(loc);
-    mlir::FuncOp func = mlir::FuncOp::create(
+    mlir::func::FuncOp func = mlir::func::FuncOp::create(
         loc, "func1", builder.getFunctionType(llvm::None, llvm::None));
     auto *entryBlock = func.addEntryBlock();
     mod.push_back(mod);
     builder.setInsertionPointToStart(entryBlock);
 
-    fir::support::loadDialects(context);
     firBuilder = std::make_unique<fir::FirOpBuilder>(mod, kindMap);
   }
 
@@ -48,8 +51,8 @@ static arith::CmpIOp createCondition(fir::FirOpBuilder &builder) {
 }
 
 static void checkIntegerConstant(mlir::Value value, mlir::Type ty, int64_t v) {
-  EXPECT_TRUE(mlir::isa<ConstantOp>(value.getDefiningOp()));
-  auto cstOp = dyn_cast<ConstantOp>(value.getDefiningOp());
+  EXPECT_TRUE(mlir::isa<mlir::arith::ConstantOp>(value.getDefiningOp()));
+  auto cstOp = dyn_cast<mlir::arith::ConstantOp>(value.getDefiningOp());
   EXPECT_EQ(ty, cstOp.getType());
   auto valueAttr = cstOp.getValue().dyn_cast_or_null<IntegerAttr>();
   EXPECT_EQ(v, valueAttr.getInt());
@@ -64,8 +67,8 @@ TEST_F(FIRBuilderTest, genIfThen) {
   auto loc = builder.getUnknownLoc();
   auto cdt = createCondition(builder);
   auto ifBuilder = builder.genIfThen(loc, cdt);
-  EXPECT_FALSE(ifBuilder.getIfOp().thenRegion().empty());
-  EXPECT_TRUE(ifBuilder.getIfOp().elseRegion().empty());
+  EXPECT_FALSE(ifBuilder.getIfOp().getThenRegion().empty());
+  EXPECT_TRUE(ifBuilder.getIfOp().getElseRegion().empty());
 }
 
 TEST_F(FIRBuilderTest, genIfThenElse) {
@@ -73,8 +76,8 @@ TEST_F(FIRBuilderTest, genIfThenElse) {
   auto loc = builder.getUnknownLoc();
   auto cdt = createCondition(builder);
   auto ifBuilder = builder.genIfThenElse(loc, cdt);
-  EXPECT_FALSE(ifBuilder.getIfOp().thenRegion().empty());
-  EXPECT_FALSE(ifBuilder.getIfOp().elseRegion().empty());
+  EXPECT_FALSE(ifBuilder.getIfOp().getThenRegion().empty());
+  EXPECT_FALSE(ifBuilder.getIfOp().getElseRegion().empty());
 }
 
 TEST_F(FIRBuilderTest, genIfWithThen) {
@@ -82,8 +85,8 @@ TEST_F(FIRBuilderTest, genIfWithThen) {
   auto loc = builder.getUnknownLoc();
   auto cdt = createCondition(builder);
   auto ifBuilder = builder.genIfOp(loc, {}, cdt, false);
-  EXPECT_FALSE(ifBuilder.getIfOp().thenRegion().empty());
-  EXPECT_TRUE(ifBuilder.getIfOp().elseRegion().empty());
+  EXPECT_FALSE(ifBuilder.getIfOp().getThenRegion().empty());
+  EXPECT_TRUE(ifBuilder.getIfOp().getElseRegion().empty());
 }
 
 TEST_F(FIRBuilderTest, genIfWithThenAndElse) {
@@ -91,8 +94,8 @@ TEST_F(FIRBuilderTest, genIfWithThenAndElse) {
   auto loc = builder.getUnknownLoc();
   auto cdt = createCondition(builder);
   auto ifBuilder = builder.genIfOp(loc, {}, cdt, true);
-  EXPECT_FALSE(ifBuilder.getIfOp().thenRegion().empty());
-  EXPECT_FALSE(ifBuilder.getIfOp().elseRegion().empty());
+  EXPECT_FALSE(ifBuilder.getIfOp().getThenRegion().empty());
+  EXPECT_FALSE(ifBuilder.getIfOp().getElseRegion().empty());
 }
 
 //===----------------------------------------------------------------------===//
@@ -107,7 +110,7 @@ TEST_F(FIRBuilderTest, genIsNotNull) {
   auto res = builder.genIsNotNull(loc, dummyValue);
   EXPECT_TRUE(mlir::isa<arith::CmpIOp>(res.getDefiningOp()));
   auto cmpOp = dyn_cast<arith::CmpIOp>(res.getDefiningOp());
-  EXPECT_EQ(arith::CmpIPredicate::ne, cmpOp.predicate());
+  EXPECT_EQ(arith::CmpIPredicate::ne, cmpOp.getPredicate());
 }
 
 TEST_F(FIRBuilderTest, genIsNull) {
@@ -118,7 +121,7 @@ TEST_F(FIRBuilderTest, genIsNull) {
   auto res = builder.genIsNull(loc, dummyValue);
   EXPECT_TRUE(mlir::isa<arith::CmpIOp>(res.getDefiningOp()));
   auto cmpOp = dyn_cast<arith::CmpIOp>(res.getDefiningOp());
-  EXPECT_EQ(arith::CmpIPredicate::eq, cmpOp.predicate());
+  EXPECT_EQ(arith::CmpIPredicate::eq, cmpOp.getPredicate());
 }
 
 TEST_F(FIRBuilderTest, createZeroConstant) {
@@ -144,10 +147,11 @@ TEST_F(FIRBuilderTest, createRealZeroConstant) {
   auto loc = builder.getUnknownLoc();
   auto realTy = mlir::FloatType::getF64(ctx);
   auto cst = builder.createRealZeroConstant(loc, realTy);
-  EXPECT_TRUE(mlir::isa<mlir::arith::ConstantOp>(cst.getDefiningOp()));
-  auto cstOp = dyn_cast<mlir::arith::ConstantOp>(cst.getDefiningOp());
+  EXPECT_TRUE(mlir::isa<arith::ConstantOp>(cst.getDefiningOp()));
+  auto cstOp = dyn_cast<arith::ConstantOp>(cst.getDefiningOp());
   EXPECT_EQ(realTy, cstOp.getType());
-  EXPECT_EQ(0u, cstOp.value().cast<FloatAttr>().getValue().convertToDouble());
+  EXPECT_EQ(
+      0u, cstOp.getValue().cast<FloatAttr>().getValue().convertToDouble());
 }
 
 TEST_F(FIRBuilderTest, createBool) {
@@ -184,13 +188,13 @@ TEST_F(FIRBuilderTest, createGlobal1) {
   auto global = builder.createGlobal(
       loc, i64Type, "global1", builder.createInternalLinkage(), {}, true);
   EXPECT_TRUE(mlir::isa<fir::GlobalOp>(global));
-  EXPECT_EQ("global1", global.sym_name());
-  EXPECT_TRUE(global.constant().hasValue());
-  EXPECT_EQ(i64Type, global.type());
-  EXPECT_TRUE(global.linkName().hasValue());
-  EXPECT_EQ(
-      builder.createInternalLinkage().getValue(), global.linkName().getValue());
-  EXPECT_FALSE(global.initVal().hasValue());
+  EXPECT_EQ("global1", global.getSymName());
+  EXPECT_TRUE(global.getConstant().hasValue());
+  EXPECT_EQ(i64Type, global.getType());
+  EXPECT_TRUE(global.getLinkName().hasValue());
+  EXPECT_EQ(builder.createInternalLinkage().getValue(),
+      global.getLinkName().getValue());
+  EXPECT_FALSE(global.getInitVal().hasValue());
 
   auto g1 = builder.getNamedGlobal("global1");
   EXPECT_EQ(global, g1);
@@ -208,16 +212,16 @@ TEST_F(FIRBuilderTest, createGlobal2) {
   auto global = builder.createGlobal(
       loc, i32Type, "global2", builder.createLinkOnceLinkage(), attr, false);
   EXPECT_TRUE(mlir::isa<fir::GlobalOp>(global));
-  EXPECT_EQ("global2", global.sym_name());
-  EXPECT_FALSE(global.constant().hasValue());
-  EXPECT_EQ(i32Type, global.type());
-  EXPECT_TRUE(global.initVal().hasValue());
-  EXPECT_TRUE(global.initVal().getValue().isa<mlir::IntegerAttr>());
+  EXPECT_EQ("global2", global.getSymName());
+  EXPECT_FALSE(global.getConstant().hasValue());
+  EXPECT_EQ(i32Type, global.getType());
+  EXPECT_TRUE(global.getInitVal().hasValue());
+  EXPECT_TRUE(global.getInitVal().getValue().isa<mlir::IntegerAttr>());
   EXPECT_EQ(
-      16, global.initVal().getValue().cast<mlir::IntegerAttr>().getValue());
-  EXPECT_TRUE(global.linkName().hasValue());
-  EXPECT_EQ(
-      builder.createLinkOnceLinkage().getValue(), global.linkName().getValue());
+      16, global.getInitVal().getValue().cast<mlir::IntegerAttr>().getValue());
+  EXPECT_TRUE(global.getLinkName().hasValue());
+  EXPECT_EQ(builder.createLinkOnceLinkage().getValue(),
+      global.getLinkName().getValue());
 }
 
 TEST_F(FIRBuilderTest, uniqueCFIdent) {
@@ -237,7 +241,7 @@ TEST_F(FIRBuilderTest, uniqueCFIdent) {
 
 TEST_F(FIRBuilderTest, locationToLineNo) {
   auto builder = getBuilder();
-  auto loc = mlir::FileLineColLoc::get(builder.getIdentifier("file1"), 10, 5);
+  auto loc = mlir::FileLineColLoc::get(builder.getStringAttr("file1"), 10, 5);
   mlir::Value line =
       fir::factory::locationToLineNo(builder, loc, builder.getI64Type());
   checkIntegerConstant(line, builder.getI64Type(), 10);
@@ -259,10 +263,10 @@ TEST_F(FIRBuilderTest, hasDynamicSize) {
 TEST_F(FIRBuilderTest, locationToFilename) {
   auto builder = getBuilder();
   auto loc =
-      mlir::FileLineColLoc::get(builder.getIdentifier("file1.f90"), 10, 5);
+      mlir::FileLineColLoc::get(builder.getStringAttr("file1.f90"), 10, 5);
   mlir::Value locToFile = fir::factory::locationToFilename(builder, loc);
   auto addrOp = dyn_cast<fir::AddrOfOp>(locToFile.getDefiningOp());
-  auto symbol = addrOp.symbol().getRootReference().getValue();
+  auto symbol = addrOp.getSymbol().getRootReference().getValue();
   auto global = builder.getNamedGlobal(symbol);
   auto stringLitOps = global.getRegion().front().getOps<fir::StringLitOp>();
   EXPECT_TRUE(llvm::hasSingleElement(stringLitOps));
@@ -304,12 +308,12 @@ TEST_F(FIRBuilderTest, createStringLiteral) {
   auto addr = charBox->getBuffer();
   EXPECT_TRUE(mlir::isa<fir::AddrOfOp>(addr.getDefiningOp()));
   auto addrOp = dyn_cast<fir::AddrOfOp>(addr.getDefiningOp());
-  auto symbol = addrOp.symbol().getRootReference().getValue();
+  auto symbol = addrOp.getSymbol().getRootReference().getValue();
   auto global = builder.getNamedGlobal(symbol);
-  EXPECT_EQ(
-      builder.createLinkOnceLinkage().getValue(), global.linkName().getValue());
+  EXPECT_EQ(builder.createLinkOnceLinkage().getValue(),
+      global.getLinkName().getValue());
   EXPECT_EQ(fir::CharacterType::get(builder.getContext(), 1, strValue.size()),
-      global.type());
+      global.getType());
 
   auto stringLitOps = global.getRegion().front().getOps<fir::StringLitOp>();
   EXPECT_TRUE(llvm::hasSingleElement(stringLitOps));
@@ -328,13 +332,13 @@ TEST_F(FIRBuilderTest, allocateLocal) {
       loc, builder.getI64Type(), "", varName, {}, {}, false);
   EXPECT_TRUE(mlir::isa<fir::AllocaOp>(var.getDefiningOp()));
   auto allocaOp = dyn_cast<fir::AllocaOp>(var.getDefiningOp());
-  EXPECT_EQ(builder.getI64Type(), allocaOp.in_type());
-  EXPECT_TRUE(allocaOp.bindc_name().hasValue());
-  EXPECT_EQ(varName, allocaOp.bindc_name().getValue());
-  EXPECT_FALSE(allocaOp.uniq_name().hasValue());
-  EXPECT_FALSE(allocaOp.pinned());
-  EXPECT_EQ(0u, allocaOp.typeparams().size());
-  EXPECT_EQ(0u, allocaOp.shape().size());
+  EXPECT_EQ(builder.getI64Type(), allocaOp.getInType());
+  EXPECT_TRUE(allocaOp.getBindcName().hasValue());
+  EXPECT_EQ(varName, allocaOp.getBindcName().getValue());
+  EXPECT_FALSE(allocaOp.getUniqName().hasValue());
+  EXPECT_FALSE(allocaOp.getPinned());
+  EXPECT_EQ(0u, allocaOp.getTypeparams().size());
+  EXPECT_EQ(0u, allocaOp.getShape().size());
 }
 
 static void checkShapeOp(mlir::Value shape, mlir::Value c10, mlir::Value c100) {
@@ -413,3 +417,114 @@ TEST_F(FIRBuilderTest, getExtents) {
   auto readExtents = fir::factory::getExtents(builder, loc, ex);
   EXPECT_EQ(2u, readExtents.size());
 }
+
+TEST_F(FIRBuilderTest, createZeroValue) {
+  auto builder = getBuilder();
+  auto loc = builder.getUnknownLoc();
+
+  mlir::Type i64Ty = mlir::IntegerType::get(builder.getContext(), 64);
+  mlir::Value zeroInt = fir::factory::createZeroValue(builder, loc, i64Ty);
+  EXPECT_TRUE(zeroInt.getType() == i64Ty);
+  auto cst =
+      mlir::dyn_cast_or_null<mlir::arith::ConstantOp>(zeroInt.getDefiningOp());
+  EXPECT_TRUE(cst);
+  auto intAttr = cst.getValue().dyn_cast<mlir::IntegerAttr>();
+  EXPECT_TRUE(intAttr && intAttr.getInt() == 0);
+
+  mlir::Type f32Ty = mlir::FloatType::getF32(builder.getContext());
+  mlir::Value zeroFloat = fir::factory::createZeroValue(builder, loc, f32Ty);
+  EXPECT_TRUE(zeroFloat.getType() == f32Ty);
+  auto cst2 = mlir::dyn_cast_or_null<mlir::arith::ConstantOp>(
+      zeroFloat.getDefiningOp());
+  EXPECT_TRUE(cst2);
+  auto floatAttr = cst2.getValue().dyn_cast<mlir::FloatAttr>();
+  EXPECT_TRUE(floatAttr && floatAttr.getValueAsDouble() == 0.);
+
+  mlir::Type boolTy = mlir::IntegerType::get(builder.getContext(), 1);
+  mlir::Value flaseBool = fir::factory::createZeroValue(builder, loc, boolTy);
+  EXPECT_TRUE(flaseBool.getType() == boolTy);
+  auto cst3 = mlir::dyn_cast_or_null<mlir::arith::ConstantOp>(
+      flaseBool.getDefiningOp());
+  EXPECT_TRUE(cst3);
+  auto intAttr2 = cst.getValue().dyn_cast<mlir::IntegerAttr>();
+  EXPECT_TRUE(intAttr2 && intAttr2.getInt() == 0);
+}
+
+TEST_F(FIRBuilderTest, getBaseTypeOf) {
+  auto builder = getBuilder();
+  auto loc = builder.getUnknownLoc();
+
+  auto makeExv = [&](mlir::Type elementType, mlir::Type arrayType)
+      -> std::tuple<llvm::SmallVector<fir::ExtendedValue, 4>,
+          llvm::SmallVector<fir::ExtendedValue, 4>> {
+    auto ptrTyArray = fir::PointerType::get(arrayType);
+    auto ptrTyScalar = fir::PointerType::get(elementType);
+    auto ptrBoxTyArray = fir::BoxType::get(ptrTyArray);
+    auto ptrBoxTyScalar = fir::BoxType::get(ptrTyScalar);
+    auto boxRefTyArray = fir::ReferenceType::get(ptrBoxTyArray);
+    auto boxRefTyScalar = fir::ReferenceType::get(ptrBoxTyScalar);
+    auto boxTyArray = fir::BoxType::get(arrayType);
+    auto boxTyScalar = fir::BoxType::get(elementType);
+
+    auto ptrValArray = builder.create<fir::UndefOp>(loc, ptrTyArray);
+    auto ptrValScalar = builder.create<fir::UndefOp>(loc, ptrTyScalar);
+    auto boxRefValArray = builder.create<fir::UndefOp>(loc, boxRefTyArray);
+    auto boxRefValScalar = builder.create<fir::UndefOp>(loc, boxRefTyScalar);
+    auto boxValArray = builder.create<fir::UndefOp>(loc, boxTyArray);
+    auto boxValScalar = builder.create<fir::UndefOp>(loc, boxTyScalar);
+
+    llvm::SmallVector<fir::ExtendedValue, 4> scalars;
+    scalars.emplace_back(fir::UnboxedValue(ptrValScalar));
+    scalars.emplace_back(fir::BoxValue(boxValScalar));
+    scalars.emplace_back(
+        fir::MutableBoxValue(boxRefValScalar, mlir::ValueRange(), {}));
+
+    llvm::SmallVector<fir::ExtendedValue, 4> arrays;
+    auto extent = builder.create<fir::UndefOp>(loc, builder.getIndexType());
+    llvm::SmallVector<mlir::Value> extents(
+        arrayType.dyn_cast<fir::SequenceType>().getDimension(),
+        extent.getResult());
+    arrays.emplace_back(fir::ArrayBoxValue(ptrValArray, extents));
+    arrays.emplace_back(fir::BoxValue(boxValArray));
+    arrays.emplace_back(
+        fir::MutableBoxValue(boxRefValArray, mlir::ValueRange(), {}));
+    return {scalars, arrays};
+  };
+
+  auto f32Ty = mlir::FloatType::getF32(builder.getContext());
+  mlir::Type f32SeqTy = builder.getVarLenSeqTy(f32Ty);
+  auto [f32Scalars, f32Arrays] = makeExv(f32Ty, f32SeqTy);
+  for (const auto &scalar : f32Scalars) {
+    EXPECT_EQ(fir::getBaseTypeOf(scalar), f32Ty);
+    EXPECT_EQ(fir::getElementTypeOf(scalar), f32Ty);
+    EXPECT_FALSE(fir::isDerivedWithLengthParameters(scalar));
+  }
+  for (const auto &array : f32Arrays) {
+    EXPECT_EQ(fir::getBaseTypeOf(array), f32SeqTy);
+    EXPECT_EQ(fir::getElementTypeOf(array), f32Ty);
+    EXPECT_FALSE(fir::isDerivedWithLengthParameters(array));
+  }
+
+  auto derivedWithLengthTy =
+      fir::RecordType::get(builder.getContext(), "derived_test");
+
+  llvm::SmallVector<std::pair<std::string, mlir::Type>> parameters;
+  llvm::SmallVector<std::pair<std::string, mlir::Type>> components;
+  parameters.emplace_back("p1", builder.getI64Type());
+  components.emplace_back("c1", f32Ty);
+  derivedWithLengthTy.finalize(parameters, components);
+  mlir::Type derivedWithLengthSeqTy =
+      builder.getVarLenSeqTy(derivedWithLengthTy);
+  auto [derivedWithLengthScalars, derivedWithLengthArrays] =
+      makeExv(derivedWithLengthTy, derivedWithLengthSeqTy);
+  for (const auto &scalar : derivedWithLengthScalars) {
+    EXPECT_EQ(fir::getBaseTypeOf(scalar), derivedWithLengthTy);
+    EXPECT_EQ(fir::getElementTypeOf(scalar), derivedWithLengthTy);
+    EXPECT_TRUE(fir::isDerivedWithLengthParameters(scalar));
+  }
+  for (const auto &array : derivedWithLengthArrays) {
+    EXPECT_EQ(fir::getBaseTypeOf(array), derivedWithLengthSeqTy);
+    EXPECT_EQ(fir::getElementTypeOf(array), derivedWithLengthTy);
+    EXPECT_TRUE(fir::isDerivedWithLengthParameters(array));
+  }
+}
