@@ -86,8 +86,6 @@ enum tool_setting_e {
 
 ompt_callbacks_active_t ompt_enabled;
 
-ompt_target_callbacks_active_t ompt_target_enabled;
-
 ompt_state_info_t ompt_state_info[] = {
 #define ompt_state_macro(state, code) {#state, state},
     FOREACH_OMPT_STATE(ompt_state_macro)
@@ -102,10 +100,6 @@ kmp_mutex_impl_info_t kmp_mutex_impl_info[] = {
 
 ompt_callbacks_internal_t ompt_callbacks;
 
-ompt_target_callbacks_internal_t ompt_target_callbacks;
-
-ompt_callbacks_internal_noemi_t ompt_callbacks_noemi;
-
 static ompt_start_tool_result_t *ompt_start_tool_result = NULL;
 
 #if KMP_OS_WINDOWS
@@ -131,51 +125,6 @@ OMPT_API_ROUTINE ompt_data_t *ompt_get_thread_data(void);
 typedef ompt_start_tool_result_t *(*ompt_start_tool_t)(unsigned int,
                                                        const char *);
 
-_OMP_EXTERN OMPT_WEAK_ATTRIBUTE bool
-libomp_start_tool(ompt_target_callbacks_active_t *libomptarget_ompt_enabled) {
-  if (!TCR_4(__kmp_init_middle)) {
-    __kmp_middle_initialize();
-  }
-  bool ret = false;
-  libomptarget_ompt_enabled->enabled = ompt_enabled.enabled;
-  if (ompt_enabled.enabled) {
-    ret = true;
-#define ompt_event_macro(event_name, callback_type, event_id)                  \
-  libomptarget_ompt_enabled->event_name = ompt_target_enabled.event_name;
-
-    FOREACH_OMPT_51_TARGET_EVENT(ompt_event_macro)
-#undef ompt_event_macro
-  }
-  return ret;
-}
-
-void ompt_callback_target_data_op_emi_wrapper(
-    ompt_scope_endpoint_t endpoint, ompt_data_t *target_task_data,
-    ompt_data_t *target_data, ompt_id_t *host_op_id,
-    ompt_target_data_op_t optype, void *src_addr, int src_device_num,
-    void *dest_addr, int dest_device_num, size_t bytes,
-    const void *codeptr_ra) {}
-
-void ompt_callback_target_emi_wrapper(ompt_target_t kind,
-                                      ompt_scope_endpoint_t endpoint,
-                                      int device_num, ompt_data_t *task_data,
-                                      ompt_data_t *target_task_data,
-                                      ompt_data_t *target_data,
-                                      const void *codeptr_ra) {}
-
-void ompt_callback_target_map_emi_wrapper(ompt_data_t *target_data,
-                                          unsigned int nitems, void **host_addr,
-                                          void **device_addr, size_t *bytes,
-                                          unsigned int *mapping_flags,
-                                          const void *codeptr_ra) {}
-
-void ompt_callback_target_submit_emi_wrapper(ompt_scope_endpoint_t endpoint,
-                                             ompt_data_t *target_data,
-                                             ompt_id_t *host_op_id,
-                                             unsigned int requested_num_teams) {
-
-}
-
 #if KMP_OS_DARWIN
 
 // While Darwin supports weak symbols, the library that wishes to provide a new
@@ -346,15 +295,23 @@ ompt_try_start_tool(unsigned int omp_version, const char *runtime_version) {
         OMPT_VERBOSE_INIT_CONTINUED_PRINT("Success. \n");
         OMPT_VERBOSE_INIT_PRINT("Searching for ompt_start_tool in %s... ",
                                 fname);
+        dlerror(); // Clear any existing error
         start_tool = (ompt_start_tool_t)dlsym(h, "ompt_start_tool");
         if (!start_tool) {
-          OMPT_VERBOSE_INIT_CONTINUED_PRINT("Failed: %s\n", dlerror());
+          char *error = dlerror();
+          if (error != NULL) {
+            OMPT_VERBOSE_INIT_CONTINUED_PRINT("Failed: %s\n", error);
+          } else {
+            OMPT_VERBOSE_INIT_CONTINUED_PRINT("Failed: %s\n",
+                                              "ompt_start_tool = NULL");
+          }
         } else
 #elif KMP_OS_WINDOWS
       OMPT_VERBOSE_INIT_PRINT("Opening %s... ", fname);
       HMODULE h = LoadLibrary(fname);
       if (!h) {
-        OMPT_VERBOSE_INIT_CONTINUED_PRINT("Failed: Error %u\n", GetLastError());
+        OMPT_VERBOSE_INIT_CONTINUED_PRINT("Failed: Error %u\n",
+                                          (unsigned)GetLastError());
       } else {
         OMPT_VERBOSE_INIT_CONTINUED_PRINT("Success. \n");
         OMPT_VERBOSE_INIT_PRINT("Searching for ompt_start_tool in %s... ",
@@ -362,7 +319,7 @@ ompt_try_start_tool(unsigned int omp_version, const char *runtime_version) {
         start_tool = (ompt_start_tool_t)GetProcAddress(h, "ompt_start_tool");
         if (!start_tool) {
           OMPT_VERBOSE_INIT_CONTINUED_PRINT("Failed: Error %u\n",
-                                            GetLastError());
+                                            (unsigned)GetLastError());
         } else
 #else
 #error Activation of OMPT is not supported on this platform.
@@ -622,37 +579,7 @@ OMPT_API_ROUTINE ompt_set_result_t ompt_set_callback(ompt_callbacks_t which,
     else                                                                       \
       return ompt_set_always;
 
-    FOREACH_OMPT_HOST_EVENT(ompt_event_macro)
-
-#undef ompt_event_macro
-
-#define ompt_event_macro(event_name, callback_type, event_id)                  \
-  case event_name:                                                             \
-    ompt_target_callbacks.ompt_callback(event_name) = (callback_type)callback; \
-    ompt_target_enabled.event_name = (callback != 0);                          \
-    if (callback)                                                              \
-      return ompt_event_implementation_status(event_name);                     \
-    else                                                                       \
-      return ompt_set_always;
-
-    FOREACH_OMPT_51_TARGET_EVENT(ompt_event_macro)
-
-#undef ompt_event_macro
-
-#define ompt_event_macro(event_name, callback_type, event_id)                  \
-  case event_name:                                                             \
-    ompt_callbacks_noemi.ompt_callback(event_name) = (callback_type)callback;  \
-    ompt_target_enabled.ompt_emi_event(event_name) = (callback != 0);          \
-    if (callback) {                                                            \
-      ompt_target_callbacks.ompt_emi_callback(event_name) =                    \
-          (ompt_emi_callback_type(event_name))(&ompt_emi_wrapper(event_name)); \
-      return ompt_event_implementation_status(event_name);                     \
-    } else {                                                                   \
-      ompt_target_callbacks.ompt_emi_callback(event_name) = NULL;              \
-      return ompt_set_always;                                                  \
-    }
-
-    FOREACH_OMPT_NOEMI_EVENT(ompt_event_macro)
+    FOREACH_OMPT_EVENT(ompt_event_macro)
 
 #undef ompt_event_macro
 
@@ -679,56 +606,7 @@ OMPT_API_ROUTINE int ompt_get_callback(ompt_callbacks_t which,
     return ompt_get_callback_failure;                                          \
   }
 
-    FOREACH_OMPT_HOST_EVENT(ompt_event_macro)
-
-#undef ompt_event_macro
-
-#define ompt_event_macro(event_name, callback_type, event_id)                  \
-  case event_name: {                                                           \
-    ompt_callback_t mycb =                                                     \
-        (ompt_callback_t)ompt_target_callbacks.ompt_callback(event_name);      \
-    if (ompt_target_enabled.event_name && mycb) {                              \
-      *callback = mycb;                                                        \
-      return ompt_get_callback_success;                                        \
-    }                                                                          \
-    return ompt_get_callback_failure;                                          \
-  }
-
-    FOREACH_OMPT_DEVICE_EVENT(ompt_event_macro)
-
-#undef ompt_event_macro
-
-#define ompt_event_macro(event_name, callback_type, event_id)                  \
-  case ompt_emi_event(event_name): {                                           \
-    ompt_callback_t mycb =                                                     \
-        (ompt_callback_t)ompt_target_callbacks.ompt_emi_callback(event_name);  \
-    if (ompt_target_enabled.ompt_emi_event(event_name) &&                      \
-        mycb != (ompt_callback_t)(&ompt_emi_wrapper(event_name))) {            \
-      *callback = mycb;                                                        \
-      return ompt_get_callback_success;                                        \
-    }                                                                          \
-    return ompt_get_callback_failure;                                          \
-  }
-
-    FOREACH_OMPT_NOEMI_EVENT(ompt_event_macro)
-
-#undef ompt_event_macro
-
-#define ompt_event_macro(event_name, callback_type, event_id)                  \
-  case event_name: {                                                           \
-    ompt_callback_t mycb =                                                     \
-        (ompt_callback_t)ompt_callbacks_noemi.ompt_callback(event_name);       \
-    ompt_callback_t wrapper =                                                  \
-        (ompt_callback_t)ompt_target_callbacks.ompt_emi_callback(event_name);  \
-    if (ompt_target_enabled.ompt_emi_event(event_name) &&                      \
-        wrapper == (ompt_callback_t)(&ompt_emi_wrapper(event_name))) {         \
-      *callback = mycb;                                                        \
-      return ompt_get_callback_success;                                        \
-    }                                                                          \
-    return ompt_get_callback_failure;                                          \
-  }
-
-    FOREACH_OMPT_NOEMI_EVENT(ompt_event_macro)
+    FOREACH_OMPT_EVENT(ompt_event_macro)
 
 #undef ompt_event_macro
 
