@@ -10,9 +10,10 @@
 //
 //===----------------------------------------------------------------------===//
 
-import TSCBasic
 import Foundation
 
+import struct TSCBasic.ByteString
+
 extension ByteString {
 
   /// Access the contents of `self` as `Data`. The contents are not copied, so it is not safe to
