@@ -10,9 +10,10 @@
 //
 //===----------------------------------------------------------------------===//
 
-import TSCBasic
 import Foundation
 
+import struct TSCBasic.AbsolutePath
+
 /// The home directory of the current user (same as returned by Foundation's `NSHomeDirectory` method).
 public var homeDirectoryForCurrentUser: AbsolutePath {
   return AbsolutePath(NSHomeDirectory())
@@ -29,3 +30,8 @@ extension AbsolutePath {
     }
   }
 }
+
+/// The directory to write generated module interfaces
+public var defaultDirectoryForGeneratedInterfaces: AbsolutePath {
+  return AbsolutePath(NSTemporaryDirectory()).appending(component: "GeneratedInterfaces")
+}
