@@ -24,11 +24,15 @@ import PackageModel
 import SourceControl
 import SKCore
 import SKSupport
-import TSCBasic
 import Workspace
 import Dispatch
 import struct Foundation.URL
 
+import func TSCBasic.resolveSymlinks
+import protocol TSCBasic.FileSystem
+import struct TSCBasic.AbsolutePath
+import var TSCBasic.localFileSystem
+
 /// Swift Package Manager build system and workspace support.
 ///
 /// This class implements the `BuildSystem` interface to provide the build settings for a Swift
@@ -50,6 +54,8 @@ public final class SwiftPMWorkspace {
 
   let workspacePath: AbsolutePath
   let packageRoot: AbsolutePath
+  /// *Public for testing*
+  public var _packageRoot: AbsolutePath { packageRoot }
   var packageGraph: PackageGraph
   let workspace: Workspace
   public let buildParameters: BuildParameters
@@ -90,7 +96,7 @@ public final class SwiftPMWorkspace {
       throw Error.noManifest(workspacePath: workspacePath)
     }
 
-    self.packageRoot = resolveSymlinks(packageRoot)
+    self.packageRoot = try resolveSymlinks(packageRoot)
 
     guard let destinationToolchainBinDir = toolchainRegistry.default?.swiftc?.parentDirectory else {
         throw Error.cannotDetermineHostToolchain
@@ -99,25 +105,19 @@ public final class SwiftPMWorkspace {
     let destination = try Destination.hostDestination(destinationToolchainBinDir)
     let toolchain = try UserToolchain(destination: destination)
 
-    var location = Workspace.Location(forRootPackage: packageRoot, fileSystem: fileSystem)
-    if let customWorkingDirectory = buildSetup.path {
-        location.workingDirectory = customWorkingDirectory
+    var location = try Workspace.Location(forRootPackage: packageRoot, fileSystem: fileSystem)
+    if let scratchDirectory = buildSetup.path {
+        location.scratchDirectory = scratchDirectory
     }
 
-      // since we are customizing the workspace location, we need to explicitly pass the mirrors configuration
-    let mirrorsConfiguration = try Workspace.Configuration.Mirrors(
-        forRootPackage: packageRoot,
-        sharedMirrorFile: location.sharedMirrorsConfigurationFile,
-        fileSystem: fileSystem
-    )
+    var configuration = WorkspaceConfiguration.default
+    configuration.skipDependenciesUpdates = true
 
     self.workspace = try Workspace(
         fileSystem: fileSystem,
         location: location,
-        mirrors: mirrorsConfiguration.mirrors,
-        customManifestLoader: ManifestLoader(toolchain: toolchain.configuration, cacheDir: location.workingDirectory),
-        resolverUpdateEnabled: false
-    )
+        configuration: configuration,
+        customHostToolchain: toolchain)
 
     let triple = toolchain.triple
 
@@ -130,13 +130,13 @@ public final class SwiftPMWorkspace {
     }
 
     self.buildParameters = BuildParameters(
-        dataPath: location.workingDirectory.appending(component: triple.platformBuildPathComponent()),
+        dataPath: location.scratchDirectory.appending(component: triple.platformBuildPathComponent()),
         configuration: buildConfiguration,
         toolchain: toolchain,
         flags: buildSetup.flags
     )
 
-    self.packageGraph = try PackageGraph(rootPackages: [], dependencies: [])
+    self.packageGraph = try PackageGraph(rootPackages: [], dependencies: [], binaryArtifacts: [:])
 
     try reloadPackage()
   }
@@ -158,7 +158,7 @@ public final class SwiftPMWorkspace {
       log("could not find manifest, or not a SwiftPM package: \(path)", level: .warning)
       return nil
     } catch {
-      log("failed to create \(SwiftPMWorkspace.self): \(error)", level: .error)
+      log("failed to create \(SwiftPMWorkspace.self) at \(url.path): \(error)", level: .error)
       return nil
     }
   }
@@ -228,6 +228,7 @@ extension SwiftPMWorkspace {
       }
     }
     delegate.fileBuildSettingsChanged(changedFiles)
+    delegate.fileHandlingCapabilityChanged()
   }
 }
 
@@ -245,6 +246,8 @@ extension SwiftPMWorkspace: SKCore.BuildSystem {
     return buildPath.appending(components: "index", "db")
   }
 
+  public var indexPrefixMappings: [PathPrefixMapping] { return [] }
+
   /// **Public for testing only**
   public func _settings(
     for uri: DocumentURI,
@@ -269,12 +272,12 @@ extension SwiftPMWorkspace: SKCore.BuildSystem {
       return nil
     }
 
-    if let td = targetDescription(for: path) {
+    if let td = try targetDescription(for: path) {
       return try settings(for: path, language, td)
     }
 
     if path.basename == "Package.swift" {
-      return settings(forPackageManifest: path)
+      return try settings(forPackageManifest: path)
     }
 
     if path.extension == "h" {
@@ -327,13 +330,13 @@ extension SwiftPMWorkspace: SKCore.BuildSystem {
 
   /// Returns the resolved target description for the given file, if one is known.
   /// Must only be called on `queue`.
-  private func targetDescription(for file: AbsolutePath) -> TargetBuildDescription? {
+  private func targetDescription(for file: AbsolutePath) throws -> TargetBuildDescription? {
     dispatchPrecondition(condition: .onQueue(queue))
     if let td = fileToTarget[file] {
       return td
     }
 
-    let realpath = resolveSymlinks(file)
+    let realpath = try resolveSymlinks(file)
     if realpath != file, let td = fileToTarget[realpath] {
       fileToTarget[file] = td
       return td
@@ -351,7 +354,7 @@ extension SwiftPMWorkspace: SKCore.BuildSystem {
     case .created, .deleted:
       return self.workspace.fileAffectsSwiftOrClangBuildSettings(filePath: AbsolutePath(fileURL.path), packageGraph: self.packageGraph)
     case .changed:
-      return false
+      return fileURL.lastPathComponent == "Package.swift"
     default: // Unknown file change type
       return false
     }
@@ -367,6 +370,19 @@ extension SwiftPMWorkspace: SKCore.BuildSystem {
       }
     }
   }
+
+  public func fileHandlingCapability(for uri: DocumentURI) -> FileHandlingCapability {
+    guard let fileUrl = uri.fileURL else {
+      return .unhandled
+    }
+    return self.queue.sync {
+      if (try? targetDescription(for: AbsolutePath(fileUrl.path))) != nil {
+        return .handled
+      } else {
+        return .unhandled
+      }
+    }
+  }
 }
 
 extension SwiftPMWorkspace {
@@ -393,7 +409,7 @@ extension SwiftPMWorkspace {
 
   /// Retrieve settings for a package manifest (Package.swift).
   /// Must only be called on `queue`.
-  private func settings(forPackageManifest path: AbsolutePath) -> FileBuildSettings? {
+  private func settings(forPackageManifest path: AbsolutePath) throws -> FileBuildSettings? {
     dispatchPrecondition(condition: .onQueue(queue))
     func impl(_ path: AbsolutePath) -> FileBuildSettings? {
       for package in packageGraph.packages where path == package.manifest.path {
@@ -407,7 +423,7 @@ extension SwiftPMWorkspace {
       return result
     }
 
-    let canonicalPath = resolveSymlinks(path)
+    let canonicalPath = try resolveSymlinks(path)
     return canonicalPath == path ? nil : impl(canonicalPath)
   }
 
@@ -430,14 +446,14 @@ extension SwiftPMWorkspace {
       return result
     }
 
-    let canonicalPath = resolveSymlinks(path)
+    let canonicalPath = try resolveSymlinks(path)
     return try canonicalPath == path ? nil : impl(canonicalPath)
   }
 
   /// Retrieve settings for the given swift file, which is part of a known target build description.
   public func settings(
     forSwiftFile path: AbsolutePath,
-    _ td: SwiftTargetBuildDescription) throws -> FileBuildSettings?
+    _ td: SwiftTargetBuildDescription) throws -> FileBuildSettings
   {
     // FIXME: this is re-implementing llbuild's constructCommandLineArgs.
     var args: [String] = [
@@ -470,7 +486,7 @@ extension SwiftPMWorkspace {
   public func settings(
     forClangFile path: AbsolutePath,
     _ language: Language,
-    _ td: ClangTargetBuildDescription) throws -> FileBuildSettings?
+    _ td: ClangTargetBuildDescription) throws -> FileBuildSettings
   {
     // FIXME: this is re-implementing things from swiftpm's createClangCompileTarget
 
@@ -480,7 +496,7 @@ extension SwiftPMWorkspace {
         URL(fileURLWithPath: path.pathString).withUnsafeFileSystemRepresentation {
           AbsolutePath(String(cString: $0!))
         }
-    let compilePath = td.compilePaths().first(where: { $0.source == nativePath })
+    let compilePath = try td.compilePaths().first(where: { $0.source == nativePath })
     if let compilePath = compilePath {
       args += [
         "-MD",
@@ -535,7 +551,15 @@ private func findPackageDirectory(
   containing path: AbsolutePath,
   _ fileSystem: FileSystem) -> AbsolutePath? {
   var path = path
-  while !fileSystem.isFile(path.appending(component: "Package.swift")) {
+  while true {
+    let packagePath = path.appending(component: "Package.swift")
+    if fileSystem.isFile(packagePath) {
+      let contents = try? fileSystem.readFileContents(packagePath)
+      if contents?.cString.contains("PackageDescription") == true {
+        return path
+      }
+    }
+
     if path.isRoot {
       return nil
     }
