@@ -16,12 +16,22 @@ import LanguageServerProtocolJSONRPC
 import LSPLogging
 import SKCore
 import SKSupport
-import TSCBasic
+
+import struct TSCBasic.AbsolutePath
 
 #if os(Windows)
 import WinSDK
 #endif
 
+extension NSLock {
+  /// NOTE: Keep in sync with SwiftPM's 'Sources/Basics/NSLock+Extensions.swift'
+  fileprivate func withLock<T>(_ body: () throws -> T) rethrows -> T {
+    lock()
+    defer { unlock() }
+    return try body()
+  }
+}
+
 /// A thin wrapper over a connection to a clangd server providing build setting handling.
 ///
 /// In addition, it also intercepts notifications and replies from clangd in order to do things
@@ -46,7 +56,7 @@ final class ClangLanguageServerShim: LanguageServer, ToolchainLanguageServer {
   private var buildSettingsByFile: [DocumentURI: ClangBuildSettings] = [:]
 
   /// Lock protecting `buildSettingsByFile`.
-  private var lock: Lock = Lock()
+  private var lock: NSLock = NSLock()
 
   /// The current state of the `clangd` language server.
   /// Changing the property automatically notified the state change handlers.
@@ -74,6 +84,10 @@ final class ClangLanguageServerShim: LanguageServer, ToolchainLanguageServer {
   /// Stored so we can replay the initialization when clangd crashes.
   private var initializeRequest: InitializeRequest?
 
+  /// The workspace this `ClangLanguageServer` was opened for.
+  /// `clangd` doesn't have support for multi-root workspaces, so we need to start a separate `clangd` instance for every workspace root.
+  private weak var workspace: Workspace?
+
   /// A callback with which `ClangLanguageServer` can request its owner to reopen all documents in case it has crashed.
   private let reopenDocuments: (ToolchainLanguageServer) -> Void
 
@@ -84,23 +98,34 @@ final class ClangLanguageServerShim: LanguageServer, ToolchainLanguageServer {
   private var clangdPid: Int32?
 #endif
 
-  /// Creates a language server for the given client referencing the clang binary at the given path.
-  public init(
-    client: Connection,
-    clangPath: AbsolutePath?,
-    clangdPath: AbsolutePath,
-    clangdOptions: [String],
+  /// Creates a language server for the given client referencing the clang binary specified in `toolchain`.
+  /// Returns `nil` if `clangd` can't be found.
+  public init?(
+    client: LocalConnection,
+    toolchain: Toolchain,
+    clientCapabilities: ClientCapabilities?,
+    options: SourceKitServer.Options,
+    workspace: Workspace,
     reopenDocuments: @escaping (ToolchainLanguageServer) -> Void
   ) throws {
-    self.clangPath = clangPath
+    guard let clangdPath = toolchain.clangd else {
+      return nil
+    }
+    self.clangPath = toolchain.clang
     self.clangdPath = clangdPath
-    self.clangdOptions = clangdOptions
+    self.clangdOptions = options.clangdOptions
+    self.workspace = workspace
     self.reopenDocuments = reopenDocuments
     self.state = .connected
     super.init(client: client)
     try startClangdProcesss()
   }
 
+  func canHandle(workspace: Workspace) -> Bool {
+    // We launch different clangd instance for each workspace because clangd doesn't have multi-root workspace support.
+    return workspace === self.workspace
+  }
+
   func addStateChangeHandler(handler: @escaping (LanguageServerState, LanguageServerState) -> Void) {
     queue.async {
       self.stateChangeHandlers.append(handler)
@@ -329,13 +354,14 @@ extension ClangLanguageServerShim {
     forwardNotificationToClangdOnQueue(initialized)
   }
 
-  public func shutdown() {
+  public func shutdown(callback: @escaping () -> Void) {
     queue.async {
       _ = self.clangd.send(ShutdownRequest(), queue: self.queue) { [weak self] _ in
         self?.clangd.send(ExitNotification())
         if let localConnection = self?.client as? LocalConnection {
           localConnection.close()
         }
+        callback()
       }
     }
   }
@@ -363,9 +389,7 @@ extension ClangLanguageServerShim {
   }
 
   public func didSaveDocument(_ note: DidSaveTextDocumentNotification) {
-    if capabilities?.textDocumentSync?.save?.isSupported == true {
-      forwardNotificationToClangdOnQueue(note)
-    }
+    forwardNotificationToClangdOnQueue(note)
   }
 
   // MARK: - Build System Integration
@@ -395,7 +419,7 @@ extension ClangLanguageServerShim {
     if let compileCommand = clangBuildSettings?.compileCommand {
       let note = DidChangeConfigurationNotification(settings: .clangd(
         ClangWorkspaceSettings(
-          compilationDatabaseChanges: [url.path: compileCommand])))
+          compilationDatabaseChanges: [AbsolutePath(url.path).pathString: compileCommand])))
       forwardNotificationToClangdOnQueue(note)
     }
   }
@@ -405,7 +429,7 @@ extension ClangLanguageServerShim {
     // with `forceRebuild` set in case any missing header files have been added.
     // This works well for us as the moment since clangd ignores the document version.
     let note = DidChangeTextDocumentNotification(
-      textDocument: VersionedTextDocumentIdentifier(uri, version: nil),
+      textDocument: VersionedTextDocumentIdentifier(uri, version: 0),
       contentChanges: [],
       forceRebuild: true)
     forwardNotificationToClangdOnQueue(note)
@@ -421,6 +445,13 @@ extension ClangLanguageServerShim {
     return true
   }
 
+  /// Returns true if the `ToolchainLanguageServer` will take ownership of the request.
+  public func declaration(_ req: Request<DeclarationRequest>) -> Bool {
+    // We handle it to provide jump-to-header support for #import/#include.
+    forwardRequestToClangdOnQueue(req)
+    return true
+  }
+
   func completion(_ req: Request<CompletionRequest>) {
     forwardRequestToClangdOnQueue(req)
   }
@@ -477,11 +508,8 @@ extension ClangLanguageServerShim {
     forwardRequestToClangdOnQueue(req)
   }
 
-  func inlayHints(_ req: Request<InlayHintsRequest>) {
-    // FIXME: Currently a Swift-specific, non-standard request.
-    // Once inlay hints have been upstreamed to LSP, forward
-    // them to clangd.
-    req.reply(.success([]))
+  func inlayHint(_ req: Request<InlayHintRequest>) {
+    forwardRequestToClangdOnQueue(req)
   }
 
   func foldingRange(_ req: Request<FoldingRangeRequest>) {
@@ -494,6 +522,10 @@ extension ClangLanguageServerShim {
     }
   }
 
+  func openInterface(_ request: Request<OpenInterfaceRequest>) {
+    request.reply(.failure(.unknown("unsupported method")))
+  }
+
   // MARK: - Other
 
   func executeCommand(_ req: Request<ExecuteCommandRequest>) {
@@ -501,30 +533,6 @@ extension ClangLanguageServerShim {
   }
 }
 
-func makeJSONRPCClangServer(
-  client: MessageHandler,
-  toolchain: Toolchain,
-  clangdOptions: [String],
-  reopenDocuments: @escaping (ToolchainLanguageServer) -> Void
-) throws -> ToolchainLanguageServer {
-  guard let clangdPath = toolchain.clangd else {
-    preconditionFailure("missing clang from toolchain \(toolchain.identifier)")
-  }
-
-  let connectionToClient = LocalConnection()
-  connectionToClient.start(handler: client)
-  
-  let shim = try ClangLanguageServerShim(
-    client: connectionToClient,
-    clangPath: toolchain.clang,
-    clangdPath: clangdPath,
-    clangdOptions: clangdOptions,
-    reopenDocuments: reopenDocuments
-  )
-  
-  return shim
-}
-
 /// Clang build settings derived from a `FileBuildSettingsChange`.
 private struct ClangBuildSettings: Equatable {
   /// The compiler arguments, including the program name, argv[0].
