@@ -10,14 +10,16 @@
 //
 //===----------------------------------------------------------------------===//
 
+import Foundation
 import Dispatch
-import struct Foundation.CharacterSet
 import LanguageServerProtocol
 import LSPLogging
 import SKCore
 import SKSupport
 import SourceKitD
-import TSCBasic
+import SwiftSyntax
+import SwiftParser
+
 #if os(Windows)
 import WinSDK
 #endif
@@ -104,6 +106,9 @@ public final class SwiftLanguageServer: ToolchainLanguageServer {
   let clientCapabilities: ClientCapabilities
 
   let serverOptions: SourceKitServer.Options
+  
+  /// Directory where generated Swift interfaces will be stored.
+  let generatedInterfacesPath: URL
 
   // FIXME: ideally we wouldn't need separate management from a parent server in the same process.
   var documentManager: DocumentManager
@@ -133,24 +138,32 @@ public final class SwiftLanguageServer: ToolchainLanguageServer {
   /// A callback with which `SwiftLanguageServer` can request its owner to reopen all documents in case it has crashed.
   private let reopenDocuments: (ToolchainLanguageServer) -> Void
 
-  /// Creates a language server for the given client using the sourcekitd dylib at the specified
-  /// path.
-  /// `reopenDocuments` is a closure that will be called if sourcekitd crashes and the
-  /// `SwiftLanguageServer` asks its parent server to reopen all of its documents.
-  public init(
+  /// Creates a language server for the given client using the sourcekitd dylib specified in `toolchain`.
+  /// `reopenDocuments` is a closure that will be called if sourcekitd crashes and the `SwiftLanguageServer` asks its parent server to reopen all of its documents.
+  /// Returns `nil` if `sourcektid` couldn't be found.
+  public init?(
     client: LocalConnection,
-    sourcekitd: AbsolutePath,
-    clientCapabilities: ClientCapabilities,
-    serverOptions: SourceKitServer.Options,
+    toolchain: Toolchain,
+    clientCapabilities: ClientCapabilities?,
+    options: SourceKitServer.Options,
+    workspace: Workspace,
     reopenDocuments: @escaping (ToolchainLanguageServer) -> Void
   ) throws {
+    guard let sourcekitd = toolchain.sourcekitd else { return nil }
     self.client = client
     self.sourcekitd = try SourceKitDImpl.getOrCreate(dylibPath: sourcekitd)
-    self.clientCapabilities = clientCapabilities
-    self.serverOptions = serverOptions
+    self.clientCapabilities = clientCapabilities ?? ClientCapabilities(workspace: nil, textDocument: nil)
+    self.serverOptions = options
     self.documentManager = DocumentManager()
     self.state = .connected
     self.reopenDocuments = reopenDocuments
+    self.generatedInterfacesPath = options.generatedInterfacesPath.asURL
+    try FileManager.default.createDirectory(at: generatedInterfacesPath, withIntermediateDirectories: true)
+  }
+
+  public func canHandle(workspace: Workspace) -> Bool {
+    // We have a single sourcekitd instance for all workspaces.
+    return true
   }
 
   public func addStateChangeHandler(handler: @escaping (_ oldState: LanguageServerState, _ newState: LanguageServerState) -> Void) {
@@ -161,14 +174,13 @@ public final class SwiftLanguageServer: ToolchainLanguageServer {
 
   /// Updates the lexical tokens for the given `snapshot`.
   /// Must be called on `self.queue`.
-  private func updateLexicalTokens(
-    response: SKDResponseDictionary,
+  private func updateSyntacticTokens(
     for snapshot: DocumentSnapshot
   ) {
     dispatchPrecondition(condition: .onQueue(queue))
 
     let uri = snapshot.document.uri
-    let docTokens = updatedLexicalTokens(response: response, for: snapshot)
+    let docTokens = updateSyntaxTree(for: snapshot)
 
     do {
       try documentManager.updateTokens(uri, tokens: docTokens)
@@ -178,31 +190,13 @@ public final class SwiftLanguageServer: ToolchainLanguageServer {
   }
 
   /// Returns the updated lexical tokens for the given `snapshot`.
-  private func updatedLexicalTokens(
-    response: SKDResponseDictionary,
+  private func updateSyntaxTree(
     for snapshot: DocumentSnapshot
   ) -> DocumentTokens {
     logExecutionTime(level: .debug) {
       var docTokens = snapshot.tokens
 
-      guard let offset: Int = response[keys.offset],
-            let length: Int = response[keys.length],
-            let start: Position = snapshot.positionOf(utf8Offset: offset),
-            let end: Position = snapshot.positionOf(utf8Offset: offset + length) else {
-        // This e.g. happens in the case of empty edits
-        log("did not update lexical/syntactic tokens, no range found", level: .debug)
-        return docTokens
-      }
-
-      let range = start..<end
-
-      if let syntaxMap: SKDResponseArray = response[keys.syntaxmap] {
-        let tokenParser = SyntaxHighlightingTokenParser(sourcekitd: sourcekitd)
-        var tokens: [SyntaxHighlightingToken] = []
-        tokenParser.parseTokens(syntaxMap, in: snapshot, into: &tokens)
-
-        docTokens.replaceLexical(in: range, with: tokens)
-      }
+      docTokens.syntaxTree = Parser.parse(source: snapshot.text)
 
       return docTokens
     }
@@ -257,6 +251,20 @@ public final class SwiftLanguageServer: ToolchainLanguageServer {
     }
   }
 
+  /// Shift the ranges of all current diagnostics in the document with the given `uri` to account for `edit`.
+  private func adjustDiagnosticRanges(of uri: DocumentURI, for edit: TextDocumentContentChangeEvent) {
+    guard let rangeAdjuster = RangeAdjuster(edit: edit) else {
+      return
+    }
+    currentDiagnostics[uri] = currentDiagnostics[uri]?.compactMap({ cachedDiag in
+      if let adjustedRange = rangeAdjuster.adjust(cachedDiag.diagnostic.range) {
+        return cachedDiag.withRange(adjustedRange)
+      } else {
+        return nil
+      }
+    })
+  }
+  
   /// Publish diagnostics for the given `snapshot`. We withhold semantic diagnostics if we are using
   /// fallback arguments.
   ///
@@ -334,21 +342,19 @@ extension SwiftLanguageServer {
     sourcekitd.addNotificationHandler(self)
 
     return InitializeResult(capabilities: ServerCapabilities(
-      textDocumentSync: TextDocumentSyncOptions(
+      textDocumentSync: .options(TextDocumentSyncOptions(
         openClose: true,
-        change: .incremental,
-        willSave: true,
-        willSaveWaitUntil: false,
-        save: .value(TextDocumentSyncOptions.SaveOptions(includeText: false))),
-      hoverProvider: true,
+        change: .incremental
+      )),
+      hoverProvider: .bool(true),
       completionProvider: CompletionOptions(
         resolveProvider: false,
         triggerCharacters: [".", "("]),
       definitionProvider: nil,
       implementationProvider: .bool(true),
       referencesProvider: nil,
-      documentHighlightProvider: true,
-      documentSymbolProvider: true,
+      documentHighlightProvider: .bool(true),
+      documentSymbolProvider: .bool(true),
       codeActionProvider: .value(CodeActionServerCapabilities(
         clientCapabilities: initialize.capabilities.textDocument?.codeAction,
         codeActionOptions: CodeActionOptions(codeActionKinds: [.quickFix, .refactor]),
@@ -362,7 +368,9 @@ extension SwiftLanguageServer {
           tokenTypes: SyntaxHighlightingToken.Kind.allCases.map(\.lspName),
           tokenModifiers: SyntaxHighlightingToken.Modifiers.allModifiers.map { $0.lspName! }),
         range: .bool(true),
-        full: .bool(true))
+        full: .bool(true)),
+      inlayHintProvider: .value(InlayHintOptions(
+        resolveProvider: false))
     ))
   }
 
@@ -370,7 +378,7 @@ extension SwiftLanguageServer {
     // Nothing to do.
   }
 
-  public func shutdown() {
+  public func shutdown(callback: @escaping () -> Void) {
     queue.async {
       if let session = self.currentCompletionSession {
         session.close()
@@ -378,6 +386,7 @@ extension SwiftLanguageServer {
       }
       self.sourcekitd.removeNotificationHandler(self)
       self.client.close()
+      callback()
     }
   }
 
@@ -415,7 +424,7 @@ extension SwiftLanguageServer {
     }
     self.publishDiagnostics(
         response: dict, for: snapshot, compileCommand: compileCmd)
-    self.updateLexicalTokens(response: dict, for: snapshot)
+    self.updateSyntacticTokens(for: snapshot)
   }
 
   public func documentUpdatedBuildSettings(_ uri: DocumentURI, change: FileBuildSettingsChange) {
@@ -480,7 +489,7 @@ extension SwiftLanguageServer {
         return
       }
       self.publishDiagnostics(response: dict, for: snapshot, compileCommand: compileCommand)
-      self.updateLexicalTokens(response: dict, for: snapshot)
+      self.updateSyntacticTokens(for: snapshot)
     }
   }
 
@@ -531,9 +540,11 @@ extension SwiftLanguageServer {
 
         req[keys.sourcetext] = edit.text
         lastResponse = try? self.sourcekitd.sendSync(req)
+
+        self.adjustDiagnosticRanges(of: note.textDocument.uri, for: edit)
       } updateDocumentTokens: { (after: DocumentSnapshot) in
-        if let dict = lastResponse {
-          return self.updatedLexicalTokens(response: dict, for: after)
+        if lastResponse != nil {
+          return self.updateSyntaxTree(for: after)
         } else {
           return DocumentTokens()
         }
@@ -562,6 +573,11 @@ extension SwiftLanguageServer {
     return false
   }
 
+  public func declaration(_ request: Request<DeclarationRequest>) -> Bool {
+    // We don't handle it.
+    return false
+  }
+
   public func completion(_ req: Request<CompletionRequest>) {
     queue.async {
       self._completion(req)
@@ -573,7 +589,7 @@ extension SwiftLanguageServer {
     let position = req.params.position
     cursorInfo(uri, position..<position) { result in
       guard let cursorInfo: CursorInfo = result.success ?? nil else {
-        if let error = result.failure, error != .responseError(.cancelled) {
+        if let error = result.failure, error != .responseError(.serverCancelled) {
           log("cursor info failed \(uri):\(position): \(error)", level: .warning)
         }
         return req.reply(nil)
@@ -636,14 +652,23 @@ extension SwiftLanguageServer {
       return completion(.failure(.unknown(msg)))
     }
 
+    let helperDocumentName = "DocumentSymbols:" + snapshot.document.uri.pseudoPath
     let skreq = SKDRequestDictionary(sourcekitd: self.sourcekitd)
     skreq[keys.request] = self.requests.editor_open
-    skreq[keys.name] = "DocumentSymbols:" + snapshot.document.uri.pseudoPath
+    skreq[keys.name] = helperDocumentName
     skreq[keys.sourcetext] = snapshot.text
     skreq[keys.syntactic_only] = 1
 
     let handle = self.sourcekitd.send(skreq, self.queue) { [weak self] result in
       guard let self = self else { return }
+
+      defer {
+        let closeHelperReq = SKDRequestDictionary(sourcekitd: self.sourcekitd)
+        closeHelperReq[self.keys.request] = self.requests.editor_close
+        closeHelperReq[self.keys.name] = helperDocumentName
+        _ = self.sourcekitd.send(closeHelperReq, .global(qos: .utility), reply: { _ in })
+      }
+
       guard let dict = result.success else {
         return completion(.failure(ResponseError(result.failure!)))
       }
@@ -733,14 +758,23 @@ extension SwiftLanguageServer {
         return
       }
 
+      let helperDocumentName = "DocumentColor:" + snapshot.document.uri.pseudoPath
       let skreq = SKDRequestDictionary(sourcekitd: self.sourcekitd)
       skreq[keys.request] = self.requests.editor_open
-      skreq[keys.name] = "DocumentColor:" + snapshot.document.uri.pseudoPath
+      skreq[keys.name] = helperDocumentName
       skreq[keys.sourcetext] = snapshot.text
       skreq[keys.syntactic_only] = 1
 
       let handle = self.sourcekitd.send(skreq, self.queue) { [weak self] result in
         guard let self = self else { return }
+
+        defer {
+          let closeHelperReq = SKDRequestDictionary(sourcekitd: self.sourcekitd)
+          closeHelperReq[keys.request] = self.requests.editor_close
+          closeHelperReq[keys.name] = helperDocumentName
+          _ = self.sourcekitd.send(closeHelperReq, .global(qos: .utility), reply: { _ in })
+        }
+
         guard let dict = result.success else {
           req.reply(.failure(ResponseError(result.failure!)))
           return
@@ -828,7 +862,7 @@ extension SwiftLanguageServer {
         return
       }
 
-      let tokens = snapshot.tokens.mergedAndSorted
+      let tokens = snapshot.mergedAndSortedTokens()
       let encodedTokens = tokens.lspEncoded
 
       req.reply(DocumentSemanticTokensResponse(data: encodedTokens))
@@ -851,7 +885,7 @@ extension SwiftLanguageServer {
         return
       }
 
-      let tokens = snapshot.tokens.mergedAndSorted.filter { $0.range.overlaps(range) }
+      let tokens = snapshot.mergedAndSortedTokens(in: range)
       let encodedTokens = tokens.lspEncoded
 
       req.reply(DocumentSemanticTokensResponse(data: encodedTokens))
@@ -930,8 +964,6 @@ extension SwiftLanguageServer {
   }
 
   public func foldingRange(_ req: Request<FoldingRangeRequest>) {
-    let keys = self.keys
-
     queue.async {
       guard let snapshot = self.documentManager.latestSnapshot(req.params.textDocument.uri) else {
         log("failed to find snapshot for url \(req.params.textDocument.uri)")
@@ -939,125 +971,193 @@ extension SwiftLanguageServer {
         return
       }
 
-      let skreq = SKDRequestDictionary(sourcekitd: self.sourcekitd)
-      skreq[keys.request] = self.requests.editor_open
-      skreq[keys.name] = "FoldingRanges:" + snapshot.document.uri.pseudoPath
-      skreq[keys.sourcetext] = snapshot.text
-      skreq[keys.syntactic_only] = 1
-
-      let handle = self.sourcekitd.send(skreq, self.queue) { [weak self] result in
-        guard let self = self else { return }
-        guard let dict = result.success else {
-          req.reply(.failure(ResponseError(result.failure!)))
-          return
-        }
-
-        guard let syntaxMap: SKDResponseArray = dict[self.keys.syntaxmap],
-              let substructure: SKDResponseArray = dict[self.keys.substructure] else {
-          return req.reply([])
-        }
+      guard let sourceFile = snapshot.tokens.syntaxTree else {
+        log("no lexical structure available for url \(req.params.textDocument.uri)")
+        req.reply(nil)
+        return
+      }
 
+      final class FoldingRangeFinder: SyntaxVisitor {
+        private let snapshot: DocumentSnapshot
         /// Some ranges might occur multiple times.
         /// E.g. for `print("hi")`, `"hi"` is both the range of all call arguments and the range the first argument in the call.
         /// It doesn't make sense to report them multiple times, so use a `Set` here.
-        var ranges: Set<FoldingRange> = []
+        private var ranges: Set<FoldingRange>
+        /// The client-imposed limit on the number of folding ranges it would
+        /// prefer to recieve from the LSP server. If the value is `nil`, there
+        /// is no preset limit.
+        private var rangeLimit: Int?
+        /// If `true`, the client is only capable of folding entire lines. If
+        /// `false` the client can handle folding ranges.
+        private var lineFoldingOnly: Bool
 
-        var hasReachedLimit: Bool {
-          let capabilities = self.clientCapabilities.textDocument?.foldingRange
-          guard let rangeLimit = capabilities?.rangeLimit else {
-            return false
+        init(snapshot: DocumentSnapshot, rangeLimit: Int?, lineFoldingOnly: Bool) {
+          self.snapshot = snapshot
+          self.ranges = []
+          self.rangeLimit = rangeLimit
+          self.lineFoldingOnly = lineFoldingOnly
+          super.init(viewMode: .sourceAccurate)
+        }
+
+        override func visit(_ node: TokenSyntax) -> SyntaxVisitorContinueKind {
+          // Index comments, so we need to see at least '/*', or '//'.
+          if node.leadingTriviaLength.utf8Length > 2 {
+            self.addTrivia(from: node, node.leadingTrivia)
           }
-          return ranges.count >= rangeLimit
-        }
 
-        // If the limit is less than one, do nothing.
-        guard hasReachedLimit == false else {
-          req.reply([])
-          return
-        }
-
-        // Merge successive comments into one big comment by adding their lengths.
-        var currentComment: (offset: Int, length: Int)? = nil
-
-        syntaxMap.forEach { _, value in
-          if let kind: sourcekitd_uid_t = value[self.keys.kind],
-             kind.isCommentKind(self.values),
-             let offset: Int = value[self.keys.offset],
-             let length: Int = value[self.keys.length]
-          {
-            if let comment = currentComment, comment.offset + comment.length == offset {
-              currentComment!.length += length
-              return true
-            }
-            if let comment = currentComment {
-              self.addFoldingRange(offset: comment.offset, length: comment.length, kind: .comment, in: snapshot, toSet: &ranges)
-            }
-            currentComment = (offset: offset, length: length)
+          if node.trailingTriviaLength.utf8Length > 2 {
+            self.addTrivia(from: node, node.trailingTrivia)
           }
-          return hasReachedLimit == false
+
+          return .visitChildren
         }
 
-        // Add the last stored comment.
-        if let comment = currentComment, hasReachedLimit == false {
-          self.addFoldingRange(offset: comment.offset, length: comment.length, kind: .comment, in: snapshot, toSet: &ranges)
-          currentComment = nil
-        }
+        private func addTrivia(from node: TokenSyntax, _ trivia: Trivia) {
+          var start = node.position.utf8Offset
+          var lineCommentStart: Int? = nil
+          func flushLineComment(_ offset: Int = 0) {
+            if let lineCommentStart = lineCommentStart {
+              _ = self.addFoldingRange(
+                start: lineCommentStart,
+                end: start + offset,
+                kind: .comment)
+            }
+            lineCommentStart = nil
+          }
 
-        var structureStack: [SKDResponseArray] = [substructure]
-        while !hasReachedLimit, let substructure = structureStack.popLast() {
-          substructure.forEach { _, value in
-            if let offset: Int = value[self.keys.bodyoffset],
-               let length: Int = value[self.keys.bodylength],
-               length > 0
-            {
-              self.addFoldingRange(offset: offset, length: length, in: snapshot, toSet: &ranges)
-              if hasReachedLimit {
-                return false
+          for piece in node.leadingTrivia {
+            defer { start += piece.sourceLength.utf8Length }
+            switch piece {
+            case .blockComment(_):
+              flushLineComment()
+              _ = self.addFoldingRange(
+                start: start,
+                end: start + piece.sourceLength.utf8Length,
+                kind: .comment)
+            case .docBlockComment(_):
+              flushLineComment()
+              _ = self.addFoldingRange(
+                start: start,
+                end: start + piece.sourceLength.utf8Length,
+                kind: .comment)
+            case .lineComment(_), .docLineComment(_):
+              if lineCommentStart == nil {
+                lineCommentStart = start
               }
+            case .newlines(1), .carriageReturns(1), .spaces(_), .tabs(_):
+              if lineCommentStart != nil {
+                continue
+              } else {
+                flushLineComment()
+              }
+            default:
+              flushLineComment()
+              continue
             }
-            if let substructure: SKDResponseArray = value[self.keys.substructure] {
-              structureStack.append(substructure)
-            }
-            return true
           }
+
+          flushLineComment()
         }
 
-        req.reply(ranges.sorted())
+        override func visit(_ node: CodeBlockSyntax) -> SyntaxVisitorContinueKind {
+          return self.addFoldingRange(
+            start: node.statements.position.utf8Offset,
+            end: node.rightBrace.positionAfterSkippingLeadingTrivia.utf8Offset)
+        }
+
+        override func visit(_ node: MemberDeclBlockSyntax) -> SyntaxVisitorContinueKind {
+          return self.addFoldingRange(
+            start: node.members.position.utf8Offset,
+            end: node.rightBrace.positionAfterSkippingLeadingTrivia.utf8Offset)
+        }
+
+        override func visit(_ node: ClosureExprSyntax) -> SyntaxVisitorContinueKind {
+          return self.addFoldingRange(
+            start: node.statements.position.utf8Offset,
+            end: node.rightBrace.positionAfterSkippingLeadingTrivia.utf8Offset)
+        }
+
+        override func visit(_ node: AccessorBlockSyntax) -> SyntaxVisitorContinueKind {
+          return self.addFoldingRange(
+            start: node.accessors.position.utf8Offset,
+            end: node.rightBrace.positionAfterSkippingLeadingTrivia.utf8Offset)
+        }
+
+        override func visit(_ node: SwitchStmtSyntax) -> SyntaxVisitorContinueKind {
+          return self.addFoldingRange(
+            start: node.cases.position.utf8Offset,
+            end: node.rightBrace.positionAfterSkippingLeadingTrivia.utf8Offset)
+        }
+
+        override func visit(_ node: FunctionCallExprSyntax) -> SyntaxVisitorContinueKind {
+          return self.addFoldingRange(
+            start: node.argumentList.position.utf8Offset,
+            end: node.argumentList.endPosition.utf8Offset)
+        }
+
+        override func visit(_ node: SubscriptExprSyntax) -> SyntaxVisitorContinueKind {
+          return self.addFoldingRange(
+            start: node.argumentList.position.utf8Offset,
+            end: node.argumentList.endPosition.utf8Offset)
+        }
+
+        __consuming func finalize() -> Set<FoldingRange> {
+          return self.ranges
+        }
+
+        private func addFoldingRange(start: Int, end: Int, kind: FoldingRangeKind? = nil) -> SyntaxVisitorContinueKind {
+          if let limit = self.rangeLimit, self.ranges.count >= limit {
+            return .skipChildren
+          }
+
+          guard let start: Position = snapshot.positionOf(utf8Offset: start),
+                let end: Position = snapshot.positionOf(utf8Offset: end) else {
+            log("folding range failed to retrieve position of \(snapshot.document.uri): \(start)-\(end)", level: .warning)
+            return .visitChildren
+          }
+          let range: FoldingRange
+          if lineFoldingOnly {
+            // Since the client cannot fold less than a single line, if the
+            // fold would span 1 line there's no point in reporting it.
+            guard end.line > start.line else {
+              return .visitChildren
+            }
+
+            // If the client only supports folding full lines, don't report
+            // the end of the range since there's nothing they could do with it.
+            range = FoldingRange(startLine: start.line,
+                                 startUTF16Index: nil,
+                                 endLine: end.line,
+                                 endUTF16Index: nil,
+                                 kind: kind)
+          } else {
+            range = FoldingRange(startLine: start.line,
+                                 startUTF16Index: start.utf16index,
+                                 endLine: end.line,
+                                 endUTF16Index: end.utf16index,
+                                 kind: kind)
+          }
+          ranges.insert(range)
+          return .visitChildren
+        }
       }
 
-      // FIXME: cancellation
-      _ = handle
-    }
-  }
-
-  func addFoldingRange(offset: Int, length: Int, kind: FoldingRangeKind? = nil, in snapshot: DocumentSnapshot, toSet ranges: inout Set<FoldingRange>) {
-    guard let start: Position = snapshot.positionOf(utf8Offset: offset),
-          let end: Position = snapshot.positionOf(utf8Offset: offset + length) else {
-      log("folding range failed to retrieve position of \(snapshot.document.uri): \(offset)-\(offset + length)", level: .warning)
-      return
-    }
-    let capabilities = clientCapabilities.textDocument?.foldingRange
-    let range: FoldingRange
-    // If the client only supports folding full lines, ignore the end character's line.
-    if capabilities?.lineFoldingOnly == true {
-      let lastLineToFold = end.line - 1
-      if lastLineToFold <= start.line {
+      let capabilities = self.clientCapabilities.textDocument?.foldingRange
+      // If the limit is less than one, do nothing.
+      if let limit = capabilities?.rangeLimit, limit <= 0 {
+        req.reply([])
         return
-      } else {
-        range = FoldingRange(startLine: start.line,
-                             startUTF16Index: nil,
-                             endLine: lastLineToFold,
-                             endUTF16Index: nil,
-                             kind: kind)
       }
-    } else {
-      range = FoldingRange(startLine: start.line,
-                           startUTF16Index: start.utf16index,
-                           endLine: end.line,
-                           endUTF16Index: end.utf16index,
-                           kind: kind)
+
+      let rangeFinder = FoldingRangeFinder(
+        snapshot: snapshot,
+        rangeLimit: capabilities?.rangeLimit,
+        lineFoldingOnly: capabilities?.lineFoldingOnly ?? false)
+      rangeFinder.walk(sourceFile)
+      let ranges = rangeFinder.finalize()
+
+      req.reply(ranges.sorted())
     }
-    ranges.insert(range)
   }
 
   public func codeAction(_ req: Request<CodeActionRequest>) {
@@ -1197,12 +1297,7 @@ extension SwiftLanguageServer {
     completion(.success(codeActions))
   }
 
-  public func inlayHints(_ req: Request<InlayHintsRequest>) {
-    guard req.params.only?.contains(.type) ?? true else {
-      req.reply([])
-      return
-    }
-
+  public func inlayHint(_ req: Request<InlayHintRequest>) {
     let uri = req.params.textDocument.uri
     variableTypeInfos(uri, req.params.range) { infosResult in
       do {
@@ -1210,11 +1305,16 @@ extension SwiftLanguageServer {
         let hints = infos
           .lazy
           .filter { !$0.hasExplicitType }
-          .map { info in
-            InlayHint(
-              position: info.range.upperBound,
-              category: .type,
-              label: info.printedType
+          .map { info -> InlayHint in
+            let position = info.range.upperBound
+            let label = ": \(info.printedType)"
+            return InlayHint(
+              position: position,
+              label: .string(label),
+              kind: .type,
+              textEdits: [
+                TextEdit(range: position..<position, newText: label)
+              ]
             )
           }
 
@@ -1335,7 +1435,7 @@ extension SwiftLanguageServer: SKDNotificationHandler {
         // unhandled.
 #if os(Windows)
         let isPath: Bool = name.withCString(encodedAs: UTF16.self) {
-          PathIsUNCW($0) || (0...25) ~= PathGetDriveNumberW($0)
+          !PathIsURLW($0)
         }
 #else
         let isPath: Bool = name.starts(with: "/")
@@ -1384,25 +1484,6 @@ extension DocumentSnapshot {
   }
 }
 
-func makeLocalSwiftServer(
-  client: MessageHandler,
-  sourcekitd: AbsolutePath,
-  clientCapabilities: ClientCapabilities?,
-  options: SourceKitServer.Options,
-  reopenDocuments: @escaping (ToolchainLanguageServer) -> Void
-) throws -> ToolchainLanguageServer {
-  let connectionToClient = LocalConnection()
-
-  let server = try SwiftLanguageServer(
-    client: connectionToClient,
-    sourcekitd: sourcekitd,
-    clientCapabilities: clientCapabilities ?? ClientCapabilities(workspace: nil, textDocument: nil),
-    serverOptions: options,
-    reopenDocuments: reopenDocuments)
-  connectionToClient.start(handler: client)
-  return server
-}
-
 extension sourcekitd_uid_t {
   func isCommentKind(_ vals: sourcekitd_values) -> Bool {
     switch self {
@@ -1500,6 +1581,8 @@ extension sourcekitd_uid_t {
       case vals.decl_extension:
         // There are no extensions in LSP, so I return something vaguely similar
         return .namespace
+      case vals.ref_module:
+        return .module
       default:
         return nil
     }
