@@ -12,9 +12,10 @@
 
 import SKCore
 import TSCBasic
-import TSCUtility
 import XCTest
 
+import enum TSCUtility.Platform
+
 final class ToolchainRegistryTests: XCTestCase {
   func testDefaultBasic() {
     let tr = ToolchainRegistry()
@@ -197,7 +198,13 @@ final class ToolchainRegistryTests: XCTestCase {
     XCTAssertNil(tr.default)
     XCTAssert(tr.toolchains.isEmpty)
 
-    try! ProcessEnv.setVar("SOURCEKIT_PATH", value: "/bogus:\(binPath):/bogus2")
+#if os(Windows)
+    let separator: String = ";"
+#else
+    let separator: String = ":"
+#endif
+
+    try! ProcessEnv.setVar("SOURCEKIT_PATH", value: ["/bogus", binPath.pathString, "/bogus2"].joined(separator: separator))
     defer { try! ProcessEnv.setVar("SOURCEKIT_PATH", value: "") }
 
     tr.scanForToolchains(fs)
@@ -216,7 +223,7 @@ final class ToolchainRegistryTests: XCTestCase {
     XCTAssertNil(tc.libIndexStore)
 
     let binPath2 = AbsolutePath("/other/my_toolchain/bin")
-    try! ProcessEnv.setVar("SOME_TEST_ENV_PATH", value: "/bogus:\(binPath2):/bogus2")
+    try! ProcessEnv.setVar("SOME_TEST_ENV_PATH", value: ["/bogus", binPath2.pathString, "bogus2"].joined(separator: separator))
     makeToolchain(binPath: binPath2, fs, sourcekitd: true)
     tr.scanForToolchains(pathVariables: ["NOPE", "SOME_TEST_ENV_PATH", "MORE_NOPE"], fs)
 
@@ -303,9 +310,17 @@ final class ToolchainRegistryTests: XCTestCase {
 
       let t1 = Toolchain(path.parentDirectory, fs)!
       XCTAssertNotNil(t1.sourcekitd)
+#if os(Windows)
+      // Windows does not have file permissions but rather checks the contents
+      // which have been written out.
+      XCTAssertNotNil(t1.clang)
+      XCTAssertNotNil(t1.clangd)
+      XCTAssertNotNil(t1.swiftc)
+#else
       XCTAssertNil(t1.clang)
       XCTAssertNil(t1.clangd)
       XCTAssertNil(t1.swiftc)
+#endif
 
 #if !os(Windows)
       func chmodRX(_ path: AbsolutePath) {
@@ -507,12 +522,25 @@ private func makeToolchain(
   precondition(!clang && !swiftc && !clangd || fs === localFileSystem || !shouldChmod,
     "Cannot make toolchain binaries exectuable with InMemoryFileSystem")
 
+  // tiny PE binary from: https://archive.is/w01DO
+  let contents: [UInt8] = [
+    0x4d, 0x5a, 0x00, 0x00, 0x50, 0x45, 0x00, 0x00, 0x4c, 0x01, 0x01, 0x00,
+    0x6a, 0x2a, 0x58, 0xc3, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
+    0x04, 0x00, 0x03, 0x01, 0x0b, 0x01, 0x08, 0x00, 0x04, 0x00, 0x00, 0x00,
+    0x00, 0x00, 0x00, 0x00, 0x04, 0x00, 0x00, 0x00, 0x0c, 0x00, 0x00, 0x00,
+    0x04, 0x00, 0x00, 0x00, 0x0c, 0x00, 0x00, 0x00, 0x00, 0x00, 0x40, 0x00,
+    0x04, 0x00, 0x00, 0x00, 0x04, 0x00, 0x00, 0x00, 0x04, 0x00, 0x00, 0x00,
+    0x00, 0x00, 0x00, 0x00, 0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
+    0x68, 0x00, 0x00, 0x00, 0x64, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
+    0x02
+  ]
+
   let libPath = binPath.parentDirectory.appending(component: "lib")
   try! fs.createDirectory(binPath, recursive: true)
   try! fs.createDirectory(libPath)
 
   let makeExec = { (path: AbsolutePath) in
-    try! fs.writeFileContents(path , bytes: "")
+    try! fs.writeFileContents(path, bytes: ByteString(contents))
 #if !os(Windows)
     if shouldChmod {
       XCTAssertEqual(chmod(path.pathString, S_IRUSR | S_IXUSR), 0)
@@ -532,16 +560,25 @@ private func makeToolchain(
     makeExec(binPath.appending(component: "swiftc\(execExt)"))
   }
 
-  let dylibExt = Platform.currentPlatform?.dynamicLibraryExtension ?? ".so"
+  let dylibSuffix = Platform.currentPlatform?.dynamicLibraryExtension ?? ".so"
 
   if sourcekitd {
     try! fs.createDirectory(libPath.appending(component: "sourcekitd.framework"))
     try! fs.writeFileContents(libPath.appending(components: "sourcekitd.framework", "sourcekitd") , bytes: "")
   }
   if sourcekitdInProc {
-    try! fs.writeFileContents(libPath.appending(component: "libsourcekitdInProc\(dylibExt)") , bytes: "")
+#if os(Windows)
+    try! fs.writeFileContents(binPath.appending(component: "sourcekitdInProc\(dylibSuffix)") , bytes: "")
+#else
+    try! fs.writeFileContents(libPath.appending(component: "libsourcekitdInProc\(dylibSuffix)") , bytes: "")
+#endif
   }
   if libIndexStore {
-    try! fs.writeFileContents(libPath.appending(component: "libIndexStore\(dylibExt)") , bytes: "")
+#if os(Windows)
+    // Windows has a prefix of `lib` on this particular library ...
+    try! fs.writeFileContents(binPath.appending(component: "libIndexStore\(dylibSuffix)") , bytes: "")
+#else
+    try! fs.writeFileContents(libPath.appending(component: "libIndexStore\(dylibSuffix)") , bytes: "")
+#endif
   }
 }
