@@ -20,9 +20,10 @@ import SKCore
 import SKSwiftPMWorkspace
 import SKTestSupport
 import TSCBasic
-import TSCUtility
 import XCTest
 
+import struct PackageModel.BuildFlags
+
 final class SwiftPMWorkspaceTests: XCTestCase {
 
   func testNoPackage() {
@@ -96,7 +97,7 @@ final class SwiftPMWorkspaceTests: XCTestCase {
               targets: [.target(name: "lib", dependencies: [])])
             """
       ])
-      let packageRoot = resolveSymlinks(tempDir.appending(component: "pkg"))
+      let packageRoot = try resolveSymlinks(tempDir.appending(component: "pkg"))
       let tr = ToolchainRegistry.shared
       let ws = try! SwiftPMWorkspace(
         workspacePath: packageRoot,
@@ -106,7 +107,7 @@ final class SwiftPMWorkspaceTests: XCTestCase {
 
       let aswift = packageRoot.appending(components: "Sources", "lib", "a.swift")
       let hostTriple = ws.buildParameters.triple
-      let build = buildPath(root: packageRoot, triple: hostTriple)
+      let build = buildPath(root: packageRoot, platform: hostTriple.platformBuildPathComponent())
 
       XCTAssertEqual(ws.buildPath, build)
       XCTAssertNotNil(ws.indexStorePath)
@@ -152,7 +153,7 @@ final class SwiftPMWorkspaceTests: XCTestCase {
       let config = BuildSetup(
           configuration: .release,
           path: packageRoot.appending(component: "non_default_build_path"),
-          flags: BuildFlags(xcc: ["-m32"], xcxx: [], xswiftc: ["-typecheck"], xlinker: []))
+          flags: BuildFlags(cCompilerFlags: ["-m32"], swiftCompilerFlags: ["-typecheck"]))
 
       let ws = try! SwiftPMWorkspace(
         workspacePath: packageRoot,
@@ -162,7 +163,7 @@ final class SwiftPMWorkspaceTests: XCTestCase {
 
       let aswift = packageRoot.appending(components: "Sources", "lib", "a.swift")
       let hostTriple = ws.buildParameters.triple
-      let build = buildPath(root: packageRoot, config: config, triple: hostTriple)
+      let build = buildPath(root: packageRoot, config: config, platform: hostTriple.platformBuildPathComponent())
 
       XCTAssertEqual(ws.buildPath, build)
       let arguments = try ws._settings(for: aswift.asURI, .swift)!.compilerArguments
@@ -194,7 +195,7 @@ final class SwiftPMWorkspaceTests: XCTestCase {
         fileSystem: fs,
         buildSetup: TestSourceKitServer.serverOptions.buildSetup)
 
-      let source = resolveSymlinks(packageRoot.appending(component: "Package.swift"))
+      let source = try resolveSymlinks(packageRoot.appending(component: "Package.swift"))
       let arguments = try ws._settings(for: source.asURI, .swift)!.compilerArguments
 
       check("-swift-version", "4.2", arguments: arguments)
@@ -216,7 +217,7 @@ final class SwiftPMWorkspaceTests: XCTestCase {
               targets: [.target(name: "lib", dependencies: [])])
             """
       ])
-      let packageRoot = resolveSymlinks(tempDir.appending(component: "pkg"))
+      let packageRoot = try resolveSymlinks(tempDir.appending(component: "pkg"))
       let tr = ToolchainRegistry.shared
       let ws = try! SwiftPMWorkspace(
         workspacePath: packageRoot,
@@ -256,7 +257,7 @@ final class SwiftPMWorkspaceTests: XCTestCase {
               ])
             """
       ])
-      let packageRoot = resolveSymlinks(tempDir.appending(component: "pkg"))
+      let packageRoot = try resolveSymlinks(tempDir.appending(component: "pkg"))
       let tr = ToolchainRegistry.shared
       let ws = try! SwiftPMWorkspace(
         workspacePath: packageRoot,
@@ -338,7 +339,7 @@ final class SwiftPMWorkspaceTests: XCTestCase {
               cxxLanguageStandard: .cxx14)
             """
       ])
-      let packageRoot = resolveSymlinks(tempDir.appending(component: "pkg"))
+      let packageRoot = try resolveSymlinks(tempDir.appending(component: "pkg"))
       let tr = ToolchainRegistry.shared
       let ws = try! SwiftPMWorkspace(
         workspacePath: packageRoot,
@@ -349,7 +350,7 @@ final class SwiftPMWorkspaceTests: XCTestCase {
       let acxx = packageRoot.appending(components: "Sources", "lib", "a.cpp")
       let bcxx = packageRoot.appending(components: "Sources", "lib", "b.cpp")
       let hostTriple = ws.buildParameters.triple
-      let build = buildPath(root: packageRoot, triple: hostTriple)
+      let build = buildPath(root: packageRoot, platform: hostTriple.platformBuildPathComponent())
 
       XCTAssertEqual(ws.buildPath, build)
       XCTAssertNotNil(ws.indexStorePath)
@@ -396,7 +397,7 @@ final class SwiftPMWorkspaceTests: XCTestCase {
       let headerArgs = try ws._settings(for: header.asURI, .cpp)!.compilerArguments
       checkArgsCommon(headerArgs)
 
-      check("-c", "-x", "c++-header", URL(fileURLWithPath: header.pathString).path,
+      check("-c", "-x", "c++-header", AbsolutePath(URL(fileURLWithPath: header.pathString).path).pathString,
             arguments: headerArgs)
     }
   }
@@ -476,13 +477,13 @@ final class SwiftPMWorkspaceTests: XCTestCase {
       XCTAssertEqual(arguments1, arguments2)
 
       checkNot(aswift1.pathString, arguments: arguments1 ?? [])
-      check(resolveSymlinks(aswift1).pathString, arguments: arguments1 ?? [])
+      check(try resolveSymlinks(aswift1).pathString, arguments: arguments1 ?? [])
 
       let argsManifest = try ws._settings(for: manifest.asURI, .swift)?.compilerArguments
       XCTAssertNotNil(argsManifest)
 
       checkNot(manifest.pathString, arguments: argsManifest ?? [])
-      check(resolveSymlinks(manifest).pathString, arguments: argsManifest ?? [])
+      check(try resolveSymlinks(manifest).pathString, arguments: argsManifest ?? [])
     }
   }
 
@@ -522,12 +523,12 @@ final class SwiftPMWorkspaceTests: XCTestCase {
       let argsCxx = try ws._settings(for: acxx.asURI, .cpp)?.compilerArguments
       XCTAssertNotNil(argsCxx)
       check(acxx.pathString, arguments: argsCxx ?? [])
-      checkNot(resolveSymlinks(acxx).pathString, arguments: argsCxx ?? [])
+      checkNot(try resolveSymlinks(acxx).pathString, arguments: argsCxx ?? [])
 
       let argsH = try ws._settings(for: ah.asURI, .cpp)?.compilerArguments
       XCTAssertNotNil(argsH)
       checkNot(ah.pathString, arguments: argsH ?? [])
-      check(resolveSymlinks(ah).pathString, arguments: argsH ?? [])
+      check(try resolveSymlinks(ah).pathString, arguments: argsH ?? [])
     }
   }
 
@@ -549,7 +550,7 @@ final class SwiftPMWorkspaceTests: XCTestCase {
                   resources: [.copy("a.txt")])])
             """
       ])
-      let packageRoot = resolveSymlinks(tempDir.appending(component: "pkg"))
+      let packageRoot = try resolveSymlinks(tempDir.appending(component: "pkg"))
       let tr = ToolchainRegistry.shared
       let ws = try! SwiftPMWorkspace(
         workspacePath: packageRoot,
@@ -565,6 +566,30 @@ final class SwiftPMWorkspaceTests: XCTestCase {
         }), "missing resource_bundle_accessor.swift from \(arguments)")
     }
   }
+
+  func testNestedInvalidPackageSwift() throws {
+    let fs = InMemoryFileSystem()
+    try! withTemporaryDirectory(removeTreeOnDeinit: true) { tempDir in
+      try! fs.createFiles(root: tempDir, files: [
+        "pkg/Sources/lib/Package.swift": "// not a valid package",
+        "pkg/Package.swift": """
+        // swift-tools-version:4.2
+        import PackageDescription
+        let package = Package(name: "a", products: [], dependencies: [],
+        targets: [.target(name: "lib", dependencies: [])])
+        """
+      ])
+      let packageRoot = try resolveSymlinks(tempDir.appending(components: "pkg", "Sources", "lib"))
+      let tr = ToolchainRegistry.shared
+      let ws = try! SwiftPMWorkspace(
+        workspacePath: packageRoot,
+        toolchainRegistry: tr,
+        fileSystem: fs,
+        buildSetup: TestSourceKitServer.serverOptions.buildSetup)
+
+      XCTAssertEqual(ws._packageRoot, try resolveSymlinks(tempDir.appending(component: "pkg")))
+    }
+  }
 }
 
 private func checkNot(
@@ -602,8 +627,8 @@ private func check(
 private func buildPath(
   root: AbsolutePath,
   config: BuildSetup = TestSourceKitServer.serverOptions.buildSetup,
-  triple: Triple) -> AbsolutePath
+  platform: String) -> AbsolutePath
 {
   let buildPath = config.path ?? root.appending(component: ".build")
-  return buildPath.appending(components: triple.platformBuildPathComponent(), "\(config.configuration)")
+  return buildPath.appending(components: platform, "\(config.configuration)")
 }
