@@ -28,9 +28,8 @@ final class LocalSwiftTests: XCTestCase {
   /// The primary interface to make requests to the SourceKitServer.
   var sk: TestClient! = nil
 
-  /// The server's workspace data. Accessing this is unsafe if the server does so concurrently.
-  var workspace: Workspace! {
-    connection.server!.workspace!
+  var documentManager: DocumentManager! {
+    connection.server!._documentManager
   }
 
   override func setUp() {
@@ -59,7 +58,7 @@ final class LocalSwiftTests: XCTestCase {
   }
 
   func testEditing() {
-    let url = URL(fileURLWithPath: "/\(#function)/a.swift")
+    let url = URL(fileURLWithPath: "/\(UUID())/a.swift")
     let uri = DocumentURI(url)
 
     sk.allowUnexpectedNotification = false
@@ -75,7 +74,7 @@ final class LocalSwiftTests: XCTestCase {
       log("Received diagnostics for open - syntactic")
       XCTAssertEqual(note.params.version, 12)
       XCTAssertEqual(note.params.diagnostics.count, 1)
-      XCTAssertEqual("func", self.workspace.documentManager.latestSnapshot(uri)!.text)
+      XCTAssertEqual("func", self.documentManager.latestSnapshot(uri)!.text)
     }, { (note: Notification<PublishDiagnosticsNotification>) in
       log("Received diagnostics for open - semantic")
       XCTAssertEqual(note.params.version, 12)
@@ -93,7 +92,7 @@ final class LocalSwiftTests: XCTestCase {
       // 0 = semantic update finished already
       XCTAssertEqual(note.params.version, 13)
       XCTAssertLessThanOrEqual(note.params.diagnostics.count, 1)
-      XCTAssertEqual("func foo() {}\n", self.workspace.documentManager.latestSnapshot(uri)!.text)
+      XCTAssertEqual("func foo() {}\n", self.documentManager.latestSnapshot(uri)!.text)
     }, { (note: Notification<PublishDiagnosticsNotification>) in
       log("Received diagnostics for edit 1 - semantic")
       XCTAssertEqual(note.params.version, 13)
@@ -111,7 +110,7 @@ final class LocalSwiftTests: XCTestCase {
         XCTAssertEqual("""
         func foo() {}
         bar()
-        """, self.workspace.documentManager.latestSnapshot(uri)!.text)
+        """, self.documentManager.latestSnapshot(uri)!.text)
     }, { (note: Notification<PublishDiagnosticsNotification>) in
       log("Received diagnostics for edit 2 - semantic")
       XCTAssertEqual(note.params.version, 14)
@@ -132,7 +131,7 @@ final class LocalSwiftTests: XCTestCase {
         XCTAssertEqual("""
         func foo() {}
         foo()
-        """, self.workspace.documentManager.latestSnapshot(uri)!.text)
+        """, self.documentManager.latestSnapshot(uri)!.text)
     }, { (note: Notification<PublishDiagnosticsNotification>) in
       log("Received diagnostics for edit 3 - semantic")
       XCTAssertEqual(note.params.version, 14)
@@ -150,7 +149,7 @@ final class LocalSwiftTests: XCTestCase {
         XCTAssertEqual("""
         func foo() {}
         fooTypo()
-        """, self.workspace.documentManager.latestSnapshot(uri)!.text)
+        """, self.documentManager.latestSnapshot(uri)!.text)
     }, { (note: Notification<PublishDiagnosticsNotification>) in
       log("Received diagnostics for edit 4 - semantic")
       XCTAssertEqual(note.params.version, 15)
@@ -173,7 +172,7 @@ final class LocalSwiftTests: XCTestCase {
         XCTAssertEqual("""
         func bar() {}
         foo()
-        """, self.workspace.documentManager.latestSnapshot(uri)!.text)
+        """, self.documentManager.latestSnapshot(uri)!.text)
     }, { (note: Notification<PublishDiagnosticsNotification>) in
       log("Received diagnostics for edit 5 - semantic")
       XCTAssertEqual(note.params.version, 16)
@@ -200,7 +199,7 @@ final class LocalSwiftTests: XCTestCase {
       log("Received diagnostics for open - syntactic")
       XCTAssertEqual(note.params.version, 12)
       XCTAssertEqual(note.params.diagnostics.count, 1)
-      XCTAssertEqual("func", self.workspace.documentManager.latestSnapshot(uri)!.text)
+      XCTAssertEqual("func", self.documentManager.latestSnapshot(uri)!.text)
     }, { (note: Notification<PublishDiagnosticsNotification>) in
       log("Received diagnostics for open - semantic")
       XCTAssertEqual(note.params.version, 12)
@@ -218,7 +217,7 @@ final class LocalSwiftTests: XCTestCase {
       // 1 = remaining semantic error
       // 0 = semantic update finished already
       XCTAssertLessThanOrEqual(note.params.diagnostics.count, 1)
-      XCTAssertEqual("func foo() {}\n", self.workspace.documentManager.latestSnapshot(uri)!.text)
+      XCTAssertEqual("func foo() {}\n", self.documentManager.latestSnapshot(uri)!.text)
     }, { (note: Notification<PublishDiagnosticsNotification>) in
       log("Received diagnostics for edit 1 - semantic")
       XCTAssertEqual(note.params.version, 13)
@@ -236,7 +235,7 @@ final class LocalSwiftTests: XCTestCase {
         XCTAssertEqual("""
         func foo() {}
         bar()
-        """, self.workspace.documentManager.latestSnapshot(uri)!.text)
+        """, self.documentManager.latestSnapshot(uri)!.text)
     }, { (note: Notification<PublishDiagnosticsNotification>) in
       log("Received diagnostics for edit 2 - semantic")
       XCTAssertEqual(note.params.version, 14)
@@ -257,7 +256,7 @@ final class LocalSwiftTests: XCTestCase {
         XCTAssertEqual("""
         func foo() {}
         foo()
-        """, self.workspace.documentManager.latestSnapshot(uri)!.text)
+        """, self.documentManager.latestSnapshot(uri)!.text)
     }, { (note: Notification<PublishDiagnosticsNotification>) in
       log("Received diagnostics for edit 3 - semantic")
       XCTAssertEqual(note.params.version, 14)
@@ -275,7 +274,7 @@ final class LocalSwiftTests: XCTestCase {
         XCTAssertEqual("""
         func foo() {}
         fooTypo()
-        """, self.workspace.documentManager.latestSnapshot(uri)!.text)
+        """, self.documentManager.latestSnapshot(uri)!.text)
     }, { (note: Notification<PublishDiagnosticsNotification>) in
       log("Received diagnostics for edit 4 - semantic")
       XCTAssertEqual(note.params.version, 15)
@@ -298,7 +297,7 @@ final class LocalSwiftTests: XCTestCase {
         XCTAssertEqual("""
         func bar() {}
         foo()
-        """, self.workspace.documentManager.latestSnapshot(uri)!.text)
+        """, self.documentManager.latestSnapshot(uri)!.text)
     }, { (note: Notification<PublishDiagnosticsNotification>) in
       log("Received diagnostics for edit 5 - semantic")
       XCTAssertEqual(note.params.version, 16)
@@ -345,8 +344,8 @@ final class LocalSwiftTests: XCTestCase {
   }
 
   func testCrossFileDiagnostics() {
-    let urlA = URL(fileURLWithPath: "/\(#function)/a.swift")
-    let urlB = URL(fileURLWithPath: "/\(#function)/b.swift")
+    let urlA = URL(fileURLWithPath: "/\(UUID())/a.swift")
+    let urlB = URL(fileURLWithPath: "/\(UUID())/b.swift")
     let uriA = DocumentURI(urlA)
     let uriB = DocumentURI(urlB)
 
@@ -406,7 +405,7 @@ final class LocalSwiftTests: XCTestCase {
   }
 
   func testDiagnosticsReopen() {
-    let urlA = URL(fileURLWithPath: "/\(#function)/a.swift")
+    let urlA = URL(fileURLWithPath: "/\(UUID())/a.swift")
     let uriA = DocumentURI(urlA)
     sk.allowUnexpectedNotification = false
 
@@ -456,7 +455,7 @@ final class LocalSwiftTests: XCTestCase {
   }
 
   func testEducationalNotesAreUsedAsDiagnosticCodes() {
-    let url = URL(fileURLWithPath: "/\(#function)/a.swift")
+    let url = URL(fileURLWithPath: "/\(UUID())/a.swift")
     let uri = DocumentURI(url)
 
     sk.allowUnexpectedNotification = false
@@ -476,7 +475,7 @@ final class LocalSwiftTests: XCTestCase {
   }
 
   func testFixitsAreIncludedInPublishDiagnostics() {
-    let url = URL(fileURLWithPath: "/\(#function)/a.swift")
+    let url = URL(fileURLWithPath: "/\(UUID())/a.swift")
     let uri = DocumentURI(url)
 
     sk.allowUnexpectedNotification = false
@@ -510,7 +509,7 @@ final class LocalSwiftTests: XCTestCase {
   }
 
   func testFixitsAreIncludedInPublishDiagnosticsNotes() {
-    let url = URL(fileURLWithPath: "/\(#function)/a.swift")
+    let url = URL(fileURLWithPath: "/\(UUID())/a.swift")
     let uri = DocumentURI(url)
 
     sk.allowUnexpectedNotification = false
@@ -563,7 +562,7 @@ final class LocalSwiftTests: XCTestCase {
   }
 
   func testFixitInsert() {
-    let url = URL(fileURLWithPath: "/\(#function)/a.swift")
+    let url = URL(fileURLWithPath: "/\(UUID())/a.swift")
     let uri = DocumentURI(url)
 
     sk.allowUnexpectedNotification = false
@@ -602,11 +601,11 @@ final class LocalSwiftTests: XCTestCase {
     XCTAssertEqual("Remove 'foo ='", CodeAction.fixitTitle(replace: "foo =", with: ""))
   }
 
-  func testFixitsAreReturnedFromCodeActions() {
-    let url = URL(fileURLWithPath: "/\(#function)/a.swift")
+  func testFixitsAreReturnedFromCodeActions() throws {
+    let url = URL(fileURLWithPath: "/\(UUID())/a.swift")
     let uri = DocumentURI(url)
 
-    var diagnostic: Diagnostic! = nil
+    var diagnostic: Diagnostic? = nil
     sk.sendNoteSync(DidOpenTextDocumentNotification(textDocument: TextDocumentItem(
       uri: uri, language: .swift, version: 12,
       text: """
@@ -619,15 +618,15 @@ final class LocalSwiftTests: XCTestCase {
     }, { (note: Notification<PublishDiagnosticsNotification>) in
       log("Received diagnostics for open - semantic")
       XCTAssertEqual(note.params.diagnostics.count, 1)
-      diagnostic = note.params.diagnostics.first!
+      diagnostic = note.params.diagnostics.first
     })
 
     let request = CodeActionRequest(
       range: Position(line: 1, utf16index: 0)..<Position(line: 1, utf16index: 11),
-      context: CodeActionContext(diagnostics: [diagnostic], only: nil),
+      context: CodeActionContext(diagnostics: [try XCTUnwrap(diagnostic, "expected diagnostic to be available")], only: nil),
       textDocument: TextDocumentIdentifier(uri)
     )
-    let response = try! sk.sendSync(request)
+    let response = try sk.sendSync(request)
 
     XCTAssertNotNil(response)
     guard case .codeActions(let codeActions) = response else {
@@ -639,7 +638,7 @@ final class LocalSwiftTests: XCTestCase {
     let fixit = quickFixes.first!
 
     // Diagnostic returned by code actions cannot be recursive
-    var expectedDiagnostic = diagnostic!
+    var expectedDiagnostic = try XCTUnwrap(diagnostic, "expected diagnostic to be available")
     expectedDiagnostic.codeActions = nil
 
     // Expected Fix-it: Replace `let a` with `_` because it's never used
@@ -652,11 +651,11 @@ final class LocalSwiftTests: XCTestCase {
       command: nil))
   }
 
-  func testFixitsAreReturnedFromCodeActionsNotes() {
-    let url = URL(fileURLWithPath: "/\(#function)/a.swift")
+  func testFixitsAreReturnedFromCodeActionsNotes() throws {
+    let url = URL(fileURLWithPath: "/\(UUID())/a.swift")
     let uri = DocumentURI(url)
 
-    var diagnostic: Diagnostic! = nil
+    var diagnostic: Diagnostic?
     sk.sendNoteSync(DidOpenTextDocumentNotification(textDocument: TextDocumentItem(
       uri: uri, language: .swift, version: 12,
       text: """
@@ -669,15 +668,15 @@ final class LocalSwiftTests: XCTestCase {
     }, { (note: Notification<PublishDiagnosticsNotification>) in
       log("Received diagnostics for open - semantic")
       XCTAssertEqual(note.params.diagnostics.count, 1)
-      diagnostic = note.params.diagnostics.first!
+      diagnostic = note.params.diagnostics.first
     })
 
     let request = CodeActionRequest(
       range: Position(line: 1, utf16index: 0)..<Position(line: 1, utf16index: 11),
-      context: CodeActionContext(diagnostics: [diagnostic], only: nil),
+      context: CodeActionContext(diagnostics: [try XCTUnwrap(diagnostic, "expected diagnostic to be available")], only: nil),
       textDocument: TextDocumentIdentifier(uri)
     )
-    let response = try! sk.sendSync(request)
+    let response = try sk.sendSync(request)
 
     XCTAssertNotNil(response)
     guard case .codeActions(let codeActions) = response else {
@@ -707,11 +706,11 @@ final class LocalSwiftTests: XCTestCase {
     }
   }
 
-  func testMuliEditFixitCodeActionPrimary() {
-    let url = URL(fileURLWithPath: "/\(#function)/a.swift")
+  func testMuliEditFixitCodeActionPrimary() throws {
+    let url = URL(fileURLWithPath: "/\(UUID())/a.swift")
     let uri = DocumentURI(url)
 
-    var diagnostic: Diagnostic! = nil
+    var diagnostic: Diagnostic?
     sk.sendNoteSync(DidOpenTextDocumentNotification(textDocument: TextDocumentItem(
       uri: uri, language: .swift, version: 12,
       text: """
@@ -723,15 +722,15 @@ final class LocalSwiftTests: XCTestCase {
     }, { (note: Notification<PublishDiagnosticsNotification>) in
       log("Received diagnostics for open - semantic")
       XCTAssertEqual(note.params.diagnostics.count, 1)
-      diagnostic = note.params.diagnostics.first!
+      diagnostic = note.params.diagnostics.first
     })
 
     let request = CodeActionRequest(
       range: Position(line: 0, utf16index: 1)..<Position(line: 0, utf16index: 10),
-      context: CodeActionContext(diagnostics: [diagnostic], only: nil),
+      context: CodeActionContext(diagnostics: [try XCTUnwrap(diagnostic, "expected diagnostic to be available")], only: nil),
       textDocument: TextDocumentIdentifier(uri)
     )
-    let response = try! sk.sendSync(request)
+    let response = try sk.sendSync(request)
 
     XCTAssertNotNil(response)
     guard case .codeActions(let codeActions) = response else {
@@ -750,11 +749,11 @@ final class LocalSwiftTests: XCTestCase {
     ])
   }
 
-  func testMuliEditFixitCodeActionNote() {
-    let url = URL(fileURLWithPath: "/\(#function)/a.swift")
+  func testMuliEditFixitCodeActionNote() throws {
+    let url = URL(fileURLWithPath: "/\(UUID())/a.swift")
     let uri = DocumentURI(url)
 
-    var diagnostic: Diagnostic! = nil
+    var diagnostic: Diagnostic?
     sk.sendNoteSync(DidOpenTextDocumentNotification(textDocument: TextDocumentItem(
       uri: uri, language: .swift, version: 12,
       text: """
@@ -774,10 +773,10 @@ final class LocalSwiftTests: XCTestCase {
 
     let request = CodeActionRequest(
       range: Position(line: 3, utf16index: 2)..<Position(line: 3, utf16index: 2),
-      context: CodeActionContext(diagnostics: [diagnostic], only: nil),
+      context: CodeActionContext(diagnostics: [try XCTUnwrap(diagnostic, "expected diagnostic to be available")], only: nil),
       textDocument: TextDocumentIdentifier(uri)
     )
-    let response = try! sk.sendSync(request)
+    let response = try sk.sendSync(request)
 
     XCTAssertNotNil(response)
     guard case .codeActions(let codeActions) = response else {
@@ -1084,7 +1083,7 @@ final class LocalSwiftTests: XCTestCase {
   }
 
   func testSymbolInfo() {
-    let url = URL(fileURLWithPath: "/\(#function)/a.swift")
+    let url = URL(fileURLWithPath: "/\(UUID())/a.swift")
     let uri = DocumentURI(url)
 
     sk.send(DidOpenTextDocumentNotification(textDocument: TextDocumentItem(
@@ -1092,6 +1091,7 @@ final class LocalSwiftTests: XCTestCase {
       language: .swift,
       version: 1,
       text: """
+      import Foundation
       struct S {
         func foo() {
           var local = 1
@@ -1106,15 +1106,16 @@ final class LocalSwiftTests: XCTestCase {
 
       XCTAssertEqual(resp.count, 1)
       if let sym = resp.first {
-        XCTAssertEqual(sym.name, "S")
+        XCTAssertEqual(sym.name, "Foundation")
         XCTAssertNil(sym.containerName)
-        XCTAssertEqual(sym.usr, "s:1a1SV")
-        XCTAssertEqual(sym.bestLocalDeclaration?.uri, uri)
-        XCTAssertEqual(sym.bestLocalDeclaration?.range.lowerBound.line, 0)
-        XCTAssertEqual(sym.bestLocalDeclaration?.range.lowerBound.utf16index, 7)
+        XCTAssertEqual(sym.usr, nil)
+        XCTAssertEqual(sym.kind, .module)
+        XCTAssertEqual(sym.bestLocalDeclaration?.uri, nil)
+        XCTAssertEqual(sym.bestLocalDeclaration?.range.lowerBound.line, nil)
+        XCTAssertEqual(sym.bestLocalDeclaration?.range.lowerBound.utf16index, nil)
       }
     }
-
+    
     do {
       let resp = try! sk.sendSync(SymbolInfoRequest(
         textDocument: TextDocumentIdentifier(url),
@@ -1122,9 +1123,9 @@ final class LocalSwiftTests: XCTestCase {
 
       XCTAssertEqual(resp.count, 1)
       if let sym = resp.first {
-        XCTAssertEqual(sym.name, "foo()")
+        XCTAssertEqual(sym.name, "S")
         XCTAssertNil(sym.containerName)
-        XCTAssertEqual(sym.usr, "s:1a1SV3fooyyF")
+        XCTAssertEqual(sym.usr, "s:1a1SV")
         XCTAssertEqual(sym.bestLocalDeclaration?.uri, uri)
         XCTAssertEqual(sym.bestLocalDeclaration?.range.lowerBound.line, 1)
         XCTAssertEqual(sym.bestLocalDeclaration?.range.lowerBound.utf16index, 7)
@@ -1134,7 +1135,23 @@ final class LocalSwiftTests: XCTestCase {
     do {
       let resp = try! sk.sendSync(SymbolInfoRequest(
         textDocument: TextDocumentIdentifier(url),
-        position: Position(line: 2, utf16index: 8)))
+        position: Position(line: 2, utf16index: 7)))
+
+      XCTAssertEqual(resp.count, 1)
+      if let sym = resp.first {
+        XCTAssertEqual(sym.name, "foo()")
+        XCTAssertNil(sym.containerName)
+        XCTAssertEqual(sym.usr, "s:1a1SV3fooyyF")
+        XCTAssertEqual(sym.bestLocalDeclaration?.uri, uri)
+        XCTAssertEqual(sym.bestLocalDeclaration?.range.lowerBound.line, 2)
+        XCTAssertEqual(sym.bestLocalDeclaration?.range.lowerBound.utf16index, 7)
+      }
+    }
+
+    do {
+      let resp = try! sk.sendSync(SymbolInfoRequest(
+        textDocument: TextDocumentIdentifier(url),
+        position: Position(line: 3, utf16index: 8)))
 
       XCTAssertEqual(resp.count, 1)
       if let sym = resp.first {
@@ -1142,7 +1159,7 @@ final class LocalSwiftTests: XCTestCase {
         XCTAssertNil(sym.containerName)
         XCTAssertEqual(sym.usr, "s:1a1SV3fooyyF5localL_Sivp")
         XCTAssertEqual(sym.bestLocalDeclaration?.uri, uri)
-        XCTAssertEqual(sym.bestLocalDeclaration?.range.lowerBound.line, 2)
+        XCTAssertEqual(sym.bestLocalDeclaration?.range.lowerBound.line, 3)
         XCTAssertEqual(sym.bestLocalDeclaration?.range.lowerBound.utf16index, 8)
       }
     }
@@ -1157,7 +1174,7 @@ final class LocalSwiftTests: XCTestCase {
   }
 
   func testHover() {
-    let url = URL(fileURLWithPath: "/\(#function)/a.swift")
+    let url = URL(fileURLWithPath: "/\(UUID())/a.swift")
     let uri = DocumentURI(url)
 
     sk.send(DidOpenTextDocumentNotification(textDocument: TextDocumentItem(
@@ -1210,7 +1227,7 @@ final class LocalSwiftTests: XCTestCase {
   }
 
   func testHoverNameEscaping() {
-    let url = URL(fileURLWithPath: "/\(#function)/a.swift")
+    let url = URL(fileURLWithPath: "/\(UUID())/a.swift")
 
     sk.send(DidOpenTextDocumentNotification(textDocument: TextDocumentItem(
       uri: DocumentURI(url),
@@ -1273,9 +1290,9 @@ final class LocalSwiftTests: XCTestCase {
       }
     }
   }
-
-  func testDocumentSymbolHighlight() {
-    let url = URL(fileURLWithPath: "/\(#function)/a.swift")
+  
+  func testDocumentSymbolHighlight() throws {
+    let url = URL(fileURLWithPath: "/\(UUID())/a.swift")
     let uri = DocumentURI(url)
 
     sk.send(DidOpenTextDocumentNotification(textDocument: TextDocumentItem(
@@ -1293,14 +1310,14 @@ final class LocalSwiftTests: XCTestCase {
       """)))
 
     do {
-      let resp = try! sk.sendSync(DocumentHighlightRequest(
+      let resp = try sk.sendSync(DocumentHighlightRequest(
         textDocument: TextDocumentIdentifier(url),
         position: Position(line: 0, utf16index: 0)))
       XCTAssertEqual(resp?.count, 0)
     }
 
     do {
-      let resp = try! sk.sendSync(DocumentHighlightRequest(
+      let resp = try sk.sendSync(DocumentHighlightRequest(
         textDocument: TextDocumentIdentifier(url),
         position: Position(line: 1, utf16index: 6)))
       XCTAssertEqual(resp?.count, 1)
@@ -1313,7 +1330,7 @@ final class LocalSwiftTests: XCTestCase {
     }
 
     do {
-      let resp = try! sk.sendSync(DocumentHighlightRequest(
+      let resp = try sk.sendSync(DocumentHighlightRequest(
         textDocument: TextDocumentIdentifier(url),
         position: Position(line: 2, utf16index: 6)))
       XCTAssertEqual(resp?.count, 2)
@@ -1332,7 +1349,7 @@ final class LocalSwiftTests: XCTestCase {
     }
 
     do {
-      let resp = try! sk.sendSync(DocumentHighlightRequest(
+      let resp = try sk.sendSync(DocumentHighlightRequest(
         textDocument: TextDocumentIdentifier(url),
         position: Position(line: 3, utf16index: 6)))
       XCTAssertEqual(resp?.count, 3)
