@@ -18,9 +18,6 @@ import XCTest
 
 private typealias Token = SyntaxHighlightingToken
 
-/// The number of seconds that we wait for a semantic token refresh.
-private let semanticRefreshTimeout: TimeInterval = 60
-
 final class SemanticTokensTests: XCTestCase {
   /// Connection and lifetime management for the service.
   private var connection: TestSourceKitServer! = nil
@@ -106,7 +103,7 @@ final class SemanticTokensTests: XCTestCase {
     )))
     version += 1
 
-    wait(for: [registerCapabilityExpectation, refreshExpectation], timeout: semanticRefreshTimeout)
+    wait(for: [registerCapabilityExpectation, refreshExpectation], timeout: defaultTimeout)
   }
 
   private func editDocument(changes: [TextDocumentContentChangeEvent], expectRefresh: Bool = true) {
@@ -128,7 +125,7 @@ final class SemanticTokensTests: XCTestCase {
     ))
     version += 1
 
-    wait(for: expectations, timeout: semanticRefreshTimeout)
+    wait(for: expectations, timeout: defaultTimeout)
   }
 
   private func editDocument(range: Range<Position>, text: String, expectRefresh: Bool = true) {
@@ -202,7 +199,7 @@ final class SemanticTokensTests: XCTestCase {
     """
     openDocument(text: text)
 
-    guard let snapshot = connection.server?.workspace?.documentManager.latestSnapshot(uri) else {
+    guard let snapshot = connection.server?._documentManager.latestSnapshot(uri) else {
       fatalError("Could not fetch document snapshot for \(#function)")
     }
 
@@ -351,6 +348,7 @@ final class SemanticTokensTests: XCTestCase {
       Token(line: 3, utf16index: 0, length: 3, kind: .keyword),
       Token(line: 3, utf16index: 4, length: 1, kind: .identifier),
       Token(line: 3, utf16index: 8, length: 1, kind: .variable),
+      Token(line: 3, utf16index: 10, length: 1, kind: .operator),
       Token(line: 3, utf16index: 12, length: 1, kind: .variable),
       // func a() {}
       Token(line: 5, utf16index: 0, length: 4, kind: .keyword),
@@ -491,6 +489,20 @@ final class SemanticTokensTests: XCTestCase {
     ])
   }
 
+
+  func testOperatorDeclaration() {
+    let text = """
+    infix operator ?= :ComparisonPrecedence
+    """
+    let tokens = openAndPerformSemanticTokensRequest(text: text)
+    XCTAssertEqual(tokens, [
+      Token(line: 0, utf16index: 0, length: 5, kind: .keyword),
+      Token(line: 0, utf16index: 6, length: 8, kind: .keyword),
+      Token(line: 0, utf16index: 15, length: 2, kind: .operator),
+      Token(line: 0, utf16index: 19, length: 20, kind: .identifier),
+    ])
+  }
+
   func testEmptyEdit() {
     let text = """
     let x: String = "test"
@@ -561,19 +573,27 @@ final class SemanticTokensTests: XCTestCase {
     """
     openDocument(text: text)
 
+    let expectedBefore = [
+      SyntaxHighlightingToken(line: 0, utf16index: 0, length: 3, kind: .keyword),
+      SyntaxHighlightingToken(line: 0, utf16index: 4, length: 1, kind: .identifier),
+      SyntaxHighlightingToken(line: 0, utf16index: 7, length: 6, kind: .struct, modifiers: [.defaultLibrary]),
+      SyntaxHighlightingToken(line: 0, utf16index: 16, length: 6, kind: .string)
+    ]
     let before = performSemanticTokensRequest()
+    XCTAssertEqual(before, expectedBefore)
 
     let pos = Position(line: 0, utf16index: 0)
     let editText = " "
     editDocument(range: pos..<pos, text: editText, expectRefresh: false)
 
     let after = performSemanticTokensRequest()
-    let expected: [Token] = before.map {
-      var token = $0
-      token.move(utf16indexDelta: editText.utf16.count)
-      return token
-    }
-    XCTAssertEqual(after, expected)
+    let expectedAfter = [
+      SyntaxHighlightingToken(line: 0, utf16index: 1, length: 3, kind: .keyword),
+      SyntaxHighlightingToken(line: 0, utf16index: 5, length: 1, kind: .identifier),
+      SyntaxHighlightingToken(line: 0, utf16index: 8, length: 6, kind: .struct, modifiers: [.defaultLibrary]),
+      SyntaxHighlightingToken(line: 0, utf16index: 17, length: 6, kind: .string)
+    ]
+    XCTAssertEqual(after, expectedAfter)
   }
 
   func testInsertSpaceAfterToken() {
@@ -598,18 +618,22 @@ final class SemanticTokensTests: XCTestCase {
     """
     openDocument(text: text)
 
+    let expectedBefore = [
+      SyntaxHighlightingToken(line: 0, utf16index: 0, length: 10, kind: .function, modifiers: [.defaultLibrary]),
+      SyntaxHighlightingToken(line: 0, utf16index: 11, length: 5, kind: .string)
+    ]
     let before = performSemanticTokensRequest()
+    XCTAssertEqual(before, expectedBefore)
 
     let pos = Position(line: 0, utf16index: 0)
     editDocument(range: pos..<pos, text: "\n", expectRefresh: false)
 
     let after = performSemanticTokensRequest()
-    let expected: [Token] = before.map {
-      var token = $0
-      token.move(lineDelta: 1)
-      return token
-    }
-    XCTAssertEqual(after, expected)
+    let expectedAfter = [
+      SyntaxHighlightingToken(line: 1, utf16index: 0, length: 10, kind: .function, modifiers: [.defaultLibrary]),
+      SyntaxHighlightingToken(line: 1, utf16index: 11, length: 5, kind: .string)
+    ]
+    XCTAssertEqual(after, expectedAfter)
   }
 
   func testRemoveNewline() {
@@ -632,11 +656,11 @@ final class SemanticTokensTests: XCTestCase {
     editDocument(range: start..<end, text: "", expectRefresh: false)
 
     let after = performSemanticTokensRequest()
-    let expectedAfter: [Token] = expectedBefore.map {
-      var token = $0
-      token.move(to: Position(line: 0, utf16index: token.start.utf16index))
-      return token
-    }
+    let expectedAfter = [
+      Token(line: 0, utf16index: 0, length: 3, kind: .keyword),
+      Token(line: 0, utf16index: 4, length: 1, kind: .identifier),
+      Token(line: 0, utf16index: 8, length: 5, kind: .string),
+    ]
     XCTAssertEqual(after, expectedAfter)
   }
 
