@@ -42,6 +42,7 @@ OBJC_EXPORT void *objc_destructInstance(id obj);
 
 #if TARGET_OS_WIN32
 #include <Shellapi.h>
+#include <pathcch.h>
 #endif
 
 enum {
@@ -442,7 +443,21 @@ CFTypeRef _CFRuntimeCreateInstance(CFAllocatorRef allocator, CFTypeID typeID, CF
     uintptr_t isa = __CFRuntimeObjCClassTable[typeID];
     CFIndex size = sizeof(CFRuntimeBase) + extraBytes;
     const CFRuntimeClass *cls = __CFRuntimeClassTable[typeID];
-    size_t align = (cls->version & _kCFRuntimeRequiresAlignment) ? cls->requiredAlignment : 16;
+
+#if !defined(__APPLE__) && (defined(__i686__) || (defined(__arm__) && !defined(__aarch64__)) || defined(_M_IX86) || defined(_M_ARM))
+    // Linux and Windows 32-bit targets perform 8-byte alignment by default.
+    static const kDefaultAlignment = 8;
+#else
+    static const kDefaultAlignment = 16;
+#endif
+
+    // Ensure that we get the alignment correct for various targets.  In the
+    // case that we are over-aligned `swift_allocObject` will go through a
+    // different allocator to ensure that the pointer is suitably aligned.  When
+    // we subsequently release the pointer we do not tag that release to go
+    // through the overalign'ed path.  This may result in a cross-domainf free
+    // and a resultant heap corruption.
+    size_t align = (cls->version & _kCFRuntimeRequiresAlignment) ? cls->requiredAlignment : kDefaultAlignment;
     
     CFRuntimeBase *memory = (CFRuntimeBase *)swift_allocObject(isa, size, align - 1);
     
@@ -771,7 +786,7 @@ CFTypeRef _CFNonObjCRetain(CFTypeRef cf) {
 
 CFTypeRef CFRetain(CFTypeRef cf) {
     if (NULL == cf) { CRSetCrashLogMessage("*** CFRetain() called with NULL ***"); HALT; }
-    if (cf) __CFGenericAssertIsCF(cf);
+    __CFGenericAssertIsCF(cf);
     return _CFRetain(cf, false);
 }
 
@@ -789,7 +804,7 @@ void _CFNonObjCRelease(CFTypeRef cf) {
 
 void CFRelease(CFTypeRef cf) {
     if (NULL == cf) { CRSetCrashLogMessage("*** CFRelease() called with NULL ***"); HALT; }
-    if (cf) __CFGenericAssertIsCF(cf);
+    __CFGenericAssertIsCF(cf);
     _CFRelease(cf);
 }
 
@@ -1156,10 +1171,20 @@ CF_EXPORT
 
 CF_PRIVATE os_unfair_recursive_lock CFPlugInGlobalDataLock;
 
+#if __HAS_DISPATCH__
+extern void libdispatch_init();
+#endif
+
 void __CFInitialize(void) {
     if (!__CFInitialized && !__CFInitializing) {
         __CFInitializing = 1;
 
+#if __HAS_DISPATCH__
+    // libdispatch has to be initialized before CoreFoundation, so to avoid
+    // issues with static initializer ordering, we are doing it explicitly.
+    libdispatch_init();
+#endif
+
     // This is a no-op on Darwin, but is needed on Linux and Windows.
     _CFPerformDynamicInitOfOSRecursiveLock(&CFPlugInGlobalDataLock);
 
@@ -1321,33 +1346,24 @@ CF_PRIVATE void __CFSocketCleanup(void);
 CF_PRIVATE void __CFStreamCleanup(void);
 
 static CFBundleRef RegisterCoreFoundationBundle(void) {
-#ifdef _DEBUG
-    // might be nice to get this from the project file at some point
-    wchar_t *DLLFileName = (wchar_t *)L"CoreFoundation_debug.dll";
-#else
-    wchar_t *DLLFileName = (wchar_t *)L"CoreFoundation.dll";
-#endif
     wchar_t path[MAX_PATH+1];
     path[0] = path[1] = 0;
     DWORD wResult;
     CFIndex idx;
-    HMODULE ourModule = GetModuleHandleW(DLLFileName);
 
-    CFAssert(ourModule, __kCFLogAssertion, "GetModuleHandle failed");
+    HMODULE ourModule = NULL;
+    GetModuleHandleExW(GET_MODULE_HANDLE_EX_FLAG_FROM_ADDRESS | GET_MODULE_HANDLE_EX_FLAG_UNCHANGED_REFCOUNT,
+                       (LPCWSTR)&RegisterCoreFoundationBundle, &ourModule);
+    CFAssert(ourModule, __kCFLogAssertion, "GetModuleHandleExW failed");
 
     wResult = GetModuleFileNameW(ourModule, path, MAX_PATH+1);
     CFAssert1(wResult > 0, __kCFLogAssertion, "GetModuleFileName failed: %d", GetLastError());
     CFAssert1(wResult < MAX_PATH+1, __kCFLogAssertion, "GetModuleFileName result truncated: %s", path);
 
     // strip off last component, the DLL name
-    for (idx = wResult - 1; idx; idx--) {
-        if ('\\' == path[idx]) {
-            path[idx] = '\0';
-            break;
-        }
-    }
+    PathCchRemoveFileSpec(path, wResult);
 
-    CFStringRef fsPath = CFStringCreateWithCharacters(kCFAllocatorSystemDefault, (UniChar*)path, idx);
+    CFStringRef fsPath = CFStringCreateWithCharacters(kCFAllocatorSystemDefault, (UniChar*)path, wcslen(path));
     CFURLRef dllURL = CFURLCreateWithFileSystemPath(kCFAllocatorSystemDefault, fsPath, kCFURLWindowsPathStyle, TRUE);
     CFURLRef bundleURL = CFURLCreateCopyAppendingPathComponent(kCFAllocatorSystemDefault, dllURL, CFSTR("CoreFoundation.resources"), TRUE);
     CFRelease(fsPath);
