@@ -1,18 +1,31 @@
 // swift-tools-version:5.5
 
-/*
- This source file is part of the Swift.org open source project
-
- Copyright (c) 2014 - 2021 Apple Inc. and the Swift project authors
- Licensed under Apache License v2.0 with Runtime Library Exception
-
- See http://swift.org/LICENSE.txt for license information
- See http://swift.org/CONTRIBUTORS.txt for Swift project authors
-*/
+//===----------------------------------------------------------------------===//
+//
+// This source file is part of the Swift open source project
+//
+// Copyright (c) 2014-2022 Apple Inc. and the Swift project authors
+// Licensed under Apache License v2.0 with Runtime Library Exception
+//
+// See http://swift.org/LICENSE.txt for license information
+// See http://swift.org/CONTRIBUTORS.txt for the list of Swift project authors
+//
+//===----------------------------------------------------------------------===//
 
 import PackageDescription
 import class Foundation.ProcessInfo
 
+// When building the toolchain on the CI for ELF platforms, remove the CI's
+// stdlib absolute runpath and add ELF's $ORIGIN relative paths before installing.
+let swiftpmLinkSettings : [LinkerSetting]
+let packageLibraryLinkSettings : [LinkerSetting]
+if let resourceDirPath = ProcessInfo.processInfo.environment["SWIFTCI_INSTALL_RPATH_OS"] {
+  swiftpmLinkSettings = [ .unsafeFlags(["-no-toolchain-stdlib-rpath", "-Xlinker", "-rpath", "-Xlinker", "$ORIGIN/../lib/swift/\(resourceDirPath)"]) ]
+  packageLibraryLinkSettings = [ .unsafeFlags(["-no-toolchain-stdlib-rpath", "-Xlinker", "-rpath", "-Xlinker", "$ORIGIN/../../\(resourceDirPath)"]) ]
+} else {
+  swiftpmLinkSettings = []
+  packageLibraryLinkSettings = []
+}
 
 /** SwiftPMDataModel is the subset of SwiftPM product that includes just its data model.
 This allows some clients (such as IDEs) that use SwiftPM's data model but not its build system
@@ -26,10 +39,10 @@ let swiftPMDataModelProduct = (
         "PackageCollectionsModel",
         "PackageGraph",
         "PackageLoading",
+        "PackageMetadata",
         "PackageModel",
         "SourceControl",
         "Workspace",
-        "Xcodeproj",
     ]
 )
 
@@ -53,8 +66,8 @@ automatic linking type with `-auto` suffix appended to product's name.
 */
 let autoProducts = [swiftPMProduct, swiftPMDataModelProduct]
 
-let useSwiftCryptoV2 = ProcessInfo.processInfo.environment["SWIFTPM_USE_SWIFT_CRYPTO_V2"] != nil
-let minimumCryptoVersion: Version = useSwiftCryptoV2 ? "2.0.2" : "1.1.4"
+let useSwiftCryptoV2 = ProcessInfo.processInfo.environment["SWIFTPM_USE_SWIFT_CRYPTO_V1"] == nil
+let minimumCryptoVersion: Version = useSwiftCryptoV2 ? "2.2.3" : "1.1.7"
 var swiftSettings: [SwiftSetting] = []
 if useSwiftCryptoV2 {
     swiftSettings.append(.define("CRYPTO_v2"))
@@ -75,7 +88,6 @@ packageCollectionsSigningTargets.append(
         dependencies: [
             .product(name: "Crypto", package: "swift-crypto"), // for CCryptoBoringSSL
         ],
-        exclude: ["CMakeLists.txt"],
         cSettings: [
             .define("WIN32_LEAN_AND_MEAN"),
         ]
@@ -89,7 +101,6 @@ packageCollectionsSigningTargets.append(
          /** Package collections signing */
          name: "PackageCollectionsSigning",
          dependencies: packageCollectionsSigningDeps,
-         exclude: ["CMakeLists.txt"],
          swiftSettings: swiftSettings
     )
 )
@@ -152,7 +163,8 @@ let package = Package(
             swiftSettings: [
                 .unsafeFlags(["-package-description-version", "999.0"]),
                 .unsafeFlags(["-enable-library-evolution"], .when(platforms: [.macOS]))
-            ]
+            ],
+            linkerSettings: packageLibraryLinkSettings
         ),
 
         // The `PackagePlugin` target provides the API that is available to
@@ -164,15 +176,21 @@ let package = Package(
             swiftSettings: [
                 .unsafeFlags(["-package-description-version", "999.0"]),
                 .unsafeFlags(["-enable-library-evolution"], .when(platforms: [.macOS]))
-            ]
+            ],
+            linkerSettings: packageLibraryLinkSettings
         ),
 
         // MARK: SwiftPM specific support libraries
 
+        .systemLibrary(name: "SPMSQLite3", pkgConfig: "sqlite3"),
+
         .target(
             name: "Basics",
             dependencies: [
+                "SPMSQLite3",
+                .product(name: "OrderedCollections", package: "swift-collections"),
                 .product(name: "SwiftToolsSupport-auto", package: "swift-tools-support-core"),
+                .product(name: "SystemPackage", package: "swift-system"),
             ],
             exclude: ["CMakeLists.txt"]
         ),
@@ -199,7 +217,10 @@ let package = Package(
         .target(
             /** Source control operations */
             name: "SourceControl",
-            dependencies: ["Basics"],
+            dependencies: [
+                "Basics",
+                "PackageModel"
+            ],
             exclude: ["CMakeLists.txt"]
         ),
 
@@ -237,9 +258,7 @@ let package = Package(
             dependencies: [
                 "Basics",
                 "PackageLoading",
-                "PackageModel",
-                "PackageRegistry",
-                "SourceControl"
+                "PackageModel"
             ],
             exclude: ["CMakeLists.txt", "README.md"]
         ),
@@ -251,7 +270,6 @@ let package = Package(
             name: "PackageCollectionsModel",
             dependencies: [],
             exclude: [
-                "CMakeLists.txt",
                 "Formats/v1.md"
             ]
         ),
@@ -265,10 +283,9 @@ let package = Package(
                 "PackageCollectionsSigning",
                 "PackageModel",
                 "SourceControl",
-            ],
-            exclude: ["CMakeLists.txt"]
+            ]
         ),
-        
+
         .target(
             name: "PackageFingerprint",
             dependencies: [
@@ -299,6 +316,15 @@ let package = Package(
                 "SPMBuildCore",
                 "SPMLLBuild",
                 .product(name: "SwiftDriver", package: "swift-driver"),
+                "DriverSupport",
+            ],
+            exclude: ["CMakeLists.txt"]
+        ),
+        .target(
+            name: "DriverSupport",
+            dependencies: [
+		"Basics",
+                .product(name: "SwiftDriver", package: "swift-driver"),
             ],
             exclude: ["CMakeLists.txt"]
         ),
@@ -306,13 +332,7 @@ let package = Package(
             /** Support for building using Xcode's build system */
             name: "XCBuildSupport",
             dependencies: ["SPMBuildCore", "PackageGraph"],
-            exclude: ["CMakeLists.txt"]
-        ),
-        .target(
-            /** Generates Xcode projects */
-            name: "Xcodeproj",
-            dependencies: ["Basics", "PackageGraph"],
-            exclude: ["CMakeLists.txt", "TODO.md"]
+            exclude: ["CMakeLists.txt", "CODEOWNERS"]
         ),
         .target(
             /** High level functionality */
@@ -322,15 +342,40 @@ let package = Package(
                 "PackageFingerprint",
                 "PackageGraph",
                 "PackageModel",
+                "PackageRegistry",
                 "SourceControl",
                 "SPMBuildCore",
-                "Xcodeproj"
             ],
             exclude: ["CMakeLists.txt"]
         ),
+        .target(
+            // ** High level interface for package discovery */
+            name: "PackageMetadata",
+            dependencies: [
+                "Basics",
+                "PackageCollections",
+                "PackageModel",
+                "PackageRegistry",
+            ]
+        ),
 
         // MARK: Commands
 
+        .target(
+            /** Minimal set of commands required for bootstrapping a new SwiftPM */
+            name: "CoreCommands",
+            dependencies: [
+                .product(name: "ArgumentParser", package: "swift-argument-parser"),
+                "Basics",
+                "Build",
+                "PackageFingerprint",
+                "PackageModel",
+                "Workspace",
+                "XCBuildSupport",
+            ],
+            exclude: ["CMakeLists.txt"]
+        ),
+
         .target(
             /** High-level commands */
             name: "Commands",
@@ -338,16 +383,58 @@ let package = Package(
                 .product(name: "ArgumentParser", package: "swift-argument-parser"),
                 "Basics",
                 "Build",
-                "PackageCollections",
-                "PackageFingerprint",
+                "CoreCommands",
                 "PackageGraph",
                 "SourceControl",
                 "Workspace",
-                "Xcodeproj",
                 "XCBuildSupport",
             ],
             exclude: ["CMakeLists.txt", "README.md"]
         ),
+
+        .target(
+            /** Interacts with cross-compilation destinations */
+            name: "CrossCompilationDestinationsTool",
+            dependencies: [
+                .product(name: "ArgumentParser", package: "swift-argument-parser"),
+                "Basics",
+                "CoreCommands",
+                "SPMBuildCore",
+                "PackageModel",
+            ]
+        ),
+
+        .target(
+            /** Interacts with package collections */
+            name: "PackageCollectionsTool",
+            dependencies: [
+                .product(name: "ArgumentParser", package: "swift-argument-parser"),
+                "Basics",
+                "Commands",
+                "CoreCommands",
+                "PackageCollections",
+                "PackageModel",
+            ]
+        ),
+
+        .target(
+            /** Interact with package registry */
+            name: "PackageRegistryTool",
+            dependencies: [
+                .product(name: "ArgumentParser", package: "swift-argument-parser"),
+                "Basics",
+                "Commands",
+                "CoreCommands",
+                "PackageGraph",
+                "PackageLoading",
+                "PackageModel",
+                "PackageRegistry",
+                "SourceControl",
+                "SPMBuildCore",
+                "Workspace",
+            ]
+        ),
+
         .executableTarget(
             /** The main executable provided by SwiftPM */
             name: "swift-package",
@@ -360,6 +447,17 @@ let package = Package(
             dependencies: ["Commands"],
             exclude: ["CMakeLists.txt"]
         ),
+        .executableTarget(
+            /** Builds SwiftPM itself for bootstrapping (minimal version of `swift-build`) */
+            name: "swift-bootstrap",
+            dependencies: ["CoreCommands"],
+            exclude: ["CMakeLists.txt"]
+        ),
+        .executableTarget(
+            /** Interacts with cross-compilation destinations */
+            name: "swift-experimental-destination",
+            dependencies: ["Commands", "CrossCompilationDestinationsTool"]
+        ),
         .executableTarget(
             /** Runs package tests */
             name: "swift-test",
@@ -375,20 +473,29 @@ let package = Package(
         .executableTarget(
             /** Interacts with package collections */
             name: "swift-package-collection",
-            dependencies: ["Commands"],
-            exclude: ["CMakeLists.txt"]
+            dependencies: ["Commands", "PackageCollectionsTool"]
+        ),
+        .executableTarget(
+            /** Multi-tool entry point for SwiftPM. */
+            name: "swift-package-manager",
+            dependencies: [
+                "Basics",
+                "Commands",
+                "CrossCompilationDestinationsTool",
+                "PackageCollectionsTool",
+                "PackageRegistryTool"
+            ],
+            linkerSettings: swiftpmLinkSettings
         ),
         .executableTarget(
             /** Interact with package registry */
             name: "swift-package-registry",
-            dependencies: ["Commands"],
-            exclude: ["CMakeLists.txt"]
+            dependencies: ["Commands", "PackageRegistryTool"]
         ),
         .executableTarget(
             /** Shim tool to find test names on OS X */
             name: "swiftpm-xctest-helper",
             dependencies: [],
-            exclude: ["CMakeLists.txt"],
             linkerSettings: [
                 .unsafeFlags(["-Xlinker", "-rpath", "-Xlinker", "@executable_path/../../../lib/swift/macosx"], .when(platforms: [.macOS])),
             ]),
@@ -407,16 +514,24 @@ let package = Package(
                 "SourceControl",
                 .product(name: "TSCTestSupport", package: "swift-tools-support-core"),
                 "Workspace",
-                "Xcodeproj",
                 "XCBuildSupport",
             ]
         ),
 
+        .target(
+            /** Test for thread-santizer. */
+            name: "tsan_utils",
+            dependencies: []),
+
         // MARK: SwiftPM tests
 
         .testTarget(
             name: "BasicsTests",
-            dependencies: ["Basics", "SPMTestSupport"]
+            dependencies: ["Basics", "SPMTestSupport", "tsan_utils"],
+            exclude: [
+                "Inputs/archive.zip",
+                "Inputs/invalid_archive.zip",
+            ]
         ),
         .testTarget(
             name: "BuildTests",
@@ -440,7 +555,8 @@ let package = Package(
             name: "WorkspaceTests",
             dependencies: ["Workspace", "SPMTestSupport"]
         ),
-        .testTarget(
+        // rdar://101868275 "error: cannot find 'XCTAssertEqual' in scope" can affect almost any functional test, so we flat out disable them all until we know what is going on
+        /*.testTarget(
             name: "FunctionalTests",
             dependencies: [
                 "swift-build",
@@ -449,7 +565,7 @@ let package = Package(
                 "PackageModel",
                 "SPMTestSupport"
             ]
-        ),
+        ),*/
         .testTarget(
             name: "FunctionalPerformanceTests",
             dependencies: [
@@ -504,7 +620,7 @@ let package = Package(
         ),
         .testTarget(
             name: "PackageCollectionsTests",
-            dependencies: ["PackageCollections", "SPMTestSupport"]
+            dependencies: ["PackageCollections", "SPMTestSupport", "tsan_utils"]
         ),
         .testTarget(
             name: "PackageFingerprintTests",
@@ -523,10 +639,6 @@ let package = Package(
             dependencies: ["SourceControl", "SPMTestSupport"],
             exclude: ["Inputs/TestRepo.tgz"]
         ),
-        .testTarget(
-            name: "XcodeprojTests",
-            dependencies: ["Xcodeproj", "SPMTestSupport"]
-        ),
         .testTarget(
             name: "XCBuildSupportTests",
             dependencies: ["XCBuildSupport", "SPMTestSupport"],
@@ -553,7 +665,7 @@ let package = Package(
 // this right now.
 
 /// When not using local dependencies, the branch to use for llbuild and TSC repositories.
-let relatedDependenciesBranch = "release/5.6"
+let relatedDependenciesBranch = "release/5.8"
 
 if ProcessInfo.processInfo.environment["SWIFTPM_LLBUILD_FWK"] == nil {
     if ProcessInfo.processInfo.environment["SWIFTCI_USE_LOCAL_DEPS"] == nil {
@@ -579,6 +691,8 @@ if ProcessInfo.processInfo.environment["SWIFTCI_USE_LOCAL_DEPS"] == nil {
         .package(url: "https://github.com/apple/swift-argument-parser.git", .upToNextMinor(from: "1.0.3")),
         .package(url: "https://github.com/apple/swift-driver.git", .branch(relatedDependenciesBranch)),
         .package(url: "https://github.com/apple/swift-crypto.git", .upToNextMinor(from: minimumCryptoVersion)),
+        .package(url: "https://github.com/apple/swift-system.git", .upToNextMinor(from: "1.1.1")),
+        .package(url: "https://github.com/apple/swift-collections.git", .upToNextMinor(from: "1.0.1")),
     ]
 } else {
     package.dependencies += [
@@ -586,5 +700,7 @@ if ProcessInfo.processInfo.environment["SWIFTCI_USE_LOCAL_DEPS"] == nil {
         .package(path: "../swift-argument-parser"),
         .package(path: "../swift-driver"),
         .package(path: "../swift-crypto"),
+        .package(path: "../swift-system"),
+        .package(path: "../swift-collections"),
     ]
 }
