@@ -1,12 +1,14 @@
-/*
- This source file is part of the Swift.org open source project
-
- Copyright (c) 2021 Apple Inc. and the Swift project authors
- Licensed under Apache License v2.0 with Runtime Library Exception
-
- See http://swift.org/LICENSE.txt for license information
- See http://swift.org/CONTRIBUTORS.txt for Swift project authors
- */
+//===----------------------------------------------------------------------===//
+//
+// This source file is part of the Swift open source project
+//
+// Copyright (c) 2021-2022 Apple Inc. and the Swift project authors
+// Licensed under Apache License v2.0 with Runtime Library Exception
+//
+// See http://swift.org/LICENSE.txt for license information
+// See http://swift.org/CONTRIBUTORS.txt for the list of Swift project authors
+//
+//===----------------------------------------------------------------------===//
 
 import struct Foundation.Data
 import struct Foundation.URL
@@ -15,10 +17,15 @@ import Security
 #endif
 
 import TSCBasic
-import TSCUtility
 
 public protocol AuthorizationProvider {
-    func authentication(for url: Foundation.URL) -> (user: String, password: String)?
+    func authentication(for url: URL) -> (user: String, password: String)?
+}
+
+public protocol AuthorizationWriter {
+    func addOrUpdate(for url: URL, user: String, password: String, persist: Bool, callback: @escaping (Result<Void, Error>) -> Void)
+
+    func remove(for url: URL, callback: @escaping (Result<Void, Error>) -> Void)
 }
 
 public enum AuthorizationProviderError: Error {
@@ -29,7 +36,7 @@ public enum AuthorizationProviderError: Error {
 }
 
 public extension AuthorizationProvider {
-    func httpAuthorizationHeader(for url: Foundation.URL) -> String? {
+    func httpAuthorizationHeader(for url: URL) -> String? {
         guard let (user, password) = self.authentication(for: url) else {
             return nil
         }
@@ -41,7 +48,7 @@ public extension AuthorizationProvider {
     }
 }
 
-private extension Foundation.URL {
+private extension URL {
     var authenticationID: String? {
         guard let host = host?.lowercased() else {
             return nil
@@ -52,11 +59,13 @@ private extension Foundation.URL {
 
 // MARK: - netrc
 
-public struct NetrcAuthorizationProvider: AuthorizationProvider {
+public class NetrcAuthorizationProvider: AuthorizationProvider, AuthorizationWriter {
     // marked internal for testing
     internal let path: AbsolutePath
     private let fileSystem: FileSystem
 
+    private let cache = ThreadSafeKeyValueStore<String, (user: String, password: String)>()
+
     public init(path: AbsolutePath, fileSystem: FileSystem) throws {
         self.path = path
         self.fileSystem = fileSystem
@@ -64,11 +73,16 @@ public struct NetrcAuthorizationProvider: AuthorizationProvider {
         _ = try Self.load(fileSystem: fileSystem, path: path)
     }
 
-    public mutating func addOrUpdate(for url: Foundation.URL, user: String, password: String, callback: @escaping (Result<Void, Error>) -> Void) {
+    public func addOrUpdate(for url: URL, user: String, password: String, persist: Bool = true, callback: @escaping (Result<Void, Error>) -> Void) {
         guard let machine = url.authenticationID else {
             return callback(.failure(AuthorizationProviderError.invalidURLHost))
         }
 
+        if !persist {
+            self.cache[machine] = (user, password)
+            return callback(.success(()))
+        }
+
         // Same entry already exists, no need to add or update
         let netrc = try? Self.load(fileSystem: self.fileSystem, path: self.path)
         guard netrc?.machines.first(where: { $0.name.lowercased() == machine && $0.login == user && $0.password == password }) == nil else {
@@ -80,8 +94,8 @@ public struct NetrcAuthorizationProvider: AuthorizationProvider {
             try self.fileSystem.withLock(on: self.path, type: .exclusive) {
                 let contents = try? self.fileSystem.readFileContents(self.path).contents
                 try self.fileSystem.writeFileContents(self.path) { stream in
-                    // File does not exist yet
-                    if let contents = contents {
+                    // Write existing contents
+                    if let contents = contents, !contents.isEmpty {
                         stream.write(contents)
                         stream.write("\n")
                     }
@@ -96,11 +110,18 @@ public struct NetrcAuthorizationProvider: AuthorizationProvider {
         }
     }
 
-    public func authentication(for url: Foundation.URL) -> (user: String, password: String)? {
-        self.machine(for: url).map { (user: $0.login, password: $0.password) }
+    public func remove(for url: URL, callback: @escaping (Result<Void, Error>) -> Void) {
+        callback(.failure(AuthorizationProviderError.other("User must edit netrc file at \(self.path) manually to remove entries")))
     }
 
-    private func machine(for url: Foundation.URL) -> TSCUtility.Netrc.Machine? {
+    public func authentication(for url: URL) -> (user: String, password: String)? {
+        if let machine = url.authenticationID, let cached = self.cache[machine] {
+            return cached
+        }
+        return self.machine(for: url).map { (user: $0.login, password: $0.password) }
+    }
+
+    private func machine(for url: URL) -> Basics.Netrc.Machine? {
         if let machine = url.authenticationID, let existing = self.machines.first(where: { $0.name.lowercased() == machine }) {
             return existing
         }
@@ -111,7 +132,7 @@ public struct NetrcAuthorizationProvider: AuthorizationProvider {
     }
 
     // marked internal for testing
-    internal var machines: [TSCUtility.Netrc.Machine] {
+    internal var machines: [Basics.Netrc.Machine] {
         // this ignores any errors reading the file
         // initial validation is done at the time of initializing the provider
         // and if the file becomes corrupt at runtime it will handle it gracefully
@@ -121,8 +142,8 @@ public struct NetrcAuthorizationProvider: AuthorizationProvider {
 
     private static func load(fileSystem: FileSystem, path: AbsolutePath) throws -> Netrc? {
         do {
-            return try Netrc.load(fromFileAtPath: path).get()
-        } catch Netrc.Error.fileNotFound, Netrc.Error.machineNotFound {
+            return try NetrcParser.parse(fileSystem: fileSystem, path: path)
+        } catch NetrcError.fileNotFound, NetrcError.machineNotFound {
             // These are recoverable errors.
             return .none
         }
@@ -132,17 +153,25 @@ public struct NetrcAuthorizationProvider: AuthorizationProvider {
 // MARK: - Keychain
 
 #if canImport(Security)
-public struct KeychainAuthorizationProvider: AuthorizationProvider {
+public class KeychainAuthorizationProvider: AuthorizationProvider, AuthorizationWriter {
     private let observabilityScope: ObservabilityScope
 
+    private let cache = ThreadSafeKeyValueStore<String, (user: String, password: String)>()
+
     public init(observabilityScope: ObservabilityScope) {
         self.observabilityScope = observabilityScope
     }
 
-    public func addOrUpdate(for url: Foundation.URL, user: String, password: String, callback: @escaping (Result<Void, Error>) -> Void) {
+    public func addOrUpdate(for url: URL, user: String, password: String, persist: Bool = true, callback: @escaping (Result<Void, Error>) -> Void) {
         guard let server = url.authenticationID else {
             return callback(.failure(AuthorizationProviderError.invalidURLHost))
         }
+
+        if !persist {
+            self.cache[server] = (user, password)
+            return callback(.success(()))
+        }
+
         guard let passwordData = password.data(using: .utf8) else {
             return callback(.failure(AuthorizationProviderError.cannotEncodePassword))
         }
@@ -158,11 +187,30 @@ public struct KeychainAuthorizationProvider: AuthorizationProvider {
         }
     }
 
-    public func authentication(for url: Foundation.URL) -> (user: String, password: String)? {
+    public func remove(for url: URL, callback: @escaping (Result<Void, Error>) -> Void) {
+        guard let server = url.authenticationID else {
+            return callback(.failure(AuthorizationProviderError.invalidURLHost))
+        }
+
+        let `protocol` = self.protocol(for: url)
+
+        do {
+            try self.delete(server: server, protocol: `protocol`)
+            callback(.success(()))
+        } catch {
+            callback(.failure(error))
+        }
+    }
+
+    public func authentication(for url: URL) -> (user: String, password: String)? {
         guard let server = url.authenticationID else {
             return nil
         }
 
+        if let cached = self.cache[server] {
+            return cached
+        }
+
         do {
             guard let existingItem = try self.search(server: server, protocol: self.protocol(for: url)) as? [String: Any],
                   let passwordData = existingItem[kSecValueData as String] as? Data,
@@ -215,6 +263,16 @@ public struct KeychainAuthorizationProvider: AuthorizationProvider {
         return true
     }
 
+    private func delete(server: String, protocol: CFString) throws {
+        let query: [String: Any] = [kSecClass as String: kSecClassInternetPassword,
+                                    kSecAttrServer as String: server,
+                                    kSecAttrProtocol as String: `protocol`]
+        let status = SecItemDelete(query as CFDictionary)
+        guard status == errSecSuccess else {
+            throw AuthorizationProviderError.other("Failed to delete credentials for server \(server) from keychain: status \(status)")
+        }
+    }
+
     private func search(server: String, protocol: CFString) throws -> CFTypeRef? {
         let query: [String: Any] = [kSecClass as String: kSecClassInternetPassword,
                                     kSecAttrServer as String: server,
@@ -236,7 +294,7 @@ public struct KeychainAuthorizationProvider: AuthorizationProvider {
         return item
     }
 
-    private func `protocol`(for url: Foundation.URL) -> CFString {
+    private func `protocol`(for url: URL) -> CFString {
         // See https://developer.apple.com/documentation/security/keychain_services/keychain_items/item_attribute_keys_and_values?language=swift
         // for a list of possible values for the `kSecAttrProtocol` attribute.
         switch url.scheme?.lowercased() {
@@ -252,7 +310,8 @@ public struct KeychainAuthorizationProvider: AuthorizationProvider {
 // MARK: - Composite
 
 public struct CompositeAuthorizationProvider: AuthorizationProvider {
-    private let providers: [AuthorizationProvider]
+    // marked internal for testing
+    internal let providers: [AuthorizationProvider]
     private let observabilityScope: ObservabilityScope
 
     public init(_ providers: AuthorizationProvider..., observabilityScope: ObservabilityScope) {
@@ -264,7 +323,7 @@ public struct CompositeAuthorizationProvider: AuthorizationProvider {
         self.observabilityScope = observabilityScope
     }
 
-    public func authentication(for url: Foundation.URL) -> (user: String, password: String)? {
+    public func authentication(for url: URL) -> (user: String, password: String)? {
         for provider in self.providers {
             if let authentication = provider.authentication(for: url) {
                 switch provider {
