@@ -1,16 +1,21 @@
-/*
- This source file is part of the Swift.org open source project
-
- Copyright (c) 2021 Apple Inc. and the Swift project authors
- Licensed under Apache License v2.0 with Runtime Library Exception
-
- See http://swift.org/LICENSE.txt for license information
- See http://swift.org/CONTRIBUTORS.txt for Swift project authors
- */
+//===----------------------------------------------------------------------===//
+//
+// This source file is part of the Swift open source project
+//
+// Copyright (c) 2021 Apple Inc. and the Swift project authors
+// Licensed under Apache License v2.0 with Runtime Library Exception
+//
+// See http://swift.org/LICENSE.txt for license information
+// See http://swift.org/CONTRIBUTORS.txt for the list of Swift project authors
+//
+//===----------------------------------------------------------------------===//
 
 import Dispatch
+import Foundation
 import TSCBasic
-import TSCUtility
+
+import enum TSCUtility.Diagnostics
+import protocol TSCUtility.DiagnosticDataConvertible
 
 // this could become a struct when we remove the "errorsReported" pattern
 
@@ -61,7 +66,7 @@ public final class ObservabilityScope: DiagnosticsEmitterProtocol, CustomStringC
     private let parent: ObservabilityScope?
     private let metadata: ObservabilityMetadata?
 
-    private var diagnosticsHandler: DiagnosticsHanderWrapper
+    private var diagnosticsHandler: DiagnosticsHandlerWrapper
 
     fileprivate init(
         description: String,
@@ -72,7 +77,7 @@ public final class ObservabilityScope: DiagnosticsEmitterProtocol, CustomStringC
         self.description = description
         self.parent = parent
         self.metadata = metadata
-        self.diagnosticsHandler = DiagnosticsHanderWrapper(diagnosticsHandler)
+        self.diagnosticsHandler = DiagnosticsHandlerWrapper(diagnosticsHandler)
     }
 
     public func makeChildScope(description: String, metadata: ObservabilityMetadata? = .none) -> Self {
@@ -122,7 +127,7 @@ public final class ObservabilityScope: DiagnosticsEmitterProtocol, CustomStringC
         self.diagnosticsHandler.handleDiagnostic(scope: self, diagnostic: diagnostic)
     }
 
-    private struct DiagnosticsHanderWrapper: DiagnosticsHandler {
+    private struct DiagnosticsHandlerWrapper: DiagnosticsHandler {
         private let underlying: DiagnosticsHandler
         private var _errorsReported = ThreadSafeBox<Bool>(false)
 
@@ -155,6 +160,12 @@ public protocol DiagnosticsEmitterProtocol {
 }
 
 extension DiagnosticsEmitterProtocol {
+    public func emit(_ diagnostics: [Diagnostic]) {
+        for diagnostic in diagnostics {
+            self.emit(diagnostic)
+        }
+    }
+
     public func emit(severity: Diagnostic.Severity, message: String, metadata: ObservabilityMetadata? = .none) {
         self.emit(.init(severity: severity, message: message, metadata: metadata))
     }
@@ -278,6 +289,8 @@ public struct Diagnostic: CustomStringConvertible {
             message = "\(diagnosticData)"
         } else if let convertible = error as? DiagnosticDataConvertible {
             message = "\(convertible.diagnosticData)"
+        } else if let localizedError = error as? LocalizedError {
+            message = localizedError.errorDescription ?? localizedError.localizedDescription
         } else {
             message = "\(error)"
         }
@@ -362,7 +375,7 @@ public struct Diagnostic: CustomStringConvertible {
 /// Note that specific baggage and context types MAY (and usually do), offer also a way to set baggage values,
 /// however in the most general case it is not required, as some frameworks may only be able to offer reading.
 
-// FIXME: we currently requires that Value conforms to CustomStringConvertible which sucks
+// FIXME: we currently require that Value conforms to CustomStringConvertible which sucks
 // ideally Value would conform to Equatable but that has generic requirement
 // luckily, this is about to change so we can clean this up soon
 public struct ObservabilityMetadata: CustomDebugStringConvertible {
@@ -457,21 +470,6 @@ public struct ObservabilityMetadata: CustomDebugStringConvertible {
         }
     }
 
-    @available(*, deprecated, message: "temporary for transition DiagnosticsEngine -> DiagnosticsEmitter")
-    public func droppingLegacyKeys() -> ObservabilityMetadata? {
-        var metadata = ObservabilityMetadata()
-        self.forEach { (key, value) in
-            if key.keyType == LegacyLocationKey.self {
-                return
-            }
-            if key.keyType == LegacyDataKey.self {
-                return
-            }
-            metadata._storage[key] = value
-        }
-        return metadata.isEmpty ? .none : metadata
-    }
-
     /// A type-erased `ObservabilityMetadataKey` used when iterating through the `ObservabilityMetadata` using its `forEach` method.
     public struct AnyKey {
         /// The key's type represented erased to an `Any.Type`.
@@ -558,53 +556,6 @@ extension Diagnostic {
     }
 }
 
-@available(*, deprecated, message: "temporary for transition DiagnosticsEngine -> DiagnosticsEmitter")
-extension ObservabilitySystem {
-    public convenience init(diagnosticEngine: DiagnosticsEngine) {
-        self.init(DiagnosticsEngineAdapter(diagnosticEngine: diagnosticEngine))
-    }
-
-    private struct DiagnosticsEngineAdapter: ObservabilityHandlerProvider, DiagnosticsHandler {
-        let diagnosticEngine: DiagnosticsEngine
-
-        var diagnosticsHandler: DiagnosticsHandler { self }
-
-        func handleDiagnostic(scope: ObservabilityScope, diagnostic: Diagnostic) {
-            diagnosticEngine.emit(.init(diagnostic))
-        }
-    }
-}
-
-@available(*, deprecated, message: "temporary for transition DiagnosticsEngine -> DiagnosticsEmitter")
-extension TSCBasic.Diagnostic {
-    public init(_ diagnostic: Diagnostic) {
-        let location: DiagnosticLocation
-        if let legacyLocation = diagnostic.metadata?.legacyDiagnosticLocation {
-            location = legacyLocation.underlying
-        } else {
-            location = UnknownLocation.location
-        }
-
-        let data: DiagnosticData
-        if let legacyData = diagnostic.metadata?.legacyDiagnosticData {
-            data = legacyData.underlying
-        } else {
-            data = StringDiagnostic(diagnostic.message)
-        }
-
-        switch diagnostic.severity {
-        case .error:
-            self = .init(message: .error(data), location: location)
-        case .warning:
-            self = .init(message: .warning(data), location: location)
-        case .info:
-            self = .init(message: .note(data), location: location)
-        case .debug:
-            self = .init(message: .note(data), location: location)
-        }
-    }
-}
-
 @available(*, deprecated, message: "temporary for transition DiagnosticsEngine -> DiagnosticsEmitter")
 extension ObservabilityMetadata {
     public var legacyDiagnosticLocation: DiagnosticLocationWrapper? {
