@@ -1,23 +1,33 @@
-/*
- This source file is part of the Swift.org open source project
-
- Copyright 2015 - 2022 Apple Inc. and the Swift project authors
- Licensed under Apache License v2.0 with Runtime Library Exception
-
- See http://swift.org/LICENSE.txt for license information
- See http://swift.org/CONTRIBUTORS.txt for Swift project authors
-*/
+//===----------------------------------------------------------------------===//
+//
+// This source file is part of the Swift open source project
+//
+// Copyright (c) 2015-2022 Apple Inc. and the Swift project authors
+// Licensed under Apache License v2.0 with Runtime Library Exception
+//
+// See http://swift.org/LICENSE.txt for license information
+// See http://swift.org/CONTRIBUTORS.txt for the list of Swift project authors
+//
+//===----------------------------------------------------------------------===//
 
 import Basics
+@_implementationOnly import DriverSupport
 import LLBuildManifest
 import PackageGraph
+import PackageLoading
 import PackageModel
 import SPMBuildCore
 import SPMLLBuild
 import TSCBasic
-import TSCUtility
 import Foundation
 
+import enum TSCUtility.Diagnostics
+import class TSCUtility.MultiLineNinjaProgressAnimation
+import class TSCUtility.NinjaProgressAnimation
+import protocol TSCUtility.ProgressAnimationProtocol
+
+@_implementationOnly import SwiftDriver
+
 public final class BuildOperation: PackageStructureDelegate, SPMBuildCore.BuildSystem, BuildErrorAdviceProvider {
 
     /// The delegate used by the build system.
@@ -28,13 +38,13 @@ public final class BuildOperation: PackageStructureDelegate, SPMBuildCore.BuildS
 
     /// The closure for loading the package graph.
     let packageGraphLoader: () throws -> PackageGraph
-    
+
     /// Entity responsible for compiling and running plugin scripts.
     let pluginScriptRunner: PluginScriptRunner
-    
+
     /// Directory where plugin intermediate files are stored.
     let pluginWorkDirectory: AbsolutePath
-    
+
     /// Whether to sandbox commands from build tool plugins.
     public let disableSandboxForPluginCommands: Bool
 
@@ -48,7 +58,17 @@ public final class BuildOperation: PackageStructureDelegate, SPMBuildCore.BuildS
     public let cacheBuildManifest: Bool
 
     /// The build plan that was computed, if any.
-    public private(set) var buildPlan: BuildPlan?
+    public private(set) var _buildPlan: BuildPlan?
+
+    public var buildPlan: SPMBuildCore.BuildPlan {
+        get throws {
+            if let buildPlan = _buildPlan {
+                return buildPlan
+            } else {
+                throw StringError("did not compute a build plan yet")
+            }
+        }
+    }
 
     /// The build description resulting from planing.
     private let buildDescription = ThreadSafeBox<BuildDescription>()
@@ -59,36 +79,50 @@ public final class BuildOperation: PackageStructureDelegate, SPMBuildCore.BuildS
     /// The output stream for the build delegate.
     private let outputStream: OutputByteStream
 
-    /// The verbosity level to print out at
+    /// The verbosity level to use for diagnostics.
     private let logLevel: Basics.Diagnostic.Severity
 
-    /// File system to operate on
+    /// File system to operate on.
     private let fileSystem: TSCBasic.FileSystem
 
-    /// ObservabilityScope with which to emit diagnostics
+    /// ObservabilityScope with which to emit diagnostics.
     private let observabilityScope: ObservabilityScope
 
     public var builtTestProducts: [BuiltTestProduct] {
         (try? getBuildDescription())?.builtTestProducts ?? []
     }
 
+    /// File rules to determine resource handling behavior.
+    private let additionalFileRules: [FileRuleDescription]
+
+    /// Alternative path to search for pkg-config `.pc` files.
+    private let pkgConfigDirectories: [AbsolutePath]
+
     public init(
         buildParameters: BuildParameters,
         cacheBuildManifest: Bool,
         packageGraphLoader: @escaping () throws -> PackageGraph,
+        additionalFileRules: [FileRuleDescription],
         pluginScriptRunner: PluginScriptRunner,
         pluginWorkDirectory: AbsolutePath,
-        disableSandboxForPluginCommands: Bool = false,
+        disableSandboxForPluginCommands: Bool,
+        pkgConfigDirectories: [AbsolutePath],
         outputStream: OutputByteStream,
         logLevel: Basics.Diagnostic.Severity,
         fileSystem: TSCBasic.FileSystem,
         observabilityScope: ObservabilityScope
     ) {
+        /// Checks if stdout stream is tty.
+        var buildParameters = buildParameters
+        buildParameters.colorizedOutput = outputStream.isTTY
+
         self.buildParameters = buildParameters
         self.cacheBuildManifest = cacheBuildManifest
         self.packageGraphLoader = packageGraphLoader
+        self.additionalFileRules = additionalFileRules
         self.pluginScriptRunner = pluginScriptRunner
         self.pluginWorkDirectory = pluginWorkDirectory
+        self.pkgConfigDirectories = pkgConfigDirectories
         self.disableSandboxForPluginCommands = disableSandboxForPluginCommands
         self.outputStream = outputStream
         self.logLevel = logLevel
@@ -101,7 +135,7 @@ public final class BuildOperation: PackageStructureDelegate, SPMBuildCore.BuildS
             try self.packageGraphLoader()
         }
     }
-    
+
     /// Compute and return the latest build description.
     ///
     /// This will try skip build planning if build manifest caching is enabled
@@ -119,7 +153,7 @@ public final class BuildOperation: PackageStructureDelegate, SPMBuildCore.BuildS
                             throw InternalError("could not find build descriptor at \(buildDescriptionPath)")
                         }
                         // return the build description that's on disk.
-                        return try BuildDescription.load(from: buildDescriptionPath)
+                        return try BuildDescription.load(fileSystem: self.fileSystem, path: buildDescriptionPath)
                     }
                 } catch {
                     // since caching is an optimization, warn about failing to load the cached version
@@ -127,21 +161,93 @@ public final class BuildOperation: PackageStructureDelegate, SPMBuildCore.BuildS
                 }
             }
             // We need to perform actual planning if we reach here.
-            return try self.plan()
+            return try self.plan().description
         }
     }
 
+    public func getBuildManifest() throws -> LLBuildManifest.BuildManifest {
+        return try self.plan().manifest
+    }
+
     /// Cancel the active build operation.
-    public func cancel() {
+    public func cancel(deadline: DispatchTime) throws {
         buildSystem?.cancel()
     }
 
+    // Emit a warning if a target imports another target in this build
+    // without specifying it as a dependency in the manifest
+    private func verifyTargetImports(in description: BuildDescription) throws {
+        let checkingMode = description.explicitTargetDependencyImportCheckingMode
+        guard checkingMode != .none else {
+            return
+        }
+        // Ensure the compiler supports the import-scan operation
+        guard DriverSupport.checkSupportedFrontendFlags(flags: ["import-prescan"], fileSystem: localFileSystem) else {
+            return
+        }
+
+        for (target, commandLine) in description.swiftTargetScanArgs {
+            do {
+                guard let dependencies = description.targetDependencyMap[target] else {
+                    // Skip target if no dependency information is present
+                    continue
+                }
+                let targetDependenciesSet = Set(dependencies)
+                guard !description.generatedSourceTargetSet.contains(target),
+                      targetDependenciesSet.intersection(description.generatedSourceTargetSet).isEmpty else {
+                    // Skip targets which contain, or depend-on-targets, with generated source-code.
+                    // Such as test discovery targets and targets with plugins.
+                    continue
+                }
+                let resolver = try ArgsResolver(fileSystem: localFileSystem)
+                let executor = SPMSwiftDriverExecutor(resolver: resolver,
+                                                      fileSystem: localFileSystem,
+                                                      env: ProcessEnv.vars)
+
+                let consumeDiagnostics: DiagnosticsEngine = DiagnosticsEngine(handlers: [])
+                var driver = try Driver(args: commandLine,
+                                        diagnosticsEngine: consumeDiagnostics,
+                                        fileSystem: localFileSystem,
+                                        executor: executor)
+                guard !consumeDiagnostics.hasErrors else {
+                  // If we could not init the driver with this command, something went wrong,
+                  // proceed without checking this target.
+                  continue
+                }
+                let imports = try driver.performImportPrescan().imports
+                let nonDependencyTargetsSet =
+                    Set(description.targetDependencyMap.keys.filter { !targetDependenciesSet.contains($0) })
+                let importedTargetsMissingDependency = Set(imports).intersection(nonDependencyTargetsSet)
+                if let missedDependency = importedTargetsMissingDependency.first {
+                    switch checkingMode {
+                        case .error:
+                            self.observabilityScope.emit(error: "Target \(target) imports another target (\(missedDependency)) in the package without declaring it a dependency.")
+                        case .warn:
+                            self.observabilityScope.emit(warning: "Target \(target) imports another target (\(missedDependency)) in the package without declaring it a dependency.")
+                        case .none:
+                            fatalError("Explicit import checking is disabled.")
+                    }
+                }
+            } catch {
+                // The above verification is a best-effort attempt to warn the user about a potential manifest
+                // error. If something went wrong during the import-prescan, proceed silently.
+                return
+            }
+        }
+    }
+
     /// Perform a build using the given build description and subset.
     public func build(subset: BuildSubset) throws {
+
         let buildStartTime = DispatchTime.now()
-        
+
         // Get the build description (either a cached one or newly created).
-        let buildDescription = try self.getBuildDescription()
+
+        // Get the build description
+        let buildDescription = try getBuildDescription()
+
+        // Verify dependency imports on the described targers
+        try verifyTargetImports(in: buildDescription)
 
         // Create the build system.
         let buildSystem = try self.createBuildSystem(buildDescription: buildDescription)
@@ -185,7 +291,7 @@ public final class BuildOperation: PackageStructureDelegate, SPMBuildCore.BuildS
             self.observabilityScope.emit(warning: "unable to create symbolic link at \(oldBuildPath): \(error)")
         }
     }
-    
+
     /// Compiles any plugins specified or implied by the build subset, returning
     /// true if the build should proceed. Throws an error in case of failure. A
     /// reason why the build might not proceed even on success is if only plugins
@@ -218,21 +324,47 @@ public final class BuildOperation: PackageStructureDelegate, SPMBuildCore.BuildS
         // build, based on the subset.
         return continueBuilding
     }
-    
+
     // Compiles a single plugin, emitting its output and throwing an error if it
     // fails.
     func compilePlugin(_ plugin: PluginDescription) throws {
         // Compile the plugin, getting back a PluginCompilationResult.
-        let preparationStepName = "Compiling plugin \(plugin.targetName)..."
-        self.buildSystemDelegate?.preparationStepStarted(preparationStepName)
-        let result = try self.pluginScriptRunner.compilePluginScript(
-            sources: plugin.sources,
-            toolsVersion: plugin.toolsVersion,
-            observabilityScope: self.observabilityScope)
-        if !result.description.isEmpty {
-            self.buildSystemDelegate?.preparationStepHadOutput(preparationStepName, output: result.description)
+        class Delegate: PluginScriptCompilerDelegate {
+            let preparationStepName: String
+            let buildSystemDelegate: BuildOperationBuildSystemDelegateHandler?
+            init(preparationStepName: String, buildSystemDelegate: BuildOperationBuildSystemDelegateHandler?) {
+                self.preparationStepName = preparationStepName
+                self.buildSystemDelegate = buildSystemDelegate
+            }
+            func willCompilePlugin(commandLine: [String], environment: EnvironmentVariables) {
+                self.buildSystemDelegate?.preparationStepStarted(preparationStepName)
+            }
+            func didCompilePlugin(result: PluginCompilationResult) {
+                if !result.description.isEmpty {
+                    self.buildSystemDelegate?.preparationStepHadOutput(preparationStepName, output: result.description)
+                }
+                self.buildSystemDelegate?.preparationStepFinished(preparationStepName, result: (result.succeeded ? .succeeded : .failed))
+            }
+            func skippedCompilingPlugin(cachedResult: PluginCompilationResult) {
+                // Historically we have emitted log info about cached plugins that are used. We should reconsider whether this is the right thing to do.
+                self.buildSystemDelegate?.preparationStepStarted(preparationStepName)
+                if !cachedResult.description.isEmpty {
+                    self.buildSystemDelegate?.preparationStepHadOutput(preparationStepName, output: cachedResult.description)
+                }
+                self.buildSystemDelegate?.preparationStepFinished(preparationStepName, result: (cachedResult.succeeded ? .succeeded : .failed))
+            }
+        }
+        let delegate = Delegate(preparationStepName: "Compiling plugin \(plugin.targetName)", buildSystemDelegate: self.buildSystemDelegate)
+        let result = try tsc_await {
+            self.pluginScriptRunner.compilePluginScript(
+                sourceFiles: plugin.sources.paths,
+                pluginName: plugin.targetName,
+                toolsVersion: plugin.toolsVersion,
+                observabilityScope: self.observabilityScope,
+                callbackQueue: DispatchQueue.sharedConcurrent,
+                delegate: delegate,
+                completion: $0)
         }
-        self.buildSystemDelegate?.preparationStepFinished(preparationStepName, result: result.wasCached ? .skipped : (result.succeeded ? .succeeded : .failed))
 
         // Throw an error on failure; we will already have emitted the compiler's output in this case.
         if !result.succeeded {
@@ -262,22 +394,24 @@ public final class BuildOperation: PackageStructureDelegate, SPMBuildCore.BuildS
     }
 
     /// Create the build plan and return the build description.
-    private func plan() throws -> BuildDescription {
+    private func plan() throws -> (description: BuildDescription, manifest: LLBuildManifest.BuildManifest) {
         // Load the package graph.
         let graph = try getPackageGraph()
-        
+
         // Invoke any build tool plugins in the graph to generate prebuild commands and build commands.
         let buildToolPluginInvocationResults = try graph.invokeBuildToolPlugins(
             outputDir: self.pluginWorkDirectory.appending(component: "outputs"),
             builtToolsDir: self.buildParameters.buildPath,
             buildEnvironment: self.buildParameters.buildEnvironment,
-            toolSearchDirectories: [self.buildParameters.toolchain.swiftCompiler.parentDirectory],
+            toolSearchDirectories: [self.buildParameters.toolchain.swiftCompilerPath.parentDirectory],
+            pkgConfigDirectories: self.pkgConfigDirectories,
             pluginScriptRunner: self.pluginScriptRunner,
             observabilityScope: self.observabilityScope,
-            fileSystem: localFileSystem)
-
+            fileSystem: self.fileSystem
+        )
 
         // Surface any diagnostics from build tool plugins.
+        var succeeded = true
         for (target, results) in buildToolPluginInvocationResults {
             // There is one result for each plugin that gets applied to a target.
             for result in results {
@@ -293,9 +427,14 @@ public final class BuildOperation: PackageStructureDelegate, SPMBuildCore.BuildS
                 for diag in result.diagnostics {
                     diagnosticsEmitter.emit(diag)
                 }
+                succeeded = succeeded && result.succeeded
             }
         }
 
+        if !succeeded {
+            throw StringError("build stopped due to build-tool plugin failures")
+        }
+
         // Run any prebuild commands provided by build tool plugins. Any failure stops the build.
         let prebuildCommandResults = try graph.reachableTargets.reduce(into: [:], { partial, target in
             partial[target] = try buildToolPluginInvocationResults[target].map { try self.runPrebuildCommands(for: $0) }
@@ -307,14 +446,14 @@ public final class BuildOperation: PackageStructureDelegate, SPMBuildCore.BuildS
                 // Get the set of unhandled files in targets.
                 var unhandledFiles = Set(target.underlyingTarget.others)
                 if unhandledFiles.isEmpty { continue }
-                
+
                 // Subtract out any that were inputs to any commands generated by plugins.
                 if let result = buildToolPluginInvocationResults[target] {
                     let handledFiles = result.flatMap{ $0.buildCommands.flatMap{ $0.inputFiles } }
                     unhandledFiles.subtract(handledFiles)
                 }
                 if unhandledFiles.isEmpty { continue }
-                
+
                 // Emit a diagnostic if any remain. This is kept the same as the previous message for now, but this could be improved.
                 let diagnosticsEmitter = self.observabilityScope.makeDiagnosticsEmitter {
                     var metadata = ObservabilityMetadata()
@@ -330,17 +469,18 @@ public final class BuildOperation: PackageStructureDelegate, SPMBuildCore.BuildS
                 diagnosticsEmitter.emit(warning: warning)
             }
         }
-        
+
         // Create the build plan based, on the graph and any information from plugins.
         let plan = try BuildPlan(
             buildParameters: buildParameters,
             graph: graph,
+            additionalFileRules: additionalFileRules,
             buildToolPluginInvocationResults: buildToolPluginInvocationResults,
             prebuildCommandResults: prebuildCommandResults,
             fileSystem: self.fileSystem,
             observabilityScope: self.observabilityScope
         )
-        self.buildPlan = plan
+        self._buildPlan = plan
 
         let (buildDescription, buildManifest) = try BuildDescription.create(
             with: plan,
@@ -349,18 +489,8 @@ public final class BuildOperation: PackageStructureDelegate, SPMBuildCore.BuildS
             observabilityScope: self.observabilityScope
         )
 
-        // FIXME: ideally this would be done outside of the planning phase,
-        // but it would require deeper changes in how we serialize BuildDescription
-        // Output a dot graph
-        if buildParameters.printManifestGraphviz {
-            // FIXME: this seems like the wrong place to print
-            var serializer = DOTManifestSerializer(manifest: buildManifest)
-            serializer.writeDOT(to: self.outputStream)
-            self.outputStream.flush()
-        }
-        
         // Finally create the llbuild manifest from the plan.
-        return buildDescription
+        return (buildDescription, buildManifest)
     }
 
     /// Build the package structure target.
@@ -385,6 +515,7 @@ public final class BuildOperation: PackageStructureDelegate, SPMBuildCore.BuildS
         let buildExecutionContext = BuildExecutionContext(
             buildParameters,
             buildDescription: buildDescription,
+            fileSystem: self.fileSystem,
             observabilityScope: self.observabilityScope,
             packageStructureDelegate: self,
             buildErrorAdviceProvider: self
@@ -410,7 +541,7 @@ public final class BuildOperation: PackageStructureDelegate, SPMBuildCore.BuildS
             schedulerLanes: buildParameters.jobs
         )
 
-        // TODO: this seems fragile, perhaps we replace commandFailureHandler by adding relevant calls in the delegates chain 
+        // TODO: this seems fragile, perhaps we replace commandFailureHandler by adding relevant calls in the delegates chain
         buildSystemDelegate.commandFailureHandler = {
             buildSystem.cancel()
             self.delegate?.buildSystemDidCancel(self)
@@ -426,7 +557,7 @@ public final class BuildOperation: PackageStructureDelegate, SPMBuildCore.BuildS
         return try pluginResults.map { pluginResult in
             // As we go we will collect a list of prebuild output directories whose contents should be input to the build,
             // and a list of the files in those directories after running the commands.
-            var derivedSourceFiles: [AbsolutePath] = []
+            var derivedFiles: [AbsolutePath] = []
             var prebuildOutputDirs: [AbsolutePath] = []
             for command in pluginResult.prebuildCommands {
                 self.observabilityScope.emit(info: "Running" + (command.configuration.displayName ?? command.configuration.executable.basename))
@@ -435,9 +566,9 @@ public final class BuildOperation: PackageStructureDelegate, SPMBuildCore.BuildS
                 // TODO: We need to also use any working directory, but that support isn't yet available on all platforms at a lower level.
                 var commandLine = [command.configuration.executable.pathString] + command.configuration.arguments
                 if !self.disableSandboxForPluginCommands {
-                    commandLine = Sandbox.apply(command: commandLine, strictness: .writableTemporaryDirectory, writableDirectories: [pluginResult.pluginOutputDirectory])
+                    commandLine = try Sandbox.apply(command: commandLine, strictness: .writableTemporaryDirectory, writableDirectories: [pluginResult.pluginOutputDirectory])
                 }
-                let processResult = try Process.popen(arguments: commandLine, environment: command.configuration.environment)
+                let processResult = try TSCBasic.Process.popen(arguments: commandLine, environment: command.configuration.environment)
                 let output = try processResult.utf8Output() + processResult.utf8stderrOutput()
                 if processResult.exitStatus != .terminated(code: 0) {
                     throw StringError("failed: \(command)\n\n\(output)")
@@ -445,8 +576,8 @@ public final class BuildOperation: PackageStructureDelegate, SPMBuildCore.BuildS
 
                 // Add any files found in the output directory declared for the prebuild command after the command ends.
                 let outputFilesDir = command.outputFilesDirectory
-                if let swiftFiles = try? localFileSystem.getDirectoryContents(outputFilesDir).sorted() {
-                    derivedSourceFiles.append(contentsOf: swiftFiles.map{ outputFilesDir.appending(component: $0) })
+                if let swiftFiles = try? self.fileSystem.getDirectoryContents(outputFilesDir).sorted() {
+                    derivedFiles.append(contentsOf: swiftFiles.map{ outputFilesDir.appending(component: $0) })
                 }
 
                 // Add the output directory to the list of directories whose structure should affect the build plan.
@@ -454,25 +585,25 @@ public final class BuildOperation: PackageStructureDelegate, SPMBuildCore.BuildS
             }
 
             // Add the results of running any prebuild commands for this invocation.
-            return PrebuildCommandResult(derivedSourceFiles: derivedSourceFiles, outputDirectories: prebuildOutputDirs)
+            return PrebuildCommandResult(derivedFiles: derivedFiles, outputDirectories: prebuildOutputDirs)
         }
     }
-    
+
     public func provideBuildErrorAdvice(for target: String, command: String, message: String) -> String? {
         // Find the target for which the error was emitted.  If we don't find it, we can't give any advice.
-        guard let _ = self.buildPlan?.targets.first(where: { $0.target.name == target }) else { return nil }
-        
+        guard let _ = self._buildPlan?.targets.first(where: { $0.target.name == target }) else { return nil }
+
         // Check for cases involving modules that cannot be found.
         if let importedModule = try? RegEx(pattern: "no such module '(.+)'").matchGroups(in: message).first?.first {
             // A target is importing a module that can't be found.  We take a look at the build plan and see if can offer any advice.
-            
+
             // Look for a target with the same module name as the one that's being imported.
-            if let importedTarget = self.buildPlan?.targets.first(where: { $0.target.c99name == importedModule }) {
+            if let importedTarget = self._buildPlan?.targets.first(where: { $0.target.c99name == importedModule }) {
                 // For the moment we just check for executables that other targets try to import.
                 if importedTarget.target.type == .executable {
                     return "module '\(importedModule)' is the main module of an executable, and cannot be imported by tests and other targets"
                 }
-                
+
                 // Here we can add more checks in the future.
             }
         }
@@ -495,7 +626,7 @@ public final class BuildOperation: PackageStructureDelegate, SPMBuildCore.BuildS
 }
 
 extension BuildDescription {
-    static func create(with plan: BuildPlan, disableSandboxForPluginCommands: Bool, fileSystem: TSCBasic.FileSystem, observabilityScope: ObservabilityScope) throws -> (BuildDescription, BuildManifest) {
+    static func create(with plan: BuildPlan, disableSandboxForPluginCommands: Bool, fileSystem: TSCBasic.FileSystem, observabilityScope: ObservabilityScope) throws -> (BuildDescription, LLBuildManifest.BuildManifest) {
         // Generate the llbuild manifest.
         let llbuild = LLBuildManifestBuilder(plan, disableSandboxForPluginCommands: disableSandboxForPluginCommands, fileSystem: fileSystem, observabilityScope: observabilityScope)
         let buildManifest = try llbuild.generateManifest(at: plan.buildParameters.llbuildManifest)
@@ -503,6 +634,7 @@ extension BuildDescription {
         let swiftCommands = llbuild.manifest.getCmdToolMap(kind: SwiftCompilerTool.self)
         let swiftFrontendCommands = llbuild.manifest.getCmdToolMap(kind: SwiftFrontendTool.self)
         let testDiscoveryCommands = llbuild.manifest.getCmdToolMap(kind: TestDiscoveryTool.self)
+        let testEntryPointCommands = llbuild.manifest.getCmdToolMap(kind: TestEntryPointTool.self)
         let copyCommands = llbuild.manifest.getCmdToolMap(kind: CopyTool.self)
 
         // Create the build description.
@@ -511,6 +643,7 @@ extension BuildDescription {
             swiftCommands: swiftCommands,
             swiftFrontendCommands: swiftFrontendCommands,
             testDiscoveryCommands: testDiscoveryCommands,
+            testEntryPointCommands: testEntryPointCommands,
             copyCommands: copyCommands,
             pluginDescriptions: plan.pluginDescriptions
         )
@@ -518,7 +651,7 @@ extension BuildDescription {
             plan.buildParameters.buildDescriptionPath.parentDirectory,
             recursive: true
         )
-        try buildDescription.write(to: plan.buildParameters.buildDescriptionPath)
+        try buildDescription.write(fileSystem: fileSystem, path: plan.buildParameters.buildDescriptionPath)
         return (buildDescription, buildManifest)
     }
 }
