@@ -1,21 +1,23 @@
-/*
- This source file is part of the Swift.org open source project
-
- Copyright 2015 - 2016 Apple Inc. and the Swift project authors
- Licensed under Apache License v2.0 with Runtime Library Exception
-
- See http://swift.org/LICENSE.txt for license information
- See http://swift.org/CONTRIBUTORS.txt for Swift project authors
-*/
+//===----------------------------------------------------------------------===//
+//
+// This source file is part of the Swift open source project
+//
+// Copyright (c) 2015-2016 Apple Inc. and the Swift project authors
+// Licensed under Apache License v2.0 with Runtime Library Exception
+//
+// See http://swift.org/LICENSE.txt for license information
+// See http://swift.org/CONTRIBUTORS.txt for the list of Swift project authors
+//
+//===----------------------------------------------------------------------===//
 
 import Basics
 import PackageGraph
 import PackageModel
 import LLBuildManifest
 import SPMBuildCore
+@_implementationOnly import DriverSupport
 @_implementationOnly import SwiftDriver
 import TSCBasic
-import TSCUtility
 
 public class LLBuildManifestBuilder {
     public enum TargetKind {
@@ -85,14 +87,15 @@ public class LLBuildManifestBuilder {
             }
         }
 
-        try self.addTestManifestGenerationCommand()
+        try self.addTestDiscoveryGenerationCommand()
+        try self.addTestEntryPointGenerationCommand()
 
         // Create command for all products in the plan.
         for (_, description) in plan.productMap {
             try self.createProductCommand(description)
         }
 
-        try ManifestWriter().write(manifest, at: path)
+        try ManifestWriter(fileSystem: self.fileSystem).write(manifest, at: path)
         return manifest
     }
 
@@ -117,7 +120,7 @@ extension LLBuildManifestBuilder {
                 derivedSourceDirPaths.append(contentsOf: result.outputDirectories)
             }
             inputs.append(contentsOf: derivedSourceDirPaths.sorted().map{ Node.directoryStructure($0) })
-            
+
             // FIXME: Need to handle version-specific manifests.
             inputs.append(file: package.manifest.path)
 
@@ -170,7 +173,7 @@ extension LLBuildManifestBuilder {
         let infoPlistDestination = RelativePath("Info.plist")
 
         // Create a copy command for each resource file.
-        for resource in target.target.underlyingTarget.resources {
+        for resource in target.resources {
             let destination = bundlePath.appending(resource.destination)
             let (_, output) = addCopyCommand(from: resource.path, to: destination)
             outputs.append(output)
@@ -201,7 +204,7 @@ extension LLBuildManifestBuilder {
         let inputs = try self.computeSwiftCompileCmdInputs(target)
 
         // Outputs.
-        let objectNodes = target.objects.map(Node.file)
+        let objectNodes = try target.objects.map(Node.file)
         let moduleNode = Node.file(target.moduleOutputPath)
         let cmdOutputs = objectNodes + [moduleNode]
 
@@ -227,7 +230,7 @@ extension LLBuildManifestBuilder {
         // jobs needed to build this Swift target.
         var commandLine = try target.emitCommandLine();
         commandLine.append("-driver-use-frontend-path")
-        commandLine.append(buildParameters.toolchain.swiftCompiler.pathString)
+        commandLine.append(buildParameters.toolchain.swiftCompilerPath.pathString)
         // FIXME: At some point SwiftPM should provide its own executor for
         // running jobs/launching processes during planning
         let resolver = try ArgsResolver(fileSystem: target.fileSystem)
@@ -265,8 +268,8 @@ extension LLBuildManifestBuilder {
                 }
             }
 
-            let jobInputs = try job.inputs.map { try $0.resolveToNode() }
-            let jobOutputs = try job.outputs.map { try $0.resolveToNode() }
+            let jobInputs = try job.inputs.map { try $0.resolveToNode(fileSystem: self.fileSystem) }
+            let jobOutputs = try job.outputs.map { try $0.resolveToNode(fileSystem: self.fileSystem) }
 
             // Add target dependencies as inputs to the main module build command.
             //
@@ -323,7 +326,7 @@ extension LLBuildManifestBuilder {
     // Consider an example SwiftPM package with two targets: target B, and target A, where A
     // depends on B:
     // SwiftPM will process targets in a topological order and “bubble-up” each target’s
-    // inter-module dependency graph to its dependees. First, SwiftPM will process B, and be
+    // inter-module dependency graph to its dependencies. First, SwiftPM will process B, and be
     // able to plan its full build because it does not have any target dependencies. Then the
     // driver is tasked with planning a build for A. SwiftPM will pass as input to the driver
     // the module dependency graph of its target’s dependencies, in this case, just the
@@ -389,7 +392,7 @@ extension LLBuildManifestBuilder {
         let inputs = try self.computeSwiftCompileCmdInputs(description)
 
         // Outputs.
-        let objectNodes = description.objects.map(Node.file)
+        let objectNodes = try description.objects.map(Node.file)
         let moduleNode = Node.file(description.moduleOutputPath)
         let cmdOutputs = objectNodes + [moduleNode]
 
@@ -409,15 +412,15 @@ extension LLBuildManifestBuilder {
         explicitDependencyJobTracker: UniqueExplicitDependencyJobTracker? = nil
     ) throws {
         // Pass the driver its external dependencies (target dependencies)
-        var dependencyModulePathMap: SwiftDriver.ExternalTargetModulePathMap = [:]
+        var dependencyModuleDetailsMap: SwiftDriver.ExternalTargetModuleDetailsMap = [:]
         // Collect paths for target dependencies of this target (direct and transitive)
-        try self.collectTargetDependencyModulePaths(for: targetDescription.target, dependencyModulePathMap: &dependencyModulePathMap)
+        try self.collectTargetDependencyModuleDetails(for: targetDescription.target, dependencyModuleDetailsMap: &dependencyModuleDetailsMap)
 
         // Compute the set of frontend
         // jobs needed to build this Swift target.
         var commandLine = try targetDescription.emitCommandLine();
         commandLine.append("-driver-use-frontend-path")
-        commandLine.append(buildParameters.toolchain.swiftCompiler.pathString)
+        commandLine.append(buildParameters.toolchain.swiftCompilerPath.pathString)
         commandLine.append("-experimental-explicit-module-build")
         let resolver = try ArgsResolver(fileSystem: self.fileSystem)
         let executor = SPMSwiftDriverExecutor(resolver: resolver,
@@ -426,7 +429,7 @@ extension LLBuildManifestBuilder {
         var driver = try Driver(args: commandLine,
                                 fileSystem: self.fileSystem,
                                 executor: executor,
-                                externalTargetModulePathMap: dependencyModulePathMap,
+                                externalTargetModuleDetailsMap: dependencyModuleDetailsMap,
                                 interModuleDependencyOracle: dependencyOracle
         )
         let jobs = try driver.planBuild()
@@ -437,11 +440,11 @@ extension LLBuildManifestBuilder {
 
     /// Collect a map from all target dependencies of the specified target to the build planning artifacts for said dependency,
     /// in the form of a path to a .swiftmodule file and the dependency's InterModuleDependencyGraph.
-    private func collectTargetDependencyModulePaths(
+    private func collectTargetDependencyModuleDetails(
         for target: ResolvedTarget,
-        dependencyModulePathMap: inout SwiftDriver.ExternalTargetModulePathMap
+        dependencyModuleDetailsMap: inout SwiftDriver.ExternalTargetModuleDetailsMap
     ) throws {
-        for dependency in target.dependencies {
+        for dependency in target.dependencies(satisfying: self.buildEnvironment) {
             switch dependency {
                 case .product:
                     // Product dependencies are broken down into the targets that make them up.
@@ -449,7 +452,7 @@ extension LLBuildManifestBuilder {
                         throw InternalError("unknown dependency product for \(dependency)")
                     }
                     for dependencyProductTarget in dependencyProduct.targets {
-                        try self.addTargetDependencyInfo(for: dependencyProductTarget, dependencyModulePathMap: &dependencyModulePathMap)
+                        try self.addTargetDependencyInfo(for: dependencyProductTarget, dependencyModuleDetailsMap: &dependencyModuleDetailsMap)
 
                     }
                 case .target:
@@ -457,21 +460,21 @@ extension LLBuildManifestBuilder {
                     guard let dependencyTarget = dependency.target else {
                         throw InternalError("unknown dependency target for \(dependency)")
                     }
-                    try self.addTargetDependencyInfo(for: dependencyTarget, dependencyModulePathMap: &dependencyModulePathMap)
+                    try self.addTargetDependencyInfo(for: dependencyTarget, dependencyModuleDetailsMap: &dependencyModuleDetailsMap)
             }
         }
     }
 
     private func addTargetDependencyInfo(
         for target: ResolvedTarget,
-        dependencyModulePathMap: inout SwiftDriver.ExternalTargetModulePathMap
+        dependencyModuleDetailsMap: inout SwiftDriver.ExternalTargetModuleDetailsMap
     ) throws {
         guard case .swift(let dependencySwiftTargetDescription) = plan.targetMap[target] else {
             return
         }
-        dependencyModulePathMap[ModuleDependencyId.swiftPlaceholder(target.c99name)] =
-            dependencySwiftTargetDescription.moduleOutputPath
-        try self.collectTargetDependencyModulePaths(for: target, dependencyModulePathMap: &dependencyModulePathMap)
+        dependencyModuleDetailsMap[ModuleDependencyId.swiftPlaceholder(target.c99name)] =
+            SwiftDriver.ExternalTargetModuleDetails(path: dependencySwiftTargetDescription.moduleOutputPath, isFramework: false)
+        try self.collectTargetDependencyModuleDetails(for: target, dependencyModuleDetailsMap: &dependencyModuleDetailsMap)
     }
 
     private func addSwiftCmdsEmitSwiftModuleSeparately(
@@ -512,12 +515,13 @@ extension LLBuildManifestBuilder {
             name: cmdName,
             inputs: inputs,
             outputs: cmdOutputs,
-            executable: buildParameters.toolchain.swiftCompiler,
+            executable: buildParameters.toolchain.swiftCompilerPath,
             moduleName: target.target.c99name,
+            moduleAliases: target.target.moduleAliases,
             moduleOutputPath: target.moduleOutputPath,
             importPath: buildParameters.buildPath,
             tempsPath: target.tempsPath,
-            objects: target.objects,
+            objects: try target.objects,
             otherArguments: try target.compileArguments(),
             sources: target.sources,
             isLibrary: isLibrary,
@@ -549,14 +553,14 @@ extension LLBuildManifestBuilder {
             // Depend on the binary for executable targets.
             if target.type == .executable {
                 // FIXME: Optimize.
-                let _product = plan.graph.allProducts.first {
-                    $0.type == .executable && $0.executableModule == target
+                let _product = try plan.graph.allProducts.first {
+                    try $0.type == .executable && $0.executableTarget == target
                 }
                 if let product = _product {
                     guard let planProduct = plan.productMap[product] else {
                         throw InternalError("unknown product \(product)")
                     }
-                    inputs.append(file: planProduct.binary)
+                    inputs.append(file: planProduct.binaryPath)
                 }
                 return
             }
@@ -565,7 +569,7 @@ extension LLBuildManifestBuilder {
             case .swift(let target)?:
                 inputs.append(file: target.moduleOutputPath)
             case .clang(let target)?:
-                for object in target.objects {
+                for object in try target.objects {
                     inputs.append(file: object)
                 }
             case nil:
@@ -585,7 +589,7 @@ extension LLBuildManifestBuilder {
                         throw InternalError("unknown product \(product)")
                     }
                     // Establish a dependency on binary of the product.
-                    inputs.append(file: planProduct.binary)
+                    inputs.append(file: planProduct.binaryPath)
 
                 // For automatic and static libraries, and plugins, add their targets as static input.
                 case .library(.automatic), .library(.static), .plugin:
@@ -601,7 +605,7 @@ extension LLBuildManifestBuilder {
 
         for binaryPath in target.libraryBinaryPaths {
             let path = destinationPath(forBinaryAt: binaryPath)
-            if localFileSystem.isDirectory(binaryPath) {
+            if self.fileSystem.isDirectory(binaryPath) {
                 inputs.append(directory: path)
             } else {
                 inputs.append(file: path)
@@ -618,12 +622,12 @@ extension LLBuildManifestBuilder {
                 let displayName = command.configuration.displayName ?? execPath.basename
                 var commandLine = [execPath.pathString] + command.configuration.arguments
                 if !self.disableSandboxForPluginCommands {
-                    commandLine = Sandbox.apply(command: commandLine, strictness: .writableTemporaryDirectory, writableDirectories: [result.pluginOutputDirectory])
+                    commandLine = try Sandbox.apply(command: commandLine, strictness: .writableTemporaryDirectory, writableDirectories: [result.pluginOutputDirectory])
                 }
                 manifest.addShellCmd(
                     name: displayName + "-" + ByteString(encodingAsUTF8: uniquedName).sha256Checksum,
                     description: displayName,
-                    inputs: [.file(execPath)] + command.inputFiles.map{ .file($0) },
+                    inputs: command.inputFiles.map{ .file($0) },
                     outputs: command.outputFiles.map{ .file($0) },
                     arguments: commandLine,
                     environment: command.configuration.environment,
@@ -646,7 +650,7 @@ extension LLBuildManifestBuilder {
             inputs: cmdOutputs,
             outputs: [targetOutput]
         )
-        if plan.graph.isInRootPackages(target.target) {
+        if plan.graph.isInRootPackages(target.target, satisfying: self.buildEnvironment) {
             if !target.isTestTarget {
                 addNode(targetOutput, toTarget: .main)
             }
@@ -655,10 +659,10 @@ extension LLBuildManifestBuilder {
     }
 
     private func addModuleWrapCmd(_ target: SwiftTargetBuildDescription) throws {
-        // Add commands to perform the module wrapping Swift modules when debugging statergy is `modulewrap`.
+        // Add commands to perform the module wrapping Swift modules when debugging strategy is `modulewrap`.
         guard buildParameters.debuggingStrategy == .modulewrap else { return }
         var moduleWrapArgs = [
-            target.buildParameters.toolchain.swiftCompiler.pathString,
+            target.buildParameters.toolchain.swiftCompilerPath.pathString,
             "-modulewrap", target.moduleOutputPath.pathString,
             "-o", target.wrappedModuleOutputPath.pathString
         ]
@@ -739,7 +743,7 @@ extension LLBuildManifestBuilder {
                         throw InternalError("unknown product \(product)")
                     }
                     // Establish a dependency on binary of the product.
-                    let binary = planProduct.binary
+                    let binary = planProduct.binaryPath
                     inputs.append(file: binary)
 
                 case .library(.automatic), .library(.static), .plugin:
@@ -754,7 +758,7 @@ extension LLBuildManifestBuilder {
 
         for binaryPath in target.libraryBinaryPaths {
             let path = destinationPath(forBinaryAt: binaryPath)
-            if localFileSystem.isDirectory(binaryPath) {
+            if self.fileSystem.isDirectory(binaryPath) {
                 inputs.append(directory: path)
             } else {
                 inputs.append(file: path)
@@ -763,9 +767,11 @@ extension LLBuildManifestBuilder {
 
         var objectFileNodes: [Node] = []
 
-        for path in target.compilePaths() {
+        for path in try target.compilePaths() {
             let isCXX = path.source.extension.map{ SupportedLanguageExtension.cppExtensions.contains($0) } ?? false
-            var args = try target.basicArguments(isCXX: isCXX)
+            let isC = path.source.extension.map { $0 == SupportedLanguageExtension.c.rawValue } ?? false
+
+            var args = try target.basicArguments(isCXX: isCXX, isC: isC)
 
             args += ["-MD", "-MT", "dependencies", "-MF", path.deps.pathString]
 
@@ -807,7 +813,7 @@ extension LLBuildManifestBuilder {
             outputs: [output]
         )
 
-        if plan.graph.isInRootPackages(target.target) {
+        if plan.graph.isInRootPackages(target.target, satisfying: self.buildEnvironment) {
             if !target.isTestTarget {
                 addNode(output, toTarget: .main)
             }
@@ -819,17 +825,11 @@ extension LLBuildManifestBuilder {
 // MARK:- Test File Generation
 
 extension LLBuildManifestBuilder {
-    fileprivate func addTestManifestGenerationCommand() throws {
-        for target in plan.targets {
-            guard case .swift(let target) = target,
-                target.isTestTarget,
-                target.testDiscoveryTarget else { continue }
-
-            let testDiscoveryTarget = target
-
+    fileprivate func addTestDiscoveryGenerationCommand() throws {
+        for testDiscoveryTarget in plan.targets.compactMap(\.testDiscoveryTargetBuildDescription) {
             let testTargets = testDiscoveryTarget.target.dependencies
                 .compactMap{ $0.target }.compactMap{ plan.targetMap[$0] }
-            let objectFiles = testTargets.flatMap{ $0.objects }.sorted().map(Node.file)
+            let objectFiles = try testTargets.flatMap{ try $0.objects }.sorted().map(Node.file)
             let outputs = testDiscoveryTarget.target.sources.paths
 
             guard let mainOutput = (outputs.first{ $0.basename == TestDiscoveryTool.mainFileName }) else {
@@ -843,6 +843,48 @@ extension LLBuildManifestBuilder {
             )
         }
     }
+
+    fileprivate func addTestEntryPointGenerationCommand() throws {
+        for target in plan.targets {
+            guard case .swift(let target) = target,
+                  case .entryPoint(let isSynthesized) = target.testTargetRole,
+                  isSynthesized else { continue }
+
+            let testEntryPointTarget = target
+
+            // Get the Swift target build descriptions of all discovery targets this synthesized entry point target depends on.
+            let discoveredTargetDependencyBuildDescriptions = testEntryPointTarget.target.dependencies
+                .compactMap(\.target)
+                .compactMap { plan.targetMap[$0] }
+                .compactMap(\.testDiscoveryTargetBuildDescription)
+
+            // The module outputs of the discovery targets this synthesized entry point target depends on are
+            // considered the inputs to the entry point command.
+            let inputs = discoveredTargetDependencyBuildDescriptions.map { $0.moduleOutputPath }
+
+            let outputs = testEntryPointTarget.target.sources.paths
+
+            guard let mainOutput = (outputs.first{ $0.basename == TestEntryPointTool.mainFileName }) else {
+                throw InternalError("main output (\(TestEntryPointTool.mainFileName)) not found")
+            }
+            let cmdName = mainOutput.pathString
+            manifest.addTestEntryPointCmd(
+                name: cmdName,
+                inputs: inputs.map(Node.file),
+                outputs: outputs.map(Node.file)
+            )
+        }
+    }
+}
+
+private extension TargetBuildDescription {
+    /// If receiver represents a Swift target build description whose test target role is Discovery,
+    /// then this returns that Swift target build description, else returns nil.
+    var testDiscoveryTargetBuildDescription: SwiftTargetBuildDescription? {
+        guard case .swift(let targetBuildDescription) = self,
+              case .discovery = targetBuildDescription.testTargetRole else { return nil }
+        return targetBuildDescription
+    }
 }
 
 // MARK: - Product Command
@@ -851,21 +893,24 @@ extension LLBuildManifestBuilder {
     private func createProductCommand(_ buildProduct: ProductBuildDescription) throws {
         let cmdName = try buildProduct.product.getCommandName(config: buildConfig)
 
-        // Create archive tool for static library and shell tool for rest of the products.
-        if buildProduct.product.type == .library(.static) {
-            manifest.addArchiveCmd(
+        switch buildProduct.product.type {
+        case .library(.static):
+            manifest.addShellCmd(
                 name: cmdName,
+                description: "Archiving \(buildProduct.binaryPath.prettyPath())",
                 inputs: buildProduct.objects.map(Node.file),
-                outputs: [.file(buildProduct.binary)]
+                outputs: [.file(buildProduct.binaryPath)],
+                arguments: try buildProduct.archiveArguments()
             )
-        } else {
-            let inputs = buildProduct.objects + buildProduct.dylibs.map({ $0.binary })
+
+        default:
+            let inputs = buildProduct.objects + buildProduct.dylibs.map({ $0.binaryPath })
 
             manifest.addShellCmd(
                 name: cmdName,
-                description: "Linking \(buildProduct.binary.prettyPath())",
+                description: "Linking \(buildProduct.binaryPath.prettyPath())",
                 inputs: inputs.map(Node.file),
-                outputs: [.file(buildProduct.binary)],
+                outputs: [.file(buildProduct.binaryPath)],
                 arguments: try buildProduct.linkArguments()
             )
         }
@@ -877,7 +922,7 @@ extension LLBuildManifestBuilder {
         manifest.addNode(output, toTarget: targetName)
         manifest.addPhonyCmd(
             name: output.name,
-            inputs: [.file(buildProduct.binary)],
+            inputs: [.file(buildProduct.binaryPath)],
             outputs: [output]
         )
 
@@ -935,7 +980,7 @@ extension LLBuildManifestBuilder {
         from source: AbsolutePath,
         to destination: AbsolutePath
     ) -> (inputNode: Node, outputNode: Node) {
-        let isDirectory = localFileSystem.isDirectory(source)
+        let isDirectory = self.fileSystem.isDirectory(source)
         let nodeType = isDirectory ? Node.directory : Node.file
         let inputNode = nodeType(source)
         let outputNode = nodeType(destination)
@@ -951,11 +996,11 @@ extension LLBuildManifestBuilder {
 extension TypedVirtualPath {
     /// Resolve a typed virtual path provided by the Swift driver to
     /// a node in the build graph.
-    func resolveToNode() throws -> Node {
+    func resolveToNode(fileSystem: FileSystem) throws -> Node {
         if let absolutePath = file.absolutePath {
             return Node.file(absolutePath)
         } else if let relativePath = file.relativePath {
-            guard let workingDirectory = localFileSystem.currentWorkingDirectory else {
+            guard let workingDirectory = fileSystem.currentWorkingDirectory else {
                 throw InternalError("unknown working directory")
             }
             return Node.file(workingDirectory.appending(relativePath))
