@@ -1,22 +1,24 @@
-/*
- This source file is part of the Swift.org open source project
-
- Copyright (c) 2020-2021 Apple Inc. and the Swift project authors
- Licensed under Apache License v2.0 with Runtime Library Exception
-
- See http://swift.org/LICENSE.txt for license information
- See http://swift.org/CONTRIBUTORS.txt for Swift project authors
- */
+//===----------------------------------------------------------------------===//
+//
+// This source file is part of the Swift open source project
+//
+// Copyright (c) 2020-2021 Apple Inc. and the Swift project authors
+// Licensed under Apache License v2.0 with Runtime Library Exception
+//
+// See http://swift.org/LICENSE.txt for license information
+// See http://swift.org/CONTRIBUTORS.txt for the list of Swift project authors
+//
+//===----------------------------------------------------------------------===//
 
 import Basics
 import Dispatch
 import struct Foundation.Data
 import class Foundation.JSONDecoder
 import class Foundation.JSONEncoder
+import class Foundation.NSLock
 import struct Foundation.URL
 import PackageModel
 import TSCBasic
-import TSCUtility
 
 final class SQLitePackageCollectionsStorage: PackageCollectionsStorage, Closable {
     private static let packageCollectionsTableName = "package_collections"
@@ -33,24 +35,25 @@ final class SQLitePackageCollectionsStorage: PackageCollectionsStorage, Closable
     private let decoder: JSONDecoder
 
     private var state = State.idle
-    private let stateLock = Lock()
+    private let stateLock = NSLock()
 
     private let cache = ThreadSafeKeyValueStore<Model.CollectionIdentifier, Model.Collection>()
 
-    // Lock helps prevent concurrency errors with transaction statements during e.g. `refreshCollections`,
+    // NSLock helps prevent concurrency errors with transaction statements during e.g. `refreshCollections`,
     // since only one transaction is allowed per SQLite connection. We need transactions to speed up bulk updates.
     // TODO: we could potentially optimize this with db connection pool
-    private let ftsLock = Lock()
+    private let ftsLock = NSLock()
     // FTS not supported on some platforms; the code falls back to "slow path" in that case
-    internal let useSearchIndices = ThreadSafeBox<Bool>() // internal for testing
+    // marked internal for testing
+    internal let useSearchIndices = ThreadSafeBox<Bool>()
 
     // Targets have in-memory trie in addition to SQLite FTS as optimization
     private let targetTrie = Trie<CollectionPackage>()
     private var targetTrieReady: Bool?
-    private let populateTargetTrieLock = Lock()
+    private let populateTargetTrieLock = NSLock()
 
     init(location: SQLite.Location? = nil, configuration: Configuration = .init(), observabilityScope: ObservabilityScope) {
-        self.location = location ?? .path(localFileSystem.swiftPMCacheDirectory.appending(components: "package-collection.db"))
+        self.location = location ?? (try? .path(localFileSystem.swiftPMCacheDirectory.appending(components: "package-collection.db"))) ?? .memory
         switch self.location {
         case .path, .temporary:
             self.fileSystem = localFileSystem
@@ -72,9 +75,8 @@ final class SQLitePackageCollectionsStorage: PackageCollectionsStorage, Closable
     }
 
     deinit {
-        // TODO: we could wrap the failure here with diagnostics if it wasn't optional throughout
         guard case .disconnected = (try? self.withStateLock { self.state }) else {
-            return assertionFailure("db should be closed")
+            return self.observabilityScope.emit(warning: "SQLitePackageCollectionsStorage de-initialized but db is not closed")
         }
     }
 
