@@ -1,14 +1,17 @@
-/*
- This source file is part of the Swift.org open source project
-
- Copyright (c) 2014 - 2021 Apple Inc. and the Swift project authors
- Licensed under Apache License v2.0 with Runtime Library Exception
-
- See http://swift.org/LICENSE.txt for license information
- See http://swift.org/CONTRIBUTORS.txt for Swift project authors
- */
+//===----------------------------------------------------------------------===//
+//
+// This source file is part of the Swift open source project
+//
+// Copyright (c) 2014-2021 Apple Inc. and the Swift project authors
+// Licensed under Apache License v2.0 with Runtime Library Exception
+//
+// See http://swift.org/LICENSE.txt for license information
+// See http://swift.org/CONTRIBUTORS.txt for the list of Swift project authors
+//
+//===----------------------------------------------------------------------===//
 
 import Basics
+import OrderedCollections
 import PackageLoading
 import PackageModel
 import TSCBasic
@@ -20,13 +23,15 @@ extension PackageGraph {
         root: PackageGraphRoot,
         identityResolver: IdentityResolver,
         additionalFileRules: [FileRuleDescription] = [],
-        externalManifests: OrderedDictionary<PackageIdentity, (manifest: Manifest, fs: FileSystem)>,
+        externalManifests: OrderedCollections.OrderedDictionary<PackageIdentity, (manifest: Manifest, fs: FileSystem)>,
         requiredDependencies: Set<PackageReference> = [],
         unsafeAllowedPackages: Set<PackageReference> = [],
-        binaryArtifacts: [BinaryArtifact] = [],
-        xcTestMinimumDeploymentTargets: [PackageModel.Platform:PlatformVersion] = MinimumDeploymentTarget.default.xcTestMinimumDeploymentTargets,
+        binaryArtifacts: [PackageIdentity: [String: BinaryArtifact]],
         shouldCreateMultipleTestProducts: Bool = false,
         createREPLProduct: Bool = false,
+        customPlatformsRegistry: PlatformRegistry? = .none,
+        customXCTestMinimumDeploymentTargets: [PackageModel.Platform: PlatformVersion]? = .none,
+        testEntryPointPath: AbsolutePath? = nil,
         fileSystem: FileSystem,
         observabilityScope: ObservabilityScope
     ) throws -> PackageGraph {
@@ -39,7 +44,6 @@ extension PackageGraph {
         root.manifests.forEach {
             manifestMap[$0.key] = ($0.value, fileSystem)
         }
-
         let successors: (GraphLoadingNode) -> [GraphLoadingNode] = { node in
             node.requiredDependencies().compactMap{ dependency in
                 return manifestMap[dependency.identity].map { (manifest, fileSystem) in
@@ -48,8 +52,7 @@ extension PackageGraph {
             }
         }
 
-        // Construct the root manifest and root dependencies set.
-        let rootManifestSet = Set(root.manifests.values)
+        // Construct the root root dependencies set.
         let rootDependencies = Set(root.dependencies.compactMap{
             manifestMap[$0.identity]?.manifest
         })
@@ -114,9 +117,9 @@ extension PackageGraph {
                     productFilter: node.productFilter,
                     path: packagePath,
                     additionalFileRules: additionalFileRules,
-                    binaryArtifacts: binaryArtifacts,
-                    xcTestMinimumDeploymentTargets: xcTestMinimumDeploymentTargets,
+                    binaryArtifacts: binaryArtifacts[node.identity] ?? [:],                
                     shouldCreateMultipleTestProducts: shouldCreateMultipleTestProducts,
+                    testEntryPointPath: testEntryPointPath,
                     createREPLProduct: manifest.packageKind.isRoot ? createREPLProduct : false,
                     fileSystem: node.fileSystem,
                     observabilityScope: nodeObservabilityScope
@@ -138,18 +141,21 @@ extension PackageGraph {
             nodes: allNodes,
             identityResolver: identityResolver,
             manifestToPackage: manifestToPackage,
-            rootManifestSet: rootManifestSet,
+            rootManifests: root.manifests,
             unsafeAllowedPackages: unsafeAllowedPackages,
+            platformRegistry: customPlatformsRegistry ?? .default,
+            xcTestMinimumDeploymentTargets: customXCTestMinimumDeploymentTargets ?? MinimumDeploymentTarget.default.xcTestMinimumDeploymentTargets,
             observabilityScope: observabilityScope
         )
 
-        let rootPackages = resolvedPackages.filter{ rootManifestSet.contains($0.manifest) }
+        let rootPackages = resolvedPackages.filter{ root.manifests.values.contains($0.manifest) }
         checkAllDependenciesAreUsed(rootPackages, observabilityScope: observabilityScope)
 
         return try PackageGraph(
             rootPackages: rootPackages,
             rootDependencies: resolvedPackages.filter{ rootDependencies.contains($0.manifest) },
-            dependencies: requiredDependencies
+            dependencies: requiredDependencies,
+            binaryArtifacts: binaryArtifacts
         )
     }
 }
@@ -212,8 +218,10 @@ private func createResolvedPackages(
     identityResolver: IdentityResolver,
     manifestToPackage: [Manifest: Package],
     // FIXME: This shouldn't be needed once <rdar://problem/33693433> is fixed.
-    rootManifestSet: Set<Manifest>,
+    rootManifests: [PackageIdentity: Manifest],
     unsafeAllowedPackages: Set<PackageReference>,
+    platformRegistry: PlatformRegistry,
+    xcTestMinimumDeploymentTargets: [PackageModel.Platform: PlatformVersion],
     observabilityScope: ObservabilityScope
 ) throws -> [ResolvedPackage] {
 
@@ -223,7 +231,8 @@ private func createResolvedPackages(
             return nil
         }
         let isAllowedToVendUnsafeProducts = unsafeAllowedPackages.contains{ $0.identity == package.identity }
-        let allowedToOverride = rootManifestSet.contains(node.manifest)
+        
+        let allowedToOverride = rootManifests.values.contains(node.manifest)
         return ResolvedPackageBuilder(
             package,
             productFilter: node.productFilter,
@@ -238,6 +247,10 @@ private func createResolvedPackages(
         return ($0.package.identity, $0)
     }
 
+    // Resolve module aliases, if specified, for targets and their dependencies
+    // across packages. Aliasing will result in target renaming.
+    let moduleAliasingUsed = try resolveModuleAliases(packageBuilders: packageBuilders, observabilityScope: observabilityScope)
+
     // Scan and validate the dependencies
     for packageBuilder in packageBuilders {
         let package = packageBuilder.package
@@ -246,8 +259,8 @@ private func createResolvedPackages(
             description: "Validating package dependencies",
             metadata: package.diagnosticsMetadata
         )
-
-        var dependencies = OrderedDictionary<PackageIdentity, ResolvedPackageBuilder>()
+        
+        var dependencies = OrderedCollections.OrderedDictionary<PackageIdentity, ResolvedPackageBuilder>()
         var dependenciesByNameForTargetDependencyResolution = [String: ResolvedPackageBuilder]()
 
         // Establish the manifest-declared package dependencies.
@@ -323,8 +336,24 @@ private func createResolvedPackages(
 
         packageBuilder.dependencies = Array(dependencies.values)
 
+        packageBuilder.defaultLocalization = package.manifest.defaultLocalization
+
+        packageBuilder.platforms = computePlatforms(
+            package: package,
+            usingXCTest: false,
+            platformRegistry: platformRegistry,
+            xcTestMinimumDeploymentTargets: xcTestMinimumDeploymentTargets
+        )
+
+        let testPlatforms = computePlatforms(
+            package: package,
+            usingXCTest: true,
+            platformRegistry: platformRegistry,
+            xcTestMinimumDeploymentTargets: xcTestMinimumDeploymentTargets
+        )
+
         // Create target builders for each target in the package.
-        let targetBuilders = package.targets.map{ ResolvedTargetBuilder(target: $0, observabilityScope: observabilityScope) }
+        let targetBuilders = package.targets.map{ ResolvedTargetBuilder(target: $0, observabilityScope: packageObservabilityScope) }
         packageBuilder.targets = targetBuilders
 
         // Establish dependencies between the targets. A target can only depend on another target present in the same package.
@@ -341,6 +370,8 @@ private func createResolvedPackages(
                     return nil
                 }
             }
+            targetBuilder.defaultLocalization = packageBuilder.defaultLocalization
+            targetBuilder.platforms = targetBuilder.target.type == .test ? testPlatforms : packageBuilder.platforms
         }
 
         // Create product builders for each product in the package. A product can only contain a target present in the same package.
@@ -354,27 +385,8 @@ private func createResolvedPackages(
         }
     }
 
-    // Find duplicate products in the package graph.
-    let duplicateProducts = packageBuilders
-        .flatMap({ $0.products })
-        .map({ $0.product })
-        .spm_findDuplicateElements(by: \.name)
-        .map({ $0[0].name })
-
-    // Emit diagnostics for duplicate products.
-    for productName in duplicateProducts {
-        let packages = packageBuilders
-            .filter({ $0.products.contains(where: { $0.product.name == productName }) })
-            .map{ $0.package.identity.description }
-            .sorted()
-
-        observabilityScope.emit(PackageGraphError.duplicateProduct(product: productName, packages: packages))
-    }
-
-    // Remove the duplicate products from the builders.
-    for packageBuilder in packageBuilders {
-        packageBuilder.products = packageBuilder.products.filter({ !duplicateProducts.contains($0.product.name) })
-    }
+    let dupProductsChecker = DuplicateProductsChecker(packageBuilders: packageBuilders)
+    try dupProductsChecker.run(lookupByProductIDs: moduleAliasingUsed, observabilityScope: observabilityScope)
 
     // The set of all target names.
     var allTargetNames = Set<String>()
@@ -401,14 +413,19 @@ private func createResolvedPackages(
                 return false
             })
 
+        let lookupByProductIDs = packageBuilder.package.manifest.disambiguateByProductIDs || moduleAliasingUsed
+
         // Get all the products from dependencies of this package.
         let productDependencies = packageBuilder.dependencies
             .flatMap({ (dependency: ResolvedPackageBuilder) -> [ResolvedProductBuilder] in
                 // Filter out synthesized products such as tests and implicit executables.
-                let explicit = Set(dependency.package.manifest.products.lazy.map({ $0.name }))
-                return dependency.products.filter({ explicit.contains($0.product.name) })
+                // Check if a dependency product is explicitly declared as a product in its package manifest
+                let manifestProducts = dependency.package.manifest.products.lazy.map { $0.name }
+                let explicitProducts = dependency.package.products.filter { manifestProducts.contains($0.name) }
+                let explicitIdsOrNames = Set(explicitProducts.lazy.map({ lookupByProductIDs ? $0.identity : $0.name }))
+                return dependency.products.filter({ lookupByProductIDs ? explicitIdsOrNames.contains($0.product.identity) : explicitIdsOrNames.contains($0.product.name) })
             })
-        let productDependencyMap = productDependencies.spm_createDictionary({ ($0.product.name, $0) })
+        let productDependencyMap = lookupByProductIDs ? productDependencies.spm_createDictionary({ ($0.product.identity, $0) }) : productDependencies.spm_createDictionary({ ($0.product.name, $0) })
 
         // Establish dependencies in each target.
         for targetBuilder in packageBuilder.targets {
@@ -421,7 +438,9 @@ private func createResolvedPackages(
             // Establish product dependencies.
             for case .product(let productRef, let conditions) in targetBuilder.target.dependencies {
                 // Find the product in this package's dependency products.
-                guard let product = productDependencyMap[productRef.name] else {
+                // Look it up by ID if module aliasing is used, otherwise by name.
+                let product = lookupByProductIDs ? productDependencyMap[productRef.identity] : productDependencyMap[productRef.name]
+                guard let product = product else {
                     // Only emit a diagnostic if there are no other diagnostics.
                     // This avoids flooding the diagnostics with product not
                     // found errors when there are more important errors to
@@ -429,7 +448,7 @@ private func createResolvedPackages(
                     if !observabilityScope.errorsReportedInAnyScope {
                         // Emit error if a product (not target) declared in the package is also a productRef (dependency)
                         let declProductsAsDependency = package.products.filter { product in
-                            product.name == productRef.name
+                            lookupByProductIDs ? product.identity == productRef.identity : product.name == productRef.name
                         }.map {$0.targets}.flatMap{$0}.filter { t in
                             t.name != productRef.name
                         }
@@ -486,9 +505,207 @@ private func createResolvedPackages(
             }
         }
     }
+
     return try packageBuilders.map{ try $0.construct() }
 }
 
+fileprivate extension Product {
+    var isDefaultLibrary: Bool {
+        return type == .library(.automatic)
+    }
+}
+
+private class DuplicateProductsChecker {
+    var packageIDToBuilder = [String: ResolvedPackageBuilder]()
+    var checkedPkgIDs = [String]()
+
+    init(packageBuilders: [ResolvedPackageBuilder]) {
+        for packageBuilder in packageBuilders {
+            let pkgID = packageBuilder.package.identity.description.lowercased()
+            self.packageIDToBuilder[pkgID] = packageBuilder
+        }
+    }
+
+    func run(lookupByProductIDs: Bool = false, observabilityScope: ObservabilityScope) throws {
+        var productToPkgMap = [String: [String]]()
+        for (pkgID, pkgBuilder) in packageIDToBuilder {
+            let useProductIDs = pkgBuilder.package.manifest.disambiguateByProductIDs || lookupByProductIDs
+            let depProductRefs = pkgBuilder.package.targets.map{$0.dependencies}.flatMap{$0}.compactMap{$0.product}
+            for depRef in depProductRefs {
+                if let depPkg = depRef.package?.lowercased() {
+                    if !checkedPkgIDs.contains(depPkg) {
+                        checkedPkgIDs.append(depPkg)
+                    }
+                    let depProductIDs = packageIDToBuilder[depPkg]?.package.products.filter { $0.identity == depRef.identity }.map { useProductIDs && $0.isDefaultLibrary ? $0.identity : $0.name } ?? []
+                    for depID in depProductIDs {
+                        productToPkgMap[depID, default: []].append(depPkg)
+                    }
+                } else {
+                    let depPkgs = pkgBuilder.dependencies.filter{$0.products.contains{$0.product.name == depRef.name}}.map{$0.package.identity.description.lowercased()}
+                    productToPkgMap[depRef.name, default: []].append(contentsOf: depPkgs)
+                    checkedPkgIDs.append(contentsOf: depPkgs)
+                }
+                if !checkedPkgIDs.contains(pkgID.lowercased()) {
+                    checkedPkgIDs.append(pkgID.lowercased())
+                }
+            }
+            for (depIDOrName, depPkgs) in productToPkgMap.filter({Set($0.value).count > 1}) {
+                let name = depIDOrName.components(separatedBy: "_").dropFirst().joined(separator: "_")
+                throw PackageGraphError.duplicateProduct(product: name.isEmpty ? depIDOrName : name, packages: depPkgs.sorted())
+            }
+        }
+
+        // Check packages that exist but are not in a dependency graph
+        let untrackedPkgs = packageIDToBuilder.filter{!checkedPkgIDs.contains($0.key.lowercased())}
+        for (pkgID, pkgBuilder) in untrackedPkgs {
+            for product in pkgBuilder.products {
+                // Check if checking product ID only is safe
+                let useIDOnly = lookupByProductIDs && product.product.isDefaultLibrary
+                if !useIDOnly {
+                    // This untracked pkg could have a product name conflicting with a
+                    // product name from another package, but since it's not depended on
+                    // by other packages, keep track of both this product's name and ID
+                    // just in case other packages are < .v5_8
+                    productToPkgMap[product.product.name, default: []].append(pkgID)
+                }
+                productToPkgMap[product.product.identity, default: []].append(pkgID)
+            }
+        }
+
+        let duplicates = productToPkgMap.filter({Set($0.value).count > 1})
+        for (productName, pkgs) in duplicates {
+            throw PackageGraphError.duplicateProduct(product: productName, packages: pkgs.sorted())
+        }
+    }
+}
+
+private func computePlatforms(
+    package: Package,
+    usingXCTest: Bool,
+    platformRegistry: PlatformRegistry,
+    xcTestMinimumDeploymentTargets: [PackageModel.Platform: PlatformVersion]
+) -> SupportedPlatforms {
+
+    // the supported platforms as declared in the manifest
+    let declaredPlatforms: [SupportedPlatform] = package.manifest.platforms.map { platform in
+        let declaredPlatform = platformRegistry.platformByName[platform.platformName]
+            ?? PackageModel.Platform.custom(name: platform.platformName, oldestSupportedVersion: platform.version)
+        return SupportedPlatform(
+            platform: declaredPlatform,
+            version: .init(platform.version),
+            options: platform.options
+        )
+    }
+
+    // the derived platforms based on known minimum deployment target logic
+    var derivedPlatforms = [SupportedPlatform]()
+
+    /// Add each declared platform to the supported platforms list.
+    for platform in package.manifest.platforms {
+        let declaredPlatform = platformRegistry.platformByName[platform.platformName]
+            ?? PackageModel.Platform.custom(name: platform.platformName, oldestSupportedVersion: platform.version)
+        var version = PlatformVersion(platform.version)
+
+        if usingXCTest, let xcTestMinimumDeploymentTarget = xcTestMinimumDeploymentTargets[declaredPlatform], version < xcTestMinimumDeploymentTarget {
+            version = xcTestMinimumDeploymentTarget
+        }
+
+        // If the declared version is smaller than the oldest supported one, we raise the derived version to that.
+        if version < declaredPlatform.oldestSupportedVersion {
+            version = declaredPlatform.oldestSupportedVersion
+        }
+
+        let supportedPlatform = SupportedPlatform(
+            platform: declaredPlatform,
+            version: version,
+            options: platform.options
+        )
+
+        derivedPlatforms.append(supportedPlatform)
+    }
+
+    // Find the undeclared platforms.
+    let remainingPlatforms = Set(platformRegistry.platformByName.keys).subtracting(derivedPlatforms.map({ $0.platform.name }))
+
+    /// Start synthesizing for each undeclared platform.
+    for platformName in remainingPlatforms.sorted() {
+        let platform = platformRegistry.platformByName[platformName]!
+
+        let minimumSupportedVersion: PlatformVersion
+        if usingXCTest, let xcTestMinimumDeploymentTarget = xcTestMinimumDeploymentTargets[platform] {
+            minimumSupportedVersion = xcTestMinimumDeploymentTarget
+        } else {
+            minimumSupportedVersion = platform.oldestSupportedVersion
+        }
+
+        let oldestSupportedVersion: PlatformVersion
+        if platform == .macCatalyst, let iOS = derivedPlatforms.first(where: { $0.platform == .iOS }) {
+            // If there was no deployment target specified for Mac Catalyst, fall back to the iOS deployment target.
+            oldestSupportedVersion = max(minimumSupportedVersion, iOS.version)
+        } else {
+            oldestSupportedVersion = minimumSupportedVersion
+        }
+
+        let supportedPlatform = SupportedPlatform(
+            platform: platform,
+            version: oldestSupportedVersion,
+            options: []
+        )
+
+        derivedPlatforms.append(supportedPlatform)
+    }
+
+    return SupportedPlatforms(
+        declared: declaredPlatforms.sorted(by: { $0.platform.name < $1.platform.name }),
+        derived: derivedPlatforms.sorted(by: { $0.platform.name < $1.platform.name })
+    )
+}
+
+// Track and override module aliases specified for targets in a package graph
+private func resolveModuleAliases(packageBuilders: [ResolvedPackageBuilder],
+                                  observabilityScope: ObservabilityScope) throws -> Bool {
+    // If there are no module aliases specified, return early
+    let hasAliases = packageBuilders.contains { $0.package.targets.contains {
+            $0.dependencies.contains { dep in
+                if case let .product(prodRef, _) = dep {
+                    return prodRef.moduleAliases != nil
+                }
+                return false
+            }
+        }
+    }
+
+    guard hasAliases else { return false }
+    let aliasTracker = ModuleAliasTracker()
+    for packageBuilder in packageBuilders {
+        try aliasTracker.addTargetAliases(targets: packageBuilder.package.targets,
+                                          package: packageBuilder.package.identity)
+    }
+
+    // Track targets that need module aliases for each package
+    for packageBuilder in packageBuilders {
+        for product in packageBuilder.package.products {
+            aliasTracker.trackTargetsPerProduct(product: product,
+                                                package: packageBuilder.package.identity)
+        }
+    }
+
+    // Override module aliases upstream if needed
+    aliasTracker.propagateAliases(observabilityScope: observabilityScope)
+
+    // Validate sources (Swift files only) for modules being aliased.
+    // Needs to be done after `propagateAliases` since aliases defined
+    // upstream can be overriden.
+    for packageBuilder in packageBuilders {
+        for product in packageBuilder.package.products {
+            try aliasTracker.validateAndApplyAliases(product: product,
+                                                     package: packageBuilder.package.identity,
+                                                     observabilityScope: observabilityScope)
+        }
+    }
+    return true
+}
+
 /// A generic builder for `Resolved` models.
 private class ResolvedBuilder<T> {
     /// The constructed object, available after the first call to `construct()`.
@@ -496,7 +713,7 @@ private class ResolvedBuilder<T> {
 
     /// Construct the object with the accumulated data.
     ///
-    /// Note that once the object is constucted, future calls to
+    /// Note that once the object is constructed, future calls to
     /// this method will return the same object.
     final func construct() throws -> T {
         if let constructedObject = _constructedObject {
@@ -560,7 +777,16 @@ private final class ResolvedTargetBuilder: ResolvedBuilder<ResolvedTarget> {
     /// The target dependencies of this target.
     var dependencies: [Dependency] = []
 
-    init(target: Target, observabilityScope: ObservabilityScope) {
+    /// The defaultLocalization for this package
+    var defaultLocalization: String? = nil
+
+    /// The platforms supported by this package.
+    var platforms: SupportedPlatforms = .init(declared: [], derived: [])
+
+    init(
+        target: Target,
+        observabilityScope: ObservabilityScope
+    ) {
         self.target = target
         self.diagnosticsEmitter = observabilityScope.makeDiagnosticsEmitter() {
             var metadata = ObservabilityMetadata()
@@ -572,12 +798,8 @@ private final class ResolvedTargetBuilder: ResolvedBuilder<ResolvedTarget> {
     func diagnoseInvalidUseOfUnsafeFlags(_ product: ResolvedProduct) throws {
         // Diagnose if any target in this product uses an unsafe flag.
         for target in try product.recursiveTargetDependencies() {
-            let declarations = target.underlyingTarget.buildSettings.assignments.keys
-            for decl in declarations {
-                if BuildSettings.Declaration.unsafeSettings.contains(decl) {
-                    self.diagnosticsEmitter.emit(.productUsesUnsafeFlags(product: product.name, target: target.name))
-                    break
-                }
+            if target.underlyingTarget.usesUnsafeFlags {
+                self.diagnosticsEmitter.emit(.productUsesUnsafeFlags(product: product.name, target: target.name))
             }
         }
     }
@@ -586,8 +808,10 @@ private final class ResolvedTargetBuilder: ResolvedBuilder<ResolvedTarget> {
         let dependencies = try self.dependencies.map { dependency -> ResolvedTarget.Dependency in
             switch dependency {
             case .target(let targetBuilder, let conditions):
+                try self.target.validateDependency(target: targetBuilder.target)
                 return .target(try targetBuilder.construct(), conditions: conditions)
             case .product(let productBuilder, let conditions):
+                try self.target.validateDependency(product: productBuilder.product, productPackage: productBuilder.packageBuilder.package.identity)
                 let product = try productBuilder.construct()
                 if !productBuilder.packageBuilder.isAllowedToVendUnsafeProducts {
                     try self.diagnoseInvalidUseOfUnsafeFlags(product)
@@ -596,10 +820,28 @@ private final class ResolvedTargetBuilder: ResolvedBuilder<ResolvedTarget> {
             }
         }
 
-        return ResolvedTarget(target: target, dependencies: dependencies)
+        return ResolvedTarget(
+            target: self.target,
+            dependencies: dependencies,
+            defaultLocalization: self.defaultLocalization,
+            platforms: self.platforms
+        )
     }
 }
 
+extension Target {
+
+  func validateDependency(target: Target) throws {
+    if self.type == .plugin && target.type == .library {
+      throw PackageGraphError.unsupportedPluginDependency(targetName: self.name, dependencyName: target.name, dependencyType: target.type.rawValue, dependencyPackage: nil)
+    }
+  }
+  func validateDependency(product: Product, productPackage: PackageIdentity) throws {
+    if self.type == .plugin && product.type.isLibrary {
+      throw PackageGraphError.unsupportedPluginDependency(targetName: self.name, dependencyName: product.name, dependencyType: product.type.description, dependencyPackage: productPackage.description)
+    }
+  }
+}
 /// Builder for resolved package.
 private final class ResolvedPackageBuilder: ResolvedBuilder<ResolvedPackage> {
 
@@ -609,6 +851,12 @@ private final class ResolvedPackageBuilder: ResolvedBuilder<ResolvedPackage> {
     /// The product filter applied to the package.
     let productFilter: ProductFilter
 
+    /// Package can vend unsafe products
+    let isAllowedToVendUnsafeProducts: Bool
+
+    /// Package can be overridden
+    let allowedToOverride: Bool
+
     /// The targets in the package.
     var targets: [ResolvedTargetBuilder] = []
 
@@ -618,9 +866,11 @@ private final class ResolvedPackageBuilder: ResolvedBuilder<ResolvedPackage> {
     /// The dependencies of this package.
     var dependencies: [ResolvedPackageBuilder] = []
 
-    let isAllowedToVendUnsafeProducts: Bool
+    /// The defaultLocalization for this package.
+    var defaultLocalization: String? = nil
 
-    let allowedToOverride: Bool
+    /// The platforms supported by this package.
+    var platforms: SupportedPlatforms = .init(declared: [], derived: [])
 
     init(_ package: Package, productFilter: ProductFilter, isAllowedToVendUnsafeProducts: Bool, allowedToOverride: Bool) {
         self.package = package
@@ -631,10 +881,12 @@ private final class ResolvedPackageBuilder: ResolvedBuilder<ResolvedPackage> {
 
     override func constructImpl() throws -> ResolvedPackage {
         return ResolvedPackage(
-            package: package,
-            dependencies: try dependencies.map{ try $0.construct() },
-            targets: try targets.map{ try $0.construct() },
-            products: try products.map{ try  $0.construct() }
+            package: self.package,
+            defaultLocalization: self.defaultLocalization,
+            platforms: self.platforms,
+            dependencies: try self.dependencies.map{ try $0.construct() },
+            targets: try self.targets.map{ try $0.construct() },
+            products: try self.products.map{ try $0.construct() }
         )
     }
 }
@@ -647,7 +899,9 @@ fileprivate func findCycle(
     successors: (GraphLoadingNode) throws -> [GraphLoadingNode]
 ) rethrows -> (path: [Manifest], cycle: [Manifest])? {
     // Ordered set to hold the current traversed path.
-    var path = OrderedSet<Manifest>()
+    var path = OrderedCollections.OrderedSet<Manifest>()
+    
+    var fullyVisitedManifests = Set<Manifest>()
 
     // Function to visit nodes recursively.
     // FIXME: Convert to stack.
@@ -655,8 +909,14 @@ fileprivate func findCycle(
       _ node: GraphLoadingNode,
       _ successors: (GraphLoadingNode) throws -> [GraphLoadingNode]
     ) rethrows -> (path: [Manifest], cycle: [Manifest])? {
+        // Once all successors have been visited, this node cannot participate
+        // in a cycle.
+        if fullyVisitedManifests.contains(node.manifest) {
+            return nil
+        }
+        
         // If this node is already in the current path then we have found a cycle.
-        if !path.append(node.manifest) {
+        if !path.append(node.manifest).inserted {
             let index = path.firstIndex(of: node.manifest)! // forced unwrap safe
             return (Array(path[path.startIndex..<index]), Array(path[index..<path.endIndex]))
         }
@@ -669,6 +929,8 @@ fileprivate func findCycle(
         // No cycle found for this node, remove it from the path.
         let item = path.removeLast()
         assert(item == node.manifest)
+        // Track fully visited nodes
+        fullyVisitedManifests.insert(node.manifest)
         return nil
     }
 
