@@ -1,18 +1,21 @@
-/*
- This source file is part of the Swift.org open source project
-
- Copyright (c) 2019 Apple Inc. and the Swift project authors
- Licensed under Apache License v2.0 with Runtime Library Exception
-
- See http://swift.org/LICENSE.txt for license information
- See http://swift.org/CONTRIBUTORS.txt for Swift project authors
- */
+//===----------------------------------------------------------------------===//
+//
+// This source file is part of the Swift open source project
+//
+// Copyright (c) 2019 Apple Inc. and the Swift project authors
+// Licensed under Apache License v2.0 with Runtime Library Exception
+//
+// See http://swift.org/LICENSE.txt for license information
+// See http://swift.org/CONTRIBUTORS.txt for the list of Swift project authors
+//
+//===----------------------------------------------------------------------===//
 
 import Basics
 import Dispatch
+import class Foundation.NSLock
+import OrderedCollections
 import PackageModel
 import TSCBasic
-import TSCUtility
 
 /// The solver that is able to transitively resolve a set of package constraints
 /// specified by a root package.
@@ -37,7 +40,7 @@ public struct PubgrubDependencyResolver {
         /// The current best guess for a solution satisfying all requirements.
         public private(set) var solution: PartialSolution
 
-        private let lock = Lock()
+        private let lock = NSLock()
 
         init(root: DependencyResolutionNode,
              overriddenPackages: [PackageReference: (version: BoundVersion, products: ProductFilter)] = [:],
@@ -77,7 +80,6 @@ public struct PubgrubDependencyResolver {
         func decide(_ node: DependencyResolutionNode, at version: Version) {
             let term = Term(node, .exact(version))
             self.lock.withLock {
-                // FIXME: Shouldn't we check this _before_ making a decision?
                 assert(term.isValidDecision(for: self.solution))
                 self.solution.decide(node, at: version)
             }
@@ -106,10 +108,10 @@ public struct PubgrubDependencyResolver {
     private let packageContainerProvider: PackageContainerProvider
 
     /// Should resolver prefetch the containers.
-    private let prefetchingEnabled: Bool
+    private let prefetchBasedOnResolvedFile: Bool
 
     /// Update containers while fetching them.
-    private let updateEnabled: Bool
+    private let skipDependenciesUpdates: Bool
 
     /// Resolver delegate
     private let delegate: DependencyResolverDelegate?
@@ -117,48 +119,58 @@ public struct PubgrubDependencyResolver {
     public init(
         provider: PackageContainerProvider,
         pinsMap: PinsStore.PinsMap = [:],
-        updateEnabled: Bool = true,
-        prefetchingEnabled: Bool = false,
+        skipDependenciesUpdates: Bool = false,
+        prefetchBasedOnResolvedFile: Bool = false,
         observabilityScope: ObservabilityScope,
         delegate: DependencyResolverDelegate? = nil
     ) {
         self.packageContainerProvider = provider
         self.pinsMap = pinsMap
-        self.updateEnabled = updateEnabled
-        self.prefetchingEnabled = prefetchingEnabled
-        self.provider = ContainerProvider(provider: self.packageContainerProvider, updateEnabled: self.updateEnabled, pinsMap: self.pinsMap, observabilityScope: observabilityScope)
+        self.skipDependenciesUpdates = skipDependenciesUpdates
+        self.prefetchBasedOnResolvedFile = prefetchBasedOnResolvedFile
+        self.provider = ContainerProvider(
+            provider: self.packageContainerProvider,
+            skipUpdate: self.skipDependenciesUpdates,
+            pinsMap: self.pinsMap,
+            observabilityScope: observabilityScope
+        )
         self.delegate = delegate
     }
 
     /// Execute the resolution algorithm to find a valid assignment of versions.
     public func solve(constraints: [Constraint]) -> Result<[DependencyResolver.Binding], Error> {
-        let root = DependencyResolutionNode.root(package: .root(
-            identity: .plain("<synthesized-root>"),
-            path: .root
-        ))
+        // the graph resolution root
+        let root: DependencyResolutionNode
+        if constraints.count == 1, let constraint = constraints.first, constraint.package.kind.isRoot {
+            // root level package, use it as our resolution root
+            root = .root(package: constraint.package)
+        } else {
+            // more complex setup requires a synthesized root
+            root = .root(package: .root(
+                identity: .plain("<synthesized-root>"),
+                path: .root
+            ))
+        }
 
         do {
             // strips state
             return .success(try self.solve(root: root, constraints: constraints).bindings)
         } catch {
-            var error = error
-
             // If version solving failing, build the user-facing diagnostic.
             if let pubGrubError = error as? PubgrubError, let rootCause = pubGrubError.rootCause, let incompatibilities = pubGrubError.incompatibilities {
-                var builder = DiagnosticReportBuilder(
-                    root: root,
-                    incompatibilities: incompatibilities,
-                    provider: self.provider
-                )
-
                 do {
+                    var builder = DiagnosticReportBuilder(
+                        root: root,
+                        incompatibilities: incompatibilities,
+                        provider: self.provider
+                    )
                     let diagnostic = try builder.makeErrorReport(for: rootCause)
-                    error = PubgrubError.unresolvable(diagnostic)
+                    return.failure(PubgrubError.unresolvable(diagnostic))
                 } catch {
                     // failed to construct the report, will report the original error
+                    return .failure(error)
                 }
             }
-
             return .failure(error)
         }
     }
@@ -171,7 +183,7 @@ public struct PubgrubDependencyResolver {
         let inputs = try self.processInputs(root: root, with: constraints)
 
         // Prefetch the containers if prefetching is enabled.
-        if self.prefetchingEnabled {
+        if self.prefetchBasedOnResolvedFile {
             // We avoid prefetching packages that are overridden since
             // otherwise we'll end up creating a repository container
             // for them.
@@ -218,7 +230,9 @@ public struct PubgrubDependencyResolver {
             let updatePackage = try container.underlying.loadPackageReference(at: boundVersion)
 
             if var existing = flattenedAssignments[updatePackage] {
-                assert(existing.binding == boundVersion, "Two products in one package resolved to different versions: \(existing.products)@\(existing.binding) vs \(products)@\(boundVersion)")
+                guard existing.binding == boundVersion else {
+                    throw InternalError("Two products in one package resolved to different versions: \(existing.products)@\(existing.binding) vs \(products)@\(boundVersion)")
+                }
                 existing.products.formUnion(products)
                 flattenedAssignments[updatePackage] = existing
             } else {
@@ -254,7 +268,7 @@ public struct PubgrubDependencyResolver {
         // The list of constraints that we'll be working with. We start with the input constraints
         // and process them in two phases. The first phase finds all unversioned constraints and
         // the second phase discovers all branch-based constraints.
-        var constraints = OrderedSet(constraints)
+        var constraints = OrderedCollections.OrderedSet(constraints)
 
         // The list of packages that are overridden in the graph. A local package reference will
         // always override any other kind of package reference and branch-based reference will override
@@ -262,20 +276,22 @@ public struct PubgrubDependencyResolver {
         var overriddenPackages: [PackageReference: (version: BoundVersion, products: ProductFilter)] = [:]
 
         // The list of version-based references reachable via local and branch-based references.
-        // These are added as top-level incompatibilities since they always need to be statisfied.
+        // These are added as top-level incompatibilities since they always need to be satisfied.
         // Some of these might be overridden as we discover local and branch-based references.
-        var versionBasedDependencies: [DependencyResolutionNode: [VersionBasedConstraint]] = [:]
+        var versionBasedDependencies = OrderedCollections.OrderedDictionary<DependencyResolutionNode, [VersionBasedConstraint]>()
 
         // Process unversioned constraints in first phase. We go through all of the unversioned packages
         // and collect them and their dependencies. This gives us the complete list of unversioned
-        // packages in the graph since unversioned packages can only be refered by other
+        // packages in the graph since unversioned packages can only be referred by other
         // unversioned packages.
         while let constraint = constraints.first(where: { $0.requirement == .unversioned }) {
             constraints.remove(constraint)
 
             // Mark the package as overridden.
             if var existing = overriddenPackages[constraint.package] {
-                assert(existing.version == .unversioned, "Overridden package is not unversioned: \(constraint.package)@\(existing.version)")
+                guard existing.version == .unversioned else {
+                    throw InternalError("Overridden package is not unversioned: \(constraint.package)@\(existing.version)")
+                }
                 existing.products.formUnion(constraint.products)
                 overriddenPackages[constraint.package] = existing
             } else {
@@ -297,7 +313,7 @@ public struct PubgrubDependencyResolver {
                         }
                     } else if !overriddenPackages.keys.contains(dependency.package) {
                         // Add the constraint if its not already present. This will ensure we don't
-                        // end up looping infinitely due to a cycle (which are diagnosed seperately).
+                        // end up looping infinitely due to a cycle (which are diagnosed separately).
                         constraints.append(dependency)
                     }
                 }
@@ -305,7 +321,7 @@ public struct PubgrubDependencyResolver {
         }
 
         // Process revision-based constraints in the second phase. Here we do the similar processing
-        // as the first phase but we also ignore the constraints that are overriden due to
+        // as the first phase but we also ignore the constraints that are overridden due to
         // presence of unversioned constraints.
         while let constraint = constraints.first(where: { $0.requirement.isRevision }) {
             guard case .revision(let revision) = constraint.requirement else {
@@ -318,7 +334,7 @@ public struct PubgrubDependencyResolver {
             switch overriddenPackages[package]?.version {
             case .excluded?, .version?:
                 // These values are not possible.
-                throw InternalError("Unexpected value for overriden package \(package) in \(overriddenPackages)")
+                throw InternalError("Unexpected value for overridden package \(package) in \(overriddenPackages)")
             case .unversioned?:
                 // This package is overridden by an unversioned package so we can ignore this constraint.
                 continue
@@ -370,7 +386,7 @@ public struct PubgrubDependencyResolver {
                     case .versionSet(let req):
                         for node in dependency.nodes() {
                             let versionedBasedConstraint = VersionBasedConstraint(node: node, req: req)
-                            versionBasedDependencies[node, default: []].append(versionedBasedConstraint)
+                            versionBasedDependencies[.root(package: constraint.package), default: []].append(versionedBasedConstraint)
                         }
                     case .revision:
                         constraints.append(dependency)
@@ -386,13 +402,11 @@ public struct PubgrubDependencyResolver {
 
         // At this point, we should be left with only version-based requirements in our constraints
         // list. Add them to our version-based dependency list.
-        for dependency in constraints {
-            switch dependency.requirement {
+        for constraint in constraints {
+            switch constraint.requirement {
             case .versionSet(let req):
-                for node in dependency.nodes() {
+                for node in constraint.nodes() {
                     let versionedBasedConstraint = VersionBasedConstraint(node: node, req: req)
-                    // FIXME: It would be better to record where this constraint came from, instead of just
-                    // using root.
                     versionBasedDependencies[root, default: []].append(versionedBasedConstraint)
                 }
             case .revision, .unversioned:
@@ -401,11 +415,11 @@ public struct PubgrubDependencyResolver {
         }
 
         // Finally, compute the root incompatibilities (which will be all version-based).
+        // note versionBasedDependencies may point to the root package dependencies, or the dependencies of root's non-versioned dependencies
         var rootIncompatibilities: [Incompatibility] = []
         for (node, constraints) in versionBasedDependencies {
             for constraint in constraints {
                 if overriddenPackages.keys.contains(constraint.node.package) { continue }
-
                 let incompat = try Incompatibility(
                     Term(root, .exact("1.0.0")),
                     Term(not: constraint.node, constraint.requirement),
@@ -444,7 +458,7 @@ public struct PubgrubDependencyResolver {
     /// If a conflict is found, the conflicting incompatibility is returned to
     /// resolve the conflict on.
     internal func propagate(state: State, node: DependencyResolutionNode) throws {
-        var changed: OrderedSet<DependencyResolutionNode> = [node]
+        var changed: OrderedCollections.OrderedSet<DependencyResolutionNode> = [node]
 
         while !changed.isEmpty {
             let package = changed.removeFirst()
@@ -513,6 +527,11 @@ public struct PubgrubDependencyResolver {
         var incompatibility = conflict
         var createdIncompatibility = false
 
+        // rdar://93335995
+        // hard protection from infinite loops
+        let maxIterations = 1000
+        var iterations: Int = 0
+
         while !isCompleteFailure(incompatibility, root: state.root) {
             var mostRecentTerm: Term?
             var mostRecentSatisfier: Assignment?
@@ -565,11 +584,15 @@ public struct PubgrubDependencyResolver {
             newTerms += priorCause.terms.filter { $0.node != _mostRecentSatisfier.term.node }
 
             if let _difference = difference {
-                newTerms.append(_difference.inverse)
+                // rdar://93335995
+                // do not add the exact inverse of a requirement as it can lead to endless loops
+                if _difference.inverse != mostRecentTerm {
+                    newTerms.append(_difference.inverse)
+                }
             }
 
             incompatibility = try Incompatibility(
-                OrderedSet(newTerms),
+                OrderedCollections.OrderedSet(newTerms),
                 root: state.root,
                 cause: .conflict(cause: .init(conflict: incompatibility, other: priorCause))
             )
@@ -582,6 +605,13 @@ public struct PubgrubDependencyResolver {
                     self.delegate?.satisfied(term: term, by: _mostRecentSatisfier, incompatibility: incompatibility)
                 }
             }
+
+            // rdar://93335995
+            // hard protection from infinite loops
+            iterations = iterations + 1
+            if iterations >= maxIterations {
+                break
+            }
         }
 
         self.delegate?.failedToResolve(incompatibility: incompatibility)
@@ -721,7 +751,7 @@ private struct DiagnosticReportBuilder {
         }
 
         let stream = BufferedOutputByteStream()
-        let padding = lineNumbers.isEmpty ? 0 : "\(lineNumbers.values.map { $0 }.last!) ".count
+        let padding = lineNumbers.isEmpty ? 0 : "\(Array(lineNumbers.values).last!) ".count
 
         for (idx, line) in lines.enumerated() {
             stream <<< Format.asRepeating(string: " ", count: padding)
@@ -850,14 +880,20 @@ private struct DiagnosticReportBuilder {
 
     private func description(for incompatibility: Incompatibility) throws -> String {
         switch incompatibility.cause {
-        case .dependency(node: _):
+        case .dependency(let causeNode):
             assert(incompatibility.terms.count == 2)
             let depender = incompatibility.terms.first!
             let dependee = incompatibility.terms.last!
             assert(depender.isPositive)
             assert(!dependee.isPositive)
 
-            let dependerDesc = try description(for: depender, normalizeRange: true)
+            let dependerDesc: String
+            // when depender is the root node, the causeNode may be different as it may represent root's indirect dependencies (e.g. dependencies of root's unversioned dependencies)
+            if depender.node == self.rootNode, causeNode != self.rootNode {
+                dependerDesc = causeNode.nameForDiagnostics
+            } else {
+                dependerDesc = try description(for: depender, normalizeRange: true)
+            }
             let dependeeDesc = try description(for: dependee)
             return "\(dependerDesc) depends on \(dependeeDesc)"
         case .noAvailableVersion:
@@ -871,6 +907,8 @@ private struct DiagnosticReportBuilder {
             let term = incompatibility.terms.first!
             assert(term.isPositive)
             return "\(term.node.nameForDiagnostics) is \(term.requirement)"
+        case .conflict where incompatibility.terms.count == 1 && incompatibility.terms.first?.node == self.rootNode:
+            return "dependencies could not be resolved"
         case .conflict:
             break
         case .versionBasedDependencyContainsUnversionedDependency(let versionedDependency, let unversionedDependency):
@@ -880,10 +918,6 @@ private struct DiagnosticReportBuilder {
             return "\(try description(for: term, normalizeRange: true)) contains incompatible tools version (\(version))"
         }
 
-        if isFailure(incompatibility) {
-            return "dependencies could not be resolved"
-        }
-
         let terms = incompatibility.terms
         if terms.count == 1 {
             let term = terms.first!
@@ -957,11 +991,6 @@ private struct DiagnosticReportBuilder {
         return !lineNumbers.keys.contains(complex)
     }
 
-    // FIXME: This is duplicated and wrong.
-    private func isFailure(_ incompatibility: Incompatibility) -> Bool {
-        return incompatibility.terms.count == 1 && incompatibility.terms.first?.node.package.identity == .plain("<synthesized-root>")
-    }
-
     private func description(for term: Term, normalizeRange: Bool = false) throws -> String {
         let name = term.node.nameForDiagnostics
 
@@ -1086,12 +1115,16 @@ internal final class PubGrubPackageContainer {
     /// Returns the best available version for a given term.
     func getBestAvailableVersion(for term: Term) throws -> Version? {
         assert(term.isPositive, "Expected term to be positive")
-        var versionSet = term.requirement
+        let versionSet = term.requirement
 
         // Restrict the selection to the pinned version if is allowed by the current requirements.
         if let pinnedVersion = self.pinnedVersion {
             if versionSet.contains(pinnedVersion) {
-                versionSet = .exact(pinnedVersion)
+                // Make sure the pinned version is still available
+                let version = try self.underlying.versionsDescending().first { pinnedVersion == $0 }
+                if version != nil {
+                    return version
+                }
             }
         }
 
@@ -1177,7 +1210,7 @@ internal final class PubGrubPackageContainer {
                 return [try Incompatibility(Term(node, .exact(version)), root: root, cause: cause)]
             }
 
-            // Skip if this package is overriden.
+            // Skip if this package is overridden.
             if overriddenPackages.keys.contains(dep.package) {
                 continue
             }
@@ -1204,7 +1237,7 @@ internal final class PubGrubPackageContainer {
                     return nil
                 }
 
-                var terms: OrderedSet<Term> = []
+                var terms: OrderedCollections.OrderedSet<Term> = []
                 // the package version requirement
                 terms.append(Term(node, .exact(version)))
                 // the dependency's version requirement
@@ -1224,7 +1257,7 @@ private final class ContainerProvider {
     private let underlying: PackageContainerProvider
 
     /// Whether to perform update (git fetch) on existing cloned repositories or not.
-    private let updateEnabled: Bool
+    private let skipUpdate: Bool
 
     /// Reference to the pins store.
     private let pinsMap: PinsStore.PinsMap
@@ -1238,9 +1271,14 @@ private final class ContainerProvider {
     //// Store prefetches synchronization
     private var prefetches = ThreadSafeKeyValueStore<PackageReference, DispatchGroup>()
 
-    init(provider underlying: PackageContainerProvider, updateEnabled: Bool, pinsMap: PinsStore.PinsMap, observabilityScope: ObservabilityScope) {
+    init(
+        provider underlying: PackageContainerProvider,
+        skipUpdate: Bool,
+        pinsMap: PinsStore.PinsMap,
+        observabilityScope: ObservabilityScope
+    ) {
         self.underlying = underlying
-        self.updateEnabled = updateEnabled
+        self.skipUpdate = skipUpdate
         self.pinsMap = pinsMap
         self.observabilityScope = observabilityScope
     }
@@ -1274,7 +1312,12 @@ private final class ContainerProvider {
             }
         } else {
             // Otherwise, fetch the container from the provider
-            self.underlying.getContainer(for: package, skipUpdate: !self.updateEnabled, observabilityScope: self.observabilityScope, on: .sharedConcurrent) { result in
+            self.underlying.getContainer(
+                for: package,
+                skipUpdate: self.skipUpdate,
+                observabilityScope: self.observabilityScope.makeChildScope(description: "getting package container", metadata: package.diagnosticsMetadata),
+                on: .sharedConcurrent
+            ) { result in
                 let result = result.tryMap { container -> PubGrubPackageContainer in
                     let pubGrubContainer = PubGrubPackageContainer(underlying: container, pinsMap: self.pinsMap)
                     // only cache positive results
@@ -1298,7 +1341,12 @@ private final class ContainerProvider {
                 return group
             }
             if needsFetching {
-                self.underlying.getContainer(for: identifier, skipUpdate: !self.updateEnabled, observabilityScope: self.observabilityScope, on: .sharedConcurrent) { result in
+                self.underlying.getContainer(
+                    for: identifier,
+                    skipUpdate: self.skipUpdate,
+                    observabilityScope: self.observabilityScope.makeChildScope(description: "prefetcing package container", metadata: identifier.diagnosticsMetadata),
+                    on: .sharedConcurrent
+                ) { result in
                     defer { self.prefetches[identifier]?.leave() }
                     // only cache positive results
                     if case .success(let container) = result {
@@ -1393,6 +1441,7 @@ private extension PackageRequirement {
 
 private extension DependencyResolutionNode {
     var nameForDiagnostics: String {
-        return "'\(package.identity)'"
+        return "'\(self.package.identity)'"
     }
 }
+
