@@ -1,18 +1,19 @@
-/*
- This source file is part of the Swift.org open source project
-
- Copyright (c) 2014 - 2021 Apple Inc. and the Swift project authors
- Licensed under Apache License v2.0 with Runtime Library Exception
-
- See http://swift.org/LICENSE.txt for license information
- See http://swift.org/CONTRIBUTORS.txt for Swift project authors
- */
+//===----------------------------------------------------------------------===//
+//
+// This source file is part of the Swift open source project
+//
+// Copyright (c) 2014-2021 Apple Inc. and the Swift project authors
+// Licensed under Apache License v2.0 with Runtime Library Exception
+//
+// See http://swift.org/LICENSE.txt for license information
+// See http://swift.org/CONTRIBUTORS.txt for the list of Swift project authors
+//
+//===----------------------------------------------------------------------===//
 
 import Basics
-import Foundation
+@_implementationOnly import Foundation
 import PackageModel
 import TSCBasic
-import TSCUtility
 
 enum ManifestJSONParser {
     private static let filePrefix = "file://"
@@ -53,8 +54,13 @@ enum ManifestJSONParser {
         manifest.swiftLanguageVersions = try Self.parseSwiftLanguageVersion(package)
         manifest.products = try package.getArray("products").map(ProductDescription.init(v4:))
         manifest.providers = try? package.getArray("providers").map(SystemPackageProviderDescription.init(v4:))
-        manifest.targets = try package.getArray("targets").map(Self.parseTarget(json:))
-        manifest.dependencies = try package.getArray("dependencies").map{
+        manifest.targets = try package.getArray("targets").map {
+            try Self.parseTarget(
+                json: $0,
+                identityResolver: identityResolver
+            )
+        }
+        manifest.dependencies = try package.getArray("dependencies").map {
             try Self.parseDependency(
                 json: $0,
                 toolsVersion: toolsVersion,
@@ -96,29 +102,35 @@ enum ManifestJSONParser {
             if type == "fileSystem" {
                 let name: String? = kindJSON.get("name")
                 let path: String = try kindJSON.get("path")
-                return try Self.makeFileSystemDependency(
+                return try Self.parseFileSystemDependency(
                     packageKind: packageKind,
                     at: path,
                     name: name,
                     identityResolver: identityResolver,
-                    fileSystem: fileSystem)
+                    fileSystem: fileSystem
+                )
             } else if type == "sourceControl" {
                 let name: String? = kindJSON.get("name")
                 let location: String = try kindJSON.get("location")
                 let requirementJSON: JSON = try kindJSON.get("requirement")
                 let requirement = try PackageDependency.SourceControl.Requirement(v4: requirementJSON)
-                return try Self.makeSourceControlDependency(
+                return try Self.parseSourceControlDependency(
                     packageKind: packageKind,
                     at: location,
                     name: name,
                     requirement: requirement,
                     identityResolver: identityResolver,
-                    fileSystem: fileSystem)
+                    fileSystem: fileSystem
+                )
             } else if type == "registry" {
                 let identity: String = try kindJSON.get("identity")
                 let requirementJSON: JSON = try kindJSON.get("requirement")
                 let requirement = try PackageDependency.Registry.Requirement(v4: requirementJSON)
-                return .registry(identity: .plain(identity), requirement: requirement, productFilter: .everything)
+                return try Self.parseRegistryDependency(
+                    identity: .plain(identity),
+                    requirement: requirement,
+                    identityResolver: identityResolver
+                )
             } else {
                 throw InternalError("Unknown dependency type \(kindJSON)")
             }
@@ -132,39 +144,40 @@ enum ManifestJSONParser {
             let requirementType: String = try requirementJSON.get(String.self, forKey: "type")
             switch requirementType {
             case "localPackage":
-                return try Self.makeFileSystemDependency(
+                return try Self.parseFileSystemDependency(
                     packageKind: packageKind,
                     at: url,
                     name: name,
                     identityResolver: identityResolver,
-                    fileSystem: fileSystem)
-
+                    fileSystem: fileSystem
+                )
             default:
                 let requirement = try PackageDependency.SourceControl.Requirement(v4: requirementJSON)
-                return try Self.makeSourceControlDependency(
+                return try Self.parseSourceControlDependency(
                     packageKind: packageKind,
                     at: url,
                     name: name,
                     requirement: requirement,
                     identityResolver: identityResolver,
-                    fileSystem: fileSystem)
+                    fileSystem: fileSystem
+                )
             }
         }
     }
 
-    private static func makeFileSystemDependency(
+    private static func parseFileSystemDependency(
         packageKind: PackageReference.Kind,
         at location: String,
         name: String?,
         identityResolver: IdentityResolver,
         fileSystem: TSCBasic.FileSystem
     ) throws -> PackageDependency {
-        let location = try fixDependencyLocation(fileSystem: fileSystem, packageKind: packageKind, dependencyLocation: location)
+        let location = try sanitizeDependencyLocation(fileSystem: fileSystem, packageKind: packageKind, dependencyLocation: location)
         let path: AbsolutePath
         do {
             path = try AbsolutePath(validating: location)
         } catch PathValidationError.invalidAbsolutePath(let path) {
-            throw ManifestParseError.invalidManifestFormat("'\(path)' is not a valid path for path-based dependencies; use relative or absolute path instead.", diagnosticFile: nil)
+            throw ManifestParseError.invalidManifestFormat("'\(path)' is not a valid path for path-based dependencies; use relative or absolute path instead.", diagnosticFile: nil, compilerCommandLine: nil)
         }
         let identity = try identityResolver.resolveIdentity(for: path)
         return .fileSystem(identity: identity,
@@ -173,7 +186,7 @@ enum ManifestJSONParser {
                            productFilter: .everything)
     }
 
-    private static func makeSourceControlDependency(
+    private static func parseSourceControlDependency(
         packageKind: PackageReference.Kind,
         at location: String,
         name: String?,
@@ -182,11 +195,28 @@ enum ManifestJSONParser {
         fileSystem: TSCBasic.FileSystem
     ) throws -> PackageDependency {
         // cleans up variants of path based location
-        var location = try fixDependencyLocation(fileSystem: fileSystem, packageKind: packageKind, dependencyLocation: location)
+        var location = try sanitizeDependencyLocation(fileSystem: fileSystem, packageKind: packageKind, dependencyLocation: location)
         // location mapping (aka mirrors) if any
         location = identityResolver.mappedLocation(for: location)
-        // a package in a git location, may be a remote URL or on disk
-        if let localPath = try? AbsolutePath(validating: location) {
+        if PackageIdentity.plain(location).scopeAndName != nil {
+            // re-mapped to registry
+            let identity = PackageIdentity.plain(location)
+            let registryRequirement: PackageDependency.Registry.Requirement
+            switch requirement {
+            case .branch, .revision:
+                throw StringError("invalid mapping of source control to registry, requirement information mismatch: cannot map branch or revision based dependencies to registry.")
+            case .exact(let value):
+                registryRequirement = .exact(value)
+            case .range(let value):
+                registryRequirement = .range(value)
+            }
+            return .registry(
+                identity: identity,
+                requirement: registryRequirement,
+                productFilter: .everything
+            )
+        } else if let localPath = try? AbsolutePath(validating: location) {
+            // a package in a git location, may be a remote URL or on disk
             // in the future this will check with the registries for the identity of the URL
             let identity = try identityResolver.resolveIdentity(for: localPath)
             return .localSourceControl(
@@ -211,26 +241,47 @@ enum ManifestJSONParser {
         }
     }
 
-    private static func fixDependencyLocation(fileSystem: TSCBasic.FileSystem, packageKind: PackageReference.Kind, dependencyLocation: String) throws -> String {
-        let packagePath: AbsolutePath
-        switch packageKind {
-        case .root(let path):
-            packagePath = path
-        case .fileSystem(let path):
-            packagePath = path
-        case .localSourceControl(let path):
-            packagePath = path
-        case .remoteSourceControl:
-            // nothing to fix
-            return dependencyLocation
-        case .registry:
-            // nothing to fix
-            return dependencyLocation
+    private static func parseRegistryDependency(
+        identity: PackageIdentity,
+        requirement: PackageDependency.Registry.Requirement,
+        identityResolver: IdentityResolver
+    ) throws -> PackageDependency {
+        // location mapping (aka mirrors) if any
+        let location = identityResolver.mappedLocation(for: identity.description)
+        if PackageIdentity.plain(location).scopeAndName != nil {
+            // re-mapped to registry
+            let identity = PackageIdentity.plain(location)
+            return .registry(
+                identity: identity,
+                requirement: requirement,
+                productFilter: .everything
+            )
+        } else if let url = URL(string: location){
+            // in the future this will check with the registries for the identity of the URL
+            let identity = try identityResolver.resolveIdentity(for: url)
+            let sourceControlRequirement: PackageDependency.SourceControl.Requirement
+            switch requirement {
+            case .exact(let value):
+                sourceControlRequirement = .exact(value)
+            case .range(let value):
+                sourceControlRequirement = .range(value)
+            }
+            return .remoteSourceControl(
+                identity: identity,
+                nameForTargetDependencyResolutionOnly: identity.description,
+                url: url,
+                requirement: sourceControlRequirement,
+                productFilter: .everything
+            )
+        } else {
+            throw StringError("invalid location: \(location)")
         }
+    }
 
+    private static func sanitizeDependencyLocation(fileSystem: TSCBasic.FileSystem, packageKind: PackageReference.Kind, dependencyLocation: String) throws -> String {
         if dependencyLocation.hasPrefix("~/") {
             // If the dependency URL starts with '~/', try to expand it.
-            return AbsolutePath(String(dependencyLocation.dropFirst(2)), relativeTo: fileSystem.homeDirectory).pathString
+            return try AbsolutePath(validating: String(dependencyLocation.dropFirst(2)), relativeTo: fileSystem.homeDirectory).pathString
         } else if dependencyLocation.hasPrefix(filePrefix) {
             // FIXME: SwiftPM can't handle file locations with file:// scheme so we need to
             // strip that. We need to design a Location data structure for SwiftPM.
@@ -239,21 +290,27 @@ enum ManifestJSONParser {
             guard hostnameComponent.isEmpty else {
               if hostnameComponent == ".." {
                 throw ManifestParseError.invalidManifestFormat(
-                  "file:// URLs cannot be relative, did you mean to use '.package(path:)'?", diagnosticFile: nil
+                  "file:// URLs cannot be relative, did you mean to use '.package(path:)'?", diagnosticFile: nil, compilerCommandLine: nil
                 )
               }
               throw ManifestParseError.invalidManifestFormat(
-                "file:// URLs with hostnames are not supported, are you missing a '/'?", diagnosticFile: nil
+                "file:// URLs with hostnames are not supported, are you missing a '/'?", diagnosticFile: nil, compilerCommandLine: nil
               )
             }
-            return AbsolutePath(location).pathString
-        } else if URL.scheme(dependencyLocation) == nil {
-            // If the dependency URL is not remote, try to "fix" it.
+            return try AbsolutePath(validating: location).pathString
+        } else if parseScheme(dependencyLocation) == nil {
             // If the URL has no scheme, we treat it as a path (either absolute or relative to the base URL).
-            return AbsolutePath(dependencyLocation, relativeTo: packagePath).pathString
+            switch packageKind {
+            case .root(let packagePath), .fileSystem(let packagePath), .localSourceControl(let packagePath):
+                return try AbsolutePath(validating: dependencyLocation, relativeTo: packagePath).pathString
+            case .remoteSourceControl, .registry:
+                // nothing to "fix"
+                return dependencyLocation
+            }
+        } else {
+            // nothing to "fix"
+            return dependencyLocation
         }
-
-        return dependencyLocation
     }
 
     private static func parsePlatforms(_ package: JSON) throws -> [PlatformDescription] {
@@ -319,7 +376,10 @@ enum ManifestJSONParser {
         return platforms
     }
 
-    private static func parseTarget(json: JSON) throws -> TargetDescription {
+    private static func parseTarget(
+        json: JSON,
+        identityResolver: IdentityResolver
+    ) throws -> TargetDescription {
         let providers = try? json
             .getArray("providers")
             .map(SystemPackageProviderDescription.init(v4:))
@@ -328,7 +388,12 @@ enum ManifestJSONParser {
 
         let dependencies = try json
             .getArray("dependencies")
-            .map(TargetDescription.Dependency.init(v4:))
+            .map {
+                try TargetDescription.Dependency.init(
+                    v4: $0,
+                    identityResolver: identityResolver
+                )
+            }
 
         let sources: [String]? = try? json.get("sources")
         try sources?.forEach{ _ = try RelativePath(validating: $0) }
@@ -362,12 +427,18 @@ enum ManifestJSONParser {
     private static func parseResources(_ json: JSON) throws -> [TargetDescription.Resource] {
         guard let resourcesJSON = try? json.getArray("resources") else { return [] }
         return try resourcesJSON.map { json in
-            let rawRule = try json.get(String.self, forKey: "rule")
-            let rule = TargetDescription.Resource.Rule(rawValue: rawRule)!
+            let rule = try json.get(String.self, forKey: "rule")
             let path = try RelativePath(validating: json.get(String.self, forKey: "path"))
-            let localizationString = try? json.get(String.self, forKey: "localization")
-            let localization = localizationString.map({ TargetDescription.Resource.Localization(rawValue: $0)! })
-            return .init(rule: rule, path: path.pathString, localization: localization)
+            switch rule {
+            case "process":
+                let localizationString = try? json.get(String.self, forKey: "localization")
+                let localization = localizationString.map({ TargetDescription.Resource.Localization(rawValue: $0)! })
+                return .init(rule: .process(localization: localization), path: path.pathString)
+            case "copy":
+                return .init(rule: .copy, path: path.pathString)
+            default:
+                throw InternalError("invalid resource rule \(rule)")
+            }
         }
     }
 
@@ -388,30 +459,94 @@ enum ManifestJSONParser {
             try Self.parseBuildSetting($0, tool: tool)
         })
     }
-
+    
     private static func parseBuildSetting(_ json: JSON, tool: TargetBuildSettingDescription.Tool) throws -> TargetBuildSettingDescription.Setting {
         let json = try json.getJSON("data")
-        let name = try TargetBuildSettingDescription.SettingName(rawValue: json.get("name"))!
+        let name = try json.get(String.self, forKey: "name")
+        let values = try json.get([String].self, forKey: "value")
         let condition = try (try? json.getJSON("condition")).flatMap(PackageConditionDescription.init(v4:))
-
-        let value = try json.get([String].self, forKey: "value")
-
+        
         // Diagnose invalid values.
-        for item in value {
+        for item in values {
             let groups = Self.invalidValueRegex.matchGroups(in: item).flatMap{ $0 }
             if !groups.isEmpty {
                 let error = "the build setting '\(name)' contains invalid component(s): \(groups.joined(separator: " "))"
                 throw ManifestParseError.runtimeManifestErrors([error])
             }
         }
-
+        
+        let kind: TargetBuildSettingDescription.Kind
+        switch name {
+        case "headerSearchPath":
+            guard let value = values.first else {
+                throw InternalError("invalid (empty) build settings value")
+            }
+            kind = .headerSearchPath(value)
+        case "define":
+            guard let value = values.first else {
+                throw InternalError("invalid (empty) build settings value")
+            }
+            kind = .define(value)
+        case "linkedLibrary":
+            guard let value = values.first else {
+                throw InternalError("invalid (empty) build settings value")
+            }
+            kind = .linkedLibrary(value)
+        case "linkedFramework":
+            guard let value = values.first else {
+                throw InternalError("invalid (empty) build settings value")
+            }
+            kind = .linkedFramework(value)
+        case "enableUpcomingFeature":
+            guard let value = values.first else {
+                throw InternalError("invalid (empty) build settings value")
+            }
+            kind = .enableUpcomingFeature(value)
+        case "enableExperimentalFeature":
+            guard let value = values.first else {
+                throw InternalError("invalid (empty) build settings value")
+            }
+            kind = .enableExperimentalFeature(value)
+        case "unsafeFlags":
+            kind = .unsafeFlags(values)
+        default:
+            throw InternalError("invalid build setting \(name)")
+        }
+        
         return .init(
-            tool: tool, name: name,
-            value: value,
+            tool: tool,
+            kind: kind,
             condition: condition
         )
     }
 
+    /// Parses the URL type of a git repository
+    /// e.g. https://github.com/apple/swift returns "https"
+    /// e.g. git@github.com:apple/swift returns "git"
+    ///
+    /// This is *not* a generic URI scheme parser!
+    private static func parseScheme(_ location: String) -> String? {
+        func prefixOfSplitBy(_ delimiter: String) -> String? {
+            let (head, tail) = location.spm_split(around: delimiter)
+            if tail == nil {
+                //not found
+                return nil
+            } else {
+                //found, return head
+                //lowercase the "scheme", as specified by the URI RFC (just in case)
+                return head.lowercased()
+            }
+        }
+
+        for delim in ["://", "@"] {
+            if let found = prefixOfSplitBy(delim), !found.contains("/") {
+                return found
+            }
+        }
+
+        return nil
+    }
+
     /// Looks for Xcode-style build setting macros "$()".
     private static let invalidValueRegex = try! RegEx(pattern: #"(\$\(.*?\))"#)
 }
@@ -559,7 +694,7 @@ extension TargetDescription.TargetType {
 }
 
 extension TargetDescription.Dependency {
-    fileprivate init(v4 json: JSON) throws {
+    fileprivate init(v4 json: JSON, identityResolver: IdentityResolver) throws {
         let type = try json.get(String.self, forKey: "type")
         let condition = try (try? json.getJSON("condition")).flatMap(PackageConditionDescription.init(v4:))
 
@@ -569,7 +704,17 @@ extension TargetDescription.Dependency {
 
         case "product":
             let name = try json.get(String.self, forKey: "name")
-            self = .product(name: name, package: json.get("package"), condition: condition)
+            let moduleAliases: [String: String]? = try? json.get("moduleAliases")
+            var package: String? = json.get("package")
+            if let packageName = package {
+                package = try identityResolver.mappedIdentity(for: .plain(packageName)).description
+            }
+            self = .product(
+                name: name,
+                package: package,
+                moduleAliases: moduleAliases,
+                condition: condition
+            )
 
         case "byname":
             self = try .byName(name: json.get("name"), condition: condition)
