@@ -1,15 +1,17 @@
-/*
- This source file is part of the Swift.org open source project
-
- Copyright (c) 2014 - 2021 Apple Inc. and the Swift project authors
- Licensed under Apache License v2.0 with Runtime Library Exception
-
- See http://swift.org/LICENSE.txt for license information
- See http://swift.org/CONTRIBUTORS.txt for Swift project authors
-*/
+//===----------------------------------------------------------------------===//
+//
+// This source file is part of the Swift open source project
+//
+// Copyright (c) 2014-2021 Apple Inc. and the Swift project authors
+// Licensed under Apache License v2.0 with Runtime Library Exception
+//
+// See http://swift.org/LICENSE.txt for license information
+// See http://swift.org/CONTRIBUTORS.txt for the list of Swift project authors
+//
+//===----------------------------------------------------------------------===//
 
 import Basics
-import Foundation
+@_implementationOnly import Foundation
 import PackageModel
 import TSCBasic
 
@@ -62,23 +64,20 @@ public struct TargetSourcesBuilder {
         target: TargetDescription,
         path: AbsolutePath,
         defaultLocalization: String?,
-        additionalFileRules: [FileRuleDescription] = [],
-        toolsVersion: ToolsVersion = .currentToolsVersion,
+        additionalFileRules: [FileRuleDescription],
+        toolsVersion: ToolsVersion,
         fileSystem: FileSystem,
         observabilityScope: ObservabilityScope
     ) {
         self.packageIdentity = packageIdentity
         self.packageKind = packageKind
-        //self.packageLocation = packageLocation
         self.packagePath = packagePath
         self.target = target
         self.defaultLocalization = defaultLocalization
         self.targetPath = path
-        // In version 5.4 and earlier, SwiftPM did not support `additionalFileRules` and always implicitly included XCBuild file types.
-        let actualAdditionalRules = (toolsVersion <= ToolsVersion.v5_4 ? FileRuleDescription.xcbuildFileTypes : additionalFileRules)
-        self.rules = FileRuleDescription.builtinRules + actualAdditionalRules
+        self.rules = Self.rules(additionalFileRules: additionalFileRules, toolsVersion: toolsVersion)
         self.toolsVersion = toolsVersion
-        let excludedPaths = target.exclude.map { AbsolutePath($0, relativeTo: path) }
+        let excludedPaths = target.exclude.compactMap { try? AbsolutePath(validating: $0, relativeTo: path) }
         self.excludedPaths = Set(excludedPaths)
         self.opaqueDirectoriesExtensions = FileRuleDescription.opaqueDirectoriesExtensions.union(
             additionalFileRules.reduce(into: Set<String>(), { partial, item in
@@ -93,7 +92,7 @@ public struct TargetSourcesBuilder {
             return metadata
         }
 
-        let declaredSources = target.sources?.map { AbsolutePath($0, relativeTo: path) }
+        let declaredSources = target.sources?.compactMap { try? AbsolutePath(validating: $0, relativeTo: path) }
         if let declaredSources = declaredSources {
             // Diagnose duplicate entries.
             let duplicates = declaredSources.spm_findDuplicateElements()
@@ -124,6 +123,12 @@ public struct TargetSourcesBuilder {
       #endif
     }
 
+    private static func rules(additionalFileRules: [FileRuleDescription], toolsVersion: ToolsVersion) -> [FileRuleDescription] {
+        // In version 5.4 and earlier, SwiftPM did not support `additionalFileRules` and always implicitly included XCBuild file types.
+        let actualAdditionalRules = (toolsVersion <= .v5_4 ? FileRuleDescription.xcbuildFileTypes : additionalFileRules)
+        return FileRuleDescription.builtinRules + actualAdditionalRules
+    }
+
     @discardableResult
     private func validTargetPath(at: AbsolutePath) -> Error? {
         // Check if paths that are enumerated in targets: [] exist
@@ -156,23 +161,22 @@ public struct TargetSourcesBuilder {
     /// Run the builder to produce the sources of the target.
     public func run() throws -> (sources: Sources, resources: [Resource], headers: [AbsolutePath], ignored: [AbsolutePath], others: [AbsolutePath]) {
         let contents = self.computeContents()
-        var pathToRule: [AbsolutePath: Rule] = [:]
+        var pathToRule: [AbsolutePath: FileRuleDescription.Rule] = [:]
 
         for path in contents {
-            pathToRule[path] = self.computeRule(for: path)
+            pathToRule[path] = try self.computeRule(for: path)
         }
 
-        let headers = pathToRule.lazy.filter { $0.value.rule == .header }.map { $0.key }.sorted()
-        let compilePaths = pathToRule.lazy.filter { $0.value.rule == .compile }.map { $0.key }
+        let headers = pathToRule.lazy.filter { $0.value == .header }.map { $0.key }.sorted()
+        let compilePaths = pathToRule.lazy.filter { $0.value == .compile }.map { $0.key }
         let sources = Sources(paths: Array(compilePaths), root: targetPath)
         let resources: [Resource] = pathToRule.compactMap { resource(for: $0.key, with: $0.value) }
-        let ignored = pathToRule.filter { $0.value.rule == .ignored }.map { $0.key }
-        let others = pathToRule.filter { $0.value.rule == .none }.map { $0.key }
+        let ignored = pathToRule.filter { $0.value == .ignored }.map { $0.key }
+        let others = pathToRule.filter { $0.value == .none }.map { $0.key }
 
         diagnoseConflictingResources(in: resources)
         diagnoseCopyConflictsWithLocalizationDirectories(in: resources)
         diagnoseLocalizedAndUnlocalizedVariants(in: resources)
-        diagnoseMissingDevelopmentRegionResource(in: resources)
         diagnoseInfoPlistConflicts(in: resources)
         diagnoseInvalidResource(in: target.resources)
 
@@ -184,43 +188,47 @@ public struct TargetSourcesBuilder {
         return (sources, resources, headers, ignored, others)
     }
 
-    private struct Rule {
-        let rule: FileRuleDescription.Rule
-        let localization: TargetDescription.Resource.Localization?
+    /// Compute the rule for the given path.
+    private static func computeRule(for path: AbsolutePath,
+                                    toolsVersion: ToolsVersion,
+                                    additionalFileRules: [FileRuleDescription],
+                                    observabilityScope: ObservabilityScope) -> FileRuleDescription.Rule {
+        let rules = Self.rules(additionalFileRules: additionalFileRules, toolsVersion: toolsVersion)
+        // For now, we are not passing in any declared resources or sources here and instead handle any generated files automatically at the callsite. Eventually, we will want the ability to declare opinions for generated files in the manifest as well.
+        return Self.computeRule(for: path, toolsVersion: toolsVersion, rules: rules, declaredResources: [], declaredSources: nil, observabilityScope: observabilityScope)
     }
 
-    /// Compute the rule for the given path.
-    private func computeRule(for path: AbsolutePath) -> Rule {
-        var matchedRule: Rule = Rule(rule: .none, localization: nil)
+    private static func computeRule(for path: AbsolutePath, toolsVersion: ToolsVersion, rules: [FileRuleDescription], declaredResources: [(path: AbsolutePath, rule: TargetDescription.Resource.Rule)], declaredSources: [AbsolutePath]?, observabilityScope: ObservabilityScope) -> FileRuleDescription.Rule {
+        var matchedRule: FileRuleDescription.Rule = .none
 
         // First match any resources explicitly declared in the manifest file.
-        for declaredResource in target.resources {
-            let resourcePath = AbsolutePath(declaredResource.path, relativeTo: self.targetPath)
+        for declaredResource in declaredResources {
+            let resourcePath = declaredResource.path
             if path.isDescendantOfOrEqual(to: resourcePath) {
-                if matchedRule.rule != .none {
-                    self.observabilityScope.emit(error: "duplicate resource rule '\(declaredResource.rule)' found for file at '\(path)'")
+                if matchedRule != .none {
+                    observabilityScope.emit(error: "duplicate resource rule '\(declaredResource.rule)' found for file at '\(path)'")
                 }
-                matchedRule = Rule(rule: declaredResource.rule.fileRule, localization: declaredResource.localization)
+                matchedRule = .init(declaredResource.rule)
             }
         }
 
         // Match any sources explicitly declared in the manifest file.
-        if let declaredSources = self.declaredSources {
+        if let declaredSources = declaredSources {
             for sourcePath in declaredSources {
                 if path.isDescendantOfOrEqual(to: sourcePath) {
-                    if matchedRule.rule != .none {
-                        self.observabilityScope.emit(error: "duplicate rule found for file at '\(path)'")
+                    if matchedRule != .none {
+                        observabilityScope.emit(error: "duplicate rule found for file at '\(path)'")
                     }
 
                     // Check for header files as they're allowed to be mixed with sources.
                     if let ext = path.extension,
                       FileRuleDescription.header.fileTypes.contains(ext) {
-                        matchedRule = Rule(rule: .header, localization: nil)
+                        matchedRule = .header
                     } else if toolsVersion >= .v5_3 {
-                        matchedRule = Rule(rule: .compile, localization: nil)
+                        matchedRule = .compile
                     } else if let ext = path.extension,
                       SupportedLanguageExtension.validExtensions(toolsVersion: toolsVersion).contains(ext) {
-                        matchedRule = Rule(rule: .compile, localization: nil)
+                        matchedRule = .compile
                     }
                     // The source file might have been declared twice so
                     // exit on first match.
@@ -232,29 +240,34 @@ public struct TargetSourcesBuilder {
 
         // We haven't found a rule using that's explicitly declared in the manifest
         // so try doing an automatic match.
-        if matchedRule.rule == .none {
+        if matchedRule == .none {
             let effectiveRules: [FileRuleDescription] = {
                 // Don't automatically match compile rules if target's sources are
                 // explicitly declared in the package manifest.
-                if target.sources != nil {
-                    return self.rules.filter { $0.rule != .compile }
+                if declaredSources != nil {
+                    return rules.filter { $0.rule != .compile }
                 }
-                return self.rules
+                return rules
             }()
 
             if let needle = effectiveRules.first(where: { $0.match(path: path, toolsVersion: toolsVersion) }) {
-                matchedRule = Rule(rule: needle.rule, localization: nil)
+                matchedRule = needle.rule
             } else if path.parentDirectory.extension == Resource.localizationDirectoryExtension {
-                matchedRule = Rule(rule: .processResource, localization: nil)
+                matchedRule = .processResource(localization: .none)
             }
         }
 
         return matchedRule
     }
 
+    private func computeRule(for path: AbsolutePath) throws -> FileRuleDescription.Rule {
+        let declaredResources = try target.resources.map { (path: try AbsolutePath(validating: $0.path, relativeTo: self.targetPath), rule: $0.rule) }
+        return Self.computeRule(for: path, toolsVersion: toolsVersion, rules: rules, declaredResources: declaredResources, declaredSources: declaredSources, observabilityScope: observabilityScope)
+    }
+
     /// Returns the `Resource` file associated with a file and a particular rule, if there is one.
-    private func resource(for path: AbsolutePath, with rule: Rule) -> Resource? {
-        switch rule.rule {
+    private static func resource(for path: AbsolutePath, with rule: FileRuleDescription.Rule, defaultLocalization: String?, targetName: String, targetPath: AbsolutePath, observabilityScope: ObservabilityScope) -> Resource? {
+        switch rule {
         case .compile, .header, .none, .modulemap, .ignored:
             return nil
         case .processResource:
@@ -267,26 +280,33 @@ public struct TargetSourcesBuilder {
             }()
 
             let explicitLocalization: String? = {
-                switch rule.localization  {
-                case .default?: return defaultLocalization ?? "en"
-                case .base?: return "Base"
-                case nil: return nil
+                switch rule  {
+                case .processResource(localization: .default):
+                    return defaultLocalization ?? "en"
+                case .processResource(localization: .base):
+                    return "Base"
+                default:
+                    return .none
                 }
             }()
 
             // If a resource is both inside a localization directory and has an explicit localization, it's ambiguous.
             guard implicitLocalization == nil || explicitLocalization == nil else {
                 let relativePath = path.relative(to: targetPath)
-                self.observabilityScope.emit(.localizationAmbiguity(path: relativePath, targetName: target.name))
+                observabilityScope.emit(.localizationAmbiguity(path: relativePath, targetName: targetName))
                 return nil
             }
 
-            return Resource(rule: .process, path: path, localization: implicitLocalization ?? explicitLocalization)
+            return Resource(rule: .process(localization: implicitLocalization ?? explicitLocalization), path: path)
         case .copy:
-            return Resource(rule: .copy, path: path, localization: nil)
+            return Resource(rule: .copy, path: path)
         }
     }
 
+    private func resource(for path: AbsolutePath, with rule: FileRuleDescription.Rule) -> Resource? {
+        return Self.resource(for: path, with: rule, defaultLocalization: defaultLocalization, targetName: target.name, targetPath: targetPath, observabilityScope: observabilityScope)
+    }
+
     private func diagnoseConflictingResources(in resources: [Resource]) {
         let duplicateResources = resources.spm_findDuplicateElements(by: \.destination)
         for resources in duplicateResources {
@@ -324,24 +344,6 @@ public struct TargetSourcesBuilder {
         }
     }
 
-    private func diagnoseMissingDevelopmentRegionResource(in resources: [Resource]) {
-        // We can't diagnose anything here without a default localization set.
-        guard let defaultLocalization = self.defaultLocalization else {
-            return
-        }
-
-        let localizedResources = resources.lazy.filter({ $0.localization != nil && $0.localization != "Base" })
-        let resourcesByBasename = Dictionary(grouping: localizedResources, by: { $0.path.basename })
-        for (basename, resources) in resourcesByBasename {
-            if !resources.contains(where: { $0.localization == defaultLocalization }) {
-                self.observabilityScope.emit(.missingDefaultLocalizationResource(
-                    resource: basename,
-                    targetName: target.name,
-                    defaultLocalization: defaultLocalization))
-            }
-        }
-    }
-
     private func diagnoseInfoPlistConflicts(in resources: [Resource]) {
         for resource in resources {
             if resource.destination == RelativePath("Info.plist") {
@@ -354,7 +356,9 @@ public struct TargetSourcesBuilder {
 
     private func diagnoseInvalidResource(in resources: [TargetDescription.Resource]) {
         resources.forEach { resource in
-            let resourcePath = AbsolutePath(resource.path, relativeTo: self.targetPath)
+            guard let resourcePath = try? AbsolutePath(validating: resource.path, relativeTo: self.targetPath) else {
+                return
+            }
             if let message = validTargetPath(at: resourcePath), self.packageKind.emitAuthorWarnings {
                 let warning = "Invalid Resource '\(resource.path)': \(message)."
                 self.observabilityScope.emit(warning: warning)
@@ -443,7 +447,7 @@ public struct TargetSourcesBuilder {
 
             // Check if the directory is marked to be copied.
             let directoryMarkedToBeCopied = target.resources.contains{ resource in
-                let resourcePath = AbsolutePath(resource.path, relativeTo: self.targetPath)
+                let resourcePath = try? AbsolutePath(validating: resource.path, relativeTo: self.targetPath)
                 if resource.rule == .copy && resourcePath == path {
                     return true
                 }
@@ -474,6 +478,49 @@ public struct TargetSourcesBuilder {
 
         return contents
     }
+
+    public static func computeContents(for generatedFiles: [AbsolutePath], toolsVersion: ToolsVersion, additionalFileRules: [FileRuleDescription], defaultLocalization: String?, targetName: String, targetPath: AbsolutePath, observabilityScope: ObservabilityScope) -> (sources: [AbsolutePath], resources: [Resource]) {
+        var sources = [AbsolutePath]()
+        var resources = [Resource]()
+
+        generatedFiles.forEach { absPath in
+            // 5.6 handled treated all generated files as sources.
+            if toolsVersion <= .v5_6 {
+                sources.append(absPath)
+                return
+            }
+
+            var rule = Self.computeRule(for: absPath, toolsVersion: toolsVersion, additionalFileRules: additionalFileRules, observabilityScope: observabilityScope)
+
+            // If we did not find a rule for a generated file, we treat it as to be processed for now. Eventually, we should handle generated files the same as other files and require explicit handling in the manifest for unknown types.
+            if rule == .none {
+                rule = .processResource(localization: .none)
+            }
+
+            switch rule {
+            case .compile:
+                if absPath.extension == "swift" {
+                    sources.append(absPath)
+                } else {
+                    observabilityScope.emit(warning: "Only Swift is supported for generated plugin source files at this time: \(absPath)")
+                }
+            case .copy, .processResource:
+                if let resource = Self.resource(for: absPath, with: rule, defaultLocalization: defaultLocalization, targetName: targetName, targetPath: targetPath, observabilityScope: observabilityScope) {
+                    resources.append(resource)
+                } else {
+                    // If this is reached, `TargetSourcesBuilder` already emitted a diagnostic, so we can ignore this case here.
+                }
+            case .header:
+                observabilityScope.emit(warning: "Headers generated by plugins are not supported at this time: \(absPath)")
+            case .modulemap:
+                observabilityScope.emit(warning: "Module maps generated by plugins are not supported at this time: \(absPath)")
+            case .ignored, .none:
+                break
+            }
+        }
+
+        return (sources, resources)
+    }
 }
 
 /// Describes a rule for including a source or resource file in a target.
@@ -481,14 +528,14 @@ public struct FileRuleDescription {
     /// A rule semantically describes a file/directory in a target.
     ///
     /// It is up to the build system to translate a rule into a build command.
-    public enum Rule {
+    public enum Rule: Equatable {
         /// The compile rule for `sources` in a package.
         case compile
 
         /// Process resource file rule for any type of platform-specific processing.
         ///
         /// This defaults to copy if there's no specialized behavior.
-        case processResource
+        case processResource(localization: TargetDescription.Resource.Localization?)
 
         /// The copy rule.
         case copy
@@ -583,7 +630,7 @@ public struct FileRuleDescription {
     /// File types related to the interface builder and storyboards.
     public static let xib: FileRuleDescription = {
         .init(
-            rule: .processResource,
+            rule: .processResource(localization: .none),
             toolsVersion: .v5_3,
             fileTypes: ["nib", "xib", "storyboard"]
         )
@@ -592,7 +639,7 @@ public struct FileRuleDescription {
     /// File types related to the asset catalog.
     public static let assetCatalog: FileRuleDescription = {
         .init(
-            rule: .processResource,
+            rule: .processResource(localization: .none),
             toolsVersion: .v5_3,
             fileTypes: ["xcassets"]
         )
@@ -601,7 +648,7 @@ public struct FileRuleDescription {
     /// File types related to the CoreData.
     public static let coredata: FileRuleDescription = {
         .init(
-            rule: .processResource,
+            rule: .processResource(localization: .none),
             toolsVersion: .v5_3,
             fileTypes: ["xcdatamodeld", "xcdatamodel", "xcmappingmodel"]
         )
@@ -610,7 +657,7 @@ public struct FileRuleDescription {
     /// File types related to Metal.
     public static let metal: FileRuleDescription = {
         .init(
-            rule: .processResource,
+            rule: .processResource(localization: .none),
             toolsVersion: .v5_3,
             fileTypes: ["metal"]
         )
@@ -656,13 +703,24 @@ public struct FileRuleDescription {
     }
 }
 
-extension TargetDescription.Resource.Rule {
-    fileprivate var fileRule: FileRuleDescription.Rule {
-        switch self {
-        case .process:
-            return .processResource
+extension FileRuleDescription.Rule {
+    init(_ seed: TargetDescription.Resource.Rule)  {
+        switch seed {
+        case .process(let localization):
+            self = .processResource(localization: localization)
         case .copy:
-            return .copy
+            self = .copy
+        }
+    }
+}
+
+extension Resource {
+    var localization: String? {
+        switch self.rule {
+        case .process(let localization):
+            return localization
+        default:
+            return .none
         }
     }
 }
@@ -695,8 +753,8 @@ extension ObservabilityMetadata {
     }
 }
 
-fileprivate extension PackageReference.Kind {
-    var emitAuthorWarnings: Bool {
+extension PackageReference.Kind {
+    fileprivate var emitAuthorWarnings: Bool {
         switch self {
         case .remoteSourceControl, .registry:
             return false
