@@ -1,16 +1,17 @@
-/*
- This source file is part of the Swift.org open source project
-
- Copyright (c) 2014 - 2017 Apple Inc. and the Swift project authors
- Licensed under Apache License v2.0 with Runtime Library Exception
-
- See http://swift.org/LICENSE.txt for license information
- See http://swift.org/CONTRIBUTORS.txt for Swift project authors
-*/
+//===----------------------------------------------------------------------===//
+//
+// This source file is part of the Swift open source project
+//
+// Copyright (c) 2014-2017 Apple Inc. and the Swift project authors
+// Licensed under Apache License v2.0 with Runtime Library Exception
+//
+// See http://swift.org/LICENSE.txt for license information
+// See http://swift.org/CONTRIBUTORS.txt for the list of Swift project authors
+//
+//===----------------------------------------------------------------------===//
 
 import Basics
 import TSCBasic
-import TSCUtility
 import Foundation
 
 /// This contains the declarative specification loaded from package manifest
@@ -23,12 +24,6 @@ public final class Manifest {
     /// The standard basename for the manifest.
     public static let basename = "Package"
 
-    /// The name of the package as it appears in the manifest
-    @available(*, deprecated, message: "use displayName instead, and only for display purposes")
-    public var name: String {
-        return self.displayName
-    }
-
     /// The name of the package as it appears in the manifest
     /// FIXME: deprecate this, there is no value in this once we have real package identifiers
     public let displayName: String
@@ -53,14 +48,6 @@ public final class Manifest {
         CanonicalPackageLocation(self.packageLocation)
     }
 
-    // FIXME: deprecated 2/2021, remove once clients migrate
-    @available(*, deprecated, message: "use packageLocation instead")
-    public var url: String {
-        get {
-            self.packageLocation
-        }
-    }
-
     /// Whether kind of package this manifest is from.
     public let packageKind: PackageReference.Kind
 
@@ -154,7 +141,7 @@ public final class Manifest {
     /// Returns the targets required for a particular product filter.
     public func targetsRequired(for productFilter: ProductFilter) -> [TargetDescription] {
         #if ENABLE_TARGET_BASED_DEPENDENCY_RESOLUTION
-        // If we have already calcualted it, returned the cached value.
+        // If we have already calculated it, returned the cached value.
         if let targets = _requiredTargets[productFilter] {
             return targets
         } else {
@@ -192,22 +179,28 @@ public final class Manifest {
             return self.dependencies
         }
 
-        var requiredDependencyURLs: Set<PackageIdentity> = []
-
-        for target in self.targetsRequired(for: products) {
-            for targetDependency in target.dependencies {
+        var requiredDependencies: Set<PackageIdentity> = []
+        for targetTriple in self.targetsRequired(for: products) {
+            for targetDependency in targetTriple.dependencies {
                 if let dependency = self.packageDependency(referencedBy: targetDependency) {
-                    requiredDependencyURLs.insert(dependency.identity)
+                    requiredDependencies.insert(dependency.identity)
+                }
+            }
+
+            targetTriple.pluginUsages?.forEach {
+                if let dependency = self.packageDependency(referencedBy: $0) {
+                    requiredDependencies.insert(dependency.identity)
                 }
             }
         }
 
-        return self.dependencies.filter { requiredDependencyURLs.contains($0.identity) }
+        return self.dependencies.filter { requiredDependencies.contains($0.identity) }
         #endif
     }
 
     /// Returns the targets required for building the provided products.
     public func targetsRequired(for products: [ProductDescription]) -> [TargetDescription] {
+        let productsByName = Dictionary(products.map({ ($0.name, $0) }), uniquingKeysWith: { $1 })
         let targetsByName = Dictionary(targets.map({ ($0.name, $0) }), uniquingKeysWith: { $1 })
         let productTargetNames = products.flatMap({ $0.targets })
 
@@ -227,7 +220,13 @@ public final class Manifest {
                 let plugins: [String] = target.pluginUsages?.compactMap { pluginUsage in
                     switch pluginUsage {
                     case .plugin(name: let name, package: nil):
-                        return targetsByName.keys.contains(name) ? name : nil
+                        if targetsByName.keys.contains(name) {
+                            return name
+                        } else if let targetName = productsByName[name]?.targets.first {
+                            return targetName
+                        } else {
+                            return nil
+                        }
                     default:
                         return nil
                     }
@@ -298,7 +297,7 @@ public final class Manifest {
         let packageName: String
 
         switch targetDependency {
-        case .product(_, package: let name?, _),
+        case .product(_, package: let name?, _, _),
              .byName(name: let name, _):
             packageName = name
         default:
@@ -307,6 +306,19 @@ public final class Manifest {
 
         return packageDependency(referencedBy: packageName)
     }
+
+    /// Finds the package dependency referenced by the specified plugin usage.
+    /// - Returns: Returns `nil` if  the used plugin is from the same package or if the package the used plugin is from cannot be found.
+    public func packageDependency(
+        referencedBy pluginUsage: TargetDescription.PluginUsage) -> PackageDependency? {
+        switch pluginUsage {
+        case .plugin(_, let .some(package)):
+            return packageDependency(referencedBy: package)
+        default:
+            return nil
+        }
+    }
+
     private func packageDependency(
         referencedBy packageName: String
     ) -> PackageDependency? {
@@ -330,7 +342,7 @@ public final class Manifest {
     ///
     /// - Parameters:
     ///   - targetDependency: The target dependency to register.
-    ///   - registry: The registry in which to record the assocation.
+    ///   - registry: The registry in which to record the association.
     ///   - availablePackages: The set of available packages.
     private func register(
         targetDependency: TargetDescription.Dependency,
@@ -340,7 +352,7 @@ public final class Manifest {
         switch targetDependency {
         case .target:
             break
-        case .product(let product, let package, _):
+        case .product(let product, let package, _, _):
             if let package = package { // ≥ 5.2
                 if !register(
                     product: product,
@@ -389,7 +401,7 @@ public final class Manifest {
     ///
     /// - Parameters:
     ///   - requiredPlugIn: The plug‐in to register.
-    ///   - registry: The registry in which to record the assocation.
+    ///   - registry: The registry in which to record the association.
     ///   - availablePackages: The set of available packages.
     private func register(
         requiredPlugIn: TargetDescription.PluginUsage,
