@@ -1,12 +1,14 @@
-/*
- This source file is part of the Swift.org open source project
-
- Copyright (c) 2014 - 2020 Apple Inc. and the Swift project authors
- Licensed under Apache License v2.0 with Runtime Library Exception
-
- See http://swift.org/LICENSE.txt for license information
- See http://swift.org/CONTRIBUTORS.txt for Swift project authors
-*/
+//===----------------------------------------------------------------------===//
+//
+// This source file is part of the Swift open source project
+//
+// Copyright (c) 2014-2020 Apple Inc. and the Swift project authors
+// Licensed under Apache License v2.0 with Runtime Library Exception
+//
+// See http://swift.org/LICENSE.txt for license information
+// See http://swift.org/CONTRIBUTORS.txt for the list of Swift project authors
+//
+//===----------------------------------------------------------------------===//
 
 /// Represents a manifest condition.
 public struct PackageConditionDescription: Codable, Equatable {
@@ -76,6 +78,10 @@ public struct ConfigurationCondition: PackageConditionProtocol {
     }
 
     public func satisfies(_ environment: BuildEnvironment) -> Bool {
-        configuration == environment.configuration
+        if environment.configuration == nil {
+            return true
+        } else {
+            return configuration == environment.configuration
+        }
     }
 }
