@@ -1,30 +1,32 @@
-/*
- This source file is part of the Swift.org open source project
-
- Copyright (c) 2014 - 2020 Apple Inc. and the Swift project authors
- Licensed under Apache License v2.0 with Runtime Library Exception
-
- See http://swift.org/LICENSE.txt for license information
- See http://swift.org/CONTRIBUTORS.txt for Swift project authors
-*/
+//===----------------------------------------------------------------------===//
+//
+// This source file is part of the Swift open source project
+//
+// Copyright (c) 2014-2020 Apple Inc. and the Swift project authors
+// Licensed under Apache License v2.0 with Runtime Library Exception
+//
+// See http://swift.org/LICENSE.txt for license information
+// See http://swift.org/CONTRIBUTORS.txt for the list of Swift project authors
+//
+//===----------------------------------------------------------------------===//
 
 import Foundation
 import TSCBasic
 
 /// Represents a package dependency.
-public enum PackageDependency: Equatable {
+public enum PackageDependency: Equatable, Hashable {
     case fileSystem(FileSystem)
     case sourceControl(SourceControl)
     case registry(Registry)
     
-    public struct FileSystem: Equatable, Encodable {
+    public struct FileSystem: Equatable, Hashable, Encodable {
         public let identity: PackageIdentity
         public let nameForTargetDependencyResolutionOnly: String?
         public let path: AbsolutePath
         public let productFilter: ProductFilter
     }
 
-    public struct SourceControl: Equatable, Encodable {
+    public struct SourceControl: Equatable, Hashable, Encodable {
         public let identity: PackageIdentity
         public let nameForTargetDependencyResolutionOnly: String?
         public let location: Location
@@ -38,13 +40,13 @@ public enum PackageDependency: Equatable {
             case branch(String)
         }
 
-        public enum Location: Equatable {
+        public enum Location: Equatable, Hashable {
             case local(AbsolutePath)
-            case remote(Foundation.URL)
+            case remote(URL)
         }
     }
 
-    public struct Registry: Equatable, Encodable {
+    public struct Registry: Equatable, Hashable, Encodable {
         public let identity: PackageIdentity
         public let requirement: Requirement
         public let productFilter: ProductFilter
@@ -167,7 +169,7 @@ public enum PackageDependency: Equatable {
     
     public static func remoteSourceControl(identity: PackageIdentity,
                                            nameForTargetDependencyResolutionOnly: String?,
-                                           url: Foundation.URL,
+                                           url: URL,
                                            requirement: SourceControl.Requirement,
                                            productFilter: ProductFilter
     ) -> Self {
