@@ -1,18 +1,20 @@
-/*
- This source file is part of the Swift.org open source project
-
- Copyright (c) 2014 - 2021 Apple Inc. and the Swift project authors
- Licensed under Apache License v2.0 with Runtime Library Exception
-
- See http://swift.org/LICENSE.txt for license information
- See http://swift.org/CONTRIBUTORS.txt for Swift project authors
-*/
+//===----------------------------------------------------------------------===//
+//
+// This source file is part of the Swift open source project
+//
+// Copyright (c) 2014-2021 Apple Inc. and the Swift project authors
+// Licensed under Apache License v2.0 with Runtime Library Exception
+//
+// See http://swift.org/LICENSE.txt for license information
+// See http://swift.org/CONTRIBUTORS.txt for the list of Swift project authors
+//
+//===----------------------------------------------------------------------===//
 
 /// The description of an individual target.
-public struct TargetDescription: Equatable, Codable {
+public struct TargetDescription: Equatable, Encodable {
 
     /// The target type.
-    public enum TargetType: String, Equatable, Codable {
+    public enum TargetType: String, Equatable, Encodable {
         case regular
         case executable
         case test
@@ -24,25 +26,25 @@ public struct TargetDescription: Equatable, Codable {
     /// Represents a target's dependency on another entity.
     public enum Dependency: Equatable {
         case target(name: String, condition: PackageConditionDescription?)
-        case product(name: String, package: String?, condition: PackageConditionDescription?)
+        case product(name: String, package: String?, moduleAliases: [String: String]? = nil, condition: PackageConditionDescription?)
         case byName(name: String, condition: PackageConditionDescription?)
 
         public static func target(name: String) -> Dependency {
             return .target(name: name, condition: nil)
         }
 
-        public static func product(name: String, package: String? = nil) -> Dependency {
-            return .product(name: name, package: package, condition: nil)
+        public static func product(name: String, package: String? = nil, moduleAliases: [String: String]? = nil) -> Dependency {
+            return .product(name: name, package: package, moduleAliases: moduleAliases, condition: nil)
         }
     }
 
-    public struct Resource: Codable, Equatable {
-        public enum Rule: String, Codable, Equatable {
-            case process
+    public struct Resource: Encodable, Equatable {
+        public enum Rule: Encodable, Equatable {
+            case process(localization: Localization?)
             case copy
         }
 
-        public enum Localization: String, Codable, Equatable {
+        public enum Localization: String, Encodable {
             case `default`
             case base
         }
@@ -53,14 +55,9 @@ public struct TargetDescription: Equatable, Codable {
         /// The path of the resource.
         public let path: String
 
-        /// The explicit localization of the resource.
-        public let localization: Localization?
-
-        public init(rule: Rule, path: String, localization: Localization? = nil) {
-            precondition(rule == .process || localization == nil)
+        public init(rule: Rule, path: String) {
             self.rule = rule
             self.path = path
-            self.localization = localization
         }
     }
 
@@ -224,11 +221,12 @@ extension TargetDescription.Dependency: Codable {
             var unkeyedContainer = container.nestedUnkeyedContainer(forKey: .target)
             try unkeyedContainer.encode(a1)
             try unkeyedContainer.encode(a2)
-        case let .product(a1, a2, a3):
+        case let .product(a1, a2, a3, a4):
             var unkeyedContainer = container.nestedUnkeyedContainer(forKey: .product)
             try unkeyedContainer.encode(a1)
             try unkeyedContainer.encode(a2)
             try unkeyedContainer.encode(a3)
+            try unkeyedContainer.encode(a4)
         case let .byName(a1, a2):
             var unkeyedContainer = container.nestedUnkeyedContainer(forKey: .byName)
             try unkeyedContainer.encode(a1)
@@ -251,8 +249,9 @@ extension TargetDescription.Dependency: Codable {
             var unkeyedValues = try values.nestedUnkeyedContainer(forKey: key)
             let a1 = try unkeyedValues.decode(String.self)
             let a2 = try unkeyedValues.decodeIfPresent(String.self)
-            let a3 = try unkeyedValues.decodeIfPresent(PackageConditionDescription.self)
-            self = .product(name: a1, package: a2, condition: a3)
+            let a3 = try unkeyedValues.decode([String: String].self)
+            let a4 = try unkeyedValues.decodeIfPresent(PackageConditionDescription.self)
+            self = .product(name: a1, package: a2, moduleAliases: a3, condition: a4)
         case .byName:
             var unkeyedValues = try values.nestedUnkeyedContainer(forKey: key)
             let a1 = try unkeyedValues.decode(String.self)
