@@ -1,14 +1,17 @@
-/*
- This source file is part of the Swift.org open source project
-
- Copyright (c) 2020 - 2021 Apple Inc. and the Swift project authors
- Licensed under Apache License v2.0 with Runtime Library Exception
-
- See http://swift.org/LICENSE.txt for license information
- See http://swift.org/CONTRIBUTORS.txt for Swift project authors
-*/
+//===----------------------------------------------------------------------===//
+//
+// This source file is part of the Swift open source project
+//
+// Copyright (c) 2020-2022 Apple Inc. and the Swift project authors
+// Licensed under Apache License v2.0 with Runtime Library Exception
+//
+// See http://swift.org/LICENSE.txt for license information
+// See http://swift.org/CONTRIBUTORS.txt for the list of Swift project authors
+//
+//===----------------------------------------------------------------------===//
 
 import Foundation
+import TSCBasic
 
 
 /// Extensions on Manifest for generating source code expressing its contents
@@ -22,19 +25,24 @@ extension Manifest {
     /// in canonical declarative form.
     /// 
     /// - Parameters:
+    ///   - packageDirectory: Directory of the manifest's package (for purposes of making strings relative).
     ///   - toolsVersionHeaderComment: Optional string to add to the `swift-tools-version` header (it will be ignored).
     ///   - additionalImportModuleNames: Names of any modules to import besides PackageDescription (would commonly contain custom product type definitions).
     ///   - customProductTypeSourceGenerator: Closure that will be called once for each custom product type in the manifest; it should return a SourceCodeFragment for the product type.
     /// 
     /// Returns: a string containing the full source code for the manifest.
     public func generateManifestFileContents(
+        packageDirectory: AbsolutePath,
         toolsVersionHeaderComment: String? = .none,
         additionalImportModuleNames: [String] = [],
         customProductTypeSourceGenerator: ManifestCustomProductTypeSourceGenerator? = .none
     ) rethrows -> String {
         // Generate the source code fragment for the top level of the package
         // expression.
-        let packageExprFragment = try SourceCodeFragment(from: self, customProductTypeSourceGenerator: customProductTypeSourceGenerator)
+        let packageExprFragment = try SourceCodeFragment(
+            from: self,
+            packageDirectory: packageDirectory,
+            customProductTypeSourceGenerator: customProductTypeSourceGenerator)
         
         // Generate the source code from the module names and code fragment.
         // We only write out the major and minor (not patch) versions of the
@@ -48,12 +56,6 @@ extension Manifest {
             let package = \(packageExprFragment.generateSourceCode())
             """
     }
-    
-    /// Generates and returns a string containing the contents of the manifest
-    /// in canonical declarative form.
-    public var generatedManifestFileContents: String {
-        return self.generateManifestFileContents(customProductTypeSourceGenerator: nil)
-    }
 }
 
 /// Constructs and returns a SourceCodeFragment that represents the instantiation of a custom product type with the specified identifer and having the given serialized parameters (the contents of whom are a private matter between the serialized form in PackageDescription and the client). The generated source code should, if evaluated as a part of a package manifest, result in the same serialized parameters.
@@ -64,7 +66,11 @@ public typealias ManifestCustomProductTypeSourceGenerator = (ProductDescription)
 fileprivate extension SourceCodeFragment {
     
     /// Instantiates a SourceCodeFragment to represent an entire manifest.
-    init(from manifest: Manifest, customProductTypeSourceGenerator: ManifestCustomProductTypeSourceGenerator?) rethrows {
+    init(
+        from manifest: Manifest,
+        packageDirectory: AbsolutePath,
+        customProductTypeSourceGenerator: ManifestCustomProductTypeSourceGenerator?
+    ) rethrows {
         var params: [SourceCodeFragment] = []
         
         params.append(SourceCodeFragment(key: "name", string: manifest.displayName))
@@ -93,7 +99,7 @@ fileprivate extension SourceCodeFragment {
         }
 
         if !manifest.dependencies.isEmpty {
-            let nodes = manifest.dependencies.map{ SourceCodeFragment(from: $0) }
+            let nodes = manifest.dependencies.map{ SourceCodeFragment(from: $0, pathAnchor: packageDirectory) }
             params.append(SourceCodeFragment(key: "dependencies", subnodes: nodes))
         }
 
@@ -144,18 +150,20 @@ fileprivate extension SourceCodeFragment {
     }
     
     /// Instantiates a SourceCodeFragment to represent a single package dependency.
-    init(from dependency: PackageDependency) {
+    init(from dependency: PackageDependency, pathAnchor: AbsolutePath) {
         var params: [SourceCodeFragment] = []
         if let explicitName = dependency.explicitNameForTargetDependencyResolutionOnly {
             params.append(SourceCodeFragment(key: "name", string: explicitName))
         }
         switch dependency {
         case .fileSystem(let settings):
-            params.append(SourceCodeFragment(key: "path", string: settings.path.pathString))
+            let relPath = settings.path.relative(to: pathAnchor)
+            params.append(SourceCodeFragment(key: "path", string: relPath.pathString))
         case .sourceControl(let settings):
             switch settings.location {
-            case .local(let path):
-                params.append(SourceCodeFragment(key: "url", string: path.pathString))
+            case .local(let absPath):
+                let relPath = absPath.relative(to: pathAnchor)
+                params.append(SourceCodeFragment(key: "url", string: relPath.pathString))
             case .remote(let url):
                 params.append(SourceCodeFragment(key: "url", string: url.absoluteString))
             }
@@ -319,11 +327,15 @@ fileprivate extension SourceCodeFragment {
             }
             self.init(enum: "target", subnodes: params)
             
-        case .product(name: let name, package: let packageName, condition: let condition):
+        case .product(name: let name, package: let packageName, moduleAliases: let aliases, condition: let condition):
             params.append(SourceCodeFragment(key: "name", string: name))
             if let packageName = packageName {
                 params.append(SourceCodeFragment(key: "package", string: packageName))
             }
+            if let aliases = aliases {
+                let vals = aliases.map { SourceCodeFragment(key: $0.key.quotedForPackageManifest, string: $0.value) }
+                params.append(SourceCodeFragment(key: "moduleAliases", subnodes: vals))
+            }
             if let condition = condition {
                 params.append(SourceCodeFragment(key: "condition", subnode: SourceCodeFragment(from: condition)))
             }
@@ -392,6 +404,9 @@ fileprivate extension SourceCodeFragment {
         case .yum(let names):
             let params = [SourceCodeFragment(strings: names)]
             self.init(enum: "yum", subnodes: params)
+        case .nuget(let names):
+            let params = [SourceCodeFragment(strings: names)]
+            self.init(enum: "nuget", subnodes: params)
         }
     }
 
@@ -399,11 +414,11 @@ fileprivate extension SourceCodeFragment {
     init(from resource: TargetDescription.Resource) {
         var params: [SourceCodeFragment] = []
         params.append(SourceCodeFragment(string: resource.path))
-        if let localization = resource.localization {
-            params.append(SourceCodeFragment(key: "localization", enum: localization.rawValue))
-        }
         switch resource.rule {
-        case .process:
+        case .process(let localization):
+            if let localization = localization {
+                params.append(SourceCodeFragment(key: "localization", enum: localization.rawValue))
+            }
             self.init(enum: "process", subnodes: params)
         case .copy:
             self.init(enum: "copy", subnodes: params)
@@ -454,17 +469,15 @@ fileprivate extension SourceCodeFragment {
     init(from setting: TargetBuildSettingDescription.Setting) {
         var params: [SourceCodeFragment] = []
 
-        switch setting.name {
-        case .headerSearchPath, .linkedLibrary, .linkedFramework:
-            assert(setting.value.count == 1)
-            params.append(SourceCodeFragment(string: setting.value[0]))
+        switch setting.kind {
+        case .headerSearchPath(let value), .linkedLibrary(let value), .linkedFramework(let value), .enableUpcomingFeature(let value), .enableExperimentalFeature(let value):
+            params.append(SourceCodeFragment(string: value))
             if let condition = setting.condition {
                 params.append(SourceCodeFragment(from: condition))
             }
-            self.init(enum: setting.name.rawValue, subnodes: params)
-        case .define:
-            assert(setting.value.count == 1)
-            let parts = setting.value[0].split(separator: "=", maxSplits: 1)
+            self.init(enum: setting.kind.name, subnodes: params)
+        case .define(let value):
+            let parts = value.split(separator: "=", maxSplits: 1)
             assert(parts.count == 1 || parts.count == 2)
             params.append(SourceCodeFragment(string: String(parts[0])))
             if parts.count == 2 {
@@ -473,13 +486,13 @@ fileprivate extension SourceCodeFragment {
             if let condition = setting.condition {
                 params.append(SourceCodeFragment(from: condition))
             }
-            self.init(enum: setting.name.rawValue, subnodes: params)
-        case .unsafeFlags:
-            params.append(SourceCodeFragment(strings: setting.value))
+            self.init(enum: setting.kind.name, subnodes: params)
+        case .unsafeFlags(let values):
+            params.append(SourceCodeFragment(strings: values))
             if let condition = setting.condition {
                 params.append(SourceCodeFragment(from: condition))
             }
-            self.init(enum: setting.name.rawValue, subnodes: params)
+            self.init(enum: setting.kind.name, subnodes: params)
         }
     }
 }
@@ -558,13 +571,13 @@ public struct SourceCodeFragment {
     /// A literal prefix to emit at the start of the source code fragment.
     var literal: String
     
-    /// The type of delimeters to use around the subfragments (if any).
+    /// The type of delimiters to use around the subfragments (if any).
     var delimiters: Delimiters
     
     /// Whether or not to emit newlines before the subfragments (if any).
     var multiline: Bool
     
-    /// Any subfragments; no delimeters are emitted if none.
+    /// Any subfragments; no delimiters are emitted if none.
     var subnodes: [SourceCodeFragment]?
     
     /// Type of delimiters to emit around any subfragments.
@@ -613,10 +626,29 @@ public struct SourceCodeFragment {
     }
 }
 
+extension TargetBuildSettingDescription.Kind {
+    fileprivate var name: String {
+        switch self {
+        case .headerSearchPath:
+            return "headerSearchPath"
+        case .define:
+            return "define"
+        case .linkedLibrary:
+            return "linkedLibrary"
+        case .linkedFramework:
+            return "linkedFramework"
+        case .unsafeFlags:
+            return "unsafeFlags"
+        case .enableUpcomingFeature:
+            return "enableUpcomingFeature"
+        case .enableExperimentalFeature:
+            return "enableExperimentalFeature"
+        }
+    }
+}
 
-fileprivate extension String {
-    
-    var quotedForPackageManifest: String {
+extension String {
+    fileprivate var quotedForPackageManifest: String {
         return "\"" + self
             .replacingOccurrences(of: "\\", with: "\\\\")
             .replacingOccurrences(of: "\"", with: "\\\"")
