@@ -1,16 +1,17 @@
-/*
- This source file is part of the Swift.org open source project
-
- Copyright (c) 2014 - 2020 Apple Inc. and the Swift project authors
- Licensed under Apache License v2.0 with Runtime Library Exception
-
- See http://swift.org/LICENSE.txt for license information
- See http://swift.org/CONTRIBUTORS.txt for Swift project authors
- */
+//===----------------------------------------------------------------------===//
+//
+// This source file is part of the Swift open source project
+//
+// Copyright (c) 2014-2020 Apple Inc. and the Swift project authors
+// Licensed under Apache License v2.0 with Runtime Library Exception
+//
+// See http://swift.org/LICENSE.txt for license information
+// See http://swift.org/CONTRIBUTORS.txt for the list of Swift project authors
+//
+//===----------------------------------------------------------------------===//
 
 import Foundation
 import TSCBasic
-import TSCUtility
 
 /// The canonical identifier for a package, based on its source location.
 public struct PackageIdentity: CustomStringConvertible {
@@ -25,7 +26,7 @@ public struct PackageIdentity: CustomStringConvertible {
 
     /// Creates a package identity from a URL.
     /// - Parameter url: The package's URL.
-    public init(url: Foundation.URL) {
+    public init(url: URL) {
         self.init(urlString: url.absoluteString)
     }
 
@@ -276,7 +277,7 @@ struct PackageIdentityParser {
     }
 
     /// Compute the default name of a package given its URL.
-    public static func computeDefaultName(fromURL url: Foundation.URL) -> String {
+    public static func computeDefaultName(fromURL url: URL) -> String {
         Self.computeDefaultName(fromLocation: url.absoluteString)
     }
 
