@@ -1,15 +1,20 @@
-/*
- This source file is part of the Swift.org open source project
-
- Copyright (c) 2014 - 2021 Apple Inc. and the Swift project authors
- Licensed under Apache License v2.0 with Runtime Library Exception
-
- See http://swift.org/LICENSE.txt for license information
- See http://swift.org/CONTRIBUTORS.txt for Swift project authors
-*/
+//===----------------------------------------------------------------------===//
+//
+// This source file is part of the Swift open source project
+//
+// Copyright (c) 2014-2021 Apple Inc. and the Swift project authors
+// Licensed under Apache License v2.0 with Runtime Library Exception
+//
+// See http://swift.org/LICENSE.txt for license information
+// See http://swift.org/CONTRIBUTORS.txt for the list of Swift project authors
+//
+//===----------------------------------------------------------------------===//
 
 import TSCBasic
-import TSCUtility
+import Dispatch
+
+import protocol TSCUtility.PolymorphicCodableProtocol
+import Basics
 
 public class Target: PolymorphicCodableProtocol {
     public static var implementations: [PolymorphicCodableProtocol.Type] = [
@@ -40,10 +45,28 @@ public class Target: PolymorphicCodableProtocol {
         /// The name of the package containing the product.
         public let package: String?
 
+        /// Module aliases for targets of this product dependency. The key is an
+        /// original target name and the value is a new unique name that also
+        /// becomes the name of its .swiftmodule binary.
+        public let moduleAliases: [String: String]?
+
+        /// Fully qualified name for this product dependency: package ID + name of the product
+        public var identity: String {
+            if let pkg = package {
+                return pkg.lowercased() + "_" + name
+            } else {
+                // this is hit only if this product is referenced `.byName(name)`
+                // which assumes the name of this product, its package, and its target
+                // all have the same name
+                return name.lowercased() + "_" + name
+            }
+        }
+
         /// Creates a product reference instance.
-        public init(name: String, package: String?) {
+        public init(name: String, package: String?, moduleAliases: [String: String]? = nil) {
             self.name = name
             self.package = package
+            self.moduleAliases = moduleAliases
         }
     }
 
@@ -104,19 +127,79 @@ public class Target: PolymorphicCodableProtocol {
     ///
     /// NOTE: This name is not the language-level target (i.e., the importable
     /// name) name in many cases, instead use c99name if you need uniqueness.
-    public let name: String
+    public private(set) var name: String
 
-    /// The default localization for resources.
-    public let defaultLocalization: String?
+    /// Module aliases needed to build this target. The key is an original name of a
+    /// dependent target and the value is a new unique name mapped to the name
+    /// of its .swiftmodule binary.
+    public private(set) var moduleAliases: [String: String]?
+    /// Used to store pre-chained / pre-overriden module aliases
+    public private(set) var prechainModuleAliases: [String: String]?
+    /// Used to store aliases that should be referenced directly in source code
+    public private(set) var directRefAliases: [String: [String]]?
+
+    /// Add module aliases (if applicable) for dependencies of this target.
+    ///
+    /// For example, adding an alias `Bar` for a target name `Foo` will result in
+    /// compiling references to `Foo` in source code of this target as `Bar.swiftmodule`.
+    /// If the name argument `Foo` is the same as this target's name, this target will be
+    /// renamed as `Bar` and the resulting binary will be `Bar.swiftmodule`.
+    ///
+    /// - Parameters:
+    ///   - name: The original name of a dependent target or this target
+    ///   - alias: A new unique name mapped to the resulting binary name
+    public func addModuleAlias(for name: String, as alias: String) {
+        if moduleAliases == nil {
+            moduleAliases = [name: alias]
+        } else {
+            moduleAliases?[name] = alias
+        }
+    }
+
+    public func removeModuleAlias(for name: String) {
+        moduleAliases?.removeValue(forKey: name)
+        if moduleAliases?.isEmpty ?? false {
+            moduleAliases = nil
+        }
+    }
+
+    public func addPrechainModuleAlias(for name: String, as alias: String) {
+        if prechainModuleAliases == nil {
+            prechainModuleAliases = [name: alias]
+        } else {
+            prechainModuleAliases?[name] = alias
+        }
+    }
+    public func addDirectRefAliases(for name: String, as aliases: [String]) {
+        if directRefAliases == nil {
+            directRefAliases = [name: aliases]
+        } else {
+            directRefAliases?[name] = aliases
+        }
+    }
+
+    @discardableResult
+    public func applyAlias() -> Bool {
+        // If there's an alias for this target, rename
+        if let alias = moduleAliases?[name] {
+            self.name = alias
+            self.c99name = alias.spm_mangledToC99ExtendedIdentifier()
+            return true
+        }
+        return false
+    }
 
     /// The dependencies of this target.
     public let dependencies: [Dependency]
 
     /// The language-level target name.
-    public let c99name: String
+    public private(set) var c99name: String
 
     /// The bundle name, if one is being generated.
-    public let bundleName: String?
+    public var bundleName: String? {
+        return resources.isEmpty ? nil : potentialBundleName
+    }
+    public let potentialBundleName: String?
 
     /// Suffix that's expected for test targets.
     public static let testModuleNameSuffix = "Tests"
@@ -124,6 +207,9 @@ public class Target: PolymorphicCodableProtocol {
     /// The kind of target.
     public let type: Kind
 
+    /// The path of the target.
+    public let path: AbsolutePath
+
     /// The sources for the target.
     public let sources: Sources
 
@@ -136,39 +222,33 @@ public class Target: PolymorphicCodableProtocol {
     /// Other kinds of files in the target.
     public let others: [AbsolutePath]
 
-    /// The list of platforms that are supported by this target.
-    public let platforms: [SupportedPlatform]
-
-    /// Returns the supported platform instance for the given platform.
-    public func getSupportedPlatform(for platform: Platform) -> SupportedPlatform? {
-        return self.platforms.first(where: { $0.platform == platform })
-    }
-
     /// The build settings assignments of this target.
     public let buildSettings: BuildSettings.AssignmentTable
 
     /// The usages of package plugins by this target.
     public let pluginUsages: [PluginUsage]
 
+    /// Whether or not this target uses any custom unsafe flags.
+    public let usesUnsafeFlags: Bool
+
     fileprivate init(
         name: String,
-        bundleName: String? = nil,
-        defaultLocalization: String?,
-        platforms: [SupportedPlatform],
+        potentialBundleName: String? = nil,
         type: Kind,
+        path: AbsolutePath,
         sources: Sources,
         resources: [Resource] = [],
         ignored: [AbsolutePath] = [],
         others: [AbsolutePath] = [],
         dependencies: [Target.Dependency],
         buildSettings: BuildSettings.AssignmentTable,
-        pluginUsages: [PluginUsage]
+        pluginUsages: [PluginUsage],
+        usesUnsafeFlags: Bool
     ) {
         self.name = name
-        self.bundleName = bundleName
-        self.defaultLocalization = defaultLocalization
-        self.platforms = platforms
+        self.potentialBundleName = potentialBundleName
         self.type = type
+        self.path = path
         self.sources = sources
         self.resources = resources
         self.ignored = ignored
@@ -177,10 +257,11 @@ public class Target: PolymorphicCodableProtocol {
         self.c99name = self.name.spm_mangledToC99ExtendedIdentifier()
         self.buildSettings = buildSettings
         self.pluginUsages = pluginUsages
+        self.usesUnsafeFlags = usesUnsafeFlags
     }
 
     private enum CodingKeys: String, CodingKey {
-        case name, bundleName, defaultLocalization, platforms, type, sources, resources, ignored, others, buildSettings, pluginUsages
+        case name, potentialBundleName, defaultLocalization, platforms, type, path, sources, resources, ignored, others, buildSettings, pluginUsages, usesUnsafeFlags
     }
 
     public func encode(to encoder: Encoder) throws {
@@ -189,10 +270,9 @@ public class Target: PolymorphicCodableProtocol {
         // FIXME: dependencies property is skipped on purpose as it points to
         // the actual target dependency object.
         try container.encode(name, forKey: .name)
-        try container.encode(bundleName, forKey: .bundleName)
-        try container.encode(defaultLocalization, forKey: .defaultLocalization)
-        try container.encode(platforms, forKey: .platforms)
+        try container.encode(potentialBundleName, forKey: .potentialBundleName)
         try container.encode(type, forKey: .type)
+        try container.encode(path, forKey: .path)
         try container.encode(sources, forKey: .sources)
         try container.encode(resources, forKey: .resources)
         try container.encode(ignored, forKey: .ignored)
@@ -200,15 +280,15 @@ public class Target: PolymorphicCodableProtocol {
         try container.encode(buildSettings, forKey: .buildSettings)
         // FIXME: pluginUsages property is skipped on purpose as it points to
         // the actual target dependency object.
+        try container.encode(usesUnsafeFlags, forKey: .usesUnsafeFlags)
     }
 
     required public init(from decoder: Decoder) throws {
         let container = try decoder.container(keyedBy: CodingKeys.self)
         self.name = try container.decode(String.self, forKey: .name)
-        self.bundleName = try container.decodeIfPresent(String.self, forKey: .bundleName)
-        self.defaultLocalization = try container.decodeIfPresent(String.self, forKey: .defaultLocalization)
-        self.platforms = try container.decode([SupportedPlatform].self, forKey: .platforms)
+        self.potentialBundleName = try container.decodeIfPresent(String.self, forKey: .potentialBundleName)
         self.type = try container.decode(Kind.self, forKey: .type)
+        self.path = try container.decode(AbsolutePath.self, forKey: .path)
         self.sources = try container.decode(Sources.self, forKey: .sources)
         self.resources = try container.decode([Resource].self, forKey: .resources)
         self.ignored = try container.decode([AbsolutePath].self, forKey: .ignored)
@@ -221,6 +301,7 @@ public class Target: PolymorphicCodableProtocol {
         // FIXME: pluginUsages property is skipped on purpose as it points to
         // the actual target dependency object.
         self.pluginUsages = []
+        self.usesUnsafeFlags = try container.decode(Bool.self, forKey: .usesUnsafeFlags)
     }
 }
 
@@ -242,21 +323,26 @@ extension Target: CustomStringConvertible {
 
 public final class SwiftTarget: Target {
 
-    /// The file name of test manifest.
-    public static let testManifestNames = ["XCTMain.swift", "LinuxMain.swift"]
+    /// The default name for the test entry point file located in a package.
+    public static let defaultTestEntryPointName = "XCTMain.swift"
 
-    public init(testDiscoverySrc: Sources, name: String, dependencies: [Target.Dependency]) {
+    /// The list of all supported names for the test entry point file located in a package.
+    public static var testEntryPointNames: [String] {
+        [defaultTestEntryPointName, "LinuxMain.swift"]
+    }
+
+    public init(name: String, dependencies: [Target.Dependency], testDiscoverySrc: Sources) {
         self.swiftVersion = .v5
 
         super.init(
             name: name,
-            defaultLocalization: nil,
-            platforms: [],
-            type: .executable,
+            type: .library,
+            path: .root,
             sources: testDiscoverySrc,
             dependencies: dependencies,
             buildSettings: .init(),
-            pluginUsages: []
+            pluginUsages: [],
+            usesUnsafeFlags: false
         )
     }
 
@@ -265,10 +351,9 @@ public final class SwiftTarget: Target {
 
     public init(
         name: String,
-        bundleName: String? = nil,
-        defaultLocalization: String? = nil,
-        platforms: [SupportedPlatform] = [],
+        potentialBundleName: String? = nil,
         type: Kind,
+        path: AbsolutePath,
         sources: Sources,
         resources: [Resource] = [],
         ignored: [AbsolutePath] = [],
@@ -276,27 +361,28 @@ public final class SwiftTarget: Target {
         dependencies: [Target.Dependency] = [],
         swiftVersion: SwiftLanguageVersion,
         buildSettings: BuildSettings.AssignmentTable = .init(),
-        pluginUsages: [PluginUsage] = []
+        pluginUsages: [PluginUsage] = [],
+        usesUnsafeFlags: Bool
     ) {
         self.swiftVersion = swiftVersion
         super.init(
             name: name,
-            bundleName: bundleName,
-            defaultLocalization: defaultLocalization,
-            platforms: platforms,
+            potentialBundleName: potentialBundleName,
             type: type,
+            path: path,
             sources: sources,
             resources: resources,
             ignored: ignored,
             others: others,
             dependencies: dependencies,
             buildSettings: buildSettings,
-            pluginUsages: pluginUsages
+            pluginUsages: pluginUsages,
+            usesUnsafeFlags: usesUnsafeFlags
         )
     }
 
-    /// Create an executable Swift target from test manifest file.
-    public init(testManifest: AbsolutePath, name: String, dependencies: [Target.Dependency]) {
+    /// Create an executable Swift target from test entry point file.
+    public init(name: String, dependencies: [Target.Dependency], testEntryPointPath: AbsolutePath) {
         // Look for the first swift test target and use the same swift version
         // for linux main target. This will need to change if we move to a model
         // where we allow per target swift language version build settings.
@@ -309,20 +395,18 @@ public final class SwiftTarget: Target {
         // We need to select the latest Swift language version that can
         // satisfy the current tools version but there is not a good way to
         // do that currently.
-        self.swiftVersion = swiftTestTarget?.swiftVersion ?? SwiftLanguageVersion(string: String(ToolsVersion.currentToolsVersion.major)) ?? .v4
-        let sources = Sources(paths: [testManifest], root: testManifest.parentDirectory)
-
-        let platforms: [SupportedPlatform] = swiftTestTarget?.platforms ?? []
+        self.swiftVersion = swiftTestTarget?.swiftVersion ?? SwiftLanguageVersion(string: String(SwiftVersion.current.major)) ?? .v4
+        let sources = Sources(paths: [testEntryPointPath], root: testEntryPointPath.parentDirectory)
 
         super.init(
             name: name,
-            defaultLocalization: nil,
-            platforms: platforms,
             type: .executable,
+            path: .root,
             sources: sources,
             dependencies: dependencies,
             buildSettings: .init(),
-            pluginUsages: []
+            pluginUsages: [],
+            usesUnsafeFlags: false
         )
     }
 
@@ -351,18 +435,12 @@ public final class SystemLibraryTarget: Target {
     /// List of system package providers, if any.
     public let providers: [SystemPackageProviderDescription]?
 
-    /// The package path.
-    public var path: AbsolutePath {
-        return sources.root
-    }
-
     /// True if this system library should become implicit target
     /// dependency of its dependent packages.
     public let isImplicit: Bool
 
     public init(
         name: String,
-        platforms: [SupportedPlatform] = [],
         path: AbsolutePath,
         isImplicit: Bool = true,
         pkgConfig: String? = nil,
@@ -374,13 +452,13 @@ public final class SystemLibraryTarget: Target {
         self.isImplicit = isImplicit
         super.init(
             name: name,
-            defaultLocalization: nil,
-            platforms: platforms,
             type: .systemModule,
+            path: sources.root,
             sources: sources,
             dependencies: [],
             buildSettings: .init(),
-            pluginUsages: []
+            pluginUsages: [],
+            usesUnsafeFlags: false
         )
     }
 
@@ -432,23 +510,25 @@ public final class ClangTarget: Target {
 
     public init(
         name: String,
-        bundleName: String? = nil,
-        defaultLocalization: String? = nil,
-        platforms: [SupportedPlatform] = [],
+        potentialBundleName: String? = nil,
         cLanguageStandard: String?,
         cxxLanguageStandard: String?,
         includeDir: AbsolutePath,
         moduleMapType: ModuleMapType,
         headers: [AbsolutePath] = [],
         type: Kind,
+        path: AbsolutePath,
         sources: Sources,
         resources: [Resource] = [],
         ignored: [AbsolutePath] = [],
         others: [AbsolutePath] = [],
         dependencies: [Target.Dependency] = [],
-        buildSettings: BuildSettings.AssignmentTable = .init()
-    ) {
-        assert(includeDir.isDescendantOfOrEqual(to: sources.root), "\(includeDir) should be contained in the source root \(sources.root)")
+        buildSettings: BuildSettings.AssignmentTable = .init(),
+        usesUnsafeFlags: Bool
+    ) throws {
+        guard includeDir.isDescendantOfOrEqual(to: sources.root) else {
+            throw StringError("\(includeDir) should be contained in the source root \(sources.root)")
+        }
         self.isCXX = sources.containsCXXFiles
         self.cLanguageStandard = cLanguageStandard
         self.cxxLanguageStandard = cxxLanguageStandard
@@ -457,17 +537,17 @@ public final class ClangTarget: Target {
         self.headers = headers
         super.init(
             name: name,
-            bundleName: bundleName,
-            defaultLocalization: defaultLocalization,
-            platforms: platforms,
+            potentialBundleName: potentialBundleName,
             type: type,
+            path: path,
             sources: sources,
             resources: resources,
             ignored: ignored,
             others: others,
             dependencies: dependencies,
             buildSettings: buildSettings,
-            pluginUsages: []
+            pluginUsages: [],
+            usesUnsafeFlags: usesUnsafeFlags
         )
     }
 
@@ -513,7 +593,6 @@ public final class BinaryTarget: Target {
     public init(
         name: String,
         kind: Kind,
-        platforms: [SupportedPlatform] = [],
         path: AbsolutePath,
         origin: Origin
     ) {
@@ -522,13 +601,13 @@ public final class BinaryTarget: Target {
         let sources = Sources(paths: [], root: path)
         super.init(
             name: name,
-            defaultLocalization: nil,
-            platforms: platforms,
             type: .binary,
+            path: .root,
             sources: sources,
             dependencies: [],
             buildSettings: .init(),
-            pluginUsages: []
+            pluginUsages: [],
+            usesUnsafeFlags: false
         )
     }
 
@@ -576,13 +655,11 @@ public final class BinaryTarget: Target {
                 return "unknown"
             }
         }
+    }
 
-        public static func forFileExtension(_ fileExtension: String) throws -> Kind {
-            guard let kind = Kind.allCases.first(where: { $0.fileExtension == fileExtension }) else {
-                throw StringError("unknown binary artifact file extension '\(fileExtension)'")
-            }
-            return kind
-        }
+    public var containsExecutable: Bool {
+        // FIXME: needs to be revisited once libraries are supported in artifact bundles
+        return self.kind == .artifactsArchive
     }
 
     public enum Origin: Equatable, Codable {
@@ -635,7 +712,6 @@ public final class PluginTarget: Target {
 
     public init(
         name: String,
-        platforms: [SupportedPlatform] = [],
         sources: Sources,
         apiVersion: ToolsVersion,
         pluginCapability: PluginCapability,
@@ -645,13 +721,13 @@ public final class PluginTarget: Target {
         self.apiVersion = apiVersion
         super.init(
             name: name,
-            defaultLocalization: nil,
-            platforms: platforms,
             type: .plugin,
+            path: .root,
             sources: sources,
             dependencies: dependencies,
             buildSettings: .init(),
-            pluginUsages: []
+            pluginUsages: [],
+            usesUnsafeFlags: false
         )
     }
 
@@ -748,3 +824,18 @@ public enum PluginPermission: Hashable, Codable {
         }
     }
 }
+
+public extension Sequence where Iterator.Element == Target {
+    var executables: [Target] {
+        return filter {
+            switch $0.type {
+            case .binary:
+                return ($0 as? BinaryTarget)?.containsExecutable == true
+            case .executable, .snippet:
+                return true
+            default:
+                return false
+            }
+        }
+    }
+}
