@@ -1,102 +1,29 @@
-/*
- This source file is part of the Swift.org open source project
-
- Copyright (c) 2021 Apple Inc. and the Swift project authors
- Licensed under Apache License v2.0 with Runtime Library Exception
-
- See http://swift.org/LICENSE.txt for license information
- See http://swift.org/CONTRIBUTORS.txt for Swift project authors
- */
+//===----------------------------------------------------------------------===//
+//
+// This source file is part of the Swift open source project
+//
+// Copyright (c) 2021-2023 Apple Inc. and the Swift project authors
+// Licensed under Apache License v2.0 with Runtime Library Exception
+//
+// See http://swift.org/LICENSE.txt for license information
+// See http://swift.org/CONTRIBUTORS.txt for the list of Swift project authors
+//
+//===----------------------------------------------------------------------===//
 
 import Basics
 import Dispatch
-import class Foundation.JSONDecoder
-import struct Foundation.URL
-import struct Foundation.URLComponents
-import struct Foundation.URLQueryItem
+import Foundation
 import PackageFingerprint
 import PackageLoading
 import PackageModel
 import TSCBasic
-import protocol TSCUtility.Archiver
-import struct TSCUtility.ZipArchiver
-
-public enum RegistryError: Error, CustomStringConvertible {
-    case registryNotConfigured(scope: PackageIdentity.Scope?)
-    case invalidPackage(PackageIdentity)
-    case invalidURL(URL)
-    case invalidResponseStatus(expected: Int, actual: Int)
-    case invalidContentVersion(expected: String, actual: String?)
-    case invalidContentType(expected: String, actual: String?)
-    case invalidResponse
-    case missingSourceArchive
-    case invalidSourceArchive
-    case unsupportedHashAlgorithm(String)
-    case failedToDetermineExpectedChecksum(Error)
-    case failedToComputeChecksum(Error)
-    case checksumChanged(latest: String, previous: String)
-    case invalidChecksum(expected: String, actual: String)
-    case pathAlreadyExists(AbsolutePath)
-    case failedRetrievingReleases(Error)
-    case failedRetrievingReleaseChecksum(Error)
-    case failedRetrievingManifest(Error)
-    case failedDownloadingSourceArchive(Error)
-
-    public var description: String {
-        switch self {
-        case .registryNotConfigured(let scope):
-            if let scope = scope {
-                return "No registry configured for '\(scope)' scope"
-            } else {
-                return "No registry configured'"
-            }
-        case .invalidPackage(let package):
-            return "Invalid package '\(package)'"
-        case .invalidURL(let url):
-            return "Invalid URL '\(url)'"
-        case .invalidResponseStatus(let expected, let actual):
-            return "Invalid registry response status '\(actual)', expected '\(expected)'"
-        case .invalidContentVersion(expected: let expected, actual: let actual):
-            return "Invalid registry response content version '\(actual ?? "")', expected '\(expected)'"
-        case .invalidContentType(let expected, let actual):
-            return "Invalid registry response content type '\(actual ?? "")', expected '\(expected)'"
-        case .invalidResponse:
-            return "Invalid registry response"
-        case .missingSourceArchive:
-            return "Missing registry source archive"
-        case .invalidSourceArchive:
-            return "Invalid registry source archive"
-        case .unsupportedHashAlgorithm(let algorithm):
-            return "Unsupported hash algorithm '\(algorithm)'"
-        case .failedToDetermineExpectedChecksum(let error):
-            return "Failed determining registry source archive checksum: \(error)"
-        case .failedToComputeChecksum(let error):
-            return "Failed computing registry source archive checksum: \(error)"
-        case .checksumChanged(let latest, let previous):
-            return "The latest checksum '\(latest)' is different from the previously recorded value '\(previous)'"
-        case .invalidChecksum(let expected, let actual):
-            return "Invalid registry source archive checksum '\(actual)', expected '\(expected)'"
-        case .pathAlreadyExists(let path):
-            return "Path already exists '\(path)'"
-        case .failedRetrievingReleases(let error):
-            return "Failed fetching releases from registry: \(error)"
-        case .failedRetrievingReleaseChecksum(let error):
-            return "Failed fetching release checksum from registry: \(error)"
-        case .failedRetrievingManifest(let error):
-            return "Failed retrieving manifest from registry: \(error)"
-        case .failedDownloadingSourceArchive(let error):
-            return "Failed downloading source archive from registry: \(error)"
-        }
-    }
-}
 
 /// Package registry client.
 /// API specification: https://github.com/apple/swift-package-manager/blob/main/Documentation/Registry.md
-public final class RegistryClient {
+public final class RegistryClient: Cancellable {
     private let apiVersion: APIVersion = .v1
 
     private let configuration: RegistryConfiguration
-    private let identityResolver: IdentityResolver
     private let archiverProvider: (FileSystem) -> Archiver
     private let httpClient: HTTPClient
     private let authorizationProvider: HTTPClientAuthorizationProvider?
@@ -106,16 +33,38 @@ public final class RegistryClient {
 
     public init(
         configuration: RegistryConfiguration,
-        identityResolver: IdentityResolver,
-        fingerprintStorage: PackageFingerprintStorage? = .none,
+        fingerprintStorage: PackageFingerprintStorage?,
         fingerprintCheckingMode: FingerprintCheckingMode,
-        authorizationProvider: HTTPClientAuthorizationProvider? = .none,
+        authorizationProvider: AuthorizationProvider? = .none,
         customHTTPClient: HTTPClient? = .none,
         customArchiverProvider: ((FileSystem) -> Archiver)? = .none
     ) {
         self.configuration = configuration
-        self.identityResolver = identityResolver
-        self.authorizationProvider = authorizationProvider
+
+        if let authorizationProvider = authorizationProvider {
+            self.authorizationProvider = { url in
+                guard let registryAuthentication = configuration.authentication(for: url) else {
+                    return .none
+                }
+                guard let (user, password) = authorizationProvider.authentication(for: url) else {
+                    return .none
+                }
+
+                switch registryAuthentication.type {
+                case .basic:
+                    let authorizationString = "\(user):\(password)"
+                    guard let authorizationData = authorizationString.data(using: .utf8) else {
+                        return nil
+                    }
+                    return "Basic \(authorizationData.base64EncodedString())"
+                case .token: // `user` value is irrelevant in this case
+                    return "Bearer \(password)"
+                }
+            }
+        } else {
+            self.authorizationProvider = .none
+        }
+
         self.httpClient = customHTTPClient ?? HTTPClient()
         self.archiverProvider = customArchiverProvider ?? { fileSystem in ZipArchiver(fileSystem: fileSystem) }
         self.fingerprintStorage = fingerprintStorage
@@ -123,12 +72,21 @@ public final class RegistryClient {
         self.jsonDecoder = JSONDecoder.makeWithDefaults()
     }
 
-    public func fetchVersions(
+    public var configured: Bool {
+        !self.configuration.isEmpty
+    }
+
+    /// Cancel any outstanding requests
+    public func cancel(deadline: DispatchTime) throws {
+        try self.httpClient.cancel(deadline: deadline)
+    }
+
+    public func getPackageMetadata(
         package: PackageIdentity,
         timeout: DispatchTimeInterval? = .none,
         observabilityScope: ObservabilityScope,
         callbackQueue: DispatchQueue,
-        completion: @escaping (Result<[Version], Error>) -> Void
+        completion: @escaping (Result<PackageMetadata, Error>) -> Void
     ) {
         let completion = self.makeAsync(completion, on: callbackQueue)
 
@@ -160,18 +118,28 @@ public final class RegistryClient {
         self.httpClient.execute(request, observabilityScope: observabilityScope, progress: nil) { result in
             completion(
                 result.tryMap { response in
-                    try self.checkResponseStatusAndHeaders(response, expectedStatusCode: 200, expectedContentType: .json)
+                    try self.checkResponseStatusAndHeaders(
+                        response,
+                        expectedStatusCode: 200,
+                        expectedContentType: .json
+                    )
 
                     guard let data = response.body else {
                         throw RegistryError.invalidResponse
                     }
 
                     let packageMetadata = try self.jsonDecoder.decode(Serialization.PackageMetadata.self, from: data)
-
                     let versions = packageMetadata.releases.filter { $0.value.problem == nil }
                         .compactMap { Version($0.key) }
                         .sorted(by: >)
-                    return versions
+
+                    let alternateLocations = try response.headers.parseAlternativeLocationLinks()
+
+                    return PackageMetadata(
+                        registry: registry,
+                        versions: versions,
+                        alternateLocations: alternateLocations?.map(\.url)
+                    )
                 }.mapError {
                     RegistryError.failedRetrievingReleases($0)
                 }
@@ -218,7 +186,11 @@ public final class RegistryClient {
         self.httpClient.execute(request, observabilityScope: observabilityScope, progress: nil) { result in
             completion(
                 result.tryMap { response in
-                    try self.checkResponseStatusAndHeaders(response, expectedStatusCode: 200, expectedContentType: .swift)
+                    try self.checkResponseStatusAndHeaders(
+                        response,
+                        expectedStatusCode: 200,
+                        expectedContentType: .swift
+                    )
 
                     guard let data = response.body else {
                         throw RegistryError.invalidResponse
@@ -228,12 +200,13 @@ public final class RegistryClient {
                     }
 
                     var result = [String: (toolsVersion: ToolsVersion, content: String?)]()
-                    let toolsVersion = try ToolsVersionLoader().load(utf8String: manifestContent)
+                    let toolsVersion = try ToolsVersionParser.parse(utf8String: manifestContent)
                     result[Manifest.filename] = (toolsVersion: toolsVersion, content: manifestContent)
 
-                    let alternativeManifests = try response.headers.get("Link").map { try parseLinkHeader($0) }.flatMap { $0 }
+                    let alternativeManifests = try response.headers.parseManifestLinks()
                     for alternativeManifest in alternativeManifests {
-                        result[alternativeManifest.filename] = (toolsVersion: alternativeManifest.toolsVersion, content: .none)
+                        result[alternativeManifest.filename] = (toolsVersion: alternativeManifest.toolsVersion,
+                                                                content: .none)
                     }
                     return result
                 }.mapError {
@@ -241,68 +214,6 @@ public final class RegistryClient {
                 }
             )
         }
-
-        func parseLinkHeader(_ value: String) throws -> [ManifestLink] {
-            let linkLines = value.split(separator: ",").map(String.init).map { $0.spm_chuzzle() ?? $0 }
-            return try linkLines.compactMap { linkLine in
-                try parseLinkLine(linkLine)
-            }
-        }
-
-        // <http://packages.example.com/mona/LinkedList/1.1.1/Package.swift?swift-version=4>; rel="alternate"; filename="Package@swift-4.swift"; swift-tools-version="4.0",
-        func parseLinkLine(_ value: String) throws -> ManifestLink? {
-            let fields = value.split(separator: ";")
-                .map(String.init)
-                .map { $0.spm_chuzzle() ?? $0 }
-
-            guard fields.count == 4 else {
-                return nil
-            }
-
-            guard let link = fields.first(where: { $0.hasPrefix("<") }).map({ String($0.dropFirst().dropLast()) }) else {
-                return nil
-            }
-
-            guard let rel = fields.first(where: { $0.hasPrefix("rel=") }).flatMap({ parseLinkFieldValue($0) }), rel == "alternate" else {
-                return nil
-            }
-
-            guard let filename = fields.first(where: { $0.hasPrefix("filename=") }).flatMap({ parseLinkFieldValue($0) }) else {
-                return nil
-            }
-
-            guard let toolsVersion = fields.first(where: { $0.hasPrefix("swift-tools-version=") }).flatMap({ parseLinkFieldValue($0) }) else {
-                return nil
-            }
-
-            guard let toolsVersion = ToolsVersion(string: toolsVersion) else {
-                throw StringError("Invalid tools version in alternate manifest link '\(value)'")
-            }
-
-            return ManifestLink(
-                value: link,
-                filename: filename,
-                toolsVersion: toolsVersion
-            )
-        }
-
-        func parseLinkFieldValue(_ field: String) -> String? {
-            let parts = field.split(separator: "=")
-                .map(String.init)
-                .map { $0.spm_chuzzle() ?? $0 }
-
-            guard parts.count == 2 else {
-                return nil
-            }
-
-            return parts[1].replacingOccurrences(of: "\"", with: "")
-        }
-
-        struct ManifestLink {
-            let value: String
-            let filename: String
-            let toolsVersion: ToolsVersion
-        }
     }
 
     public func getManifestContent(
@@ -351,7 +262,12 @@ public final class RegistryClient {
         self.httpClient.execute(request, observabilityScope: observabilityScope, progress: nil) { result in
             completion(
                 result.tryMap { response -> String in
-                    try self.checkResponseStatusAndHeaders(response, expectedStatusCode: 200, expectedContentType: .swift)
+                    try self.checkResponseStatusAndHeaders(
+                        response,
+                        expectedStatusCode: 200,
+                        expectedContentType: .swift,
+                        isContentVersionOptional: true
+                    )
 
                     guard let data = response.body else {
                         throw RegistryError.invalidResponse
@@ -408,14 +324,19 @@ public final class RegistryClient {
             switch result {
             case .success(let response):
                 do {
-                    try self.checkResponseStatusAndHeaders(response, expectedStatusCode: 200, expectedContentType: .json)
+                    try self.checkResponseStatusAndHeaders(
+                        response,
+                        expectedStatusCode: 200,
+                        expectedContentType: .json
+                    )
 
                     guard let data = response.body else {
                         throw RegistryError.invalidResponse
                     }
 
                     let versionMetadata = try self.jsonDecoder.decode(Serialization.VersionMetadata.self, from: data)
-                    guard let sourceArchive = versionMetadata.resources.first(where: { $0.name == "source-archive" }) else {
+                    guard let sourceArchive = versionMetadata.resources.first(where: { $0.name == "source-archive" })
+                    else {
                         throw RegistryError.missingSourceArchive
                     }
 
@@ -428,16 +349,21 @@ public final class RegistryClient {
                                                version: version,
                                                fingerprint: .init(origin: .registry(registry.url), value: checksum),
                                                observabilityScope: observabilityScope,
-                                               callbackQueue: callbackQueue) { storageResult in
+                                               callbackQueue: callbackQueue)
+                        { storageResult in
                             switch storageResult {
                             case .success:
                                 completion(.success(checksum))
                             case .failure(PackageFingerprintStorageError.conflict(_, let existing)):
                                 switch self.fingerprintCheckingMode {
                                 case .strict:
-                                    completion(.failure(RegistryError.checksumChanged(latest: checksum, previous: existing.value)))
+                                    completion(.failure(RegistryError
+                                            .checksumChanged(latest: checksum, previous: existing.value)))
                                 case .warn:
-                                    observabilityScope.emit(warning: "The checksum \(checksum) from \(registry.url.absoluteString) does not match previously recorded value \(existing.value) from \(String(describing: existing.origin.url?.absoluteString))")
+                                    observabilityScope
+                                        .emit(
+                                            warning: "The checksum \(checksum) from \(registry.url.absoluteString) does not match previously recorded value \(existing.value) from \(String(describing: existing.origin.url?.absoluteString))"
+                                        )
                                     completion(.success(checksum))
                                 }
                             case .failure(let error):
@@ -500,7 +426,6 @@ public final class RegistryClient {
             guard !fileSystem.exists(destinationPath) else {
                 throw RegistryError.pathAlreadyExists(destinationPath)
             }
-            try fileSystem.createDirectory(destinationPath, recursive: true)
         } catch {
             return completion(.failure(error))
         }
@@ -519,7 +444,12 @@ public final class RegistryClient {
             switch result {
             case .success(let response):
                 do {
-                    try self.checkResponseStatusAndHeaders(response, expectedStatusCode: 200, expectedContentType: .zip)
+                    try self.checkResponseStatusAndHeaders(
+                        response,
+                        expectedStatusCode: 200,
+                        expectedContentType: .zip,
+                        isContentVersionOptional: true
+                    )
                 } catch {
                     return completion(.failure(RegistryError.failedDownloadingSourceArchive(error)))
                 }
@@ -534,12 +464,21 @@ public final class RegistryClient {
                             if expectedChecksum != actualChecksum {
                                 switch self.fingerprintCheckingMode {
                                 case .strict:
-                                    return completion(.failure(RegistryError.invalidChecksum(expected: expectedChecksum, actual: actualChecksum)))
+                                    return completion(.failure(RegistryError
+                                            .invalidChecksum(expected: expectedChecksum, actual: actualChecksum)))
                                 case .warn:
-                                    observabilityScope.emit(warning: "The checksum \(actualChecksum) does not match previously recorded value \(expectedChecksum)")
+                                    observabilityScope
+                                        .emit(
+                                            warning: "The checksum \(actualChecksum) does not match previously recorded value \(expectedChecksum)"
+                                        )
                                 }
                             }
-
+                            // validate that the destination does not already exist (again, as this is async)
+                            guard !fileSystem.exists(destinationPath) else {
+                                throw RegistryError.pathAlreadyExists(destinationPath)
+                            }
+                            try fileSystem.createDirectory(destinationPath, recursive: true)
+                            // extract the content
                             let archiver = self.archiverProvider(fileSystem)
                             // TODO: Bail if archive contains relative paths or overlapping files
                             archiver.extract(from: downloadPath, to: destinationPath) { result in
@@ -572,13 +511,17 @@ public final class RegistryClient {
                                        version: version,
                                        kind: .registry,
                                        observabilityScope: observabilityScope,
-                                       callbackQueue: callbackQueue) { result in
+                                       callbackQueue: callbackQueue)
+                { result in
                     switch result {
                     case .success(let fingerprint):
                         body(.success(fingerprint.value))
                     case .failure(let error):
                         if error as? PackageFingerprintStorageError != .notFound {
-                            observabilityScope.emit(error: "Failed to get registry fingerprint for \(package) \(version) from storage: \(error)")
+                            observabilityScope
+                                .emit(
+                                    error: "Failed to get registry fingerprint for \(package) \(version) from storage: \(error)"
+                                )
                         }
                         // Try fetching checksum from registry again no matter which kind of error it is
                         self.fetchSourceArchiveChecksum(package: package,
@@ -599,7 +542,7 @@ public final class RegistryClient {
     }
 
     public func lookupIdentities(
-        url: Foundation.URL,
+        url: URL,
         timeout: DispatchTimeInterval? = .none,
         observabilityScope: ObservabilityScope,
         callbackQueue: DispatchQueue,
@@ -635,28 +578,61 @@ public final class RegistryClient {
 
         self.httpClient.execute(request, observabilityScope: observabilityScope, progress: nil) { result in
             completion(result.tryMap { response in
+                // 404 is valid, no identities mapped
+                if response.statusCode == 404 {
+                    return []
+                }
+
                 try self.checkResponseStatusAndHeaders(response, expectedStatusCode: 200, expectedContentType: .json)
 
-                guard let data = response.body,
-                      case .dictionary(let payload) = try? JSON(data: data),
-                      case .array(let identifiers) = payload["identifiers"]
-                else {
+                guard let data = response.body else {
                     throw RegistryError.invalidResponse
                 }
 
-                let packageIdentities: [PackageIdentity] = identifiers.compactMap {
-                    guard case .string(let string) = $0 else {
-                        return nil
-                    }
-                    return PackageIdentity.plain(string)
-                }
-
-                return Set(packageIdentities)
+                let packageIdentities = try self.jsonDecoder.decode(Serialization.PackageIdentifiers.self, from: data)
+                return Set(packageIdentities.identifiers.map {
+                    PackageIdentity.plain($0)
+                })
             })
         }
     }
 
-    private func makeAsync<T>(_ closure: @escaping (Result<T, Error>) -> Void, on queue: DispatchQueue) -> (Result<T, Error>) -> Void {
+    public func login(
+        url: URL,
+        timeout: DispatchTimeInterval? = .none,
+        observabilityScope: ObservabilityScope,
+        callbackQueue: DispatchQueue,
+        completion: @escaping (Result<Void, Error>) -> Void
+    ) {
+        let completion = self.makeAsync(completion, on: callbackQueue)
+
+        let request = HTTPClient.Request(
+            method: .post,
+            url: url,
+            options: self.defaultRequestOptions(timeout: timeout, callbackQueue: callbackQueue)
+        )
+
+        self.httpClient.execute(request, observabilityScope: observabilityScope, progress: nil) { result in
+            completion(
+                result.tryMap { response in
+                    switch response.statusCode {
+                    case 200:
+                        return ()
+                    case 401:
+                        throw RegistryError.unauthorized
+                    case 501:
+                        throw RegistryError.authenticationMethodNotSupported
+                    default:
+                        throw RegistryError.invalidResponseStatus(expected: 200, actual: response.statusCode)
+                    }
+                }
+            )
+        }
+    }
+
+    private func makeAsync<T>(_ closure: @escaping (Result<T, Error>) -> Void,
+                              on queue: DispatchQueue) -> (Result<T, Error>) -> Void
+    {
         { result in queue.async { closure(result) } }
     }
 
@@ -672,38 +648,126 @@ public final class RegistryClient {
     }
 }
 
-private extension RegistryClient {
-    enum APIVersion: String {
+public enum RegistryError: Error, CustomStringConvertible {
+    case registryNotConfigured(scope: PackageIdentity.Scope?)
+    case invalidPackage(PackageIdentity)
+    case invalidURL(URL)
+    case invalidResponseStatus(expected: Int, actual: Int)
+    case invalidContentVersion(expected: String, actual: String?)
+    case invalidContentType(expected: String, actual: String?)
+    case invalidResponse
+    case missingSourceArchive
+    case invalidSourceArchive
+    case unsupportedHashAlgorithm(String)
+    case failedToDetermineExpectedChecksum(Error)
+    case failedToComputeChecksum(Error)
+    case checksumChanged(latest: String, previous: String)
+    case invalidChecksum(expected: String, actual: String)
+    case pathAlreadyExists(AbsolutePath)
+    case failedRetrievingReleases(Error)
+    case failedRetrievingReleaseChecksum(Error)
+    case failedRetrievingManifest(Error)
+    case failedDownloadingSourceArchive(Error)
+    case unauthorized
+    case authenticationMethodNotSupported
+
+    public var description: String {
+        switch self {
+        case .registryNotConfigured(let scope):
+            if let scope = scope {
+                return "No registry configured for '\(scope)' scope"
+            } else {
+                return "No registry configured'"
+            }
+        case .invalidPackage(let package):
+            return "Invalid package '\(package)'"
+        case .invalidURL(let url):
+            return "Invalid URL '\(url)'"
+        case .invalidResponseStatus(let expected, let actual):
+            return "Invalid registry response status '\(actual)', expected '\(expected)'"
+        case .invalidContentVersion(expected: let expected, actual: let actual):
+            return "Invalid registry response content version '\(actual ?? "")', expected '\(expected)'"
+        case .invalidContentType(let expected, let actual):
+            return "Invalid registry response content type '\(actual ?? "")', expected '\(expected)'"
+        case .invalidResponse:
+            return "Invalid registry response"
+        case .missingSourceArchive:
+            return "Missing registry source archive"
+        case .invalidSourceArchive:
+            return "Invalid registry source archive"
+        case .unsupportedHashAlgorithm(let algorithm):
+            return "Unsupported hash algorithm '\(algorithm)'"
+        case .failedToDetermineExpectedChecksum(let error):
+            return "Failed determining registry source archive checksum: \(error)"
+        case .failedToComputeChecksum(let error):
+            return "Failed computing registry source archive checksum: \(error)"
+        case .checksumChanged(let latest, let previous):
+            return "The latest checksum '\(latest)' is different from the previously recorded value '\(previous)'"
+        case .invalidChecksum(let expected, let actual):
+            return "Invalid registry source archive checksum '\(actual)', expected '\(expected)'"
+        case .pathAlreadyExists(let path):
+            return "Path already exists '\(path)'"
+        case .failedRetrievingReleases(let error):
+            return "Failed fetching releases from registry: \(error)"
+        case .failedRetrievingReleaseChecksum(let error):
+            return "Failed fetching release checksum from registry: \(error)"
+        case .failedRetrievingManifest(let error):
+            return "Failed retrieving manifest from registry: \(error)"
+        case .failedDownloadingSourceArchive(let error):
+            return "Failed downloading source archive from registry: \(error)"
+        case .unauthorized:
+            return "Missing or invalid authentication credentials"
+        case .authenticationMethodNotSupported:
+            return "Authentication method not supported"
+        }
+    }
+}
+
+extension RegistryClient {
+    fileprivate enum APIVersion: String {
         case v1 = "1"
     }
 }
 
-private extension RegistryClient {
-    enum MediaType: String {
+extension RegistryClient {
+    fileprivate enum MediaType: String {
         case json
         case swift
         case zip
     }
 
-    enum ContentType: String {
+    fileprivate enum ContentType: String {
         case json = "application/json"
         case swift = "text/x-swift"
         case zip = "application/zip"
     }
 
-    func acceptHeader(mediaType: MediaType) -> String {
+    private func acceptHeader(mediaType: MediaType) -> String {
         "application/vnd.swift.registry.v\(self.apiVersion.rawValue)+\(mediaType)"
     }
+    
+    private func checkContentVersion(_ response: HTTPClient.Response, isOptional: Bool) throws {
+        let contentVersion = response.headers.get("Content-Version").first
+        if isOptional, contentVersion == nil {
+            return
+        }
+        // Check API version as long as `Content-Version` is set
+        guard contentVersion == self.apiVersion.rawValue else {
+            throw RegistryError.invalidContentVersion(expected: self.apiVersion.rawValue, actual: contentVersion)
+        }
+    }
 
-    func checkResponseStatusAndHeaders(_ response: HTTPClient.Response, expectedStatusCode: Int, expectedContentType: ContentType) throws {
+    private func checkResponseStatusAndHeaders(
+        _ response: HTTPClient.Response,
+        expectedStatusCode: Int,
+        expectedContentType: ContentType,
+        isContentVersionOptional: Bool = false
+    ) throws {
         guard response.statusCode == expectedStatusCode else {
             throw RegistryError.invalidResponseStatus(expected: expectedStatusCode, actual: response.statusCode)
         }
 
-        let contentVersion = response.headers.get("Content-Version").first
-        guard contentVersion == self.apiVersion.rawValue else {
-            throw RegistryError.invalidContentVersion(expected: self.apiVersion.rawValue, actual: contentVersion)
-        }
+        try checkContentVersion(response, isOptional: isContentVersionOptional)
 
         let contentType = response.headers.get("Content-Type").first
         guard contentType?.hasPrefix(expectedContentType.rawValue) == true else {
@@ -712,12 +776,154 @@ private extension RegistryClient {
     }
 }
 
+extension RegistryClient {
+    public struct PackageMetadata {
+        public let registry: Registry
+        public let versions: [Version]
+        public let alternateLocations: [URL]?
+    }
+}
+
+extension RegistryClient {
+    fileprivate struct AlternativeLocationLink {
+        let url: URL
+        let kind: Kind
+
+        enum Kind: String {
+            case canonical
+            case alternate
+        }
+    }
+}
+
+extension RegistryClient {
+    fileprivate struct ManifestLink {
+        let url: URL
+        let filename: String
+        let toolsVersion: ToolsVersion
+    }
+}
+
+extension HTTPClientHeaders {
+    /*
+     <https://github.com/mona/LinkedList>; rel="canonical",
+     <ssh://git@github.com:mona/LinkedList.git>; rel="alternate",
+      */
+    fileprivate func parseAlternativeLocationLinks() throws -> [RegistryClient.AlternativeLocationLink]? {
+        try self.get("Link").map { header -> [RegistryClient.AlternativeLocationLink] in
+            let linkLines = header.split(separator: ",").map(String.init).map { $0.spm_chuzzle() ?? $0 }
+            return try linkLines.compactMap { linkLine in
+                try parseAlternativeLocationLine(linkLine)
+            }
+        }.flatMap { $0 }
+    }
+
+    private func parseAlternativeLocationLine(_ value: String) throws -> RegistryClient.AlternativeLocationLink? {
+        let fields = value.split(separator: ";")
+            .map(String.init)
+            .map { $0.spm_chuzzle() ?? $0 }
+
+        guard fields.count == 2 else {
+            return nil
+        }
+
+        guard let link = fields.first(where: { $0.hasPrefix("<") }).map({ String($0.dropFirst().dropLast()) }),
+              let url = URL(string: link)
+        else {
+            return nil
+        }
+
+        guard let rel = fields.first(where: { $0.hasPrefix("rel=") }).flatMap({ parseLinkFieldValue($0) }),
+              let kind = RegistryClient.AlternativeLocationLink.Kind(rawValue: rel)
+        else {
+            return nil
+        }
+
+        return RegistryClient.AlternativeLocationLink(
+            url: url,
+            kind: kind
+        )
+    }
+}
+
+extension HTTPClientHeaders {
+    /*
+     <http://packages.example.com/mona/LinkedList/1.1.1/Package.swift?swift-version=4>; rel="alternate"; filename="Package@swift-4.swift"; swift-tools-version="4.0"
+     */
+    fileprivate func parseManifestLinks() throws -> [RegistryClient.ManifestLink] {
+        try self.get("Link").map { header -> [RegistryClient.ManifestLink] in
+            let linkLines = header.split(separator: ",").map(String.init).map { $0.spm_chuzzle() ?? $0 }
+            return try linkLines.compactMap { linkLine in
+                try parseManifestLinkLine(linkLine)
+            }
+        }.flatMap { $0 }
+    }
+
+    private func parseManifestLinkLine(_ value: String) throws -> RegistryClient.ManifestLink? {
+        let fields = value.split(separator: ";")
+            .map(String.init)
+            .map { $0.spm_chuzzle() ?? $0 }
+
+        guard fields.count == 4 else {
+            return nil
+        }
+
+        guard let link = fields.first(where: { $0.hasPrefix("<") }).map({ String($0.dropFirst().dropLast()) }),
+              let url = URL(string: link)
+        else {
+            return nil
+        }
+
+        guard let rel = fields.first(where: { $0.hasPrefix("rel=") }).flatMap({ parseLinkFieldValue($0) }),
+              rel == "alternate"
+        else {
+            return nil
+        }
+
+        guard let filename = fields.first(where: { $0.hasPrefix("filename=") }).flatMap({ parseLinkFieldValue($0) })
+        else {
+            return nil
+        }
+
+        guard let toolsVersion = fields.first(where: { $0.hasPrefix("swift-tools-version=") })
+            .flatMap({ parseLinkFieldValue($0) })
+        else {
+            return nil
+        }
+
+        guard let toolsVersion = ToolsVersion(string: toolsVersion) else {
+            throw StringError("Invalid tools version in alternate manifest link '\(value)'")
+        }
+
+        return RegistryClient.ManifestLink(
+            url: url,
+            filename: filename,
+            toolsVersion: toolsVersion
+        )
+    }
+}
+
+extension HTTPClientHeaders {
+    private func parseLinkFieldValue(_ field: String) -> String? {
+        let parts = field.split(separator: "=")
+            .map(String.init)
+            .map { $0.spm_chuzzle() ?? $0 }
+
+        guard parts.count == 2 else {
+            return nil
+        }
+
+        return parts[1].replacingOccurrences(of: "\"", with: "")
+    }
+}
+
 // MARK: - Serialization
 
-public extension RegistryClient {
-    enum Serialization {
+// marked public for cross module visibility
+extension RegistryClient {
+    public enum Serialization {
         public struct PackageMetadata: Codable {
-            public var releases: [String: Release]
+            public let releases: [String: Release]
 
             public init(releases: [String: Release]) {
                 self.releases = releases
@@ -746,17 +952,18 @@ public extension RegistryClient {
             }
         }
 
+        // marked public for cross module visibility
         public struct VersionMetadata: Codable {
-            public var id: String
-            public var version: String
-            public var resources: [Resource]
-            public var metadata: AdditionalMetadata
+            public let id: String
+            public let version: String
+            public let resources: [Resource]
+            public let metadata: AdditionalMetadata?
 
             public init(
                 id: String,
                 version: String,
                 resources: [Resource],
-                metadata: AdditionalMetadata
+                metadata: AdditionalMetadata?
             ) {
                 self.id = id
                 self.version = version
@@ -784,31 +991,30 @@ public extension RegistryClient {
                 }
             }
         }
+
+        // marked public for cross module visibility
+        public struct PackageIdentifiers: Codable {
+            public let identifiers: [String]
+
+            public init(identifiers: [String]) {
+                self.identifiers = identifiers
+            }
+        }
     }
 }
 
 // MARK: - Utilities
 
-private extension String {
-    /// Drops the given suffix from the string, if present.
-    func spm_dropPrefix(_ prefix: String) -> String {
-        if hasPrefix(prefix) {
-            return String(dropFirst(prefix.count))
-        }
-        return self
-    }
-}
-
-private extension AbsolutePath {
-    func withExtension(_ extension: String) -> AbsolutePath {
+extension AbsolutePath {
+    fileprivate func withExtension(_ extension: String) -> AbsolutePath {
         guard !self.isRoot else { return self }
         let `extension` = `extension`.spm_dropPrefix(".")
-        return AbsolutePath(self, RelativePath("..")).appending(component: "\(basename).\(`extension`)")
+        return self.parentDirectory.appending(component: "\(basename).\(`extension`)")
     }
 }
 
-private extension URLComponents {
-    mutating func appendPathComponents(_ components: String...) {
+extension URLComponents {
+    fileprivate mutating func appendPathComponents(_ components: String...) {
         path += (path.last == "/" ? "" : "/") + components.joined(separator: "/")
     }
 }
