@@ -1,12 +1,14 @@
-/*
- This source file is part of the Swift.org open source project
-
- Copyright (c) 2014 - 2021 Apple Inc. and the Swift project authors
- Licensed under Apache License v2.0 with Runtime Library Exception
-
- See http://swift.org/LICENSE.txt for license information
- See http://swift.org/CONTRIBUTORS.txt for Swift project authors
- */
+//===----------------------------------------------------------------------===//
+//
+// This source file is part of the Swift open source project
+//
+// Copyright (c) 2014-2021 Apple Inc. and the Swift project authors
+// Licensed under Apache License v2.0 with Runtime Library Exception
+//
+// See http://swift.org/LICENSE.txt for license information
+// See http://swift.org/CONTRIBUTORS.txt for the list of Swift project authors
+//
+//===----------------------------------------------------------------------===//
 
 import Basics
 import PackageGraph
@@ -23,69 +25,70 @@ public typealias Diagnostic = TSCBasic.Diagnostic
 public final class MockWorkspace {
     let sandbox: AbsolutePath
     let fileSystem: InMemoryFileSystem
-    public let httpClient: HTTPClient
-    public var registryClient: RegistryClient
-    public let archiver: MockArchiver
-    public let checksumAlgorithm: MockHashAlgorithm
-    public let fingerprintStorage: MockPackageFingerprintStorage
-    public let customPackageContainerProvider: MockPackageContainerProvider?
     let roots: [MockPackage]
     let packages: [MockPackage]
-    public let mirrors: DependencyMirrors
-    let identityResolver: IdentityResolver
-    public var manifestLoader: MockManifestLoader
-    public var repositoryProvider: InMemoryGitRepositoryProvider
+    let customToolsVersion: ToolsVersion?
+    let fingerprints: MockPackageFingerprintStorage
+    let mirrors: DependencyMirrors
+    public var registryClient: RegistryClient
     let registry: MockRegistry
+    let customBinaryArtifactsManager: Workspace.CustomBinaryArtifactsManager
+    public var checksumAlgorithm: MockHashAlgorithm
+    public private(set) var manifestLoader: MockManifestLoader
+    public let repositoryProvider: InMemoryGitRepositoryProvider
+    let identityResolver: IdentityResolver
+    let customPackageContainerProvider: MockPackageContainerProvider?
     public let delegate = MockWorkspaceDelegate()
-    let toolsVersion: ToolsVersion
-    let resolverUpdateEnabled: Bool
+    let skipDependenciesUpdates: Bool
+    public var sourceControlToRegistryDependencyTransformation: WorkspaceConfiguration.SourceControlToRegistryDependencyTransformation
 
     public init(
         sandbox: AbsolutePath,
         fileSystem: InMemoryFileSystem,
-        mirrors: DependencyMirrors? = nil,
         roots: [MockPackage],
-        packages: [MockPackage],
-        toolsVersion: ToolsVersion = ToolsVersion.currentToolsVersion,
-        customHttpClient: HTTPClient? = .none,
-        customRegistryClient: RegistryClient? = .none,
-        customBinaryArchiver: MockArchiver? = .none,
-        customChecksumAlgorithm: MockHashAlgorithm? = .none,
-        customFingerprintStorage: MockPackageFingerprintStorage? = .none,
+        packages: [MockPackage] = [],
+        toolsVersion customToolsVersion: ToolsVersion? = .none,
+        fingerprints customFingerprints: MockPackageFingerprintStorage? = .none,
+        mirrors customMirrors: DependencyMirrors? = nil,
+        registryClient customRegistryClient: RegistryClient? = .none,
+        binaryArtifactsManager customBinaryArtifactsManager: Workspace.CustomBinaryArtifactsManager? = .none,
+        checksumAlgorithm customChecksumAlgorithm: MockHashAlgorithm? = .none,
         customPackageContainerProvider: MockPackageContainerProvider? = .none,
-        resolverUpdateEnabled: Bool = true
+        skipDependenciesUpdates: Bool = false,
+        sourceControlToRegistryDependencyTransformation: WorkspaceConfiguration.SourceControlToRegistryDependencyTransformation = .disabled
     ) throws {
-        let archiver = customBinaryArchiver ?? MockArchiver()
-        let httpClient = customHttpClient ?? HTTPClient.mock(fileSystem: fileSystem)
-
         self.sandbox = sandbox
         self.fileSystem = fileSystem
-        self.httpClient = httpClient
-        self.archiver = archiver
-        self.checksumAlgorithm = customChecksumAlgorithm ?? MockHashAlgorithm()
-        self.fingerprintStorage = customFingerprintStorage ?? MockPackageFingerprintStorage()
-        self.mirrors = mirrors ?? DependencyMirrors()
-        self.identityResolver = DefaultIdentityResolver(locationMapper: self.mirrors.effectiveURL(for:))
-        self.customPackageContainerProvider = customPackageContainerProvider
         self.roots = roots
         self.packages = packages
-
+        self.fingerprints = customFingerprints ?? MockPackageFingerprintStorage()
+        self.mirrors = customMirrors ?? DependencyMirrors()
+        self.identityResolver = DefaultIdentityResolver(
+            locationMapper: self.mirrors.effectiveURL(for:),
+            identityMapper: self.mirrors.effectiveIdentity(for:)
+        )
         self.manifestLoader = MockManifestLoader(manifests: [:])
+        self.customPackageContainerProvider = customPackageContainerProvider
         self.repositoryProvider = InMemoryGitRepositoryProvider()
+        self.checksumAlgorithm = customChecksumAlgorithm ?? MockHashAlgorithm()
         self.registry = MockRegistry(
+            filesystem: self.fileSystem,
             identityResolver: self.identityResolver,
             checksumAlgorithm: self.checksumAlgorithm,
-            filesystem: self.fileSystem,
-            fingerprintStorage: self.fingerprintStorage
+            fingerprintStorage: self.fingerprints
         )
         self.registryClient = customRegistryClient ?? self.registry.registryClient
-        self.toolsVersion = toolsVersion
-        self.resolverUpdateEnabled = resolverUpdateEnabled
-
+        self.customToolsVersion = customToolsVersion
+        self.skipDependenciesUpdates = skipDependenciesUpdates
+        self.sourceControlToRegistryDependencyTransformation = sourceControlToRegistryDependencyTransformation
+        self.customBinaryArtifactsManager = customBinaryArtifactsManager ?? .init(
+            httpClient: HTTPClient.mock(fileSystem: fileSystem),
+            archiver: MockArchiver()
+        )
         try self.create()
     }
 
-    private var rootsDir: AbsolutePath {
+    public var rootsDir: AbsolutePath {
         return self.sandbox.appending(component: "roots")
     }
 
@@ -97,8 +100,12 @@ public final class MockWorkspace {
         return self.sandbox.appending(components: ".build", "artifacts")
     }
 
-    public func pathToPackage(withName name: String) -> AbsolutePath {
-        return AbsolutePath(name, relativeTo: self.packagesDir)
+    public func pathToRoot(withName name: String) throws -> AbsolutePath {
+        return try AbsolutePath(validating: name, relativeTo: self.rootsDir)
+    }
+
+    public func pathToPackage(withName name: String) throws -> AbsolutePath {
+        return try AbsolutePath(validating: name, relativeTo: self.packagesDir)
     }
 
     private func create() throws {
@@ -129,7 +136,7 @@ public final class MockWorkspace {
                 } else {
                     packagePath = basePath.appending(components: "sourceControl", url.absoluteString.spm_mangledToC99ExtendedIdentifier())
                 }
-            case .registry(let identity):
+            case .registry(let identity, _):
                 packagePath = basePath.appending(components: "registry", identity.description.spm_mangledToC99ExtendedIdentifier())
             }
 
@@ -139,6 +146,7 @@ public final class MockWorkspace {
 
             var sourceControlSpecifier: RepositorySpecifier? = nil
             var registryIdentity: PackageIdentity? = nil
+            var registryAlternativeURLs: [URL]? = nil
 
             switch (isRoot, package.location) {
             case (true, _):
@@ -153,20 +161,20 @@ public final class MockWorkspace {
                 packageLocation = url.absoluteString
                 packageKind = .remoteSourceControl(url)
                 sourceControlSpecifier = RepositorySpecifier(url: url)
-            case (_, .registry(let identity)):
+            case (_, .registry(let identity, let alternativeURLs)):
                 packageLocation = identity.description
                 packageKind = .registry(identity)
                 registryIdentity = identity
+                registryAlternativeURLs = alternativeURLs
             }
 
-            let toolsVersion = package.toolsVersion ?? .currentToolsVersion
-
             // Create targets on disk.
+            let packageToolsVersion = package.toolsVersion ?? .current
             if let specifier = sourceControlSpecifier {
                 let repository = self.repositoryProvider.specifierMap[specifier] ?? .init(path: packagePath, fs: self.fileSystem)
-                try writePackageContent(fileSystem: repository, root: .root, toolsVersion: toolsVersion)
+                try writePackageContent(fileSystem: repository, root: .root, toolsVersion: packageToolsVersion)
                 
-                let versions = packageVersions.compactMap({ $0 })
+                let versions = packageVersions.compactMap{ $0 }
                 if versions.isEmpty {
                     try repository.commit()
                 } else {
@@ -178,9 +186,14 @@ public final class MockWorkspace {
 
                 self.repositoryProvider.add(specifier: specifier, repository: repository)
             } else if let identity = registryIdentity {
-                let source = InMemoryRegistrySource(path: packagePath, fileSystem: self.fileSystem)
-                try writePackageContent(fileSystem: source.fileSystem, root: source.path, toolsVersion: toolsVersion)
-                self.registry.addPackage(identity: identity, versions: packageVersions.compactMap({ $0 }), source: source)
+                let source = InMemoryRegistryPackageSource(fileSystem: self.fileSystem, path: packagePath, writeContent: false)
+                try writePackageContent(fileSystem: source.fileSystem, root: source.path, toolsVersion: packageToolsVersion)
+                self.registry.addPackage(
+                    identity: identity,
+                    versions: packageVersions.compactMap{ $0 },
+                    sourceControlURLs: registryAlternativeURLs ?? [],
+                    source: source
+                )
             } else {
                 throw InternalError("unknown package type")
             }
@@ -195,10 +208,10 @@ public final class MockWorkspace {
                     packageLocation: packageLocation,
                     platforms: package.platforms,
                     version: v,
-                    toolsVersion: toolsVersion,
+                    toolsVersion: packageToolsVersion,
                     dependencies: package.dependencies.map { try $0.convert(baseURL: packagesDir, identityResolver: self.identityResolver) },
-                    products: package.products.map { ProductDescription(name: $0.name, type: .library(.automatic), targets: $0.targets) },
-                    targets: try package.targets.map { try $0.convert() }
+                    products: package.products.map { try ProductDescription(name: $0.name, type: .library(.automatic), targets: $0.targets) },
+                    targets: try package.targets.map { try $0.convert(identityResolver: self.identityResolver) }
                 )
             }
 
@@ -211,7 +224,11 @@ public final class MockWorkspace {
                 }
                 let manifestPath = root.appending(component: Manifest.filename)
                 try fileSystem.writeFileContents(manifestPath, bytes: "")
-                try rewriteToolsVersionSpecification(toDefaultManifestIn: root, specifying: toolsVersion, fileSystem: fileSystem)
+                try ToolsVersionSpecificationWriter.rewriteSpecification(
+                    manifestDirectory: root,
+                    toolsVersion: toolsVersion,
+                    fileSystem: fileSystem
+                )
             }
         }
 
@@ -233,30 +250,38 @@ public final class MockWorkspace {
             return workspace
         }
 
-        let workspace = try Workspace(
+        let workspace = try Workspace._init(
             fileSystem: self.fileSystem,
             location: .init(
-                workingDirectory: self.sandbox.appending(component: ".build"),
+                scratchDirectory: self.sandbox.appending(component: ".build"),
                 editsDirectory: self.sandbox.appending(component: "edits"),
                 resolvedVersionsFile: self.sandbox.appending(component: "Package.resolved"),
-                sharedSecurityDirectory: self.fileSystem.swiftPMSecurityDirectory,  
-                sharedCacheDirectory: self.fileSystem.swiftPMCacheDirectory,
-                sharedConfigurationDirectory: self.fileSystem.swiftPMConfigurationDirectory
+                localConfigurationDirectory: Workspace.DefaultLocations.configurationDirectory(forRootPackage: self.sandbox),
+                sharedConfigurationDirectory: self.fileSystem.swiftPMConfigurationDirectory,
+                sharedSecurityDirectory: self.fileSystem.swiftPMSecurityDirectory,
+                sharedCacheDirectory: self.fileSystem.swiftPMCacheDirectory
             ),
-            mirrors: self.mirrors,
-            customToolsVersion: self.toolsVersion,
+            configuration: .init(
+                skipDependenciesUpdates: self.skipDependenciesUpdates,
+                prefetchBasedOnResolvedFile: WorkspaceConfiguration.default.prefetchBasedOnResolvedFile,
+                shouldCreateMultipleTestProducts: WorkspaceConfiguration.default.shouldCreateMultipleTestProducts,
+                createREPLProduct: WorkspaceConfiguration.default.createREPLProduct,
+                additionalFileRules: WorkspaceConfiguration.default.additionalFileRules,
+                sharedDependenciesCacheEnabled: WorkspaceConfiguration.default.sharedDependenciesCacheEnabled,
+                fingerprintCheckingMode: .strict,
+                sourceControlToRegistryDependencyTransformation: self.sourceControlToRegistryDependencyTransformation,
+                restrictImports: .none
+            ),
+            customFingerprints: self.fingerprints,
+            customMirrors: self.mirrors,
+            customToolsVersion: self.customToolsVersion,
             customManifestLoader: self.manifestLoader,
             customPackageContainerProvider: self.customPackageContainerProvider,
             customRepositoryProvider: self.repositoryProvider,
             customRegistryClient: self.registryClient,
+            customBinaryArtifactsManager: self.customBinaryArtifactsManager,
             customIdentityResolver: self.identityResolver,
-            customHTTPClient: self.httpClient,
-            customArchiver: self.archiver,
             customChecksumAlgorithm: self.checksumAlgorithm,
-            customFingerprintStorage: self.fingerprintStorage,
-            resolverUpdateEnabled: self.resolverUpdateEnabled,
-            resolverPrefetchingEnabled: true,
-            resolverFingerprintCheckingMode: .strict,
             delegate: self.delegate
         )
 
@@ -267,12 +292,15 @@ public final class MockWorkspace {
 
     private var _workspace: Workspace?
 
-    public func closeWorkspace() {
+    public func closeWorkspace(resetState: Bool = true) throws {
+        if resetState {
+            try self._workspace?.resetState()
+        }
         self._workspace = nil
     }
 
-    public func rootPaths(for packages: [String]) -> [AbsolutePath] {
-        return packages.map { AbsolutePath($0, relativeTo: rootsDir) }
+    public func rootPaths(for packages: [String]) throws -> [AbsolutePath] {
+        return try packages.map { try AbsolutePath(validating: $0, relativeTo: rootsDir) }
     }
 
     public func checkEdit(
@@ -303,8 +331,8 @@ public final class MockWorkspace {
         _ result: ([Basics.Diagnostic]) -> Void
     ) {
         let observability = ObservabilitySystem.makeForTesting()
-        let rootInput = PackageGraphRootInput(packages: rootPaths(for: roots))
         observability.topScope.trap {
+            let rootInput = PackageGraphRootInput(packages: try rootPaths(for: roots))
             let ws = try self.getOrCreateWorkspace()
             try ws.unedit(packageName: packageName, forceRemove: forceRemove, root: rootInput, observabilityScope: observability.topScope)
         }
@@ -313,8 +341,8 @@ public final class MockWorkspace {
 
     public func checkResolve(pkg: String, roots: [String], version: TSCUtility.Version, _ result: ([Basics.Diagnostic]) -> Void) {
         let observability = ObservabilitySystem.makeForTesting()
-        let rootInput = PackageGraphRootInput(packages: rootPaths(for: roots))
         observability.topScope.trap {
+            let rootInput = PackageGraphRootInput(packages: try rootPaths(for: roots))
             let workspace = try self.getOrCreateWorkspace()
             try workspace.resolve(packageName: pkg, root: rootInput, version: version, branch: nil, revision: nil, observabilityScope: observability.topScope)
         }
@@ -350,7 +378,7 @@ public final class MockWorkspace {
         let observability = ObservabilitySystem.makeForTesting()
         observability.topScope.trap {
             let rootInput = PackageGraphRootInput(
-                packages: rootPaths(for: roots),
+                packages: try rootPaths(for: roots),
                 dependencies: dependencies
             )
             let workspace = try self.getOrCreateWorkspace()
@@ -366,7 +394,7 @@ public final class MockWorkspace {
     ) throws {
         let dependencies = try deps.map { try $0.convert(baseURL: packagesDir, identityResolver: self.identityResolver) }
         let rootInput = PackageGraphRootInput(
-            packages: rootPaths(for: roots),
+            packages: try rootPaths(for: roots),
             dependencies: dependencies
         )
 
@@ -391,19 +419,27 @@ public final class MockWorkspace {
         roots: [String] = [],
         dependencies: [PackageDependency] = [],
         forceResolvedVersions: Bool = false,
-        _ result: (PackageGraph, [Basics.Diagnostic]) -> Void
+        _ result: (PackageGraph, [Basics.Diagnostic]) throws -> Void
     ) throws {
         let observability = ObservabilitySystem.makeForTesting()
         let rootInput = PackageGraphRootInput(
-            packages: rootPaths(for: roots), dependencies: dependencies
+            packages: try rootPaths(for: roots), dependencies: dependencies
         )
         let workspace = try self.getOrCreateWorkspace()
-        let graph = try workspace.loadPackageGraph(
-            rootInput: rootInput,
-            forceResolvedVersions: forceResolvedVersions,
-            observabilityScope: observability.topScope
-        )
-        result(graph, observability.diagnostics)
+        do {
+            let graph = try workspace.loadPackageGraph(
+                rootInput: rootInput,
+                forceResolvedVersions: forceResolvedVersions,
+                observabilityScope: observability.topScope
+            )
+            try result(graph, observability.diagnostics)
+        } catch {
+            // helpful when graph fails to load
+            if observability.hasErrorDiagnostics {
+                print(observability.diagnostics.map{ $0.description }.joined(separator: "\n"))
+            }
+            throw error
+        }
     }
 
     public func checkPackageGraphFailure(
@@ -422,11 +458,11 @@ public final class MockWorkspace {
         _ result: ([Basics.Diagnostic]) -> Void
     ) {
         let observability = ObservabilitySystem.makeForTesting()
-        let rootInput = PackageGraphRootInput(
-            packages: rootPaths(for: roots),
-            dependencies: dependencies
-        )
         observability.topScope.trap {
+            let rootInput = PackageGraphRootInput(
+                packages: try rootPaths(for: roots),
+                dependencies: dependencies
+            )
             let workspace = try self.getOrCreateWorkspace()
             try workspace.loadPackageGraph(
                 rootInput: rootInput,
@@ -447,7 +483,7 @@ public final class MockWorkspace {
         let workspace = try self.getOrCreateWorkspace()
         let pinsStore = try workspace.pinsStore.load()
 
-        let rootInput = PackageGraphRootInput(packages: rootPaths(for: roots.map { $0.name }), dependencies: [])
+        let rootInput = PackageGraphRootInput(packages: try rootPaths(for: roots.map { $0.name }), dependencies: [])
         let rootManifests = try temp_await { workspace.loadRootManifests(packages: rootInput.packages, observabilityScope: observability.topScope, completion: $0) }
         let root = PackageGraphRoot(input: rootInput, manifests: rootManifests)
 
@@ -466,7 +502,7 @@ public final class MockWorkspace {
 
     public func set(
         pins: [PackageReference: CheckoutState] = [:],
-        managedDependencies: [Workspace.ManagedDependency] = [],
+        managedDependencies: [AbsolutePath: Workspace.ManagedDependency] = [:],
         managedArtifacts: [Workspace.ManagedArtifact] = []
     ) throws {
         let pins = pins.mapValues { checkoutState -> PinsStore.PinState in
@@ -483,8 +519,8 @@ public final class MockWorkspace {
     }
 
     public func set(
-        pins: [PackageReference: PinsStore.PinState] = [:],
-        managedDependencies: [Workspace.ManagedDependency] = [],
+        pins: [PackageReference: PinsStore.PinState],
+        managedDependencies: [AbsolutePath: Workspace.ManagedDependency] = [:],
         managedArtifacts: [Workspace.ManagedArtifact] = []
     ) throws {
         let workspace = try self.getOrCreateWorkspace()
@@ -495,13 +531,20 @@ public final class MockWorkspace {
         }
 
         for dependency in managedDependencies {
-            try self.fileSystem.createDirectory(workspace.path(to: dependency), recursive: true)
-            workspace.state.dependencies.add(dependency)
+            // copy the package content to expected managed path
+            let managedPath = workspace.path(to: dependency.value)
+            if managedPath != dependency.key, self.fileSystem.exists(dependency.key) {
+                try self.fileSystem.createDirectory(managedPath.parentDirectory, recursive: true)
+                try self.fileSystem.copy(from: dependency.key, to: managedPath)
+            } else {
+                try self.fileSystem.createDirectory(managedPath, recursive: true)
+            }
+            workspace.state.dependencies.add(dependency.value)
         }
 
         for artifact in managedArtifacts {
+            // create an empty directory representing the artifact
             try self.fileSystem.createDirectory(artifact.path, recursive: true)
-
             workspace.state.artifacts.add(artifact)
         }
 
@@ -630,7 +673,7 @@ public final class MockWorkspace {
             line: UInt = #line
         ) {
             guard let artifact = managedArtifacts[packageIdentity: packageIdentity, targetName: targetName] else {
-                XCTFail("\(packageIdentity).\(targetName) does not exists", file: file, line: line)
+                XCTFail("managed artifact '\(packageIdentity).\(targetName)' does not exists", file: file, line: line)
                 return
             }
             XCTAssertEqual(artifact.path, path, file: file, line: line)
@@ -655,7 +698,7 @@ public final class MockWorkspace {
         let dependencies = try deps.map { try $0.convert(baseURL: packagesDir, identityResolver: self.identityResolver) }
         let workspace = try self.getOrCreateWorkspace()
         let rootInput = PackageGraphRootInput(
-            packages: rootPaths(for: roots), dependencies: dependencies
+            packages: try rootPaths(for: roots), dependencies: dependencies
         )
         let rootManifests = try tsc_await { workspace.loadRootManifests(packages: rootInput.packages, observabilityScope: observability.topScope, completion: $0) }
         let graphRoot = PackageGraphRoot(input: rootInput, manifests: rootManifests)
@@ -740,66 +783,66 @@ public final class MockWorkspace {
 }
 
 public final class MockWorkspaceDelegate: WorkspaceDelegate {
-    private let lock = Lock()
+    private let lock = NSLock()
     private var _events = [String]()
     private var _manifest: Manifest?
     private var _manifestLoadingDiagnostics: [Basics.Diagnostic]?
 
     public init() {}
 
-    public func repositoryWillUpdate(_ repository: String) {
-        self.append("updating repo: \(repository)")
+    public func willUpdateRepository(package: PackageIdentity, repository url: String) {
+        self.append("updating repo: \(url)")
     }
 
-    public func repositoryDidUpdate(_ repository: String, duration: DispatchTimeInterval) {
-        self.append("finished updating repo: \(repository)")
+    public func didUpdateRepository(package: PackageIdentity, repository url: String, duration: DispatchTimeInterval) {
+        self.append("finished updating repo: \(url)")
     }
 
     public func dependenciesUpToDate() {
         self.append("Everything is already up-to-date")
     }
 
-    public func fetchingWillBegin(repository: String, fetchDetails: RepositoryManager.FetchDetails?) {
-        self.append("fetching repo: \(repository)")
+    public func willFetchPackage(package: PackageIdentity, packageLocation: String?, fetchDetails: PackageFetchDetails) {
+        self.append("fetching package: \(packageLocation ?? package.description)")
     }
 
-    public func fetchingRepository(from repository: String, objectsFetched: Int, totalObjectsToFetch: Int) {
+    public func fetchingPackage(package: PackageIdentity, packageLocation: String?, progress: Int64, total: Int64?) {
     }
 
-    public func fetchingDidFinish(repository: String, fetchDetails: RepositoryManager.FetchDetails?, diagnostic: Basics.Diagnostic?, duration: DispatchTimeInterval) {
-        self.append("finished fetching repo: \(repository)")
+    public func didFetchPackage(package: PackageIdentity, packageLocation: String?, result: Result<PackageFetchDetails, Error>, duration: DispatchTimeInterval) {
+        self.append("finished fetching package: \(packageLocation ?? package.description)")
     }
 
-    public func willCreateWorkingCopy(repository url: String, at path: AbsolutePath) {
+    public func willCreateWorkingCopy(package: PackageIdentity, repository url: String, at path: AbsolutePath) {
         self.append("creating working copy for: \(url)")
     }
 
-    public func didCreateWorkingCopy(repository url: String, at path: AbsolutePath, error: Basics.Diagnostic?) {
+    public func didCreateWorkingCopy(package: PackageIdentity, repository url: String, at path: AbsolutePath, duration: DispatchTimeInterval) {
         self.append("finished creating working copy for: \(url)")
     }
 
-    public func willCheckOut(repository url: String, revision: String, at path: AbsolutePath) {
+    public func willCheckOut(package: PackageIdentity, repository url: String, revision: String, at path: AbsolutePath) {
         self.append("checking out repo: \(url)")
     }
 
-    public func didCheckOut(repository url: String, revision: String, at path: AbsolutePath, error: Basics.Diagnostic?) {
+    public func didCheckOut(package: PackageIdentity, repository url: String, revision: String, at path: AbsolutePath, duration: DispatchTimeInterval) {
         self.append("finished checking out repo: \(url)")
     }
 
-    public func removing(repository: String) {
-        self.append("removing repo: \(repository)")
+    public func removing(package: PackageIdentity, packageLocation: String?) {
+        self.append("removing repo: \(packageLocation ?? package.description)")
     }
 
     public func willResolveDependencies(reason: WorkspaceResolveReason) {
         self.append("will resolve dependencies")
     }
 
-    public func willLoadManifest(packagePath: AbsolutePath, url: String, version: Version?, packageKind: PackageReference.Kind) {
-        self.append("will load manifest for \(packageKind.displayName) package: \(url)")
+    public func willLoadManifest(packageIdentity: PackageIdentity, packagePath: AbsolutePath, url: String, version: Version?, packageKind: PackageReference.Kind) {
+        self.append("will load manifest for \(packageKind.displayName) package: \(url) (identity: \(packageIdentity))")
     }
 
-    public func didLoadManifest(packagePath: AbsolutePath, url: String, version: Version?, packageKind: PackageReference.Kind, manifest: Manifest?, diagnostics: [Basics.Diagnostic]) {
-        self.append("did load manifest for \(packageKind.displayName) package: \(url)")
+    public func didLoadManifest(packageIdentity: PackageIdentity, packagePath: AbsolutePath, url: String, version: Version?, packageKind: PackageReference.Kind, manifest: Manifest?, diagnostics: [Basics.Diagnostic], duration: DispatchTimeInterval) {
+        self.append("did load manifest for \(packageKind.displayName) package: \(url) (identity: \(packageIdentity))")
         self.lock.withLock {
             self._manifest = manifest
             self._manifestLoadingDiagnostics = diagnostics
@@ -818,11 +861,19 @@ public final class MockWorkspaceDelegate: WorkspaceDelegate {
         // noop
     }
 
+    public func willDownloadBinaryArtifact(from url: String) {
+        self.append("downloading binary artifact package: \(url)")
+    }
+
+    public func didDownloadBinaryArtifact(from url: String, result: Result<AbsolutePath, Error>, duration: DispatchTimeInterval) {
+        self.append("finished downloading binary artifact package: \(url)")
+    }
+
     public func downloadingBinaryArtifact(from url: String, bytesDownloaded: Int64, totalBytesToDownload: Int64?) {
         // noop
     }
 
-    public func didDownloadBinaryArtifacts() {
+    public func didDownloadAllBinaryArtifacts() {
         // noop
     }
 
