@@ -1,22 +1,26 @@
-/*
- This source file is part of the Swift.org open source project
-
- Copyright (c) 2014 - 2021 Apple Inc. and the Swift project authors
- Licensed under Apache License v2.0 with Runtime Library Exception
-
- See http://swift.org/LICENSE.txt for license information
- See http://swift.org/CONTRIBUTORS.txt for Swift project authors
-*/
+//===----------------------------------------------------------------------===//
+//
+// This source file is part of the Swift open source project
+//
+// Copyright (c) 2014-2021 Apple Inc. and the Swift project authors
+// Licensed under Apache License v2.0 with Runtime Library Exception
+//
+// See http://swift.org/LICENSE.txt for license information
+// See http://swift.org/CONTRIBUTORS.txt for the list of Swift project authors
+//
+//===----------------------------------------------------------------------===//
 
 import Basics
 import Dispatch
+import class Foundation.NSLock
 import PackageFingerprint
 import PackageGraph
 import PackageLoading
 import PackageModel
 import SourceControl
 import TSCBasic
-import TSCUtility
+
+import enum TSCUtility.Git
 
 /// Adaptor to expose an individual repository as a package container.
 internal final class SourceControlPackageContainer: PackageContainer, CustomStringConvertible {
@@ -25,7 +29,6 @@ internal final class SourceControlPackageContainer: PackageContainer, CustomStri
     // A wrapper for getDependencies() errors. This adds additional information
     // about the container to identify it for diagnostics.
     public struct GetDependenciesError: Error, CustomStringConvertible {
-
         /// The repository  that encountered the error.
         public let repository: RepositorySpecifier
 
@@ -53,15 +56,14 @@ internal final class SourceControlPackageContainer: PackageContainer, CustomStri
     private let repository: Repository
     private let identityResolver: IdentityResolver
     private let manifestLoader: ManifestLoaderProtocol
-    private let toolsVersionLoader: ToolsVersionLoaderProtocol
     private let currentToolsVersion: ToolsVersion
     private let fingerprintStorage: PackageFingerprintStorage?
     private let fingerprintCheckingMode: FingerprintCheckingMode
     private let observabilityScope: ObservabilityScope
 
     /// The cached dependency information.
-    private var dependenciesCache = [String: [ProductFilter: (Manifest, [Constraint])]] ()
-    private var dependenciesCacheLock = Lock()
+    private var dependenciesCache = [String: [ProductFilter: (Manifest, [Constraint])]]()
+    private var dependenciesCacheLock = NSLock()
 
     private var knownVersionsCache = ThreadSafeBox<[Version: String]>()
     private var manifestsCache = ThreadSafeKeyValueStore<String, Manifest>()
@@ -77,7 +79,6 @@ internal final class SourceControlPackageContainer: PackageContainer, CustomStri
         repositorySpecifier: RepositorySpecifier,
         repository: Repository,
         manifestLoader: ManifestLoaderProtocol,
-        toolsVersionLoader: ToolsVersionLoaderProtocol,
         currentToolsVersion: ToolsVersion,
         fingerprintStorage: PackageFingerprintStorage?,
         fingerprintCheckingMode: FingerprintCheckingMode,
@@ -88,19 +89,20 @@ internal final class SourceControlPackageContainer: PackageContainer, CustomStri
         self.repositorySpecifier = repositorySpecifier
         self.repository = repository
         self.manifestLoader = manifestLoader
-        self.toolsVersionLoader = toolsVersionLoader
         self.currentToolsVersion = currentToolsVersion
         self.fingerprintStorage = fingerprintStorage
         self.fingerprintCheckingMode = fingerprintCheckingMode
-        self.observabilityScope = observabilityScope
+        self.observabilityScope = observabilityScope.makeChildScope(
+            description: "SourceControlPackageContainer",
+            metadata: package.diagnosticsMetadata)
     }
 
     // Compute the map of known versions.
     private func knownVersions() throws -> [Version: String] {
-        try self.knownVersionsCache.memoize() {
-            let knownVersionsWithDuplicates = Git.convertTagsToVersionMap(try repository.getTags())
+        try self.knownVersionsCache.memoize {
+            let knownVersionsWithDuplicates = Git.convertTagsToVersionMap(tags: try repository.getTags(), toolsVersion: self.currentToolsVersion)
 
-            return knownVersionsWithDuplicates.mapValues({ tags -> String in
+            return knownVersionsWithDuplicates.mapValues { tags -> String in
                 if tags.count > 1 {
                     // FIXME: Warn if the two tags point to different git references.
 
@@ -110,9 +112,9 @@ internal final class SourceControlPackageContainer: PackageContainer, CustomStri
                     let tagsSortedBySpecificity = tags.sorted {
                         let componentCounts = ($0.components(separatedBy: ".").count, $1.components(separatedBy: ".").count)
                         if componentCounts.0 == componentCounts.1 {
-                            //if they are both have the same number of components, favor the one without a v prefix.
-                            //this matches previously defined behavior
-                            //this assumes we can only enter this situation because one tag has a v prefix and the other does not.
+                            // if they are both have the same number of components, favor the one without a v prefix.
+                            // this matches previously defined behavior
+                            // this assumes we can only enter this situation because one tag has a v prefix and the other does not.
                             return $0.hasPrefix("v")
                         }
                         return componentCounts.0 < componentCounts.1
@@ -121,7 +123,7 @@ internal final class SourceControlPackageContainer: PackageContainer, CustomStri
                 }
                 assert(tags.count == 1, "Unexpected number of tags")
                 return tags[0]
-            })
+            }
         }
     }
 
@@ -132,7 +134,7 @@ internal final class SourceControlPackageContainer: PackageContainer, CustomStri
     /// The available version list (in reverse order).
     public func toolsVersionsAppropriateVersionsDescending() throws -> [Version] {
         let reversedVersions = try self.versionsDescending()
-        return reversedVersions.lazy.filter({
+        return reversedVersions.lazy.filter {
             // If we have the result cached, return that.
             if let result = self.validToolsVersionsCache[$0] {
                 return result
@@ -142,7 +144,7 @@ internal final class SourceControlPackageContainer: PackageContainer, CustomStri
             let isValid = (try? self.toolsVersion(for: $0)).flatMap(self.isValidToolsVersion(_:)) ?? false
             self.validToolsVersionsCache[$0] = isValid
             return isValid
-        })
+        }
     }
 
     public func getTag(for version: Version) -> String? {
@@ -157,13 +159,12 @@ internal final class SourceControlPackageContainer: PackageContainer, CustomStri
         guard case .remoteSourceControl(let sourceControlURL) = self.package.kind else {
             return
         }
-        
-        let packageIdentity = self.package.identity
+
         let fingerprint: Fingerprint
         do {
             fingerprint = try temp_await {
                 fingerprintStorage.get(
-                    package: packageIdentity,
+                    package: self.package,
                     version: version,
                     kind: .sourceControl,
                     observabilityScope: self.observabilityScope,
@@ -177,7 +178,7 @@ internal final class SourceControlPackageContainer: PackageContainer, CustomStri
             do {
                 try temp_await {
                     fingerprintStorage.put(
-                        package: packageIdentity,
+                        package: self.package,
                         version: version,
                         fingerprint: fingerprint,
                         observabilityScope: self.observabilityScope,
@@ -198,7 +199,7 @@ internal final class SourceControlPackageContainer: PackageContainer, CustomStri
             self.observabilityScope.emit(error: "Failed to get source control fingerprint for \(self.package) version \(version) from storage: \(error)")
             throw error
         }
-        
+
         // The revision (i.e., hash) must match that in fingerprint storage otherwise the integrity check fails
         if revision.identifier != fingerprint.value {
             let message = "Revision \(revision.identifier) for \(self.package) version \(version) does not match previously recorded value \(fingerprint.value)"
@@ -228,7 +229,9 @@ internal final class SourceControlPackageContainer: PackageContainer, CustomStri
                 throw StringError("unknown tag \(version)")
             }
             let fileSystem = try repository.openFileView(tag: tag)
-            return try toolsVersionLoader.load(at: .root, fileSystem: fileSystem)
+            // find the manifest path and parse it's tools-version
+            let manifestPath = try ManifestLoader.findManifest(packagePath: .root, fileSystem: fileSystem, currentToolsVersion: self.currentToolsVersion)
+            return try ToolsVersionParser.parse(manifestPath: manifestPath, fileSystem: fileSystem)
         }
     }
 
@@ -269,11 +272,10 @@ internal final class SourceControlPackageContainer: PackageContainer, CustomStri
                         underlyingError: error,
                         suggestion: .none
                     )
-                }
-                else {
+                } else {
                     // Revision does not exist, so we customize the error.
                     let sha1RegEx = try! RegEx(pattern: #"\A[:xdigit:]{40}\Z"#)
-                    let isBranchRev = sha1RegEx.matchGroups(in: revision).compactMap{ $0 }.isEmpty
+                    let isBranchRev = sha1RegEx.matchGroups(in: revision).compactMap { $0 }.isEmpty
                     let errorMessage = "could not find " + (isBranchRev ? "a branch named ‘\(revision)’" : "the commit \(revision)")
                     let mainBranchExists = (try? repository.resolveRevision(identifier: "main")) != nil
                     let suggestion = (revision == "master" && mainBranchExists) ? "did you mean ‘main’?" : nil
@@ -374,40 +376,35 @@ internal final class SourceControlPackageContainer: PackageContainer, CustomStri
     private func loadManifest(tag: String, version: Version?) throws -> Manifest {
         try self.manifestsCache.memoize(tag) {
             let fileSystem = try repository.openFileView(tag: tag)
-            return try self.loadManifest(fileSystem: fileSystem, version: version, packageVersion: tag)
+            return try self.loadManifest(fileSystem: fileSystem, version: version, revision: tag)
         }
     }
 
     private func loadManifest(at revision: Revision, version: Version?) throws -> Manifest {
         try self.manifestsCache.memoize(revision.identifier) {
             let fileSystem = try self.repository.openFileView(revision: revision)
-            return try self.loadManifest(fileSystem: fileSystem, version: version, packageVersion: revision.identifier)
+            return try self.loadManifest(fileSystem: fileSystem, version: version, revision: revision.identifier)
         }
     }
 
-    private func loadManifest(fileSystem: FileSystem, version: Version?, packageVersion: String) throws -> Manifest {
-        // Load the tools version.
-        let toolsVersion = try self.toolsVersionLoader.load(at: .root, fileSystem: fileSystem)
-
-        // Validate the tools version.
-        try toolsVersion.validateToolsVersion(
-            self.currentToolsVersion, packageIdentity: self.package.identity, packageVersion: packageVersion)
-
+    private func loadManifest(fileSystem: FileSystem, version: Version?, revision: String) throws -> Manifest {
         // Load the manifest.
         // FIXME: this should not block
         return try temp_await {
-            self.manifestLoader.load(at: AbsolutePath.root,
-                                     packageIdentity: self.package.identity,
-                                     packageKind: self.package.kind,
-                                     packageLocation: self.package.locationString,
-                                     version: version,
-                                     revision: nil,
-                                     toolsVersion: toolsVersion,
-                                     identityResolver: self.identityResolver,
-                                     fileSystem: fileSystem,
-                                     observabilityScope: self.observabilityScope,
-                                     on: .sharedConcurrent,
-                                     completion: $0)
+            self.manifestLoader.load(
+                packagePath: .root,
+                packageIdentity: self.package.identity,
+                packageKind: self.package.kind,
+                packageLocation: self.package.locationString,
+                packageVersion: (version: version, revision: revision),
+                currentToolsVersion: self.currentToolsVersion,
+                identityResolver: self.identityResolver,
+                fileSystem: fileSystem,
+                observabilityScope: self.observabilityScope,
+                delegateQueue: .sharedConcurrent,
+                callbackQueue: .sharedConcurrent,
+                completion: $0
+            )
         }
     }
 
@@ -420,14 +417,14 @@ internal final class SourceControlPackageContainer: PackageContainer, CustomStri
     }
 }
 
-fileprivate extension Git {
-    static func convertTagsToVersionMap(_ tags: [String]) -> [Version: [String]] {
+extension Git {
+    fileprivate static func convertTagsToVersionMap(tags: [String], toolsVersion: ToolsVersion) -> [Version: [String]] {
         // First, check if we need to restrict the tag set to version-specific tags.
         var knownVersions: [Version: [String]] = [:]
         var versionSpecificKnownVersions: [Version: [String]] = [:]
 
         for tag in tags {
-            for versionSpecificKey in SwiftVersion.currentVersion.versionSpecificKeys {
+            for versionSpecificKey in toolsVersion.versionSpecificKeys {
                 if tag.hasSuffix(versionSpecificKey) {
                     let trimmedTag = String(tag.dropLast(versionSpecificKey.count))
                     if let version = Version(tag: trimmedTag) {
