@@ -1,44 +1,39 @@
-/*
- This source file is part of the Swift.org open source project
-
- Copyright (c) 2014 - 2021 Apple Inc. and the Swift project authors
- Licensed under Apache License v2.0 with Runtime Library Exception
-
- See http://swift.org/LICENSE.txt for license information
- See http://swift.org/CONTRIBUTORS.txt for Swift project authors
-*/
+//===----------------------------------------------------------------------===//
+//
+// This source file is part of the Swift open source project
+//
+// Copyright (c) 2014-2021 Apple Inc. and the Swift project authors
+// Licensed under Apache License v2.0 with Runtime Library Exception
+//
+// See http://swift.org/LICENSE.txt for license information
+// See http://swift.org/CONTRIBUTORS.txt for the list of Swift project authors
+//
+//===----------------------------------------------------------------------===//
 
 import Foundation
 
 import TSCBasic
-import TSCUtility
 
 import Basics
 import PackageModel
 import PackageLoading
 import PackageGraph
+import SPMBuildCore
 
 /// The parameters required by `PIFBuilder`.
-public struct PIFBuilderParameters {
+struct PIFBuilderParameters {
 
     /// Whether or not build for testability is enabled.
-    public let enableTestability: Bool
+    let enableTestability: Bool
 
     /// Whether to create dylibs for dynamic library products.
-    public let shouldCreateDylibForDynamicProducts: Bool
+    let shouldCreateDylibForDynamicProducts: Bool
 
     /// The path to the library directory of the active toolchain.
-    public let toolchainLibDir: AbsolutePath
+    let toolchainLibDir: AbsolutePath
 
-    /// Creates a `PIFBuilderParameters` instance.
-    /// - Parameters:
-    ///   - enableTestability: Whether or not build for testability is enabled.
-    ///   - shouldCreateDylibForDynamicProducts: Whether to create dylibs for dynamic library products.
-    public init(enableTestability: Bool, shouldCreateDylibForDynamicProducts: Bool, toolchainLibDir: AbsolutePath) {
-        self.enableTestability = enableTestability
-        self.shouldCreateDylibForDynamicProducts = shouldCreateDylibForDynamicProducts
-        self.toolchainLibDir = toolchainLibDir
-    }
+    /// An array of paths to search for pkg-config `.pc` files.
+    let pkgConfigDirectories: [AbsolutePath]
 }
 
 /// PIF object builder for a package graph.
@@ -51,16 +46,16 @@ public final class PIFBuilder {
     public static let allIncludingTestsTargetName = "AllIncludingTests"
 
     /// The package graph to build from.
-    public let graph: PackageGraph
+    let graph: PackageGraph
 
     /// The parameters used to configure the PIF.
-    public let parameters: PIFBuilderParameters
+    let parameters: PIFBuilderParameters
 
     /// The ObservabilityScope to emit diagnostics to.
-    public let observabilityScope: ObservabilityScope
+    let observabilityScope: ObservabilityScope
 
     /// The file system to read from.
-    public let fileSystem: FileSystem
+    let fileSystem: FileSystem
 
     private var pif: PIF.TopLevelObject?
 
@@ -70,7 +65,7 @@ public final class PIFBuilder {
     ///   - parameters: The parameters used to configure the PIF.
     ///   - fileSystem: The file system to read from.
     ///   - observabilityScope: The ObservabilityScope to emit diagnostics to.
-    public init(
+    init(
         graph: PackageGraph,
         parameters: PIFBuilderParameters,
         fileSystem: FileSystem,
@@ -86,7 +81,7 @@ public final class PIFBuilder {
     /// - Parameters:
     ///   - prettyPrint: Whether to return a formatted JSON.
     /// - Returns: The package graph in the JSON PIF format.
-    public func generatePIF(
+    func generatePIF(
         prettyPrint: Bool = true,
         preservePIFModelStructure: Bool = false
     ) throws -> String {
@@ -132,6 +127,18 @@ public final class PIFBuilder {
             return PIF.TopLevelObject(workspace: workspace)
         }
     }
+
+    // Convenience method for generating PIF.
+    public static func generatePIF(buildParameters: BuildParameters, packageGraph: PackageGraph, fileSystem: FileSystem, observabilityScope: ObservabilityScope, preservePIFModelStructure: Bool) throws -> String {
+        let parameters = PIFBuilderParameters(buildParameters)
+        let builder = Self.init(
+            graph: packageGraph,
+            parameters: parameters,
+            fileSystem: fileSystem,
+            observabilityScope: observabilityScope
+        )
+        return try builder.generatePIF(preservePIFModelStructure: preservePIFModelStructure)
+    }
 }
 
 class PIFProjectBuilder {
@@ -161,9 +168,10 @@ class PIFProjectBuilder {
     @discardableResult
     func addBuildConfiguration(
         name: String,
-        settings: PIF.BuildSettings = PIF.BuildSettings()
+        settings: PIF.BuildSettings = PIF.BuildSettings(),
+        impartedBuildProperties: PIF.ImpartedBuildProperties = PIF.ImpartedBuildProperties(settings: PIF.BuildSettings())
     ) -> PIFBuildConfigurationBuilder {
-        let builder = PIFBuildConfigurationBuilder(name: name, settings: settings)
+        let builder = PIFBuildConfigurationBuilder(name: name, settings: settings, impartedBuildProperties: impartedBuildProperties)
         buildConfigurations.append(builder)
         return builder
     }
@@ -239,7 +247,7 @@ final class PackagePIFProjectBuilder: PIFProjectBuilder {
             metadata: package.underlyingPackage.diagnosticsMetadata
         )
 
-        executableTargetProductMap = Dictionary(uniqueKeysWithValues:
+        executableTargetProductMap = try Dictionary(throwingUniqueKeysWithValues:
             package.products.filter { $0.type == .executable }.map { ($0.mainTarget, $0) }
         )
 
@@ -260,13 +268,12 @@ final class PackagePIFProjectBuilder: PIFProjectBuilder {
         settings[.SDKROOT] = "auto"
         settings[.SDK_VARIANT] = "auto"
         settings[.SKIP_INSTALL] = "YES"
-        let firstTarget = package.targets.first(where: { $0.type != .test })?.underlyingTarget ?? package.targets.first?.underlyingTarget
-        settings[.MACOSX_DEPLOYMENT_TARGET] = firstTarget?.deploymentTarget(for: .macOS)
-        settings[.IPHONEOS_DEPLOYMENT_TARGET] = firstTarget?.deploymentTarget(for: .iOS)
-        settings[.IPHONEOS_DEPLOYMENT_TARGET, for: .macCatalyst] = firstTarget?.deploymentTarget(for: .macCatalyst)
-        settings[.TVOS_DEPLOYMENT_TARGET] = firstTarget?.deploymentTarget(for: .tvOS)
-        settings[.WATCHOS_DEPLOYMENT_TARGET] = firstTarget?.deploymentTarget(for: .watchOS)
-        settings[.DRIVERKIT_DEPLOYMENT_TARGET] = firstTarget?.deploymentTarget(for: .driverKit)
+        settings[.MACOSX_DEPLOYMENT_TARGET] = package.platforms.deploymentTarget(for: .macOS)
+        settings[.IPHONEOS_DEPLOYMENT_TARGET] = package.platforms.deploymentTarget(for: .iOS)
+        settings[.IPHONEOS_DEPLOYMENT_TARGET, for: .macCatalyst] = package.platforms.deploymentTarget(for: .macCatalyst)
+        settings[.TVOS_DEPLOYMENT_TARGET] = package.platforms.deploymentTarget(for: .tvOS)
+        settings[.WATCHOS_DEPLOYMENT_TARGET] = package.platforms.deploymentTarget(for: .watchOS)
+        settings[.DRIVERKIT_DEPLOYMENT_TARGET] = package.platforms.deploymentTarget(for: .driverKit)
         settings[.DYLIB_INSTALL_NAME_BASE] = "@rpath"
         settings[.USE_HEADERMAP] = "NO"
         settings[.SWIFT_ACTIVE_COMPILATION_CONDITIONS] = ["$(inherited)", "SWIFT_PACKAGE"]
@@ -290,7 +297,7 @@ final class PackagePIFProjectBuilder: PIFProjectBuilder {
 
         PlatformRegistry.default.knownPlatforms.forEach {
             guard let platform = PIF.BuildSettings.Platform.from(platform: $0) else { return }
-            guard let supportedPlatform = firstTarget?.getSupportedPlatform(for: $0) else { return }
+            guard let supportedPlatform = package.platforms.getDerived(for: $0) else { return }
             if !supportedPlatform.options.isEmpty {
                 settings[.SPECIALIZATION_SDK_OPTIONS, for: platform] = supportedPlatform.options
             }
@@ -327,7 +334,7 @@ final class PackagePIFProjectBuilder: PIFProjectBuilder {
         addBuildConfiguration(name: "Release", settings: releaseSettings)
 
         for product in package.products.sorted(by: { $0.name < $1.name }) {
-            addTarget(for: product)
+            try addTarget(for: product)
         }
 
         for target in package.targets.sorted(by: { $0.name < $1.name }) {
@@ -339,10 +346,10 @@ final class PackagePIFProjectBuilder: PIFProjectBuilder {
         }
     }
 
-    private func addTarget(for product: ResolvedProduct) {
+    private func addTarget(for product: ResolvedProduct) throws {
         switch product.type {
         case .executable, .snippet, .test:
-            addMainModuleTarget(for: product)
+            try addMainModuleTarget(for: product)
         case .library:
             addLibraryTarget(for: product)
         case .plugin:
@@ -377,7 +384,7 @@ final class PackagePIFProjectBuilder: PIFProjectBuilder {
         return "\(productName)_\(String(productName.hash, radix: 16, uppercase: true))_PackageProduct"
     }
 
-    private func addMainModuleTarget(for product: ResolvedProduct) {
+    private func addMainModuleTarget(for product: ResolvedProduct) throws {
         let productType: PIF.Target.ProductType = product.type == .executable ? .executable : .unitTest
         let pifTarget = addTarget(
             guid: product.pifTargetGUID,
@@ -416,10 +423,10 @@ final class PackagePIFProjectBuilder: PIFProjectBuilder {
 
         // Tests can have a custom deployment target based on the minimum supported by XCTest.
         if mainTarget.underlyingTarget.type == .test {
-            settings[.MACOSX_DEPLOYMENT_TARGET] = mainTarget.underlyingTarget.deploymentTarget(for: .macOS)
-            settings[.IPHONEOS_DEPLOYMENT_TARGET] = mainTarget.underlyingTarget.deploymentTarget(for: .iOS)
-            settings[.TVOS_DEPLOYMENT_TARGET] = mainTarget.underlyingTarget.deploymentTarget(for: .tvOS)
-            settings[.WATCHOS_DEPLOYMENT_TARGET] = mainTarget.underlyingTarget.deploymentTarget(for: .watchOS)
+            settings[.MACOSX_DEPLOYMENT_TARGET] = mainTarget.platforms.deploymentTarget(for: .macOS)
+            settings[.IPHONEOS_DEPLOYMENT_TARGET] = mainTarget.platforms.deploymentTarget(for: .iOS)
+            settings[.TVOS_DEPLOYMENT_TARGET] = mainTarget.platforms.deploymentTarget(for: .tvOS)
+            settings[.WATCHOS_DEPLOYMENT_TARGET] = mainTarget.platforms.deploymentTarget(for: .watchOS)
         }
 
         if product.type == .executable {
@@ -460,15 +467,16 @@ final class PackagePIFProjectBuilder: PIFProjectBuilder {
         var releaseSettings = settings
 
         var impartedSettings = PIF.BuildSettings()
-        addManifestBuildSettings(
+        try addManifestBuildSettings(
             from: mainTarget.underlyingTarget,
             debugSettings: &debugSettings,
             releaseSettings: &releaseSettings,
             impartedSettings: &impartedSettings
         )
 
-        pifTarget.addBuildConfiguration(name: "Debug", settings: debugSettings)
-        pifTarget.addBuildConfiguration(name: "Release", settings: releaseSettings)
+        let impartedBuildProperties = PIF.ImpartedBuildProperties(settings: impartedSettings)
+        pifTarget.addBuildConfiguration(name: "Debug", settings: debugSettings, impartedBuildProperties: impartedBuildProperties)
+        pifTarget.addBuildConfiguration(name: "Release", settings: releaseSettings, impartedBuildProperties: impartedBuildProperties)
     }
 
     private func addLibraryTarget(for product: ResolvedProduct) {
@@ -583,6 +591,9 @@ final class PackagePIFProjectBuilder: PIFProjectBuilder {
         // Disable code coverage linker flags since we're producing .o files. Otherwise, we will run into duplicated
         // symbols when there are more than one targets that produce .o as their product.
         settings[.CLANG_COVERAGE_MAPPING_LINKER_ARGS] = "NO"
+        if let aliases = target.moduleAliases {
+            settings[.SWIFT_MODULE_ALIASES] = aliases.map{ $0.key + "=" + $0.value }
+        }
 
         // Create a set of build settings that will be imparted to any target that depends on this one.
         var impartedSettings = PIF.BuildSettings()
@@ -668,15 +679,16 @@ final class PackagePIFProjectBuilder: PIFProjectBuilder {
         var debugSettings = settings
         var releaseSettings = settings
 
-        addManifestBuildSettings(
+        try addManifestBuildSettings(
             from: target.underlyingTarget,
             debugSettings: &debugSettings,
             releaseSettings: &releaseSettings,
             impartedSettings: &impartedSettings
         )
 
-        pifTarget.addBuildConfiguration(name: "Debug", settings: debugSettings)
-        pifTarget.addBuildConfiguration(name: "Release", settings: releaseSettings)
+        let impartedBuildProperties = PIF.ImpartedBuildProperties(settings: impartedSettings)
+        pifTarget.addBuildConfiguration(name: "Debug", settings: debugSettings, impartedBuildProperties: impartedBuildProperties)
+        pifTarget.addBuildConfiguration(name: "Release", settings: releaseSettings, impartedBuildProperties: impartedBuildProperties)
         pifTarget.impartedBuildSettings = impartedSettings
     }
 
@@ -685,16 +697,16 @@ final class PackagePIFProjectBuilder: PIFProjectBuilder {
             throw InternalError("unexpected target type")
         }
 
-        // Create an aggregate PIF target (which doesn't have an actual product).
-        let pifTarget = addAggregateTarget(guid: target.pifTargetGUID, name: target.name)
-        pifTarget.addBuildConfiguration(name: "Debug", settings: PIF.BuildSettings())
-        pifTarget.addBuildConfiguration(name: "Release", settings: PIF.BuildSettings())
-
         // Impart the header search path to all direct and indirect clients.
         var impartedSettings = PIF.BuildSettings()
 
         var cFlags: [String] = []
-        for result in pkgConfigArgs(for: systemTarget, fileSystem: fileSystem, observabilityScope: self.observabilityScope) {
+        for result in try pkgConfigArgs(
+            for: systemTarget,
+            pkgConfigDirectories: parameters.pkgConfigDirectories,
+            fileSystem: fileSystem,
+            observabilityScope: observabilityScope
+        ) {
             if let error = result.error {
                 self.observabilityScope.emit(
                     warning: "\(error)",
@@ -709,6 +721,12 @@ final class PackagePIFProjectBuilder: PIFProjectBuilder {
         impartedSettings[.OTHER_LDRFLAGS] = []
         impartedSettings[.OTHER_CFLAGS, default: ["$(inherited)"]] += ["-fmodule-map-file=\(systemTarget.moduleMapPath)"] + cFlags
         impartedSettings[.OTHER_SWIFT_FLAGS, default: ["$(inherited)"]] += ["-Xcc", "-fmodule-map-file=\(systemTarget.moduleMapPath)"] + cFlags
+        let impartedBuildProperties = PIF.ImpartedBuildProperties(settings: impartedSettings)
+
+        // Create an aggregate PIF target (which doesn't have an actual product).
+        let pifTarget = addAggregateTarget(guid: target.pifTargetGUID, name: target.name)
+        pifTarget.addBuildConfiguration(name: "Debug", settings: PIF.BuildSettings(), impartedBuildProperties: impartedBuildProperties)
+        pifTarget.addBuildConfiguration(name: "Release", settings: PIF.BuildSettings(), impartedBuildProperties: impartedBuildProperties)
         pifTarget.impartedBuildSettings = impartedSettings
     }
 
@@ -832,6 +850,9 @@ final class PackagePIFProjectBuilder: PIFProjectBuilder {
             resourcesTarget.addResourceFile(resourceFile)
         }
 
+        let targetGroup = groupTree.addGroup(path: "/", sourceTree: .group)
+        pifTarget.addResourceFile(targetGroup.addFileReference(path: "\(bundleName).bundle", sourceTree: .builtProductsDir))
+
         return bundleName
     }
 
@@ -856,12 +877,12 @@ final class PackagePIFProjectBuilder: PIFProjectBuilder {
         debugSettings: inout PIF.BuildSettings,
         releaseSettings: inout PIF.BuildSettings,
         impartedSettings: inout PIF.BuildSettings
-    ) {
+    ) throws {
         for (setting, assignments) in target.buildSettings.pifAssignments {
             for assignment in assignments {
                 var value = assignment.value
                 if setting == .HEADER_SEARCH_PATHS {
-                    value = value.map { AbsolutePath($0, relativeTo: target.sources.root).pathString }
+                    value = try value.map { try AbsolutePath(validating: $0, relativeTo: target.sources.root).pathString }
                 }
 
                 if let platforms = assignment.platforms {
@@ -1066,9 +1087,10 @@ class PIFBaseTargetBuilder {
     @discardableResult
     public func addBuildConfiguration(
         name: String,
-        settings: PIF.BuildSettings = PIF.BuildSettings()
+        settings: PIF.BuildSettings = PIF.BuildSettings(),
+        impartedBuildProperties: PIF.ImpartedBuildProperties = PIF.ImpartedBuildProperties(settings: PIF.BuildSettings())
     ) -> PIFBuildConfigurationBuilder {
-        let builder = PIFBuildConfigurationBuilder(name: name, settings: settings)
+        let builder = PIFBuildConfigurationBuilder(name: name, settings: settings, impartedBuildProperties: impartedBuildProperties)
         buildConfigurations.append(builder)
         return builder
     }
@@ -1312,18 +1334,20 @@ final class PIFBuildFileBuilder {
 final class PIFBuildConfigurationBuilder {
     let name: String
     let settings: PIF.BuildSettings
+    let impartedBuildProperties: PIF.ImpartedBuildProperties
 
     @DelayedImmutable
     var guid: PIF.GUID
 
-    public init(name: String, settings: PIF.BuildSettings) {
+    public init(name: String, settings: PIF.BuildSettings, impartedBuildProperties: PIF.ImpartedBuildProperties) {
         precondition(!name.isEmpty)
         self.name = name
         self.settings = settings
+        self.impartedBuildProperties = impartedBuildProperties
     }
 
     func construct() -> PIF.BuildConfiguration {
-        PIF.BuildConfiguration(guid: guid, name: name, buildSettings: settings)
+        PIF.BuildConfiguration(guid: guid, name: name, buildSettings: settings, impartedBuildProperties: impartedBuildProperties)
     }
 }
 
@@ -1378,9 +1402,9 @@ extension Array where Element == ResolvedTarget.Dependency {
     }
 }
 
-extension Target {
+extension SupportedPlatforms {
     func deploymentTarget(for platform: PackageModel.Platform) -> String? {
-        if let supportedPlatform = getSupportedPlatform(for: platform) {
+        if let supportedPlatform = self.getDerived(for: platform) {
             return supportedPlatform.version.versionString
         } else if platform == .macCatalyst {
             // If there is no deployment target specified for Mac Catalyst, fall back to the iOS deployment target.
@@ -1391,14 +1415,12 @@ extension Target {
             return nil
         }
     }
+}
 
+extension Target {
     var isCxx: Bool {
         (self as? ClangTarget)?.isCXX ?? false
     }
-
-    var usesUnsafeFlags: Bool {
-        Set(buildSettings.assignments.keys).contains(where: BuildSettings.Declaration.unsafeSettings.contains)
-    }
 }
 
 extension ProductType {
@@ -1441,16 +1463,16 @@ private extension BuildSettings.AssignmentTable {
                 switch declaration {
                 case .LINK_LIBRARIES:
                     setting = .OTHER_LDFLAGS
-                    value = assignment.value.map { "-l\($0)" }
+                    value = assignment.values.map { "-l\($0)" }
                 case .LINK_FRAMEWORKS:
                     setting = .OTHER_LDFLAGS
-                    value = assignment.value.flatMap { ["-framework", $0] }
+                    value = assignment.values.flatMap { ["-framework", $0] }
                 default:
                     guard let parsedSetting = PIF.BuildSettings.MultipleValueSetting(rawValue: declaration.name) else {
                         continue
                     }
                     setting = parsedSetting
-                    value = assignment.value
+                    value = assignment.values
                 }
 
                 let pifAssignment = PIFBuildSettingAssignment(
@@ -1589,7 +1611,7 @@ extension PIF.PlatformFilter {
         .init(platform: "windows", environment: "gnu"),
     ]
 
-    /// Andriod platform filters.
+    /// Android platform filters.
     public static let androidFilters: [PIF.PlatformFilter] = [
         .init(platform: "linux", environment: "android"),
         .init(platform: "linux", environment: "androideabi"),
