@@ -1,21 +1,22 @@
-/*
- This source file is part of the Swift.org open source project
+//===----------------------------------------------------------------------===//
+//
+// This source file is part of the Swift open source project
+//
+// Copyright (c) 2014-2021 Apple Inc. and the Swift project authors
+// Licensed under Apache License v2.0 with Runtime Library Exception
+//
+// See http://swift.org/LICENSE.txt for license information
+// See http://swift.org/CONTRIBUTORS.txt for the list of Swift project authors
+//
+//===----------------------------------------------------------------------===//
 
- Copyright (c) 2014 - 2021 Apple Inc. and the Swift project authors
- Licensed under Apache License v2.0 with Runtime Library Exception
-
- See http://swift.org/LICENSE.txt for license information
- See http://swift.org/CONTRIBUTORS.txt for Swift project authors
-*/
-
-import Commands
+@testable import Commands
 import PackageGraph
 import PackageLoading
 import PackageModel
 import SPMBuildCore
 import SPMTestSupport
 import TSCBasic
-import TSCUtility
 import Workspace
 import XCTest
 
@@ -39,7 +40,7 @@ final class BuildToolTests: CommandsTestCase {
         let (output, _) = try execute(args, packagePath: packagePath)
         defer { try! SwiftPMProduct.SwiftPackage.execute(["clean"], packagePath: packagePath) }
         let (binPathOutput, _) = try execute(["--show-bin-path"], packagePath: packagePath)
-        let binPath = AbsolutePath(binPathOutput.trimmingCharacters(in: .whitespacesAndNewlines))
+        let binPath = try AbsolutePath(validating: binPathOutput.trimmingCharacters(in: .whitespacesAndNewlines))
         let binContents = try localFileSystem.getDirectoryContents(binPath)
         return BuildResult(binPath: binPath, output: output, binContents: binContents)
     }
@@ -74,10 +75,53 @@ final class BuildToolTests: CommandsTestCase {
         }
     }
 
+    func testImportOfMissedDepWarning() throws {
+        #if swift(<5.5)
+        try XCTSkipIf(true, "skipping because host compiler doesn't support '-import-prescan'")
+        #endif
+        // Verify the warning flow
+        try fixture(name: "Miscellaneous/ImportOfMissingDependency") { path in
+            let fullPath = try resolveSymlinks(path)
+            XCTAssertThrowsError(try build(["--explicit-target-dependency-import-check=warn"], packagePath: fullPath)) { error in
+                guard case SwiftPMProductError.executionFailure(_, _, let stderr) = error else {
+                    XCTFail()
+                    return
+                }
+
+                XCTAssertTrue(stderr.contains("warning: Target A imports another target (B) in the package without declaring it a dependency."))
+            }
+        }
+
+        // Verify the error flow
+        try fixture(name: "Miscellaneous/ImportOfMissingDependency") { path in
+            let fullPath = try resolveSymlinks(path)
+            XCTAssertThrowsError(try build(["--explicit-target-dependency-import-check=error"], packagePath: fullPath)) { error in
+                guard case SwiftPMProductError.executionFailure(_, _, let stderr) = error else {
+                    XCTFail()
+                    return
+                }
+
+                XCTAssertTrue(stderr.contains("error: Target A imports another target (B) in the package without declaring it a dependency."))
+            }
+        }
+
+        // Verify that the default does not run the check
+        try fixture(name: "Miscellaneous/ImportOfMissingDependency") { path in
+            let fullPath = try resolveSymlinks(path)
+            XCTAssertThrowsError(try build([], packagePath: fullPath)) { error in
+                guard case SwiftPMProductError.executionFailure(_, _, let stderr) = error else {
+                    XCTFail()
+                    return
+                }
+                XCTAssertFalse(stderr.contains("warning: Target A imports another target (B) in the package without declaring it a dependency."))
+            }
+        }
+    }
+
     func testBinPathAndSymlink() throws {
-        fixture(name: "ValidLayouts/SingleModule/ExecutableNew") { path in
-            let fullPath = resolveSymlinks(path)
-            let targetPath = fullPath.appending(components: ".build", UserToolchain.default.triple.platformBuildPathComponent())
+        try fixture(name: "ValidLayouts/SingleModule/ExecutableNew") { fixturePath in
+            let fullPath = try resolveSymlinks(fixturePath)
+            let targetPath = fullPath.appending(components: ".build", try UserToolchain.default.triple.platformBuildPathComponent())
             let xcbuildTargetPath = fullPath.appending(components: ".build", "apple")
             XCTAssertEqual(try execute(["--show-bin-path"], packagePath: fullPath).stdout,
                            "\(targetPath.appending(component: "debug").pathString)\n")
@@ -97,24 +141,22 @@ final class BuildToolTests: CommandsTestCase {
 
             // Test symlink.
             _ = try execute([], packagePath: fullPath)
-            XCTAssertEqual(resolveSymlinks(fullPath.appending(components: ".build", "debug")),
+            XCTAssertEqual(try resolveSymlinks(fullPath.appending(components: ".build", "debug")),
                            targetPath.appending(component: "debug"))
             _ = try execute(["-c", "release"], packagePath: fullPath)
-            XCTAssertEqual(resolveSymlinks(fullPath.appending(components: ".build", "release")),
+            XCTAssertEqual(try resolveSymlinks(fullPath.appending(components: ".build", "release")),
                            targetPath.appending(component: "release"))
         }
     }
 
     func testProductAndTarget() throws {
-        fixture(name: "Miscellaneous/MultipleExecutables") { path in
-            let fullPath = resolveSymlinks(path)
+        try fixture(name: "Miscellaneous/MultipleExecutables") { fixturePath in
+            let fullPath = try resolveSymlinks(fixturePath)
 
             do {
                 let result = try build(["--product", "exec1"], packagePath: fullPath)
                 XCTAssertMatch(result.binContents, ["exec1"])
                 XCTAssertNoMatch(result.binContents, ["exec2.build"])
-            } catch SwiftPMProductError.executionFailure(_, _, let stderr) {
-                XCTFail(stderr)
             }
 
             do {
@@ -127,83 +169,58 @@ final class BuildToolTests: CommandsTestCase {
                 let result = try build(["--target", "exec2"], packagePath: fullPath)
                 XCTAssertMatch(result.binContents, ["exec2.build"])
                 XCTAssertNoMatch(result.binContents, ["exec1"])
-            } catch SwiftPMProductError.executionFailure(_, _, let stderr) {
-                XCTFail(stderr)
             }
 
-            do {
-                _ = try execute(["--product", "exec1", "--target", "exec2"], packagePath: path)
-                XCTFail("Expected to fail")
-            } catch SwiftPMProductError.executionFailure(_, _, let stderr) {
-                XCTAssertMatch(stderr, .contains("error: '--product' and '--target' are mutually exclusive"))
+            XCTAssertThrowsCommandExecutionError(try execute(["--product", "exec1", "--target", "exec2"], packagePath: fixturePath)) { error in
+                XCTAssertMatch(error.stderr, .contains("error: '--product' and '--target' are mutually exclusive"))
             }
 
-            do {
-                _ = try execute(["--product", "exec1", "--build-tests"], packagePath: path)
-                XCTFail("Expected to fail")
-            } catch SwiftPMProductError.executionFailure(_, _, let stderr) {
-                XCTAssertMatch(stderr, .contains("error: '--product' and '--build-tests' are mutually exclusive"))
+            XCTAssertThrowsCommandExecutionError(try execute(["--product", "exec1", "--build-tests"], packagePath: fixturePath)) { error in
+                XCTAssertMatch(error.stderr, .contains("error: '--product' and '--build-tests' are mutually exclusive"))
             }
 
-            do {
-                _ = try execute(["--build-tests", "--target", "exec2"], packagePath: path)
-                XCTFail("Expected to fail")
-            } catch SwiftPMProductError.executionFailure(_, _, let stderr) {
-                XCTAssertMatch(stderr, .contains("error: '--target' and '--build-tests' are mutually exclusive"))
+            XCTAssertThrowsCommandExecutionError(try execute(["--build-tests", "--target", "exec2"], packagePath: fixturePath)) { error in
+                XCTAssertMatch(error.stderr, .contains("error: '--target' and '--build-tests' are mutually exclusive"))
             }
 
-            do {
-                _ = try execute(["--build-tests", "--target", "exec2", "--product", "exec1"], packagePath: path)
-                XCTFail("Expected to fail")
-            } catch SwiftPMProductError.executionFailure(_, _, let stderr) {
-                XCTAssertMatch(stderr, .contains("error: '--product', '--target', and '--build-tests' are mutually exclusive"))
+            XCTAssertThrowsCommandExecutionError(try execute(["--build-tests", "--target", "exec2", "--product", "exec1"], packagePath: fixturePath)) { error in
+                XCTAssertMatch(error.stderr, .contains("error: '--product', '--target', and '--build-tests' are mutually exclusive"))
             }
 
-            do {
-                _ = try execute(["--product", "UnkownProduct"], packagePath: path)
-                XCTFail("Expected to fail")
-            } catch SwiftPMProductError.executionFailure(_, _, let stderr) {
-                XCTAssertMatch(stderr, .contains("error: no product named 'UnkownProduct'"))
+            XCTAssertThrowsCommandExecutionError(try execute(["--product", "UnkownProduct"], packagePath: fixturePath)){ error in
+                XCTAssertMatch(error.stderr, .contains("error: no product named 'UnkownProduct'"))
             }
 
-            do {
-                _ = try execute(["--target", "UnkownTarget"], packagePath: path)
-                XCTFail("Expected to fail")
-            } catch SwiftPMProductError.executionFailure(_, _, let stderr) {
-                XCTAssertMatch(stderr, .contains("error: no target named 'UnkownTarget'"))
+            XCTAssertThrowsCommandExecutionError(try execute(["--target", "UnkownTarget"], packagePath: fixturePath)) { error in
+                XCTAssertMatch(error.stderr, .contains("error: no target named 'UnkownTarget'"))
             }
         }
     }
     
-    func testAtMainSupport() {
-        fixture(name: "Miscellaneous/AtMainSupport") { path in
-            let fullPath = resolveSymlinks(path)
+    func testAtMainSupport() throws {
+        try fixture(name: "Miscellaneous/AtMainSupport") { fixturePath in
+            let fullPath = try resolveSymlinks(fixturePath)
+
             do {
                 let result = try build(["--product", "ClangExecSingleFile"], packagePath: fullPath)
                 XCTAssertMatch(result.binContents, ["ClangExecSingleFile"])
-            } catch SwiftPMProductError.executionFailure(_, let stdout, let stderr) {
-                XCTFail(stdout + "\n" + stderr)
             }
 
             do {
                 let result = try build(["--product", "SwiftExecSingleFile"], packagePath: fullPath)
                 XCTAssertMatch(result.binContents, ["SwiftExecSingleFile"])
-            } catch SwiftPMProductError.executionFailure(_, let stdout, let stderr) {
-                XCTFail(stdout + "\n" + stderr)
             }
 
             do {
                 let result = try build(["--product", "SwiftExecMultiFile"], packagePath: fullPath)
                 XCTAssertMatch(result.binContents, ["SwiftExecMultiFile"])
-            } catch SwiftPMProductError.executionFailure(_, let stdout, let stderr) {
-                XCTFail(stdout + "\n" + stderr)
             }
         }
     }
 
-    func testNonReachableProductsAndTargetsFunctional() {
-        fixture(name: "Miscellaneous/UnreachableTargets") { path in
-            let aPath = path.appending(component: "A")
+    func testNonReachableProductsAndTargetsFunctional() throws {
+        try fixture(name: "Miscellaneous/UnreachableTargets") { fixturePath in
+            let aPath = fixturePath.appending(component: "A")
 
             do {
                 let result = try build([], packagePath: aPath)
@@ -211,8 +228,6 @@ final class BuildToolTests: CommandsTestCase {
                 XCTAssertNoMatch(result.binContents, ["BTarget2.build"])
                 XCTAssertNoMatch(result.binContents, ["cexec"])
                 XCTAssertNoMatch(result.binContents, ["CTarget.build"])
-            } catch SwiftPMProductError.executionFailure(_, _, let stderr) {
-                XCTFail(stderr)
             }
 
             // Dependency contains a dependent product
@@ -238,16 +253,14 @@ final class BuildToolTests: CommandsTestCase {
                 XCTAssertNoMatch(result.binContents, ["ATarget.swiftinterface"])
                 XCTAssertNoMatch(result.binContents, ["BTarget.swiftinterface"])
                 XCTAssertNoMatch(result.binContents, ["CTarget.swiftinterface"])
-            } catch SwiftPMProductError.executionFailure(_, _, let stderr) {
-                XCTFail(stderr)
             }
         }
     }
 
-    func testParseableInterfaces() {
-        fixture(name: "Miscellaneous/ParseableInterfaces") { path in
+    func testParseableInterfaces() throws {
+        try fixture(name: "Miscellaneous/ParseableInterfaces") { fixturePath in
             do {
-                let result = try build(["--enable-parseable-module-interfaces"], packagePath: path)
+                let result = try build(["--enable-parseable-module-interfaces"], packagePath: fixturePath)
                 XCTAssertMatch(result.binContents, ["A.swiftinterface"])
                 XCTAssertMatch(result.binContents, ["B.swiftinterface"])
             } catch SwiftPMProductError.executionFailure(_, _, let stderr) {
@@ -256,30 +269,35 @@ final class BuildToolTests: CommandsTestCase {
         }
     }
 
-    func testAutomaticParseableInterfacesWithLibraryEvolution() {
-        fixture(name: "Miscellaneous/LibraryEvolution") { path in
+    func testAutomaticParseableInterfacesWithLibraryEvolution() throws {
+        try fixture(name: "Miscellaneous/LibraryEvolution") { fixturePath in
             do {
-                let result = try build([], packagePath: path)
+                let result = try build([], packagePath: fixturePath)
                 XCTAssertMatch(result.binContents, ["A.swiftinterface"])
                 XCTAssertMatch(result.binContents, ["B.swiftinterface"])
-            } catch SwiftPMProductError.executionFailure(_, _, let stderr) {
-                XCTFail(stderr)
             }
         }
     }
 
-    func testBuildCompleteMessage() {
-        fixture(name: "DependencyResolution/Internal/Simple") { path in
+    func testBuildCompleteMessage() throws {
+        throw XCTSkip("This test fails to match the 'Compiling' regex; rdar://101815761")
+
+        try fixture(name: "DependencyResolution/Internal/Simple") { fixturePath in
             do {
-                let result = try execute([], packagePath: path)
+                let result = try execute([], packagePath: fixturePath)
                 XCTAssertMatch(result.stdout, .regex("\\[[1-9][0-9]*\\/[1-9][0-9]*\\] Compiling"))
                 let lines = result.stdout.split(separator: "\n")
                 XCTAssertMatch(String(lines.last!), .regex("Build complete! \\([0-9]*\\.[0-9]*s\\)"))
             }
 
             do {
-                // test second time, to make sure message is presented even when nothing to build (cached)
-                let result = try execute([], packagePath: path)
+                // test second time, to stabilize the cache
+                let _ = try execute([], packagePath: fixturePath)
+            }
+
+            do {
+                // test third time, to make sure message is presented even when nothing to build (cached)
+                let result = try execute([], packagePath: fixturePath)
                 XCTAssertNoMatch(result.stdout, .regex("\\[[1-9][0-9]*\\/[1-9][0-9]*\\] Compiling"))
                 let lines = result.stdout.split(separator: "\n")
                 XCTAssertMatch(String(lines.last!), .regex("Build complete! \\([0-9]*\\.[0-9]*s\\)"))
@@ -287,15 +305,15 @@ final class BuildToolTests: CommandsTestCase {
         }
     }
 
-    func testBuildStartMessage() {
-        fixture(name: "DependencyResolution/Internal/Simple") { path in
+    func testBuildStartMessage() throws {
+        try fixture(name: "DependencyResolution/Internal/Simple") { fixturePath in
             do {
-                let result = try execute(["-c", "debug"], packagePath: path)
+                let result = try execute(["-c", "debug"], packagePath: fixturePath)
                 XCTAssertMatch(result.stdout, .prefix("Building for debugging"))
             }
 
             do {
-                let result = try execute(["-c", "release"], packagePath: path)
+                let result = try execute(["-c", "release"], packagePath: fixturePath)
                 XCTAssertMatch(result.stdout, .prefix("Building for production"))
             }
         }
@@ -305,12 +323,40 @@ final class BuildToolTests: CommandsTestCase {
         #if !os(macOS)
         try XCTSkipIf(true, "test requires `xcbuild` and is therefore only supported on macOS")
         #endif
-        fixture(name: "ValidLayouts/SingleModule/ExecutableNew") { path in
+        try fixture(name: "ValidLayouts/SingleModule/ExecutableNew") { fixturePath in
             // Try building using XCBuild with default parameters.  This should succeed.  We build verbosely so we get full command lines.
-            let defaultOutput = try execute(["-c", "debug", "-v"], packagePath: path).stdout
+            let defaultOutput = try execute(["-c", "debug", "-v"], packagePath: fixturePath).stdout
             
             // Look for certain things in the output from XCBuild.
-            XCTAssertMatch(defaultOutput, .contains("-target \(UserToolchain.default.triple.tripleString(forPlatformVersion: ""))"))
+            XCTAssertMatch(defaultOutput, .contains("-target \(try UserToolchain.default.triple.tripleString(forPlatformVersion: ""))"))
+        }
+    }
+
+    func testXcodeBuildSystemWithAdditionalBuildFlags() throws {
+        try XCTSkipIf(true, "Disabled for now because it is hitting 'IR generation failure: Cannot read legacy layout file' in CI (rdar://88828632)")
+
+        #if !os(macOS)
+        try XCTSkipIf(true, "test requires `xcbuild` and is therefore only supported on macOS")
+        #endif
+        try fixture(name: "ValidLayouts/SingleModule/ExecutableMixed") { fixturePath in
+            // Try building using XCBuild with additional flags.  This should succeed.  We build verbosely so we get full command lines.
+            let defaultOutput = try execute(
+                [
+                    "--build-system", "xcode",
+                    "-c", "debug", "-v",
+                    "-Xlinker", "-rpath", "-Xlinker", "/fakerpath",
+                    "-Xcc", "-I/cfakepath",
+                    "-Xcxx", "-I/cxxfakepath",
+                    "-Xswiftc", "-I/swiftfakepath",
+                ],
+                packagePath: fixturePath
+            ).stdout
+
+            // Look for certain things in the output from XCBuild.
+            XCTAssertMatch(defaultOutput, .contains("/fakerpath"))
+            XCTAssertMatch(defaultOutput, .contains("-I/cfakepath"))
+            XCTAssertMatch(defaultOutput, .contains("-I/cxxfakepath"))
+            XCTAssertMatch(defaultOutput, .contains("-I/swiftfakepath"))
         }
     }
 
@@ -318,31 +364,42 @@ final class BuildToolTests: CommandsTestCase {
         #if !os(macOS)
         try XCTSkipIf(true, "test requires `xcbuild` and is therefore only supported on macOS")
         #endif
-        fixture(name: "ValidLayouts/SingleModule/ExecutableNew") { path in
+        try fixture(name: "ValidLayouts/SingleModule/ExecutableNew") { fixturePath in
             // Try building using XCBuild without specifying overrides.  This should succeed, and should use the default compiler path.
-            let defaultOutput = try execute(["-c", "debug", "-v"], packagePath: path).stdout
-            XCTAssertMatch(defaultOutput, .contains(ToolchainConfiguration.default.swiftCompilerPath.pathString))
+            let defaultOutput = try execute(["-c", "debug", "--vv"], packagePath: fixturePath).stdout
+            XCTAssertMatch(defaultOutput, .contains(try UserToolchain.default.swiftCompilerPath.pathString))
 
             // Now try building using XCBuild while specifying a faulty compiler override.  This should fail.  Note that we need to set the executable to use for the manifest itself to the default one, since it defaults to SWIFT_EXEC if not provided.
             var overriddenOutput = ""
-            do {
-                overriddenOutput = try execute(["-c", "debug", "-v"], environment: ["SWIFT_EXEC": "/usr/bin/false", "SWIFT_EXEC_MANIFEST": ToolchainConfiguration.default.swiftCompilerPath.pathString], packagePath: path).stdout
-                XCTFail("unexpected success (was SWIFT_EXEC not overridden properly?)")
-            }
-            catch SwiftPMProductError.executionFailure(let error, let stdout, _) {
-                switch error {
-                case ProcessResult.Error.nonZeroExit(let result) where result.exitStatus != .terminated(code: 0):
-                    overriddenOutput = stdout
-                    break
-                default:
-                    XCTFail("`swift build' failed in an unexpected manner")
-                }
-            }
-            catch {
-                XCTFail("`swift build' failed in an unexpected manner")
+            XCTAssertThrowsCommandExecutionError(
+                try execute(
+                    ["-c", "debug", "--vv"],
+                    environment: ["SWIFT_EXEC": "/usr/bin/false", "SWIFT_EXEC_MANIFEST": UserToolchain.default.swiftCompilerPath.pathString],
+                    packagePath: fixturePath
+                )
+            ) { error in
+                overriddenOutput = error.stderr
             }
             XCTAssertMatch(overriddenOutput, .contains("/usr/bin/false"))
         }
     }
 
+    func testPrintLLBuildManifestJobGraph() throws {
+        try fixture(name: "DependencyResolution/Internal/Simple") { fixturePath in
+            let output = try execute(["--print-manifest-job-graph"], packagePath: fixturePath).stdout
+            XCTAssertMatch(output, .prefix("digraph Jobs {"))
+        }
+    }
+    
+    func testSwiftDriverRawOutputGetsNewlines() throws {
+        try fixture(name: "DependencyResolution/Internal/Simple") { fixturePath in
+            // Building with `-wmo` should result in a `remark: Incremental compilation has been disabled: it is not compatible with whole module optimization` message, which should have a trailing newline.  Since that message won't be there at all when the legacy compiler driver is used, we gate this check on whether the remark is there in the first place.
+            let result = try execute(["-c", "release", "-Xswiftc", "-wmo"], packagePath: fixturePath)
+            if result.stdout.contains("remark: Incremental compilation has been disabled: it is not compatible with whole module optimization") {
+                XCTAssertMatch(result.stdout, .contains("optimization\n"))
+                XCTAssertNoMatch(result.stdout, .contains("optimization["))
+                XCTAssertNoMatch(result.stdout, .contains("optimizationremark"))
+            }
+        }
+    }
 }
