@@ -1,12 +1,14 @@
-/*
- This source file is part of the Swift.org open source project
-
- Copyright (c) 2014 - 2019 Apple Inc. and the Swift project authors
- Licensed under Apache License v2.0 with Runtime Library Exception
-
- See http://swift.org/LICENSE.txt for license information
- See http://swift.org/CONTRIBUTORS.txt for Swift project authors
-*/
+//===----------------------------------------------------------------------===//
+//
+// This source file is part of the Swift open source project
+//
+// Copyright (c) 2014-2019 Apple Inc. and the Swift project authors
+// Licensed under Apache License v2.0 with Runtime Library Exception
+//
+// See http://swift.org/LICENSE.txt for license information
+// See http://swift.org/CONTRIBUTORS.txt for the list of Swift project authors
+//
+//===----------------------------------------------------------------------===//
 
 import Basics
 import Commands
@@ -22,7 +24,7 @@ final class MultiRootSupportTests: CommandsTestCase {
             "/tmp/test/dep/Package.swift",
             "/tmp/test/local/Package.swift",
         ])
-        let path = AbsolutePath("/tmp/test/Workspace.xcworkspace")
+        let path = AbsolutePath(path: "/tmp/test/Workspace.xcworkspace")
         try fs.writeFileContents(path.appending(component: "contents.xcworkspacedata")) {
             $0 <<< """
                 <?xml version="1.0" encoding="UTF-8"?>
