@@ -1,12 +1,14 @@
-/*
- This source file is part of the Swift.org open source project
-
- Copyright (c) 2021-2022 Apple Inc. and the Swift project authors
- Licensed under Apache License v2.0 with Runtime Library Exception
-
- See http://swift.org/LICENSE.txt for license information
- See http://swift.org/CONTRIBUTORS.txt for Swift project authors
- */
+//===----------------------------------------------------------------------===//
+//
+// This source file is part of the Swift open source project
+//
+// Copyright (c) 2021-2022 Apple Inc. and the Swift project authors
+// Licensed under Apache License v2.0 with Runtime Library Exception
+//
+// See http://swift.org/LICENSE.txt for license information
+// See http://swift.org/CONTRIBUTORS.txt for the list of Swift project authors
+//
+//===----------------------------------------------------------------------===//
 
 import Basics
 import Dispatch
@@ -19,23 +21,23 @@ import XCTest
 
 class PackageCollectionSigningTests: XCTestCase {
     func test_RSA_signAndValidate_happyCase() throws {
-        try skipIfUnsupportedPlatform()
+        try PackageCollectionsSigningTests_skipIfUnsupportedPlatform()
 
-        fixture(name: "Collections", createGitRepo: false) { directoryPath in
+        try fixture(name: "Collections", createGitRepo: false) { fixturePath in
             let jsonDecoder = JSONDecoder.makeWithDefaults()
 
-            let collectionPath = directoryPath.appending(components: "JSON", "good.json")
-            let collectionData = Data(try localFileSystem.readFileContents(collectionPath).contents)
+            let collectionPath = fixturePath.appending(components: "JSON", "good.json")
+            let collectionData: Data = try localFileSystem.readFileContents(collectionPath)
             let collection = try jsonDecoder.decode(PackageCollectionModel.V1.Collection.self, from: collectionData)
 
-            let certPath = directoryPath.appending(components: "Signing", "Test_rsa.cer")
-            let intermediateCAPath = directoryPath.appending(components: "Signing", "TestIntermediateCA.cer")
-            let rootCAPath = directoryPath.appending(components: "Signing", "TestRootCA.cer")
+            let certPath = fixturePath.appending(components: "Signing", "Test_rsa.cer")
+            let intermediateCAPath = fixturePath.appending(components: "Signing", "TestIntermediateCA.cer")
+            let rootCAPath = fixturePath.appending(components: "Signing", "TestRootCA.cer")
             let certChainPaths = [certPath, intermediateCAPath, rootCAPath].map { $0.asURL }
 
-            let privateKeyPath = directoryPath.appending(components: "Signing", "Test_rsa_key.pem")
+            let privateKeyPath = fixturePath.appending(components: "Signing", "Test_rsa_key.pem")
 
-            let rootCA = try Certificate(derEncoded: Data(try localFileSystem.readFileContents(rootCAPath).contents))
+            let rootCA = try Certificate(derEncoded: try localFileSystem.readFileContents(rootCAPath))
             // Trust the self-signed root cert
             let certPolicy = TestCertificatePolicy(anchorCerts: [rootCA])
             let signing = PackageCollectionSigning(certPolicy: certPolicy, callbackQueue: callbackQueue)
@@ -51,9 +53,9 @@ class PackageCollectionSigningTests: XCTestCase {
     }
 
     func test_RSA_signAndValidate_collectionMismatch() throws {
-        try skipIfUnsupportedPlatform()
+        try PackageCollectionsSigningTests_skipIfUnsupportedPlatform()
 
-        fixture(name: "Collections", createGitRepo: false) { directoryPath in
+        try fixture(name: "Collections", createGitRepo: false) { fixturePath in
             let collection1 = PackageCollectionModel.V1.Collection(
                 name: "Test Package Collection 1",
                 overview: nil,
@@ -75,14 +77,14 @@ class PackageCollectionSigningTests: XCTestCase {
                 generatedBy: nil
             )
 
-            let certPath = directoryPath.appending(components: "Signing", "Test_rsa.cer")
-            let intermediateCAPath = directoryPath.appending(components: "Signing", "TestIntermediateCA.cer")
-            let rootCAPath = directoryPath.appending(components: "Signing", "TestRootCA.cer")
+            let certPath = fixturePath.appending(components: "Signing", "Test_rsa.cer")
+            let intermediateCAPath = fixturePath.appending(components: "Signing", "TestIntermediateCA.cer")
+            let rootCAPath = fixturePath.appending(components: "Signing", "TestRootCA.cer")
             let certChainPaths = [certPath, intermediateCAPath, rootCAPath].map { $0.asURL }
 
-            let privateKeyPath = directoryPath.appending(components: "Signing", "Test_rsa_key.pem")
+            let privateKeyPath = fixturePath.appending(components: "Signing", "Test_rsa_key.pem")
 
-            let rootCA = try Certificate(derEncoded: Data(try localFileSystem.readFileContents(rootCAPath).contents))
+            let rootCA = try Certificate(derEncoded: try localFileSystem.readFileContents(rootCAPath))
             // Trust the self-signed root cert
             let certPolicy = TestCertificatePolicy(anchorCerts: [rootCA])
             let signing = PackageCollectionSigning(certPolicy: certPolicy, callbackQueue: callbackQueue)
@@ -107,23 +109,23 @@ class PackageCollectionSigningTests: XCTestCase {
     }
 
     func test_EC_signAndValidate_happyCase() throws {
-        try skipIfUnsupportedPlatform()
+        try PackageCollectionsSigningTests_skipIfUnsupportedPlatform()
 
-        fixture(name: "Collections", createGitRepo: false) { directoryPath in
+        try fixture(name: "Collections", createGitRepo: false) { fixturePath in
             let jsonDecoder = JSONDecoder.makeWithDefaults()
 
-            let collectionPath = directoryPath.appending(components: "JSON", "good.json")
-            let collectionData = Data(try localFileSystem.readFileContents(collectionPath).contents)
+            let collectionPath = fixturePath.appending(components: "JSON", "good.json")
+            let collectionData: Data = try localFileSystem.readFileContents(collectionPath)
             let collection = try jsonDecoder.decode(PackageCollectionModel.V1.Collection.self, from: collectionData)
 
-            let certPath = directoryPath.appending(components: "Signing", "Test_ec.cer")
-            let intermediateCAPath = directoryPath.appending(components: "Signing", "TestIntermediateCA.cer")
-            let rootCAPath = directoryPath.appending(components: "Signing", "TestRootCA.cer")
+            let certPath = fixturePath.appending(components: "Signing", "Test_ec.cer")
+            let intermediateCAPath = fixturePath.appending(components: "Signing", "TestIntermediateCA.cer")
+            let rootCAPath = fixturePath.appending(components: "Signing", "TestRootCA.cer")
             let certChainPaths = [certPath, intermediateCAPath, rootCAPath].map { $0.asURL }
 
-            let privateKeyPath = directoryPath.appending(components: "Signing", "Test_ec_key.pem")
+            let privateKeyPath = fixturePath.appending(components: "Signing", "Test_ec_key.pem")
 
-            let rootCA = try Certificate(derEncoded: Data(try localFileSystem.readFileContents(rootCAPath).contents))
+            let rootCA = try Certificate(derEncoded: try localFileSystem.readFileContents(rootCAPath))
             // Trust the self-signed root cert
             let certPolicy = TestCertificatePolicy(anchorCerts: [rootCA])
             let signing = PackageCollectionSigning(certPolicy: certPolicy, callbackQueue: callbackQueue)
@@ -139,9 +141,9 @@ class PackageCollectionSigningTests: XCTestCase {
     }
 
     func test_EC_signAndValidate_collectionMismatch() throws {
-        try skipIfUnsupportedPlatform()
+        try PackageCollectionsSigningTests_skipIfUnsupportedPlatform()
 
-        fixture(name: "Collections", createGitRepo: false) { directoryPath in
+        try fixture(name: "Collections", createGitRepo: false) { fixturePath in
             let collection1 = PackageCollectionModel.V1.Collection(
                 name: "Test Package Collection 1",
                 overview: nil,
@@ -163,14 +165,14 @@ class PackageCollectionSigningTests: XCTestCase {
                 generatedBy: nil
             )
 
-            let certPath = directoryPath.appending(components: "Signing", "Test_ec.cer")
-            let intermediateCAPath = directoryPath.appending(components: "Signing", "TestIntermediateCA.cer")
-            let rootCAPath = directoryPath.appending(components: "Signing", "TestRootCA.cer")
+            let certPath = fixturePath.appending(components: "Signing", "Test_ec.cer")
+            let intermediateCAPath = fixturePath.appending(components: "Signing", "TestIntermediateCA.cer")
+            let rootCAPath = fixturePath.appending(components: "Signing", "TestRootCA.cer")
             let certChainPaths = [certPath, intermediateCAPath, rootCAPath].map { $0.asURL }
 
-            let privateKeyPath = directoryPath.appending(components: "Signing", "Test_ec_key.pem")
+            let privateKeyPath = fixturePath.appending(components: "Signing", "Test_ec_key.pem")
 
-            let rootCA = try Certificate(derEncoded: Data(try localFileSystem.readFileContents(rootCAPath).contents))
+            let rootCA = try Certificate(derEncoded: try localFileSystem.readFileContents(rootCAPath))
             // Trust the self-signed root cert
             let certPolicy = TestCertificatePolicy(anchorCerts: [rootCA])
             let signing = PackageCollectionSigning(certPolicy: certPolicy, callbackQueue: callbackQueue)
@@ -200,22 +202,22 @@ class PackageCollectionSigningTests: XCTestCase {
         try XCTSkipIf(true)
         #endif
 
-        try skipIfUnsupportedPlatform()
+        try PackageCollectionsSigningTests_skipIfUnsupportedPlatform()
 
-        fixture(name: "Collections", createGitRepo: false) { directoryPath in
+        try fixture(name: "Collections", createGitRepo: false) { fixturePath in
             let jsonDecoder = JSONDecoder.makeWithDefaults()
 
-            let collectionPath = directoryPath.appending(components: "JSON", "good.json")
-            let collectionData = Data(try localFileSystem.readFileContents(collectionPath).contents)
+            let collectionPath = fixturePath.appending(components: "JSON", "good.json")
+            let collectionData: Data = try localFileSystem.readFileContents(collectionPath)
             let collection = try jsonDecoder.decode(PackageCollectionModel.V1.Collection.self, from: collectionData)
 
-            let certPath = directoryPath.appending(components: "Signing", "development.cer")
-            let intermediateCAPath = directoryPath.appending(components: "Signing", "AppleWWDRCAG3.cer")
-            let rootCAPath = directoryPath.appending(components: "Signing", "AppleIncRoot.cer")
-            let rootCAData = Data(try localFileSystem.readFileContents(rootCAPath).contents)
+            let certPath = fixturePath.appending(components: "Signing", "development.cer")
+            let intermediateCAPath = fixturePath.appending(components: "Signing", "AppleWWDRCAG3.cer")
+            let rootCAPath = fixturePath.appending(components: "Signing", "AppleIncRoot.cer")
+            let rootCAData: Data = try localFileSystem.readFileContents(rootCAPath)
             let certChainPaths = [certPath, intermediateCAPath, rootCAPath].map { $0.asURL }
 
-            let privateKeyPath = directoryPath.appending(components: "Signing", "development-key.pem")
+            let privateKeyPath = fixturePath.appending(components: "Signing", "development-key.pem")
             let certPolicyKey: CertificatePolicyKey = .default
 
             #if os(macOS)
@@ -309,23 +311,23 @@ class PackageCollectionSigningTests: XCTestCase {
         try XCTSkipIf(true)
         #endif
 
-        try skipIfUnsupportedPlatform()
+        try PackageCollectionsSigningTests_skipIfUnsupportedPlatform()
 
-        fixture(name: "Collections", createGitRepo: false) { directoryPath in
+        try fixture(name: "Collections", createGitRepo: false) { fixturePath in
             let jsonDecoder = JSONDecoder.makeWithDefaults()
 
-            let collectionPath = directoryPath.appending(components: "JSON", "good.json")
-            let collectionData = Data(try localFileSystem.readFileContents(collectionPath).contents)
+            let collectionPath = fixturePath.appending(components: "JSON", "good.json")
+            let collectionData: Data = try localFileSystem.readFileContents(collectionPath)
             let collection = try jsonDecoder.decode(PackageCollectionModel.V1.Collection.self, from: collectionData)
 
             // This must be an Apple Distribution cert
-            let certPath = directoryPath.appending(components: "Signing", "development.cer")
-            let intermediateCAPath = directoryPath.appending(components: "Signing", "AppleWWDRCAG3.cer")
-            let rootCAPath = directoryPath.appending(components: "Signing", "AppleIncRoot.cer")
-            let rootCAData = Data(try localFileSystem.readFileContents(rootCAPath).contents)
+            let certPath = fixturePath.appending(components: "Signing", "development.cer")
+            let intermediateCAPath = fixturePath.appending(components: "Signing", "AppleWWDRCAG3.cer")
+            let rootCAPath = fixturePath.appending(components: "Signing", "AppleIncRoot.cer")
+            let rootCAData: Data = try localFileSystem.readFileContents(rootCAPath)
             let certChainPaths = [certPath, intermediateCAPath, rootCAPath].map { $0.asURL }
 
-            let privateKeyPath = directoryPath.appending(components: "Signing", "development-key.pem")
+            let privateKeyPath = fixturePath.appending(components: "Signing", "development-key.pem")
             let certPolicyKey: CertificatePolicyKey = .appleDistribution
 
             #if os(macOS)
@@ -403,23 +405,23 @@ class PackageCollectionSigningTests: XCTestCase {
         try XCTSkipIf(true)
         #endif
 
-        try skipIfUnsupportedPlatform()
+        try PackageCollectionsSigningTests_skipIfUnsupportedPlatform()
 
-        fixture(name: "Collections", createGitRepo: false) { directoryPath in
+        try fixture(name: "Collections", createGitRepo: false) { fixturePath in
             let jsonDecoder = JSONDecoder.makeWithDefaults()
 
-            let collectionPath = directoryPath.appending(components: "JSON", "good.json")
-            let collectionData = Data(try localFileSystem.readFileContents(collectionPath).contents)
+            let collectionPath = fixturePath.appending(components: "JSON", "good.json")
+            let collectionData: Data = try localFileSystem.readFileContents(collectionPath)
             let collection = try jsonDecoder.decode(PackageCollectionModel.V1.Collection.self, from: collectionData)
 
             // This must be an Apple Swift Package Collection cert
-            let certPath = directoryPath.appending(components: "Signing", "swift_package_collection.cer")
-            let intermediateCAPath = directoryPath.appending(components: "Signing", "AppleWWDRCA.cer")
-            let rootCAPath = directoryPath.appending(components: "Signing", "AppleIncRoot.cer")
-            let rootCAData = Data(try localFileSystem.readFileContents(rootCAPath).contents)
+            let certPath = fixturePath.appending(components: "Signing", "swift_package_collection.cer")
+            let intermediateCAPath = fixturePath.appending(components: "Signing", "AppleWWDRCA.cer")
+            let rootCAPath = fixturePath.appending(components: "Signing", "AppleIncRoot.cer")
+            let rootCAData: Data = try localFileSystem.readFileContents(rootCAPath)
             let certChainPaths = [certPath, intermediateCAPath, rootCAPath].map { $0.asURL }
 
-            let privateKeyPath = directoryPath.appending(components: "Signing", "development-key.pem")
+            let privateKeyPath = fixturePath.appending(components: "Signing", "development-key.pem")
             let certPolicyKey: CertificatePolicyKey = .appleSwiftPackageCollection
 
             #if os(macOS)
@@ -497,21 +499,21 @@ class PackageCollectionSigningTests: XCTestCase {
         try XCTSkipIf(true)
         #endif
 
-        try skipIfUnsupportedPlatform()
+        try PackageCollectionsSigningTests_skipIfUnsupportedPlatform()
 
-        fixture(name: "Collections", createGitRepo: false) { directoryPath in
+        try fixture(name: "Collections", createGitRepo: false) { fixturePath in
             let jsonDecoder = JSONDecoder.makeWithDefaults()
 
-            let collectionPath = directoryPath.appending(components: "JSON", "good.json")
-            let collectionData = Data(try localFileSystem.readFileContents(collectionPath).contents)
+            let collectionPath = fixturePath.appending(components: "JSON", "good.json")
+            let collectionData: Data = try localFileSystem.readFileContents(collectionPath)
             let collection = try jsonDecoder.decode(PackageCollectionModel.V1.Collection.self, from: collectionData)
 
-            let certPath = directoryPath.appending(components: "Signing", "development.cer")
-            let intermediateCAPath = directoryPath.appending(components: "Signing", "AppleWWDRCAG3.cer")
-            let rootCAPath = directoryPath.appending(components: "Signing", "AppleIncRoot.cer")
+            let certPath = fixturePath.appending(components: "Signing", "development.cer")
+            let intermediateCAPath = fixturePath.appending(components: "Signing", "AppleWWDRCAG3.cer")
+            let rootCAPath = fixturePath.appending(components: "Signing", "AppleIncRoot.cer")
             let certChainPaths = [certPath, intermediateCAPath, rootCAPath].map { $0.asURL }
 
-            let privateKeyPath = directoryPath.appending(components: "Signing", "development-key.pem")
+            let privateKeyPath = fixturePath.appending(components: "Signing", "development-key.pem")
             let certPolicyKey: CertificatePolicyKey = .default(subjectUserID: expectedSubjectUserID)
 
             #if os(macOS)
@@ -551,22 +553,22 @@ class PackageCollectionSigningTests: XCTestCase {
         try XCTSkipIf(true)
         #endif
 
-        try skipIfUnsupportedPlatform()
+        try PackageCollectionsSigningTests_skipIfUnsupportedPlatform()
 
-        fixture(name: "Collections", createGitRepo: false) { directoryPath in
+        try fixture(name: "Collections", createGitRepo: false) { fixturePath in
             let jsonDecoder = JSONDecoder.makeWithDefaults()
 
-            let collectionPath = directoryPath.appending(components: "JSON", "good.json")
-            let collectionData = Data(try localFileSystem.readFileContents(collectionPath).contents)
+            let collectionPath = fixturePath.appending(components: "JSON", "good.json")
+            let collectionData: Data = try localFileSystem.readFileContents(collectionPath)
             let collection = try jsonDecoder.decode(PackageCollectionModel.V1.Collection.self, from: collectionData)
 
             // This must be an Apple Distribution cert
-            let certPath = directoryPath.appending(components: "Signing", "development.cer")
-            let intermediateCAPath = directoryPath.appending(components: "Signing", "AppleWWDRCAG3.cer")
-            let rootCAPath = directoryPath.appending(components: "Signing", "AppleIncRoot.cer")
+            let certPath = fixturePath.appending(components: "Signing", "development.cer")
+            let intermediateCAPath = fixturePath.appending(components: "Signing", "AppleWWDRCAG3.cer")
+            let rootCAPath = fixturePath.appending(components: "Signing", "AppleIncRoot.cer")
             let certChainPaths = [certPath, intermediateCAPath, rootCAPath].map { $0.asURL }
 
-            let privateKeyPath = directoryPath.appending(components: "Signing", "development-key.pem")
+            let privateKeyPath = fixturePath.appending(components: "Signing", "development-key.pem")
             let certPolicyKey: CertificatePolicyKey = .appleDistribution(subjectUserID: expectedSubjectUserID)
 
             #if os(macOS)
@@ -616,22 +618,22 @@ class PackageCollectionSigningTests: XCTestCase {
         try XCTSkipIf(true)
         #endif
 
-        try skipIfUnsupportedPlatform()
+        try PackageCollectionsSigningTests_skipIfUnsupportedPlatform()
 
-        fixture(name: "Collections", createGitRepo: false) { directoryPath in
+        try fixture(name: "Collections", createGitRepo: false) { fixturePath in
             let jsonDecoder = JSONDecoder.makeWithDefaults()
 
-            let collectionPath = directoryPath.appending(components: "JSON", "good.json")
-            let collectionData = Data(try localFileSystem.readFileContents(collectionPath).contents)
+            let collectionPath = fixturePath.appending(components: "JSON", "good.json")
+            let collectionData: Data = try localFileSystem.readFileContents(collectionPath)
             let collection = try jsonDecoder.decode(PackageCollectionModel.V1.Collection.self, from: collectionData)
 
             // This must be an Apple Distribution cert
-            let certPath = directoryPath.appending(components: "Signing", "swift_package_collection.cer")
-            let intermediateCAPath = directoryPath.appending(components: "Signing", "AppleWWDRCA.cer")
-            let rootCAPath = directoryPath.appending(components: "Signing", "AppleIncRoot.cer")
+            let certPath = fixturePath.appending(components: "Signing", "swift_package_collection.cer")
+            let intermediateCAPath = fixturePath.appending(components: "Signing", "AppleWWDRCA.cer")
+            let rootCAPath = fixturePath.appending(components: "Signing", "AppleIncRoot.cer")
             let certChainPaths = [certPath, intermediateCAPath, rootCAPath].map { $0.asURL }
 
-            let privateKeyPath = directoryPath.appending(components: "Signing", "development-key.pem")
+            let privateKeyPath = fixturePath.appending(components: "Signing", "development-key.pem")
             let certPolicyKey: CertificatePolicyKey = .appleSwiftPackageCollection(subjectUserID: expectedSubjectUserID)
 
             #if os(macOS)
