@@ -1,12 +1,14 @@
-/*
- This source file is part of the Swift.org open source project
-
- Copyright (c) 2020-2022 Apple Inc. and the Swift project authors
- Licensed under Apache License v2.0 with Runtime Library Exception
-
- See http://swift.org/LICENSE.txt for license information
- See http://swift.org/CONTRIBUTORS.txt for Swift project authors
- */
+//===----------------------------------------------------------------------===//
+//
+// This source file is part of the Swift open source project
+//
+// Copyright (c) 2020-2022 Apple Inc. and the Swift project authors
+// Licensed under Apache License v2.0 with Runtime Library Exception
+//
+// See http://swift.org/LICENSE.txt for license information
+// See http://swift.org/CONTRIBUTORS.txt for the list of Swift project authors
+//
+//===----------------------------------------------------------------------===//
 
 import Foundation
 import XCTest
@@ -18,14 +20,13 @@ import PackageModel
 import SourceControl
 import SPMTestSupport
 import TSCBasic
-import TSCUtility
 
 class JSONPackageCollectionProviderTests: XCTestCase {
     func testGood() throws {
-        fixture(name: "Collections", createGitRepo: false) { directoryPath in
-            let path = directoryPath.appending(components: "JSON", "good.json")
+        try fixture(name: "Collections", createGitRepo: false) { fixturePath in
+            let path = fixturePath.appending(components: "JSON", "good.json")
             let url = URL(string: "https://www.test.com/collection.json")!
-            let data = Data(try localFileSystem.readFileContents(path).contents)
+            let data: Data = try localFileSystem.readFileContents(path)
 
             let handler: HTTPClient.Handler = { request, _, completion in
                 XCTAssertEqual(request.url, url, "url should match")
@@ -84,8 +85,8 @@ class JSONPackageCollectionProviderTests: XCTestCase {
     }
 
     func testLocalFile() throws {
-        fixture(name: "Collections", createGitRepo: false) { directoryPath in
-            let path = directoryPath.appending(components: "JSON", "good.json")
+        try fixture(name: "Collections", createGitRepo: false) { fixturePath in
+            let path = fixturePath.appending(components: "JSON", "good.json")
 
             var httpClient = HTTPClient(handler: { (_, _, _) -> Void in fatalError("should not be called") })
             httpClient.configuration.circuitBreakerStrategy = .none
@@ -361,10 +362,10 @@ class JSONPackageCollectionProviderTests: XCTestCase {
     func testSignedGood() throws {
         try skipIfSignatureCheckNotSupported()
 
-        fixture(name: "Collections", createGitRepo: false) { directoryPath in
-            let path = directoryPath.appending(components: "JSON", "good_signed.json")
+        try fixture(name: "Collections", createGitRepo: false) { fixturePath in
+            let path = fixturePath.appending(components: "JSON", "good_signed.json")
             let url = URL(string: "https://www.test.com/collection.json")!
-            let data = Data(try localFileSystem.readFileContents(path).contents)
+            let data: Data = try localFileSystem.readFileContents(path)
 
             let handler: HTTPClient.Handler = { request, _, completion in
                 XCTAssertEqual(request.url, url, "url should match")
@@ -430,10 +431,10 @@ class JSONPackageCollectionProviderTests: XCTestCase {
     }
 
     func testSigned_skipSignatureCheck() throws {
-        fixture(name: "Collections", createGitRepo: false) { directoryPath in
-            let path = directoryPath.appending(components: "JSON", "good_signed.json")
+        try fixture(name: "Collections", createGitRepo: false) { fixturePath in
+            let path = fixturePath.appending(components: "JSON", "good_signed.json")
             let url = URL(string: "https://www.test.com/collection.json")!
-            let data = Data(try localFileSystem.readFileContents(path).contents)
+            let data: Data = try localFileSystem.readFileContents(path)
 
             let handler: HTTPClient.Handler = { request, _, completion in
                 XCTAssertEqual(request.url, url, "url should match")
@@ -497,10 +498,10 @@ class JSONPackageCollectionProviderTests: XCTestCase {
     func testSigned_noTrustedRootCertsConfigured() throws {
         try skipIfSignatureCheckNotSupported()
 
-        fixture(name: "Collections", createGitRepo: false) { directoryPath in
-            let path = directoryPath.appending(components: "JSON", "good_signed.json")
+        try fixture(name: "Collections", createGitRepo: false) { fixturePath in
+            let path = fixturePath.appending(components: "JSON", "good_signed.json")
             let url = URL(string: "https://www.test.com/collection.json")!
-            let data = Data(try localFileSystem.readFileContents(path).contents)
+            let data: Data = try localFileSystem.readFileContents(path)
 
             let handler: HTTPClient.Handler = { request, _, completion in
                 XCTAssertEqual(request.url, url, "url should match")
@@ -539,10 +540,10 @@ class JSONPackageCollectionProviderTests: XCTestCase {
     func testSignedBad() throws {
         try skipIfSignatureCheckNotSupported()
 
-        fixture(name: "Collections", createGitRepo: false) { directoryPath in
-            let path = directoryPath.appending(components: "JSON", "good_signed.json")
+        try fixture(name: "Collections", createGitRepo: false) { fixturePath in
+            let path = fixturePath.appending(components: "JSON", "good_signed.json")
             let url = URL(string: "https://www.test.com/collection.json")!
-            let data = Data(try localFileSystem.readFileContents(path).contents)
+            let data: Data = try localFileSystem.readFileContents(path)
 
             let handler: HTTPClient.Handler = { request, _, completion in
                 XCTAssertEqual(request.url, url, "url should match")
@@ -582,8 +583,8 @@ class JSONPackageCollectionProviderTests: XCTestCase {
     func testSignedLocalFile() throws {
         try skipIfSignatureCheckNotSupported()
 
-        fixture(name: "Collections", createGitRepo: false) { directoryPath in
-            let path = directoryPath.appending(components: "JSON", "good_signed.json")
+        try fixture(name: "Collections", createGitRepo: false) { fixturePath in
+            let path = fixturePath.appending(components: "JSON", "good_signed.json")
 
             var httpClient = HTTPClient(handler: { (_, _, _) -> Void in fatalError("should not be called") })
             httpClient.configuration.circuitBreakerStrategy = .none
@@ -629,10 +630,10 @@ class JSONPackageCollectionProviderTests: XCTestCase {
     }
 
     func testRequiredSigningGood() throws {
-        fixture(name: "Collections", createGitRepo: false) { directoryPath in
-            let path = directoryPath.appending(components: "JSON", "good_signed.json")
+        try fixture(name: "Collections", createGitRepo: false) { fixturePath in
+            let path = fixturePath.appending(components: "JSON", "good_signed.json")
             let url = URL(string: "https://www.test.com/collection.json")!
-            let data = Data(try localFileSystem.readFileContents(path).contents)
+            let data: Data = try localFileSystem.readFileContents(path)
 
             let handler: HTTPClient.Handler = { request, _, completion in
                 XCTAssertEqual(request.url, url, "url should match")
@@ -699,10 +700,10 @@ class JSONPackageCollectionProviderTests: XCTestCase {
     }
 
     func testRequiredSigningMultiplePoliciesGood() throws {
-        fixture(name: "Collections", createGitRepo: false) { directoryPath in
-            let path = directoryPath.appending(components: "JSON", "good_signed.json")
+        try fixture(name: "Collections", createGitRepo: false) { fixturePath in
+            let path = fixturePath.appending(components: "JSON", "good_signed.json")
             let url = URL(string: "https://www.test.com/collection.json")!
-            let data = Data(try localFileSystem.readFileContents(path).contents)
+            let data: Data = try localFileSystem.readFileContents(path)
 
             let handler: HTTPClient.Handler = { request, _, completion in
                 XCTAssertEqual(request.url, url, "url should match")
@@ -774,10 +775,10 @@ class JSONPackageCollectionProviderTests: XCTestCase {
     }
 
     func testMissingRequiredSignature() throws {
-        fixture(name: "Collections", createGitRepo: false) { directoryPath in
-            let path = directoryPath.appending(components: "JSON", "good.json")
+        try fixture(name: "Collections", createGitRepo: false) { fixturePath in
+            let path = fixturePath.appending(components: "JSON", "good.json")
             let url = URL(string: "https://www.test.com/collection.json")!
-            let data = Data(try localFileSystem.readFileContents(path).contents)
+            let data: Data = try localFileSystem.readFileContents(path)
 
             let handler: HTTPClient.Handler = { request, _, completion in
                 XCTAssertEqual(request.url, url, "url should match")
@@ -838,11 +839,11 @@ internal extension JSONPackageCollectionProvider {
     ) {
         self.init(
             configuration: configuration,
+            fileSystem: fileSystem,
             observabilityScope: ObservabilitySystem.NOOP,
-            httpClient: httpClient,
-            signatureValidator: signatureValidator,
             sourceCertPolicy: sourceCertPolicy,
-            fileSystem: fileSystem
+            customHTTPClient: httpClient ,
+            customSignatureValidator: signatureValidator
         )
     }
 }
