@@ -1,12 +1,14 @@
-/*
- This source file is part of the Swift.org open source project
-
- Copyright (c) 2020-2021 Apple Inc. and the Swift project authors
- Licensed under Apache License v2.0 with Runtime Library Exception
-
- See http://swift.org/LICENSE.txt for license information
- See http://swift.org/CONTRIBUTORS.txt for Swift project authors
- */
+//===----------------------------------------------------------------------===//
+//
+// This source file is part of the Swift open source project
+//
+// Copyright (c) 2020-2021 Apple Inc. and the Swift project authors
+// Licensed under Apache License v2.0 with Runtime Library Exception
+//
+// See http://swift.org/LICENSE.txt for license information
+// See http://swift.org/CONTRIBUTORS.txt for the list of Swift project authors
+//
+//===----------------------------------------------------------------------===//
 
 import Foundation
 import XCTest
@@ -16,7 +18,6 @@ import Basics
 import PackageModel
 import SourceControl
 import TSCBasic
-import TSCUtility
 
 final class PackageCollectionsTests: XCTestCase {
     func testUpdateAuthTokens() throws {
@@ -29,7 +30,7 @@ final class PackageCollectionsTests: XCTestCase {
         defer { XCTAssertNoThrow(try storage.close()) }
 
         // Disable cache for this test to avoid setup/cleanup
-        let metadataProviderConfig = GitHubPackageMetadataProvider.Configuration(authTokens: configuration.authTokens, cacheTTLInSeconds: -1)
+        let metadataProviderConfig = GitHubPackageMetadataProvider.Configuration(authTokens: configuration.authTokens, disableCache: true)
         let metadataProvider = GitHubPackageMetadataProvider(configuration: metadataProviderConfig)
         let packageCollections = PackageCollections(configuration: configuration, storage: storage, collectionProviders: [:], metadataProvider: metadataProvider)
 
@@ -55,7 +56,7 @@ final class PackageCollectionsTests: XCTestCase {
     }
 
     func testBasicRegistration() throws {
-        try skipIfUnsupportedPlatform()
+        try PackageCollectionsTests_skipIfUnsupportedPlatform()
 
         let configuration = PackageCollections.Configuration()
         let storage = makeMockStorage()
@@ -82,7 +83,7 @@ final class PackageCollectionsTests: XCTestCase {
     }
 
     func testAddDuplicates() throws {
-        try skipIfUnsupportedPlatform()
+        try PackageCollectionsTests_skipIfUnsupportedPlatform()
 
         let configuration = PackageCollections.Configuration()
         let storage = makeMockStorage()
@@ -110,7 +111,7 @@ final class PackageCollectionsTests: XCTestCase {
     }
 
     func testAddUnsigned() throws {
-        try skipIfUnsupportedPlatform()
+        try PackageCollectionsTests_skipIfUnsupportedPlatform()
 
         let configuration = PackageCollections.Configuration()
         let storage = makeMockStorage()
@@ -153,7 +154,7 @@ final class PackageCollectionsTests: XCTestCase {
     }
 
     func testInvalidCollectionNotAdded() throws {
-        try skipIfUnsupportedPlatform()
+        try PackageCollectionsTests_skipIfUnsupportedPlatform()
 
         let configuration = PackageCollections.Configuration()
         // This test doesn't use search at all and finishes quickly so disable target trie to prevent race
@@ -191,7 +192,7 @@ final class PackageCollectionsTests: XCTestCase {
     }
 
     func testCollectionPendingTrustConfirmIsKeptOnAdd() throws {
-        try skipIfUnsupportedPlatform()
+        try PackageCollectionsTests_skipIfUnsupportedPlatform()
 
         let configuration = PackageCollections.Configuration()
         // This test doesn't use search at all and finishes quickly so disable target trie to prevent race
@@ -229,7 +230,7 @@ final class PackageCollectionsTests: XCTestCase {
     }
 
     func testCollectionWithInvalidSignatureNotAdded() throws {
-        try skipIfUnsupportedPlatform()
+        try PackageCollectionsTests_skipIfUnsupportedPlatform()
 
         let configuration = PackageCollections.Configuration()
         // This test doesn't use search at all and finishes quickly so disable target trie to prevent race
@@ -267,7 +268,7 @@ final class PackageCollectionsTests: XCTestCase {
     }
 
     func testDelete() throws {
-        try skipIfUnsupportedPlatform()
+        try PackageCollectionsTests_skipIfUnsupportedPlatform()
 
         let configuration = PackageCollections.Configuration()
         let storage = makeMockStorage()
@@ -318,7 +319,7 @@ final class PackageCollectionsTests: XCTestCase {
     }
 
     func testDeleteFromBothStorages() throws {
-        try skipIfUnsupportedPlatform()
+        try PackageCollectionsTests_skipIfUnsupportedPlatform()
 
         let configuration = PackageCollections.Configuration()
         let storage = makeMockStorage()
@@ -353,7 +354,7 @@ final class PackageCollectionsTests: XCTestCase {
     }
 
     func testOrdering() throws {
-        try skipIfUnsupportedPlatform()
+        try PackageCollectionsTests_skipIfUnsupportedPlatform()
 
         let configuration = PackageCollections.Configuration()
         let storage = makeMockStorage()
@@ -424,7 +425,7 @@ final class PackageCollectionsTests: XCTestCase {
     }
 
     func testReorder() throws {
-        try skipIfUnsupportedPlatform()
+        try PackageCollectionsTests_skipIfUnsupportedPlatform()
 
         let configuration = PackageCollections.Configuration()
         let storage = makeMockStorage()
@@ -526,7 +527,7 @@ final class PackageCollectionsTests: XCTestCase {
     }
 
     func testUpdateTrust() throws {
-        try skipIfUnsupportedPlatform()
+        try PackageCollectionsTests_skipIfUnsupportedPlatform()
 
         let configuration = PackageCollections.Configuration()
         let storage = makeMockStorage()
@@ -577,7 +578,7 @@ final class PackageCollectionsTests: XCTestCase {
     }
 
     func testList() throws {
-        try skipIfUnsupportedPlatform()
+        try PackageCollectionsTests_skipIfUnsupportedPlatform()
 
         let configuration = PackageCollections.Configuration()
         let storage = makeMockStorage()
@@ -599,7 +600,7 @@ final class PackageCollectionsTests: XCTestCase {
     }
 
     func testListSubset() throws {
-        try skipIfUnsupportedPlatform()
+        try PackageCollectionsTests_skipIfUnsupportedPlatform()
 
         let configuration = PackageCollections.Configuration()
         let storage = makeMockStorage()
@@ -627,7 +628,7 @@ final class PackageCollectionsTests: XCTestCase {
         try XCTSkipIf(true)
         #endif
 
-        try skipIfUnsupportedPlatform()
+        try PackageCollectionsTests_skipIfUnsupportedPlatform()
 
         let configuration = PackageCollections.Configuration()
         let storage = makeMockStorage()
@@ -657,7 +658,7 @@ final class PackageCollectionsTests: XCTestCase {
     }
 
     func testPackageSearch() throws {
-        try skipIfUnsupportedPlatform()
+        try PackageCollectionsTests_skipIfUnsupportedPlatform()
 
         let configuration = PackageCollections.Configuration()
         let storage = makeMockStorage()
@@ -794,7 +795,7 @@ final class PackageCollectionsTests: XCTestCase {
         try XCTSkipIf(true)
         #endif
 
-        try skipIfUnsupportedPlatform()
+        try PackageCollectionsTests_skipIfUnsupportedPlatform()
 
         let configuration = PackageCollections.Configuration()
         let storage = makeMockStorage()
@@ -824,7 +825,7 @@ final class PackageCollectionsTests: XCTestCase {
     }
 
     func testTargetsSearch() throws {
-        try skipIfUnsupportedPlatform()
+        try PackageCollectionsTests_skipIfUnsupportedPlatform()
 
         let configuration = PackageCollections.Configuration()
         let storage = makeMockStorage()
@@ -928,7 +929,7 @@ final class PackageCollectionsTests: XCTestCase {
         try XCTSkipIf(true)
         #endif
 
-        try skipIfUnsupportedPlatform()
+        try PackageCollectionsTests_skipIfUnsupportedPlatform()
 
         let configuration = PackageCollections.Configuration()
         let storage = makeMockStorage()
@@ -958,7 +959,7 @@ final class PackageCollectionsTests: XCTestCase {
     }
 
     func testHappyRefresh() throws {
-        try skipIfUnsupportedPlatform()
+        try PackageCollectionsTests_skipIfUnsupportedPlatform()
 
         let configuration = PackageCollections.Configuration()
         let storage = makeMockStorage()
@@ -980,7 +981,7 @@ final class PackageCollectionsTests: XCTestCase {
     }
 
     func testBrokenRefresh() throws {
-        try skipIfUnsupportedPlatform()
+        try PackageCollectionsTests_skipIfUnsupportedPlatform()
 
         struct BrokenProvider: PackageCollectionProvider {
             let brokenSources: [PackageCollectionsModel.CollectionSource]
@@ -1054,7 +1055,7 @@ final class PackageCollectionsTests: XCTestCase {
     }
 
     func testRefreshOne() throws {
-        try skipIfUnsupportedPlatform()
+        try PackageCollectionsTests_skipIfUnsupportedPlatform()
 
         let configuration = PackageCollections.Configuration()
         let storage = makeMockStorage()
@@ -1076,7 +1077,7 @@ final class PackageCollectionsTests: XCTestCase {
     }
 
     func testRefreshOneTrustedUnsigned() throws {
-        try skipIfUnsupportedPlatform()
+        try PackageCollectionsTests_skipIfUnsupportedPlatform()
 
         let configuration = PackageCollections.Configuration()
         let storage = makeMockStorage()
@@ -1096,7 +1097,7 @@ final class PackageCollectionsTests: XCTestCase {
     }
 
     func testRefreshOneNotFound() throws {
-        try skipIfUnsupportedPlatform()
+        try PackageCollectionsTests_skipIfUnsupportedPlatform()
 
         let configuration = PackageCollections.Configuration()
         let storage = makeMockStorage()
@@ -1114,7 +1115,7 @@ final class PackageCollectionsTests: XCTestCase {
     }
 
     func testListTargets() throws {
-        try skipIfUnsupportedPlatform()
+        try PackageCollectionsTests_skipIfUnsupportedPlatform()
 
         let configuration = PackageCollections.Configuration()
         let storage = makeMockStorage()
@@ -1152,7 +1153,7 @@ final class PackageCollectionsTests: XCTestCase {
     }
 
     func testFetchMetadataHappy() throws {
-        try skipIfUnsupportedPlatform()
+        try PackageCollectionsTests_skipIfUnsupportedPlatform()
 
         let configuration = PackageCollections.Configuration()
         let storage = makeMockStorage()
@@ -1186,12 +1187,11 @@ final class PackageCollectionsTests: XCTestCase {
         let expectedMetadata = PackageCollections.mergedPackageMetadata(package: mockPackage, basicMetadata: mockMetadata)
         XCTAssertEqual(metadata.package, expectedMetadata, "package should match")
 
-        let expectedProvider = PackageMetadataProviderContext(authTokenType: nil, isAuthTokenConfigured: false, error: nil)
-        XCTAssertEqual(metadata.provider, expectedProvider, "provider should match")
+        XCTAssertNil(metadata.provider, "provider should be nil")
     }
 
     func testFetchMetadataInOrder() throws {
-        try skipIfUnsupportedPlatform()
+        try PackageCollectionsTests_skipIfUnsupportedPlatform()
 
         let configuration = PackageCollections.Configuration()
         let storage = makeMockStorage()
@@ -1229,7 +1229,7 @@ final class PackageCollectionsTests: XCTestCase {
     }
 
     func testFetchMetadataInCollections() throws {
-        try skipIfUnsupportedPlatform()
+        try PackageCollectionsTests_skipIfUnsupportedPlatform()
 
         let configuration = PackageCollections.Configuration()
         let storage = makeMockStorage()
@@ -1266,7 +1266,7 @@ final class PackageCollectionsTests: XCTestCase {
     }
 
     func testMergedPackageMetadata() throws {
-        try skipIfUnsupportedPlatform()
+        try PackageCollectionsTests_skipIfUnsupportedPlatform()
 
         let packageId = UUID().uuidString
 
@@ -1313,13 +1313,12 @@ final class PackageCollectionsTests: XCTestCase {
 
         let mockMetadata = PackageCollectionsModel.PackageBasicMetadata(summary: "\(mockPackage.summary!) 2",
                                                                         keywords: mockPackage.keywords.flatMap { $0.map { "\($0)-2" } },
-                                                                        versions: mockPackage.versions.map { PackageCollectionsModel.PackageBasicVersionMetadata(version: $0.version, title: "\($0.title!) 2", summary: "\($0.summary!) 2", createdAt: Date(), publishedAt: nil) },
+                                                                        versions: mockPackage.versions.map { PackageCollectionsModel.PackageBasicVersionMetadata(version: $0.version, title: "\($0.title!) 2", summary: "\($0.summary!) 2", createdAt: Date()) },
                                                                         watchersCount: mockPackage.watchersCount! + 1,
                                                                         readmeURL: URL(string: "\(mockPackage.readmeURL!.absoluteString)-2")!,
                                                                         license: PackageCollectionsModel.License(type: .Apache2_0, url: URL(string: "\(mockPackage.license!.url.absoluteString)-2")!),
                                                                         authors: mockPackage.authors.flatMap { $0.map { .init(username: "\($0.username + "2")", url: nil, service: nil) } },
-                                                                        languages: ["Swift"],
-                                                                        processedAt: Date())
+                                                                        languages: ["Swift"])
 
         let metadata = PackageCollections.mergedPackageMetadata(package: mockPackage, basicMetadata: mockMetadata)
 
@@ -1356,7 +1355,7 @@ final class PackageCollectionsTests: XCTestCase {
     }
 
     func testFetchMetadataNotFoundInCollections() throws {
-        try skipIfUnsupportedPlatform()
+        try PackageCollectionsTests_skipIfUnsupportedPlatform()
 
         let configuration = PackageCollections.Configuration()
         let storage = makeMockStorage()
@@ -1379,7 +1378,7 @@ final class PackageCollectionsTests: XCTestCase {
     }
 
     func testFetchMetadataNotFoundByProvider() throws {
-        try skipIfUnsupportedPlatform()
+        try PackageCollectionsTests_skipIfUnsupportedPlatform()
 
         let configuration = PackageCollections.Configuration()
         let storage = makeMockStorage()
@@ -1417,21 +1416,19 @@ final class PackageCollectionsTests: XCTestCase {
     }
 
     func testFetchMetadataProviderError() throws {
-        try skipIfUnsupportedPlatform()
+        try PackageCollectionsTests_skipIfUnsupportedPlatform()
 
         struct BrokenMetadataProvider: PackageMetadataProvider {
             var name: String = "BrokenMetadataProvider"
 
-            func get(identity: PackageIdentity, location: String, callback: @escaping (Result<PackageCollectionsModel.PackageBasicMetadata, Error>) -> Void) {
-                callback(.failure(TerribleThing()))
+            func get(
+                identity: PackageIdentity,
+                location: String,
+                callback: @escaping (Result<PackageCollectionsModel.PackageBasicMetadata, Error>, PackageMetadataProviderContext?) -> Void
+            ) {
+                callback(.failure(TerribleThing()), nil)
             }
 
-            func getAuthTokenType(for location: String) -> AuthTokenType? {
-                nil
-            }
-
-            func close() throws {}
-
             struct TerribleThing: Error {}
         }
 
@@ -1473,7 +1470,7 @@ final class PackageCollectionsTests: XCTestCase {
         try XCTSkipIf(true)
         #endif
 
-        try skipIfUnsupportedPlatform()
+        try PackageCollectionsTests_skipIfUnsupportedPlatform()
 
         let configuration = PackageCollections.Configuration()
         let storage = makeMockStorage()
@@ -1503,7 +1500,7 @@ final class PackageCollectionsTests: XCTestCase {
     }
 
     func testListPackages() throws {
-        try skipIfUnsupportedPlatform()
+        try PackageCollectionsTests_skipIfUnsupportedPlatform()
 
         let configuration = PackageCollections.Configuration()
         let storage = makeMockStorage()
@@ -1610,6 +1607,7 @@ private extension PackageCollections {
     ) {
         self.init(
             configuration: configuration,
+            fileSystem: localFileSystem,
             observabilityScope: ObservabilitySystem.NOOP,
             storage: storage,
             collectionProviders: collectionProviders,
@@ -1618,8 +1616,8 @@ private extension PackageCollections {
     }
 }
 
-private extension XCTestCase {
-    func skipIfUnsupportedPlatform() throws {
+extension XCTestCase {
+    func PackageCollectionsTests_skipIfUnsupportedPlatform() throws {
         if !PackageCollections.isSupportedPlatform {
             throw XCTSkip("Skipping test on unsupported platform")
         }
