@@ -1,22 +1,35 @@
-/*
- This source file is part of the Swift.org open source project
-
- Copyright (c) 2014 - 2017 Apple Inc. and the Swift project authors
- Licensed under Apache License v2.0 with Runtime Library Exception
-
- See http://swift.org/LICENSE.txt for license information
- See http://swift.org/CONTRIBUTORS.txt for Swift project authors
-*/
+//===----------------------------------------------------------------------===//
+//
+// This source file is part of the Swift open source project
+//
+// Copyright (c) 2014-2017 Apple Inc. and the Swift project authors
+// Licensed under Apache License v2.0 with Runtime Library Exception
+//
+// See http://swift.org/LICENSE.txt for license information
+// See http://swift.org/CONTRIBUTORS.txt for the list of Swift project authors
+//
+//===----------------------------------------------------------------------===//
 
 import XCTest
 
 import TSCBasic
 @testable import SourceControl
-import TSCUtility
 
 import SPMTestSupport
 
+import enum TSCUtility.Git
+
 class GitRepositoryTests: XCTestCase {
+
+    override func setUp() {
+        // needed for submodule tests
+        Git.environment = ["GIT_ALLOW_PROTOCOL": "file"]
+    }
+
+    override func tearDown() {
+        Git.environment = ProcessInfo.processInfo.environment
+    }
+
     /// Test the basic provider functions.
     func testRepositorySpecifier() {
         do {
@@ -32,9 +45,9 @@ class GitRepositoryTests: XCTestCase {
         }
 
         do {
-            let s1 = RepositorySpecifier(path: .init("/a"))
-            let s2 = RepositorySpecifier(path: .init("/a"))
-            let s3 = RepositorySpecifier(path: .init("/b"))
+            let s1 = RepositorySpecifier(path: .init(path: "/a"))
+            let s2 = RepositorySpecifier(path: .init(path: "/a"))
+            let s3 = RepositorySpecifier(path: .init(path: "/b"))
 
             XCTAssertEqual(s1, s1)
             XCTAssertEqual(s1, s2)
@@ -56,7 +69,7 @@ class GitRepositoryTests: XCTestCase {
             let provider = GitRepositoryProvider()
             XCTAssertTrue(try provider.workingCopyExists(at: testRepoPath))
             let repoSpec = RepositorySpecifier(path: testRepoPath)
-            try! provider.fetch(repository: repoSpec, to: testCheckoutPath)
+            try provider.fetch(repository: repoSpec, to: testCheckoutPath)
 
             // Verify the checkout was made.
             XCTAssertDirectoryExists(testCheckoutPath)
@@ -110,10 +123,17 @@ class GitRepositoryTests: XCTestCase {
     /// `Inputs`, which has known commit hashes. See the `construct.sh` script
     /// contained within it for more information.
     func testRawRepository() throws {
+#if os(Windows)
+        try XCTSkipIf(true, "test repository has non-portable file names")
+#endif
         try testWithTemporaryDirectory { path in
             // Unarchive the static test repository.
-            let inputArchivePath = AbsolutePath(#file).parentDirectory.appending(components: "Inputs", "TestRepo.tgz")
+            let inputArchivePath = AbsolutePath(path: #file).parentDirectory.appending(components: "Inputs", "TestRepo.tgz")
+#if os(Windows)
+            try systemQuietly(["tar.exe", "-x", "-v", "-C", path.pathString, "-f", inputArchivePath.pathString])
+#else
             try systemQuietly(["tar", "-x", "-v", "-C", path.pathString, "-f", inputArchivePath.pathString])
+#endif
             let testRepoPath = path.appending(component: "TestRepo")
 
             // Check hash resolution.
@@ -174,7 +194,9 @@ class GitRepositoryTests: XCTestCase {
             initGitRepo(repoPath)
 
             try Process.checkNonZeroExit(
-                args: Git.tool, "-C", repoPath.pathString, "submodule", "add", testRepoPath.pathString)
+                args: Git.tool, "-C", repoPath.pathString, "submodule", "add", testRepoPath.pathString,
+                environment: Git.environment
+            )
             let repo = GitRepository(path: repoPath)
             try repo.stageEverything()
             try repo.commit()
@@ -202,7 +224,7 @@ class GitRepositoryTests: XCTestCase {
             try localFileSystem.createDirectory(testRepoPath.appending(component: "subdir"))
             try localFileSystem.writeFileContents(testRepoPath.appending(components: "subdir", "test-file-2.txt"), bytes: test2FileContents)
             try localFileSystem.writeFileContents(testRepoPath.appending(component: "test-file-3.sh"), bytes: test3FileContents)
-            try! Process.checkNonZeroExit(args: "chmod", "+x", testRepoPath.appending(component: "test-file-3.sh").pathString)
+            try localFileSystem.chmod(.executable, path: testRepoPath.appending(component: "test-file-3.sh"), options: [])
             let testRepo = GitRepository(path: testRepoPath)
             try testRepo.stage(files: "test-file-1.txt", "subdir/test-file-2.txt", "test-file-3.sh")
             try testRepo.commit()
@@ -219,19 +241,21 @@ class GitRepositoryTests: XCTestCase {
             let view = try repository.openFileView(revision: repository.resolveRevision(tag: "test-tag"))
 
             // Check basic predicates.
-            XCTAssert(view.isDirectory(AbsolutePath("/")))
-            XCTAssert(view.isDirectory(AbsolutePath("/subdir")))
-            XCTAssert(!view.isDirectory(AbsolutePath("/does-not-exist")))
-            XCTAssert(view.exists(AbsolutePath("/test-file-1.txt")))
-            XCTAssert(!view.exists(AbsolutePath("/does-not-exist")))
-            XCTAssert(view.isFile(AbsolutePath("/test-file-1.txt")))
-            XCTAssert(!view.isSymlink(AbsolutePath("/test-file-1.txt")))
-            XCTAssert(!view.isExecutableFile(AbsolutePath("/does-not-exist")))
-            XCTAssert(view.isExecutableFile(AbsolutePath("/test-file-3.sh")))
+            XCTAssert(view.isDirectory(AbsolutePath(path: "/")))
+            XCTAssert(view.isDirectory(AbsolutePath(path: "/subdir")))
+            XCTAssert(!view.isDirectory(AbsolutePath(path: "/does-not-exist")))
+            XCTAssert(view.exists(AbsolutePath(path: "/test-file-1.txt")))
+            XCTAssert(!view.exists(AbsolutePath(path: "/does-not-exist")))
+            XCTAssert(view.isFile(AbsolutePath(path: "/test-file-1.txt")))
+            XCTAssert(!view.isSymlink(AbsolutePath(path: "/test-file-1.txt")))
+            XCTAssert(!view.isExecutableFile(AbsolutePath(path: "/does-not-exist")))
+#if !os(Windows)
+            XCTAssert(view.isExecutableFile(AbsolutePath(path: "/test-file-3.sh")))
+#endif
 
             // Check read of a directory.
-            let subdirPath = AbsolutePath("/subdir")
-            XCTAssertEqual(try view.getDirectoryContents(AbsolutePath("/")).sorted(), ["file.swift", "subdir", "test-file-1.txt", "test-file-3.sh"])
+            let subdirPath = AbsolutePath(path: "/subdir")
+            XCTAssertEqual(try view.getDirectoryContents(AbsolutePath(path: "/")).sorted(), ["file.swift", "subdir", "test-file-1.txt", "test-file-3.sh"])
             XCTAssertEqual(try view.getDirectoryContents(subdirPath).sorted(), ["test-file-2.txt"])
             XCTAssertThrows(FileSystemError(.isDirectory, subdirPath)) {
                 _ = try view.readFileContents(subdirPath)
@@ -243,28 +267,28 @@ class GitRepositoryTests: XCTestCase {
             }
 
             // Check read through a non-directory.
-            let notDirectoryPath1 = AbsolutePath("/test-file-1.txt")
+            let notDirectoryPath1 = AbsolutePath(path: "/test-file-1.txt")
             XCTAssertThrows(FileSystemError(.notDirectory, notDirectoryPath1)) {
                 _ = try view.getDirectoryContents(notDirectoryPath1)
             }
-            let notDirectoryPath2 = AbsolutePath("/test-file-1.txt/thing")
+            let notDirectoryPath2 = AbsolutePath(path: "/test-file-1.txt/thing")
             XCTAssertThrows(FileSystemError(.notDirectory, notDirectoryPath2)) {
                 _ = try view.readFileContents(notDirectoryPath2)
             }
 
             // Check read/write into a missing directory.
-            let noEntryPath1 = AbsolutePath("/does-not-exist")
+            let noEntryPath1 = AbsolutePath(path: "/does-not-exist")
             XCTAssertThrows(FileSystemError(.noEntry, noEntryPath1)) {
                 _ = try view.getDirectoryContents(noEntryPath1)
             }
-            let noEntryPath2 = AbsolutePath("/does/not/exist")
+            let noEntryPath2 = AbsolutePath(path: "/does/not/exist")
             XCTAssertThrows(FileSystemError(.noEntry, noEntryPath2)) {
                 _ = try view.readFileContents(noEntryPath2)
             }
 
             // Check read of a file.
-            XCTAssertEqual(try view.readFileContents(AbsolutePath("/test-file-1.txt")), test1FileContents)
-            XCTAssertEqual(try view.readFileContents(AbsolutePath("/subdir/test-file-2.txt")), test2FileContents)
+            XCTAssertEqual(try view.readFileContents(AbsolutePath(path: "/test-file-1.txt")), test1FileContents)
+            XCTAssertEqual(try view.readFileContents(AbsolutePath(path: "/subdir/test-file-2.txt")), test2FileContents)
         }
     }
 
@@ -559,7 +583,11 @@ class GitRepositoryTests: XCTestCase {
 
             // Add submodule to foo and tag it as 1.0.1
             try foo.checkout(newBranch: "submodule")
-            try systemQuietly([Git.tool, "-C", fooPath.pathString, "submodule", "add", barPath.pathString, "bar"])
+            try Process.checkNonZeroExit(
+                args: Git.tool, "-C", fooPath.pathString, "submodule", "add", barPath.pathString, "bar",
+                environment: Git.environment
+            )
+
             try foo.stageEverything()
             try foo.commit()
             try foo.tag(name: "1.0.1")
@@ -576,8 +604,12 @@ class GitRepositoryTests: XCTestCase {
 
             // Add something to bar.
             try localFileSystem.writeFileContents(barPath.appending(component: "bar.txt"), bytes: "hello")
-            // Add a submodule too to check for recusive submodules.
-            try systemQuietly([Git.tool, "-C", barPath.pathString, "submodule", "add", bazPath.pathString, "baz"])
+            // Add a submodule too to check for recursive submodules.
+            try Process.checkNonZeroExit(
+                args: Git.tool, "-C", barPath.pathString, "submodule", "add", bazPath.pathString, "baz",
+                environment: Git.environment
+            )
+
             try bar.stageEverything()
             try bar.commit()
 
@@ -595,7 +627,7 @@ class GitRepositoryTests: XCTestCase {
             XCTAssertFileExists(fooWorkingPath.appending(components: "bar", "bar.txt"))
             XCTAssertFileExists(fooWorkingPath.appending(components: "bar", "baz", "hello.txt"))
 
-            // Sanity check.
+            // Double check.
             try fooWorkingRepo.checkout(tag: "1.0.0")
             XCTAssertNoSuchPath(fooWorkingPath.appending(components: "bar"))
         }
@@ -675,11 +707,11 @@ class GitRepositoryTests: XCTestCase {
             try makeDirectories(testRepoPath)
             initGitRepo(testRepoPath)
             let repo = GitRepository(path: testRepoPath)
-            
+
             // Create a `newMain` branch and remove `main`.
             try repo.checkout(newBranch: "newMain")
             try systemQuietly([Git.tool, "-C", testRepoPath.pathString, "branch", "-D", "main"])
-            
+
             // Change the branch name to something non-existent.
             try systemQuietly([Git.tool, "-C", testRepoPath.pathString, "symbolic-ref", "HEAD", "refs/heads/_non_existent_branch_"])
 
@@ -704,5 +736,4 @@ class GitRepositoryTests: XCTestCase {
             XCTAssertNoThrow(try checkoutRepo.getCurrentRevision())
         }
     }
-
 }
