@@ -1,12 +1,14 @@
-/*
- This source file is part of the Swift.org open source project
-
- Copyright (c) 2020 - 2021 Apple Inc. and the Swift project authors
- Licensed under Apache License v2.0 with Runtime Library Exception
-
- See http://swift.org/LICENSE.txt for license information
- See http://swift.org/CONTRIBUTORS.txt for Swift project authors
- */
+//===----------------------------------------------------------------------===//
+//
+// This source file is part of the Swift open source project
+//
+// Copyright (c) 2020-2021 Apple Inc. and the Swift project authors
+// Licensed under Apache License v2.0 with Runtime Library Exception
+//
+// See http://swift.org/LICENSE.txt for license information
+// See http://swift.org/CONTRIBUTORS.txt for the list of Swift project authors
+//
+//===----------------------------------------------------------------------===//
 
 import Basics
 import PackageGraph
@@ -17,6 +19,21 @@ import TSCBasic
 import Workspace
 import XCTest
 
+extension String {
+    fileprivate func nativePathString(escaped: Bool) -> String {
+#if _runtime(_ObjC)
+        return self
+#else
+        let fsr = self.fileSystemRepresentation
+        defer { fsr.deallocate() }
+        if escaped {
+            return String(cString: fsr).replacingOccurrences(of: "\\", with: "\\\\")
+        }
+        return String(cString: fsr)
+#endif
+    }
+}
+
 class ManifestSourceGenerationTests: XCTestCase {
 
     /// Private function that writes the contents of a package manifest to a temporary package directory and then loads it, then serializes the loaded manifest back out again and loads it once again, after which it compares that no information was lost. Return the source of the newly generated manifest.
@@ -32,28 +49,32 @@ class ManifestSourceGenerationTests: XCTestCase {
             let observability = ObservabilitySystem.makeForTesting()
 
             // Write the original manifest file contents, and load it.
-            try fs.writeFileContents(packageDir.appending(component: Manifest.filename), bytes: ByteString(encodingAsUTF8: manifestContents))
-            let manifestLoader = ManifestLoader(toolchain: ToolchainConfiguration.default)
+            let manifestPath = packageDir.appending(component: Manifest.filename)
+            try fs.writeFileContents(manifestPath, string: manifestContents)
+            let manifestLoader = ManifestLoader(toolchain: try UserToolchain.default)
             let identityResolver = DefaultIdentityResolver()
             let manifest = try tsc_await {
-                manifestLoader.load(at: packageDir,
-                                    packageIdentity: .plain("Root"),
-                                    packageKind: .root(packageDir),
-                                    packageLocation: packageDir.pathString,
-                                    version: nil,
-                                    revision: nil,
-                                    toolsVersion: toolsVersion,
-                                    identityResolver: identityResolver,
-                                    fileSystem: fs,
-                                    observabilityScope: observability.topScope,
-                                    on: .global(),
-                                    completion: $0)
+                manifestLoader.load(
+                    manifestPath: manifestPath,
+                    manifestToolsVersion: toolsVersion,
+                    packageIdentity: .plain("Root"),
+                    packageKind: .root(packageDir),
+                    packageLocation: packageDir.pathString,
+                    packageVersion: nil,
+                    identityResolver: identityResolver,
+                    fileSystem: fs,
+                    observabilityScope: observability.topScope,
+                    delegateQueue: .sharedConcurrent,
+                    callbackQueue: .sharedConcurrent,
+                    completion: $0
+                )
             }
 
             XCTAssertNoDiagnostics(observability.diagnostics)
 
             // Generate source code for the loaded manifest,
             let newContents = try manifest.generateManifestFileContents(
+                packageDirectory: packageDir,
                 toolsVersionHeaderComment: toolsVersionHeaderComment,
                 additionalImportModuleNames: additionalImportModuleNames)
 
@@ -62,20 +83,22 @@ class ManifestSourceGenerationTests: XCTestCase {
             XCTAssertMatch(newContents, .prefix("// swift-tools-version:\(versionSpacing)\(toolsVersion.major).\(toolsVersion.minor)"))
 
             // Write out the generated manifest to replace the old manifest file contents, and load it again.
-            try fs.writeFileContents(packageDir.appending(component: Manifest.filename), bytes: ByteString(encodingAsUTF8: newContents))
+            try fs.writeFileContents(manifestPath, string: newContents)
             let newManifest = try tsc_await {
-                manifestLoader.load(at: packageDir,
-                                    packageIdentity: .plain("Root"),
-                                    packageKind: .root(packageDir),
-                                    packageLocation: packageDir.pathString,
-                                    version: nil,
-                                    revision: nil,
-                                    toolsVersion: toolsVersion,
-                                    identityResolver: identityResolver,
-                                    fileSystem: fs,
-                                    observabilityScope: observability.topScope,
-                                    on: .global(),
-                                    completion: $0)
+                manifestLoader.load(
+                    manifestPath: manifestPath,
+                    manifestToolsVersion: toolsVersion,
+                    packageIdentity: .plain("Root"),
+                    packageKind: .root(packageDir),
+                    packageLocation: packageDir.pathString,
+                    packageVersion: nil,
+                    identityResolver: identityResolver,
+                    fileSystem: fs,
+                    observabilityScope: observability.topScope,
+                    delegateQueue: .sharedConcurrent,
+                    callbackQueue: .sharedConcurrent,
+                    completion: $0
+                )
             }
 
             XCTAssertNoDiagnostics(observability.diagnostics)
@@ -229,27 +252,39 @@ class ManifestSourceGenerationTests: XCTestCase {
             // swift-tools-version:5.4
             import PackageDescription
 
+            #if os(Windows)
+            let absolutePath = "file:///C:/Users/user/SourceCache/path/to/MyPkg16"
+            #else
+            let absolutePath = "file:///path/to/MyPkg16"
+            #endif
+
             let package = Package(
                 name: "MyPackage",
                 dependencies: [
-                   .package(url: "/foo1", from: "1.0.0"),
-                   .package(url: "/foo2", .revision("58e9de4e7b79e67c72a46e164158e3542e570ab6")),
-                   .package(path: "../foo3"),
-                   .package(path: "/path/to/foo4"),
-                   .package(url: "/foo5", .exact("1.2.3")),
-                   .package(url: "/foo6", "1.2.3"..<"2.0.0"),
-                   .package(url: "/foo7", .branch("master")),
-                   .package(url: "/foo8", .upToNextMinor(from: "1.3.4")),
-                   .package(url: "/foo9", .upToNextMajor(from: "1.3.4")),
-                   .package(path: "~/path/to/foo10"),
-                   .package(path: "~foo11"),
-                   .package(path: "~/path/to/~/foo12"),
+                   .package(url: "https://example.com/MyPkg1", from: "1.0.0"),
+                   .package(url: "https://example.com/MyPkg2", .revision("58e9de4e7b79e67c72a46e164158e3542e570ab6")),
+                   .package(url: "https://example.com/MyPkg5", .exact("1.2.3")),
+                   .package(url: "https://example.com/MyPkg6", "1.2.3"..<"2.0.0"),
+                   .package(url: "https://example.com/MyPkg7", .branch("main")),
+                   .package(url: "https://example.com/MyPkg8", .upToNextMinor(from: "1.3.4")),
+                   .package(url: "ssh://git@example.com/MyPkg9", .branch("my branch with spaces")),
+                   .package(url: "../MyPkg10", from: "0.1.0"),
+                   .package(path: "../MyPkg11"),
+                   .package(path: "packages/path/to/MyPkg12"),
+                   .package(path: "~/path/to/MyPkg13"),
+                   .package(path: "~MyPkg14"),
+                   .package(path: "~/path/to/~/MyPkg15"),
                    .package(path: "~"),
-                   .package(path: "file:///path/to/foo13"),
+                   .package(path: absolutePath),
                 ]
             )
             """
-        try testManifestWritingRoundTrip(manifestContents: manifestContents, toolsVersion: .v5_3)
+        let newContents = try testManifestWritingRoundTrip(manifestContents: manifestContents, toolsVersion: .v5_3)
+
+        // Check some things about the contents of the manifest.
+        XCTAssertTrue(newContents.contains("url: \"\("../MyPkg10".nativePathString(escaped: true))\""), newContents)
+        XCTAssertTrue(newContents.contains("path: \"\("../MyPkg11".nativePathString(escaped: true))\""), newContents)
+        XCTAssertTrue(newContents.contains("path: \"\("packages/path/to/MyPkg12".nativePathString(escaped: true))"), newContents)
     }
 
     func testResources() throws {
@@ -390,20 +425,21 @@ class ManifestSourceGenerationTests: XCTestCase {
 
     func testCustomProductSourceGeneration() throws {
         // Create a manifest containing a product for which we'd like to do custom source fragment generation.
+        let packageDir = AbsolutePath(path: "/tmp/MyLibrary")
         let manifest = Manifest(
             displayName: "MyLibrary",
-            path: AbsolutePath("/tmp/MyLibrary/Package.swift"),
-            packageKind: .root(AbsolutePath("/tmp/MyLibrary")),
-            packageLocation: "/tmp/MyLibrary",
+            path: packageDir.appending(component: "Package.swift"),
+            packageKind: .root(AbsolutePath(path: "/tmp/MyLibrary")),
+            packageLocation: packageDir.pathString,
             platforms: [],
             toolsVersion: .v5_5,
             products: [
-                .init(name: "Foo", type: .library(.static), targets: ["Bar"])
+                try .init(name: "Foo", type: .library(.static), targets: ["Bar"])
             ]
         )
 
         // Generate the manifest contents, using a custom source generator for the product type.
-        let contents = manifest.generateManifestFileContents(customProductTypeSourceGenerator: { product in
+        let contents = manifest.generateManifestFileContents(packageDirectory: packageDir, customProductTypeSourceGenerator: { product in
             // This example handles library types in a custom way, for testing purposes.
             var params: [SourceCodeFragment] = []
             params.append(SourceCodeFragment(key: "name", string: product.name))
@@ -425,4 +461,48 @@ class ManifestSourceGenerationTests: XCTestCase {
         // Check that we generated what we expected.
         XCTAssertTrue(contents.contains(".library(name: \"Foo\", targets: [\"Bar\"], type: .static)"), "contents: \(contents)")
     }
+
+    func testModuleAliasGeneration() throws {
+        let manifest = Manifest.createRootManifest(
+            name: "thisPkg",
+            path: .init(path: "/thisPkg"),
+            toolsVersion: .v5_7,
+            dependencies: [
+                .localSourceControl(path: .init(path: "/fooPkg"), requirement: .upToNextMajor(from: "1.0.0")),
+                .localSourceControl(path: .init(path: "/barPkg"), requirement: .upToNextMajor(from: "2.0.0")),
+            ],
+            targets: [
+                try TargetDescription(name: "exe",
+                                  dependencies: ["Logging",
+                                                 .product(name: "Foo",
+                                                          package: "fooPkg",
+                                                          moduleAliases: ["Logging": "FooLogging"]
+                                                         ),
+                                                 .product(name: "Bar",
+                                                          package: "barPkg",
+                                                          moduleAliases: ["Logging": "BarLogging"]
+                                                         )
+                                                ]),
+                try TargetDescription(name: "Logging", dependencies: []),
+            ])
+        let contents = try manifest.generateManifestFileContents(packageDirectory: manifest.path.parentDirectory)
+        let parts =
+        """
+            dependencies: [
+                "Logging",
+                .product(name: "Foo", package: "fooPkg", moduleAliases: [
+                    "Logging": "FooLogging"
+                ]),
+                .product(name: "Bar", package: "barPkg", moduleAliases: [
+                    "Logging": "BarLogging"
+                ])
+            ]
+        """
+        let trimmedContents = contents.components(separatedBy: CharacterSet.whitespacesAndNewlines)
+        let trimmedParts = parts.components(separatedBy: CharacterSet.whitespacesAndNewlines)
+        let isContained = trimmedParts.allSatisfy(trimmedContents.contains(_:))
+        XCTAssertTrue(isContained)
+
+        try testManifestWritingRoundTrip(manifestContents: contents, toolsVersion: .v5_8)
+    }
 }
