@@ -1,12 +1,14 @@
-/*
- This source file is part of the Swift.org open source project
-
- Copyright (c) 2014 - 2020 Apple Inc. and the Swift project authors
- Licensed under Apache License v2.0 with Runtime Library Exception
-
- See http://swift.org/LICENSE.txt for license information
- See http://swift.org/CONTRIBUTORS.txt for Swift project authors
- */
+//===----------------------------------------------------------------------===//
+//
+// This source file is part of the Swift open source project
+//
+// Copyright (c) 2014-2020 Apple Inc. and the Swift project authors
+// Licensed under Apache License v2.0 with Runtime Library Exception
+//
+// See http://swift.org/LICENSE.txt for license information
+// See http://swift.org/CONTRIBUTORS.txt for the list of Swift project authors
+//
+//===----------------------------------------------------------------------===//
 
 import Basics
 import Foundation
@@ -16,10 +18,11 @@ import PackageModel
 import SourceControl
 import SPMTestSupport
 import TSCBasic
-import TSCUtility
 @testable import Workspace
 import XCTest
 
+import enum TSCUtility.Git
+
 private class MockRepository: Repository {
     /// The fake location of the repository.
     let location: RepositorySpecifier.Location
@@ -38,7 +41,7 @@ private class MockRepository: Repository {
         self.versions = versions
     }
 
-    init(fs: FileSystem, url: Foundation.URL, versions: [Version: Manifest]) {
+    init(fs: FileSystem, url: URL, versions: [Version: Manifest]) {
         self.fs = fs
         self.location = .url(url)
         self.packageRef = .remoteSourceControl(identity: PackageIdentity(url: url), url: url)
@@ -89,6 +92,7 @@ private class MockRepository: Repository {
 private class MockRepositories: RepositoryProvider {
     /// The known repositories, as a map of URL to repository.
     let repositories: [RepositorySpecifier.Location: MockRepository]
+    var fetchRepositories = ThreadSafeKeyValueStore<RepositorySpecifier, AbsolutePath>()
 
     /// A mock manifest loader for all repositories.
     let manifestLoader: MockManifestLoader
@@ -109,8 +113,12 @@ private class MockRepositories: RepositoryProvider {
     }
 
     func fetch(repository: RepositorySpecifier, to path: AbsolutePath, progressHandler: FetchProgress.Handler? = nil) throws {
-        // No-op.
         assert(self.repositories.index(forKey: repository.location) != nil)
+        self.fetchRepositories[repository] = path
+    }
+
+    func repositoryExists(at path: AbsolutePath) throws -> Bool {
+        self.fetchRepositories.get().contains(where: { key, value in value == path })
     }
 
     func copy(from sourcePath: AbsolutePath, to destinationPath: AbsolutePath) throws {
@@ -140,24 +148,26 @@ private class MockRepositories: RepositoryProvider {
     func isValidRefFormat(_ ref: String) -> Bool {
         return true
     }
+
+    func cancel(deadline: DispatchTime) throws {
+        // noop
+    }
 }
 
-private class MockResolverDelegate: RepositoryManagerDelegate {
-    var fetched = [RepositorySpecifier]()
+private class MockResolverDelegate: RepositoryManager.Delegate {
+    var fetched = ThreadSafeArrayStore<RepositorySpecifier>()
 
-    func fetchingWillBegin(handle: RepositoryManager.RepositoryHandle, fetchDetails: RepositoryManager.FetchDetails?) {
-        self.fetched += [handle.repository]
+    func willFetch(package: PackageIdentity, repository: RepositorySpecifier, details: RepositoryManager.FetchDetails) {
+        self.fetched.append(repository)
     }
 
-    func fetchingRepository(from repository: String, objectsFetched: Int, totalObjectsToFetch: Int) {
-    }
+    func fetching(package: PackageIdentity, repository: RepositorySpecifier, objectsFetched: Int, totalObjectsToFetch: Int) {}
 
-    func fetchingDidFinish(handle: RepositoryManager.RepositoryHandle, fetchDetails: RepositoryManager.FetchDetails?, error: Swift.Error?, duration: DispatchTimeInterval) {
-    }
+    func didFetch(package: PackageIdentity, repository: RepositorySpecifier, result: Result<RepositoryManager.FetchDetails, Error>, duration: DispatchTimeInterval) {}
 
-    func handleWillUpdate(handle: RepositoryManager.RepositoryHandle) {}
+    func willUpdate(package: PackageIdentity, repository: RepositorySpecifier) {}
 
-    func handleDidUpdate(handle: RepositoryManager.RepositoryHandle, duration: DispatchTimeInterval) {}
+    func didUpdate(package: PackageIdentity, repository: RepositorySpecifier, duration: DispatchTimeInterval) {}
 }
 
 // Some handy versions & ranges.
@@ -179,7 +189,7 @@ class SourceControlPackageContainerTests: XCTestCase {
         let specifier = RepositorySpecifier(path: repoPath)
         let repo = InMemoryGitRepository(path: repoPath, fs: fs)
         try repo.createDirectory(repoPath, recursive: true)
-        try repo.writeFileContents(filePath, bytes: ByteString(encodingAsUTF8: "// swift-tools-version:\(ToolsVersion.currentToolsVersion)\n"))
+        try repo.writeFileContents(filePath, bytes: ByteString(encodingAsUTF8: "// swift-tools-version:\(ToolsVersion.current)\n"))
         try repo.commit()
         try repo.tag(name: "v1.0.0")
         try repo.tag(name: "v1.0.1")
@@ -199,7 +209,7 @@ class SourceControlPackageContainerTests: XCTestCase {
             delegate: MockResolverDelegate()
         )
 
-        let provider = try Workspace(
+        let provider = try Workspace._init(
             fileSystem: fs,
             location: .init(forRootPackage: repoPath, fileSystem: fs),
             customManifestLoader: MockManifestLoader(manifests: [:]),
@@ -208,7 +218,7 @@ class SourceControlPackageContainerTests: XCTestCase {
 
         let ref = PackageReference.localSourceControl(identity: PackageIdentity(path: repoPath), path: repoPath)
         let container = try provider.getContainer(for: ref, skipUpdate: false)
-        let v = try container.toolsVersionsAppropriateVersionsDescending().map { $0 }
+        let v = try container.toolsVersionsAppropriateVersionsDescending()
         XCTAssertEqual(v, ["2.0.3", "1.0.3", "1.0.2", "1.0.1", "1.0.0"])
     }
 
@@ -252,7 +262,7 @@ class SourceControlPackageContainerTests: XCTestCase {
         )
 
         func createProvider(_ currentToolsVersion: ToolsVersion) throws -> PackageContainerProvider {
-            return try Workspace(
+            return try Workspace._init(
                 fileSystem: fs,
                 location: .init(forRootPackage: repoPath, fileSystem: fs),
                 customToolsVersion: currentToolsVersion,
@@ -265,7 +275,7 @@ class SourceControlPackageContainerTests: XCTestCase {
             let provider = try createProvider(ToolsVersion(version: "4.0.0"))
             let ref = PackageReference.localSourceControl(identity: PackageIdentity(path: repoPath), path: repoPath)
             let container = try provider.getContainer(for: ref, skipUpdate: false)
-            let v = try container.toolsVersionsAppropriateVersionsDescending().map { $0 }
+            let v = try container.toolsVersionsAppropriateVersionsDescending()
             XCTAssertEqual(v, ["1.0.1"])
         }
 
@@ -274,7 +284,7 @@ class SourceControlPackageContainerTests: XCTestCase {
             let ref = PackageReference.localSourceControl(identity: PackageIdentity(path: repoPath), path: repoPath)
             let container = try provider.getContainer(for: ref, skipUpdate: false) as! SourceControlPackageContainer
             XCTAssertTrue(container.validToolsVersionsCache.isEmpty)
-            let v = try container.toolsVersionsAppropriateVersionsDescending().map { $0 }
+            let v = try container.toolsVersionsAppropriateVersionsDescending()
             XCTAssertEqual(container.validToolsVersionsCache["1.0.0"], false)
             XCTAssertEqual(container.validToolsVersionsCache["1.0.1"], true)
             XCTAssertEqual(container.validToolsVersionsCache["1.0.2"], true)
@@ -286,7 +296,7 @@ class SourceControlPackageContainerTests: XCTestCase {
             let provider = try createProvider(ToolsVersion(version: "3.0.0"))
             let ref = PackageReference.localSourceControl(identity: PackageIdentity(path: repoPath), path: repoPath)
             let container = try provider.getContainer(for: ref, skipUpdate: false)
-            let v = try container.toolsVersionsAppropriateVersionsDescending().map { $0 }
+            let v = try container.toolsVersionsAppropriateVersionsDescending()
             XCTAssertEqual(v, [])
         }
 
@@ -314,7 +324,7 @@ class SourceControlPackageContainerTests: XCTestCase {
         let specifier = RepositorySpecifier(path: repoPath)
         let repo = InMemoryGitRepository(path: repoPath, fs: fs)
         try repo.createDirectory(repoPath, recursive: true)
-        try repo.writeFileContents(filePath, bytes: ByteString(encodingAsUTF8: "// swift-tools-version:\(ToolsVersion.currentToolsVersion)\n"))
+        try repo.writeFileContents(filePath, bytes: ByteString(encodingAsUTF8: "// swift-tools-version:\(ToolsVersion.current)\n"))
         try repo.commit()
         try repo.tag(name: "1.0.0-alpha.1")
         try repo.tag(name: "1.0.0-beta.1")
@@ -336,7 +346,7 @@ class SourceControlPackageContainerTests: XCTestCase {
             delegate: MockResolverDelegate()
         )
 
-        let provider = try Workspace(
+        let provider = try Workspace._init(
             fileSystem: fs,
             location: .init(forRootPackage: repoPath, fileSystem: fs),
             customManifestLoader: MockManifestLoader(manifests: [:]),
@@ -345,7 +355,7 @@ class SourceControlPackageContainerTests: XCTestCase {
 
         let ref = PackageReference.localSourceControl(identity: PackageIdentity(path: repoPath), path: repoPath)
         let container = try provider.getContainer(for: ref, skipUpdate: false)
-        let v = try container.toolsVersionsAppropriateVersionsDescending().map { $0 }
+        let v = try container.toolsVersionsAppropriateVersionsDescending()
         XCTAssertEqual(v, ["1.0.4-alpha", "1.0.2-dev.2", "1.0.2-dev", "1.0.1", "1.0.0", "1.0.0-beta.1", "1.0.0-alpha.1"])
     }
 
@@ -358,7 +368,7 @@ class SourceControlPackageContainerTests: XCTestCase {
         let specifier = RepositorySpecifier(path: repoPath)
         let repo = InMemoryGitRepository(path: repoPath, fs: fs)
         try repo.createDirectory(repoPath, recursive: true)
-        try repo.writeFileContents(filePath, bytes: ByteString(encodingAsUTF8: "// swift-tools-version:\(ToolsVersion.currentToolsVersion)\n"))
+        try repo.writeFileContents(filePath, bytes: ByteString(encodingAsUTF8: "// swift-tools-version:\(ToolsVersion.current)\n"))
         try repo.commit()
         try repo.tag(name: "v1.0.0")
         try repo.tag(name: "1.0.0")
@@ -385,7 +395,7 @@ class SourceControlPackageContainerTests: XCTestCase {
             delegate: MockResolverDelegate()
         )
 
-        let provider = try Workspace(
+        let provider = try Workspace._init(
             fileSystem: fs,
             location: .init(forRootPackage: repoPath, fileSystem: fs),
             customManifestLoader: MockManifestLoader(manifests: [:]),
@@ -393,7 +403,7 @@ class SourceControlPackageContainerTests: XCTestCase {
         )
         let ref = PackageReference.localSourceControl(identity: PackageIdentity(path: repoPath), path: repoPath)
         let container = try provider.getContainer(for: ref, skipUpdate: false)
-        let v = try container.toolsVersionsAppropriateVersionsDescending().map { $0 }
+        let v = try container.toolsVersionsAppropriateVersionsDescending()
         XCTAssertEqual(v, ["2.0.1", "1.3.0", "1.2.0", "1.1.0", "1.0.4", "1.0.2", "1.0.1", "1.0.0"])
     }
 
@@ -404,13 +414,13 @@ class SourceControlPackageContainerTests: XCTestCase {
 #endif
 
         let dependencies: [PackageDependency] = [
-            .localSourceControl(path: .init("/Bar1"), requirement: .upToNextMajor(from: "1.0.0")),
-            .localSourceControl(path: .init("/Bar2"), requirement: .upToNextMajor(from: "1.0.0")),
-            .localSourceControl(path: .init("/Bar3"), requirement: .upToNextMajor(from: "1.0.0")),
+            .localSourceControl(path: .init(path: "/Bar1"), requirement: .upToNextMajor(from: "1.0.0")),
+            .localSourceControl(path: .init(path: "/Bar2"), requirement: .upToNextMajor(from: "1.0.0")),
+            .localSourceControl(path: .init(path: "/Bar3"), requirement: .upToNextMajor(from: "1.0.0")),
         ]
 
         let products = [
-            ProductDescription(name: "Foo", type: .library(.automatic), targets: ["Foo1"]),
+            try ProductDescription(name: "Foo", type: .library(.automatic), targets: ["Foo1"]),
         ]
 
         let targets = [
@@ -447,7 +457,7 @@ class SourceControlPackageContainerTests: XCTestCase {
         do {
             let manifest = Manifest.createRootManifest(
                 name: "Foo",
-                path: .init("/Foo"),
+                path: .init(path: "/Foo"),
                 toolsVersion: .v5,
                 dependencies: dependencies,
                 products: products,
@@ -469,7 +479,7 @@ class SourceControlPackageContainerTests: XCTestCase {
         do {
             let manifest = Manifest.createFileSystemManifest(
                 name: "Foo",
-                path: .init("/Foo"),
+                path: .init(path: "/Foo"),
                 toolsVersion: .v5,
                 dependencies: dependencies,
                 products: products,
@@ -491,7 +501,7 @@ class SourceControlPackageContainerTests: XCTestCase {
         do {
             let manifest = Manifest.createRootManifest(
                 name: "Foo",
-                path: .init("/Foo"),
+                path: .init(path: "/Foo"),
                 toolsVersion: .v5_2,
                 dependencies: dependencies,
                 products: products,
@@ -513,7 +523,7 @@ class SourceControlPackageContainerTests: XCTestCase {
         do {
             let manifest = Manifest.createFileSystemManifest(
                 name: "Foo",
-                path: .init("/Foo"),
+                path: .init(path: "/Foo"),
                 toolsVersion: .v5_2,
                 dependencies: dependencies,
                 products: products,
@@ -554,7 +564,12 @@ class SourceControlPackageContainerTests: XCTestCase {
 
             // Create a repository manager for it.
             let repoProvider = GitRepositoryProvider()
-            let repositoryManager = RepositoryManager(fileSystem: localFileSystem, path: packageDir, provider: repoProvider, delegate: nil)
+            let repositoryManager = RepositoryManager(
+                fileSystem: localFileSystem,
+                path: packageDir,
+                provider: repoProvider,
+                delegate: .none
+            )
 
             // Create a container provider, configured with a mock manifest loader that will return the package manifest.
             let manifest = Manifest.createRootManifest(
@@ -564,7 +579,7 @@ class SourceControlPackageContainerTests: XCTestCase {
                     try TargetDescription(name: packageDir.basename, path: packageDir.pathString),
                 ]
             )
-            let containerProvider = try Workspace(
+            let containerProvider = try Workspace._init(
                 fileSystem: localFileSystem,
                 location: .init(forRootPackage: packageDir, fileSystem: localFileSystem),
                 customManifestLoader: MockManifestLoader(manifests: [.init(url: packageDir.pathString, version: nil): manifest]),
@@ -580,7 +595,7 @@ class SourceControlPackageContainerTests: XCTestCase {
             catch let error as SourceControlPackageContainer.GetDependenciesError {
                 // We expect to get an error message that mentions main.
                 XCTAssertMatch(error.description, .and(.prefix("could not find a branch named ‘master’"), .suffix("(did you mean ‘main’?)")))
-                XCTAssertMatch(error.repository.description, .suffix("/SomePackage"))
+                XCTAssertMatch(error.repository.description, .suffix("SomePackage"))
                 XCTAssertMatch(error.reference, "master")
             }
 
@@ -589,18 +604,17 @@ class SourceControlPackageContainerTests: XCTestCase {
             catch let error as SourceControlPackageContainer.GetDependenciesError {
                 // We expect to get an error message about the specific commit.
                 XCTAssertMatch(error.description, .prefix("could not find the commit 535f4cb5b4a0872fa691473e82d7b27b9894df00"))
-                XCTAssertMatch(error.repository.description, .suffix("/SomePackage"))
+                XCTAssertMatch(error.repository.description, .suffix("SomePackage"))
                 XCTAssertMatch(error.reference, "535f4cb5b4a0872fa691473e82d7b27b9894df00")
             }
         }
     }
 
+    // From rdar://problem/65284674
+    // RepositoryPackageContainer used to erroneously cache dependencies based only on version,
+    // storing the result of the first product filter and then continually returning it for other filters too.
+    // This lead to corrupt graph states.
     func testRepositoryPackageContainerCache() throws {
-        // From rdar://problem/65284674
-        // RepositoryPackageContainer used to erroneously cache dependencies based only on version,
-        // storing the result of the first product filter and then continually returning it for other filters too.
-        // This lead to corrupt graph states.
-
         try testWithTemporaryDirectory { temporaryDirectory in
             let packageDirectory = temporaryDirectory.appending(component: "Package")
             try localFileSystem.createDirectory(packageDirectory)
@@ -619,7 +633,7 @@ class SourceControlPackageContainerTests: XCTestCase {
                 fileSystem: localFileSystem,
                 path: packageDirectory,
                 provider: repositoryProvider,
-                delegate: nil
+                delegate: .none
             )
 
             let version = Version(1, 0, 0)
@@ -629,12 +643,14 @@ class SourceControlPackageContainerTests: XCTestCase {
                 toolsVersion: .v5_2,
                 dependencies: [
                     .localSourceControl(
-                        path: .init("/Somewhere/Dependency"),
+                        path: .init(path: "/Somewhere/Dependency"),
                         requirement: .exact(version),
                         productFilter: .specific([])
                     )
                 ],
-                products: [ProductDescription(name: "Product", type: .library(.automatic), targets: ["Target"])],
+                products: [
+                    try ProductDescription(name: "Product", type: .library(.automatic), targets: ["Target"])
+                ],
                 targets: [
                     try TargetDescription(
                         name: "Target",
@@ -642,7 +658,7 @@ class SourceControlPackageContainerTests: XCTestCase {
                     ),
                 ]
             )
-            let containerProvider = try Workspace(
+            let containerProvider = try Workspace._init(
                 fileSystem: localFileSystem,
                 location: .init(forRootPackage: packageDirectory, fileSystem: localFileSystem),
                 customManifestLoader: MockManifestLoader(
@@ -669,3 +685,24 @@ extension PackageContainerProvider {
         try tsc_await { self.getContainer(for: package, skipUpdate: skipUpdate, observabilityScope: ObservabilitySystem.NOOP, on: .global(), completion: $0)  }
     }
 }
+
+extension RepositoryManager {
+    fileprivate convenience init(
+        fileSystem: FileSystem,
+        path: AbsolutePath,
+        provider: RepositoryProvider,
+        cachePath: AbsolutePath? =  .none,
+        cacheLocalPackages: Bool = false,
+        delegate: RepositoryManagerDelegate? = .none
+    ) {
+        self.init(
+            fileSystem: fileSystem,
+            path: path,
+            provider: provider,
+            cachePath: cachePath,
+            cacheLocalPackages: cacheLocalPackages,
+            initializationWarningHandler: { _ in },
+            delegate: delegate
+        )
+    }
+}
