@@ -1,25 +1,28 @@
-/*
- This source file is part of the Swift.org open source project
+//===----------------------------------------------------------------------===//
+//
+// This source file is part of the Swift open source project
+//
+// Copyright (c) 2021 Apple Inc. and the Swift project authors
+// Licensed under Apache License v2.0 with Runtime Library Exception
+//
+// See http://swift.org/LICENSE.txt for license information
+// See http://swift.org/CONTRIBUTORS.txt for the list of Swift project authors
+//
+//===----------------------------------------------------------------------===//
 
- Copyright (c) 2021 Apple Inc. and the Swift project authors
- Licensed under Apache License v2.0 with Runtime Library Exception
-
- See http://swift.org/LICENSE.txt for license information
- See http://swift.org/CONTRIBUTORS.txt for Swift project authors
- */
-
-import TSCBasic
+import Basics
 @testable import Workspace
+import TSCBasic
 import XCTest
 
 final class WorkspaceStateTests: XCTestCase {
     func testV4Format() throws {
         let fs = InMemoryFileSystem()
 
-        let buildDir = AbsolutePath("/.build")
-        let statePath = buildDir.appending(component: "workspace-state.json")
-
+        let buildDir = AbsolutePath(path: "/.build")
         try fs.createDirectory(buildDir, recursive: true)
+
+        let statePath = buildDir.appending(component: "workspace-state.json")
         try fs.writeFileContents(statePath) {
             """
             {
@@ -81,7 +84,7 @@ final class WorkspaceStateTests: XCTestCase {
             """
         }
 
-        let state = WorkspaceState(dataPath: buildDir, fileSystem: fs)
+        let state = WorkspaceState(fileSystem: fs, storageDirectory: buildDir)
         XCTAssertTrue(state.dependencies.contains(where: { $0.packageRef.identity == .plain("yams") }))
         XCTAssertTrue(state.dependencies.contains(where: { $0.packageRef.identity == .plain("swift-tools-support-core") }))
         XCTAssertTrue(state.dependencies.contains(where: { $0.packageRef.identity == .plain("swift-argument-parser") }))
@@ -90,10 +93,10 @@ final class WorkspaceStateTests: XCTestCase {
     func testV4FormatWithPath() throws {
         let fs = InMemoryFileSystem()
 
-        let buildDir = AbsolutePath("/.build")
-        let statePath = buildDir.appending(component: "workspace-state.json")
-
+        let buildDir = AbsolutePath(path: "/.build")
         try fs.createDirectory(buildDir, recursive: true)
+
+        let statePath = buildDir.appending(component: "workspace-state.json")
         try fs.writeFileContents(statePath) {
             """
             {
@@ -155,7 +158,7 @@ final class WorkspaceStateTests: XCTestCase {
             """
         }
 
-        let state = WorkspaceState(dataPath: buildDir, fileSystem: fs)
+        let state = WorkspaceState(fileSystem: fs, storageDirectory: buildDir)
         XCTAssertTrue(state.dependencies.contains(where: { $0.packageRef.identity == .plain("yams") }))
         XCTAssertTrue(state.dependencies.contains(where: { $0.packageRef.identity == .plain("swift-tools-support-core") }))
         XCTAssertTrue(state.dependencies.contains(where: { $0.packageRef.identity == .plain("swift-argument-parser") }))
@@ -164,10 +167,10 @@ final class WorkspaceStateTests: XCTestCase {
     func testV5Format() throws {
         let fs = InMemoryFileSystem()
 
-        let buildDir = AbsolutePath("/.build")
-        let statePath = buildDir.appending(component: "workspace-state.json")
-
+        let buildDir = AbsolutePath(path: "/.build")
         try fs.createDirectory(buildDir, recursive: true)
+
+        let statePath = buildDir.appending(component: "workspace-state.json")
         try fs.writeFileContents(statePath) {
             """
             {
@@ -229,7 +232,7 @@ final class WorkspaceStateTests: XCTestCase {
             """
         }
 
-        let state = WorkspaceState(dataPath: buildDir, fileSystem: fs)
+        let state = WorkspaceState(fileSystem: fs, storageDirectory: buildDir)
         XCTAssertTrue(state.dependencies.contains(where: { $0.packageRef.identity == .plain("yams") }))
         XCTAssertTrue(state.dependencies.contains(where: { $0.packageRef.identity == .plain("swift-tools-support-core") }))
         XCTAssertTrue(state.dependencies.contains(where: { $0.packageRef.identity == .plain("swift-argument-parser") }))
@@ -238,10 +241,10 @@ final class WorkspaceStateTests: XCTestCase {
     func testSavedDependenciesAreSorted() throws {
         let fs = InMemoryFileSystem()
 
-        let buildDir = AbsolutePath("/.build")
-        let statePath = buildDir.appending(component: "workspace-state.json")
-
+        let buildDir = AbsolutePath(path: "/.build")
         try fs.createDirectory(buildDir, recursive: true)
+
+        let statePath = buildDir.appending(component: "workspace-state.json")
         try fs.writeFileContents(statePath) {
             """
             {
@@ -289,10 +292,10 @@ final class WorkspaceStateTests: XCTestCase {
             """
         }
 
-        let state = WorkspaceState(dataPath: buildDir, fileSystem: fs)
+        let state = WorkspaceState(fileSystem: fs, storageDirectory: buildDir)
         try state.save()
 
-        let serialized = try fs.readFileContents(statePath).description
+        let serialized: String = try fs.readFileContents(statePath)
 
         let argpRange = try XCTUnwrap(serialized.range(of: "swift-argument-parser"))
         let yamsRange = try XCTUnwrap(serialized.range(of: "yams"))
@@ -303,10 +306,10 @@ final class WorkspaceStateTests: XCTestCase {
     func testArtifacts() throws {
         let fs = InMemoryFileSystem()
 
-        let buildDir = AbsolutePath("/.build")
-        let statePath = buildDir.appending(component: "workspace-state.json")
-
+        let buildDir = AbsolutePath(path: "/.build")
         try fs.createDirectory(buildDir, recursive: true)
+
+        let statePath = buildDir.appending(component: "workspace-state.json")
         try fs.writeFileContents(statePath) {
             """
             {
@@ -365,7 +368,7 @@ final class WorkspaceStateTests: XCTestCase {
             """
         }
 
-        let state = WorkspaceState(dataPath: buildDir, fileSystem: fs)
+        let state = WorkspaceState(fileSystem: fs, storageDirectory: buildDir)
         XCTAssertTrue(state.artifacts.contains(where: { $0.packageRef.identity == .plain("foo") && $0.targetName == "foo" }))
         XCTAssertTrue(state.artifacts.contains(where: { $0.packageRef.identity == .plain("foo") && $0.targetName == "bar" }))
         XCTAssertTrue(state.artifacts.contains(where: { $0.packageRef.identity == .plain("bar") && $0.targetName == "bar" }))
@@ -375,10 +378,10 @@ final class WorkspaceStateTests: XCTestCase {
     func testDuplicateDependenciesDoNotCrash() throws {
         let fs = InMemoryFileSystem()
 
-        let buildDir = AbsolutePath("/.build")
-        let statePath = buildDir.appending(component: "workspace-state.json")
-
+        let buildDir = AbsolutePath(path: "/.build")
         try fs.createDirectory(buildDir, recursive: true)
+
+        let statePath = buildDir.appending(component: "workspace-state.json")
         try fs.writeFileContents(statePath) {
             """
             {
@@ -426,7 +429,7 @@ final class WorkspaceStateTests: XCTestCase {
             """
         }
 
-        let state = WorkspaceState(dataPath: buildDir, fileSystem: fs)
+        let state = WorkspaceState(fileSystem: fs, storageDirectory: buildDir)
         // empty since we have dups so we warn and fail the loading
         // TODO: test for diagnostics when we can get them from the WorkspaceState initializer
         XCTAssertTrue(state.dependencies.isEmpty)
@@ -436,10 +439,10 @@ final class WorkspaceStateTests: XCTestCase {
     func testDuplicateArtifactsDoNotCrash() throws {
         let fs = InMemoryFileSystem()
 
-        let buildDir = AbsolutePath("/.build")
-        let statePath = buildDir.appending(component: "workspace-state.json")
-
+        let buildDir = AbsolutePath(path: "/.build")
         try fs.createDirectory(buildDir, recursive: true)
+
+        let statePath = buildDir.appending(component: "workspace-state.json")
         try fs.writeFileContents(statePath) {
             """
             {
@@ -483,9 +486,15 @@ final class WorkspaceStateTests: XCTestCase {
             """
         }
 
-        let state = WorkspaceState(dataPath: buildDir, fileSystem: fs)
+        let state = WorkspaceState(fileSystem: fs, storageDirectory: buildDir)
         // empty since we have dups so we warn and fail the loading
         // TODO: test for diagnostics when we can get them from the WorkspaceState initializer
         XCTAssertTrue(state.artifacts.isEmpty)
     }
 }
+
+extension WorkspaceState {
+    fileprivate convenience init(fileSystem: FileSystem, storageDirectory: AbsolutePath) {
+        self.init(fileSystem: fileSystem, storageDirectory: storageDirectory, initializationWarningHandler: { _ in })
+    }
+}
