@@ -1,17 +1,18 @@
-/*
- This source file is part of the Swift.org open source project
-
- Copyright (c) 2014 - 2020 Apple Inc. and the Swift project authors
- Licensed under Apache License v2.0 with Runtime Library Exception
-
- See http://swift.org/LICENSE.txt for license information
- See http://swift.org/CONTRIBUTORS.txt for Swift project authors
-*/
+//===----------------------------------------------------------------------===//
+//
+// This source file is part of the Swift open source project
+//
+// Copyright (c) 2014-2020 Apple Inc. and the Swift project authors
+// Licensed under Apache License v2.0 with Runtime Library Exception
+//
+// See http://swift.org/LICENSE.txt for license information
+// See http://swift.org/CONTRIBUTORS.txt for the list of Swift project authors
+//
+//===----------------------------------------------------------------------===//
 
 import XCTest
 import Basics
 import TSCBasic
-import TSCUtility
 import PackageModel
 import SPMBuildCore
 import XCBuildSupport
@@ -22,13 +23,13 @@ class PIFTests: XCTestCase {
         PIF.Workspace(
             guid: "workspace",
             name: "MyWorkspace",
-            path: AbsolutePath("/path/to/workspace"),
+            path: AbsolutePath(path: "/path/to/workspace"),
             projects: [
                 PIF.Project(
                     guid: "project",
                     name: "MyProject",
-                    path: AbsolutePath("/path/to/workspace/project"),
-                    projectDirectory: AbsolutePath("/path/to/workspace/project"),
+                    path: AbsolutePath(path: "/path/to/workspace/project"),
+                    projectDirectory: AbsolutePath(path: "/path/to/workspace/project"),
                     developmentRegion: "fr",
                     buildConfigurations: [
                         PIF.BuildConfiguration(
@@ -122,6 +123,11 @@ class PIFTests: XCTestCase {
                                         var settings = PIF.BuildSettings()
                                         settings[.TARGET_NAME] = "MyLibrary-Debug"
                                         return settings
+                                    }(),
+                                    impartedBuildProperties: {
+                                        var settings = PIF.BuildSettings()
+                                        settings[.OTHER_CFLAGS] = ["-fmodule-map-file=modulemap", "$(inherited)"]
+                                        return PIF.ImpartedBuildProperties(settings: settings)
                                     }()
                                 ),
                                 PIF.BuildConfiguration(
@@ -131,6 +137,11 @@ class PIFTests: XCTestCase {
                                         var settings = PIF.BuildSettings()
                                         settings[.TARGET_NAME] = "MyLibrary"
                                         return settings
+                                    }(),
+                                    impartedBuildProperties: {
+                                        var settings = PIF.BuildSettings()
+                                        settings[.OTHER_CFLAGS] = ["-fmodule-map-file=modulemap", "$(inherited)"]
+                                        return PIF.ImpartedBuildProperties(settings: settings)
                                     }()
                                 ),
                             ],
@@ -147,11 +158,7 @@ class PIFTests: XCTestCase {
                                 )
                             ],
                             dependencies: [],
-                            impartedBuildSettings: {
-                                var settings = PIF.BuildSettings()
-                                settings[.OTHER_CFLAGS] = ["-fmodule-map-file=modulemap", "$(inherited)"]
-                                return settings
-                            }()
+                            impartedBuildSettings: PIF.BuildSettings()
                         ),
                         PIF.AggregateTarget(
                             guid: "aggregate-target-guid",
@@ -160,12 +167,22 @@ class PIFTests: XCTestCase {
                                 PIF.BuildConfiguration(
                                     guid: "aggregate-target-config-debug-guid",
                                     name: "Debug",
-                                    buildSettings: PIF.BuildSettings()
+                                    buildSettings: PIF.BuildSettings(),
+                                    impartedBuildProperties: {
+                                        var settings = PIF.BuildSettings()
+                                        settings[.OTHER_CFLAGS] = ["-fmodule-map-file=modulemap", "$(inherited)"]
+                                        return PIF.ImpartedBuildProperties(settings: settings)
+                                    }()
                                 ),
                                 PIF.BuildConfiguration(
                                     guid: "aggregate-target-config-release-guid",
                                     name: "Release",
-                                    buildSettings: PIF.BuildSettings()
+                                    buildSettings: PIF.BuildSettings(),
+                                    impartedBuildProperties: {
+                                        var settings = PIF.BuildSettings()
+                                        settings[.OTHER_CFLAGS] = ["-fmodule-map-file=modulemap", "$(inherited)"]
+                                        return PIF.ImpartedBuildProperties(settings: settings)
+                                    }()
                                 ),
                             ],
                             buildPhases: [],
@@ -173,11 +190,7 @@ class PIFTests: XCTestCase {
                                 .init(targetGUID: "target-lib-guid"),
                                 .init(targetGUID: "target-exe-guid"),
                             ],
-                            impartedBuildSettings: {
-                                var settings = PIF.BuildSettings()
-                                settings[.OTHER_CFLAGS] = ["-fmodule-map-file=modulemap", "$(inherited)"]
-                                return settings
-                            }()
+                            impartedBuildSettings: PIF.BuildSettings()
                         )
                     ],
                     groupTree: PIF.Group(guid: "main-group-guid", path: "", children: [
@@ -262,14 +275,14 @@ class PIFTests: XCTestCase {
 
         XCTAssertEqual(workspace["type"]?.string, "workspace")
         XCTAssertEqual(workspaceContents["guid"]?.string, "workspace@11")
-        XCTAssertEqual(workspaceContents["path"]?.string, "/path/to/workspace")
+        XCTAssertEqual(workspaceContents["path"]?.string, AbsolutePath(path: "/path/to/workspace").pathString)
         XCTAssertEqual(workspaceContents["name"]?.string, "MyWorkspace")
         XCTAssertEqual(workspaceContents["projects"]?.array, [project["signature"]!])
 
         XCTAssertEqual(project["type"]?.string, "project")
         XCTAssertEqual(projectContents["guid"]?.string, "project@11")
-        XCTAssertEqual(projectContents["path"]?.string, "/path/to/workspace/project")
-        XCTAssertEqual(projectContents["projectDirectory"]?.string, "/path/to/workspace/project")
+        XCTAssertEqual(projectContents["path"]?.string, AbsolutePath(path: "/path/to/workspace/project").pathString)
+        XCTAssertEqual(projectContents["projectDirectory"]?.string, AbsolutePath(path: "/path/to/workspace/project").pathString)
         XCTAssertEqual(projectContents["projectName"]?.string, "MyProject")
         XCTAssertEqual(projectContents["projectIsPackage"]?.string, "true")
         XCTAssertEqual(projectContents["developmentRegion"]?.string, "fr")
@@ -330,7 +343,6 @@ class PIFTests: XCTestCase {
         XCTAssertEqual(exeTargetContents["type"]?.string, "standard")
         XCTAssertEqual(exeTargetContents["productTypeIdentifier"]?.string, "com.apple.product-type.tool")
         XCTAssertEqual(exeTargetContents["buildRules"]?.array, [])
-        XCTAssertEqual(exeTargetContents["impartedBuildProperties"]?.dictionary, ["buildSettings": JSON([:])])
 
         XCTAssertEqual(exeTargetContents["productReference"], JSON([
             "type": "file",
@@ -345,6 +357,7 @@ class PIFTests: XCTestCase {
             let debugSettings = debugConfiguration["buildSettings"]
             XCTAssertEqual(debugSettings?["TARGET_NAME"]?.string, "MyExecutable")
             XCTAssertEqual(debugSettings?["EXECUTABLE_NAME"]?.string, "my-exe")
+            XCTAssertEqual(debugConfiguration["impartedBuildProperties"]?.dictionary, ["buildSettings": JSON([:])])
 
             let releaseConfiguration = configurations[1]
             XCTAssertEqual(releaseConfiguration["guid"]?.string, "target-exe-config-release-guid")
@@ -353,6 +366,7 @@ class PIFTests: XCTestCase {
             XCTAssertEqual(releaseSettings?["TARGET_NAME"]?.string, "MyExecutable")
             XCTAssertEqual(releaseSettings?["EXECUTABLE_NAME"]?.string, "my-exe")
             XCTAssertEqual(releaseSettings?["SKIP_INSTALL"]?.string, "NO")
+            XCTAssertEqual(releaseConfiguration["impartedBuildProperties"]?.dictionary, ["buildSettings": JSON([:])])
         } else {
             XCTFail("invalid number of build configurations")
         }
@@ -395,23 +409,26 @@ class PIFTests: XCTestCase {
             "name": "MyLibrary"
         ]))
 
-        XCTAssertEqual(
-            libTargetContents["impartedBuildProperties"]?["buildSettings"]?["OTHER_CFLAGS"]?.array,
-            [.string("-fmodule-map-file=modulemap"), .string("$(inherited)")]
-        )
-
         if let configurations = libTargetContents["buildConfigurations"]?.array, configurations.count == 2 {
             let debugConfiguration = configurations[0]
             XCTAssertEqual(debugConfiguration["guid"]?.string, "target-lib-config-debug-guid")
             XCTAssertEqual(debugConfiguration["name"]?.string, "Debug")
             let debugSettings = debugConfiguration["buildSettings"]
             XCTAssertEqual(debugSettings?["TARGET_NAME"]?.string, "MyLibrary-Debug")
+            XCTAssertEqual(
+                debugConfiguration["impartedBuildProperties"]?["buildSettings"]?["OTHER_CFLAGS"]?.array,
+                [.string("-fmodule-map-file=modulemap"), .string("$(inherited)")]
+            )
 
             let releaseConfiguration = configurations[1]
             XCTAssertEqual(releaseConfiguration["guid"]?.string, "target-lib-config-release-guid")
             XCTAssertEqual(releaseConfiguration["name"]?.string, "Release")
             let releaseSettings = releaseConfiguration["buildSettings"]
             XCTAssertEqual(releaseSettings?["TARGET_NAME"]?.string, "MyLibrary")
+            XCTAssertEqual(
+                releaseConfiguration["impartedBuildProperties"]?["buildSettings"]?["OTHER_CFLAGS"]?.array,
+                [.string("-fmodule-map-file=modulemap"), .string("$(inherited)")]
+            )
         } else {
             XCTFail("invalid number of build configurations")
         }
@@ -440,23 +457,26 @@ class PIFTests: XCTestCase {
         ])
         XCTAssertEqual(aggregateTargetContents["buildRules"], nil)
 
-        XCTAssertEqual(
-            aggregateTargetContents["impartedBuildProperties"]?["buildSettings"]?["OTHER_CFLAGS"]?.array,
-            [.string("-fmodule-map-file=modulemap"), .string("$(inherited)")]
-        )
-
         if let configurations = aggregateTargetContents["buildConfigurations"]?.array, configurations.count == 2 {
             let debugConfiguration = configurations[0]
             XCTAssertEqual(debugConfiguration["guid"]?.string, "aggregate-target-config-debug-guid")
             XCTAssertEqual(debugConfiguration["name"]?.string, "Debug")
             let debugSettings = debugConfiguration["buildSettings"]
             XCTAssertNotNil(debugSettings)
+            XCTAssertEqual(
+                debugConfiguration["impartedBuildProperties"]?["buildSettings"]?["OTHER_CFLAGS"]?.array,
+                [.string("-fmodule-map-file=modulemap"), .string("$(inherited)")]
+            )
 
             let releaseConfiguration = configurations[1]
             XCTAssertEqual(releaseConfiguration["guid"]?.string, "aggregate-target-config-release-guid")
             XCTAssertEqual(releaseConfiguration["name"]?.string, "Release")
             let releaseSettings = releaseConfiguration["buildSettings"]
             XCTAssertNotNil(releaseSettings)
+            XCTAssertEqual(
+                releaseConfiguration["impartedBuildProperties"]?["buildSettings"]?["OTHER_CFLAGS"]?.array,
+                [.string("-fmodule-map-file=modulemap"), .string("$(inherited)")]
+            )
         } else {
             XCTFail("invalid number of build configurations")
         }
